class Qml::Widget < Object
  attr_reader :widget
  attr_accessor :parent
  attr_accessor :children
  attr_property :label
  attr_accessor :tooltip
  attr_accessor :layoutOpts
  attr_property :extern
  attr_accessor :layer
  attr_accessor :x
  attr_accessor :y
  attr_accessor :w
  attr_accessor :h
  def draw(vg);;end
  def onSetup(old=nil);
    ;end
  def layout(l, selfBox);
        widget.children.each do |child|
            if(child.respond_to?(:layout))
                nbox = l.genConstBox(0, 0,
                    selfBox.w, selfBox.h, child)
                child.layout(l, nbox)
            end
        end

        selfBox
    ;end
  def fixed(l, pbox, xc, yc, wc, hc);
        box = nil
        selfBox = l.genConstBox(pbox.w*xc, pbox.h*yc, pbox.w*wc, pbox.h*hc, self)
        box = self.layout(l, selfBox)
        box
    ;end
  def fixed_long(l, pbox, xc, yc, wc, hc, xf, yf, wf, hf);
        l.fixed_long(self, pbox, xc, yc, wc, hc, xf, yf, wf, hf)
    ;end
  def damage_self(all=nil);
        rt = self.root
        rt.damage_item(self, all) if rt
    ;end
  def background(c);
        $vg.path do |v|
            v.rect(0,0,w,h)
            v.fill_color c
            v.fill
        end
    ;end
  def class_name();
        "Widget"
    ;end
  def window();
        if(parent.respond_to?(:root))
            parent.window
        else
            self
        end
    ;end
  def root();
        if(parent.respond_to?(:root))
            parent.root
        else
            parent
        end
    ;end
  def global_x();
        par = 0
        if(parent.respond_to?(:global_x))
            par = parent.global_x
        end
        par + widget.x
    ;end
  def global_y();
        par = 0
        if(parent.respond_to?(:global_y))
            par = parent.global_y
        end
        par + widget.y
    ;end
  def to_s(i=0);
        out = ""
        i.times do
            out += " |"
        end

        out += "- <"+widget.class_name()+"##{widget.ui_path}:" + widget.label
        out += ">\n"
        widget.children.each do |x|
            out += x.to_s(i+1)
        end

        out
    ;end
  def abs_x();
        p = parent
        if(p.respond_to?(:root))
            widget.x + p.abs_x
        else
            widget.x
        end
    ;end
  def abs_y();
        p = parent
        if(p.respond_to?(:root))
            widget.y + p.abs_y
        else
            widget.y
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super()
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @widget = self

    
context = Hash.new
context["widget"] = self
Qml::prop_add(self, "label")
Qml::prop_add(self, "extern")
self.parent = nil
self.children = []
self.label = ""
self.tooltip = ""
self.layoutOpts = []
self.extern = ""
self.layer = 0
self.x = nil
self.y = nil
self.w = nil
self.h = nil

    #puts "Widget, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Button < Qml::Widget
  attr_reader :button
  attr_accessor :action
  attr_property :value
  attr_accessor :renderer
  attr_accessor :textScale
  attr_accessor :whenValue
  attr_accessor :pad
  attr_property :active
  def onMousePress(ev);
        button.value = !button.value
        damage_self
        whenValue.call if whenValue
    ;end
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def onMerge(val);
        button.value = val.value if(val.respond_to? :value)
    ;end
  def class_name();
        "Button"
    ;end
  def aspect2(box, ww, hh);
        provided    = box.w/box.h
        recommended = ww/hh
        if(recommended > provided)
            #Decrease height
            new_height = box.h*provided/recommended
            box.y += (box.h-new_height)/2
            box.h = new_height
        else
            new_width = box.w*recommended/provided
            box.x += (box.w-new_width)/2
            box.w  = new_width
        end
    ;end
  def layout(l, selfBox);
        if(!self.layoutOpts.include?(:no_constraint))
            if(label.length == 1)
                l.aspect(selfBox, 1, 1)
            else
                scale = 100
                $vg.font_size scale
                bb = $vg.text_bounds(0, 0, label.upcase)
                if(bb != 0)
                    #Width cannot be so small that letters overflow
                    self.aspect2(selfBox, bb, scale)
                end
            end
        end
        selfBox
    ;end
  def draw_text(vg);
        text_color1   = Theme::TextActiveColor
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*self.textScale
        if(value == true)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        if(layoutOpts.include? :left_text)
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(8,h/2,button.label.upcase)
        else
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h/2,button.label.upcase)
        end
    ;end
  def draw_inactive(vg);
        strike_color = Theme::TextColor
        vg.path do
            vg.move_to(w*pad, h*pad)
            vg.line_to(w*(1-2*pad), h*(1-2*pad))
            vg.stroke_width 1.0
            vg.stroke_color strike_color
            vg.stroke
        end
    ;end
  def draw_button(vg);
        off_color     = Theme::ButtonInactive
        on_color      = Theme::ButtonActive
        cs = 0
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 2)
            if(button.value == true)
                cs = 1
                v.fill_color on_color
            elsif(button.value.class == Float && button.value != 0)
                cs = 2
                t = button.value
                on = on_color
                of = Theme::ButtonGrad1
                v.fill_color(color_rgb(on.r*t + of.r*(1-t),
                                       on.g*t + of.g*(1-t),
                                       on.b*t + of.b*(1-t)))
            else
                paint = v.linear_gradient(0,0,0,h,
                Theme::ButtonGrad1, Theme::ButtonGrad2)
                v.fill_paint paint
            end
            v.fill
            v.stroke_width 1
            v.stroke

        end

        vg.path do |v|
            hh = h/20
            v.move_to(w*pad+1,       h*pad+hh)
            v.line_to(w*(1-2*pad)+1, h*pad+hh)
            if(cs == 0)
                v.stroke_color color("5c5c5c")
            elsif(cs == 1)
                v.stroke_color color("16a39c")
            end
            if([0,1].include?(cs))
                v.stroke_width hh
                v.stroke
            end
        end
    ;end
  def draw(vg);
        draw_button(vg)

        draw_text(vg)

        draw_inactive(vg) if !self.active
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @button = self

    
context = Hash.new
context["button"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "active")
self.action = nil
self.value = false
self.renderer = nil
self.textScale = 0.75
self.whenValue = nil
self.pad = 1/64
self.active = true

    #puts "Button, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Col < Qml::Widget
  attr_reader :col
  attr_property :spacer
  def layout(l, selfBox);
        #Create A list of boxes
        bbs = []
        begin
            n = children.length
            children.each_with_index do |ch, i|
                sh = selfBox.h-spacer*n
                box = l.genConstBox(0, selfBox.h*i/n,
                selfBox.w, sh/n, ch)
                bb = ch.layout(l, box)
                bbs << bb
            end
        end

        minheight = 99999
        widths = Hash.new
        bbs.each do |b|
            tp = b.info.class == Qml::Knob ? :knob : :other
            minheight     = b.h if b.h < minheight
            widths[tp]  ||= b.w
            widths[tp]    = b.w if b.w < widths[tp]
        end

        bbs.each do |b|
            tp = b.info.class == Qml::Knob ? :knob : :other
            b.y  += (b.h-minheight)/2
            b.h   = minheight
            b.x   = 0
            b.w   = selfBox.w
        end

        #make longer lists compact
        if(bbs.length > 3)
            py = spacer
            bbs.each do |b|
                b.y  = py
                py  += b.h + spacer
            end
        end

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @col = self

    
context = Hash.new
context["col"] = self
Qml::prop_add(self, "spacer")
self.spacer = 5

    #puts "Col, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::DropDown < Qml::Widget
  attr_accessor :options
  attr_property :callback
  attr_property :hl
  def prime(root);
        @ctime = Time.new
        root.set_modal(self)
    ;end
  def draw(vg);
        self.options = [self.options] if self.options.class == String
        n  = options.length
        dy = h/n

        if self.hl.nil?
            self.hl = 0   if self.y == 0
            self.hl = n-1 if self.y != 0
        end

        (0...n).each do |row|
            vg.path do |v|
                v.rect(1, row*dy, w-1, dy+1)
                v.fill_color Theme::BankEven     if row%2 == 0
                v.fill_color Theme::BankOdd      if row%2 == 1
                v.fill_color Theme::ButtonActive if hl == row
                v.stroke_color color(:black)
                v.stroke_width  2.0
                v.stroke
                v.fill
            end
        end


        text_color = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*0.8/n
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color text_color

        pad  = 1/64
        options.each_with_index do |opt, i|
            vg.text(3+w*pad*2,dy*(i+0.5),opt.upcase)
        end
        vg.path do |v|
            v.rect(1, 1, w-1, h-1)
            v.stroke_color color(:black)
            v.stroke_width  2.0
            v.stroke
        end
    ;end
  def onMousePress(ev);
        #Allow Tap within 200ms
        return if @ctime && Time.new-@ctime < 0.200
        n   = options.length
        opt = (n*(ev.pos.y-global_y)/h).to_i
        xsp = (ev.pos.x-global_x)/w
        inx = (0 <= xsp && xsp <= 1) && (0 <= opt && opt < n)


        callback.call opt if callback &&  inx
        callback.call nil if callback && !inx
        rt = self.root
        return if rt.nil?
        rt.set_modal(nil)
        rt.ego_death self
    ;end
  def onMouseRelease(ev); onMousePress(ev) ;end
  def onMouseHover(ev); onMouseMove(ev)  ;end
  def onMouseMove(ev);
        n   = options.length
        opt = (n*(ev.pos.y-global_y)/h).to_i
        self.hl = opt
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous3 = self

    
context = Hash.new
Qml::prop_add(self, "callback")
Qml::prop_add(self, "hl")
self.options = []
self.callback = nil
self.hl = nil

    #puts "DropDown, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Envelope < Qml::Widget
  attr_reader :env
  attr_reader :run_view
  attr_accessor :prev
  attr_accessor :selected
  attr_property :xpoints
  attr_property :ypoints
  attr_property :points
  attr_property :sustain_point
  attr_accessor :valueRef
  attr_property :mouse_enable
  attr_property :whenTime
  attr_property :emode
  def setup_valuerefs();
        ext = env.extern
        xpts = OSC::RemoteParam.new($remote, ext + "envdt")
        ypts = OSC::RemoteParam.new($remote, ext + "envval")
        pts = OSC::RemoteParam.new($remote, ext + "Penvpoints")
        pts.mode = :selector
        sus = OSC::RemoteParam.new($remote, ext + "Penvsustain")
        sus.mode = :full
        free = OSC::RemoteParam.new($remote, ext + "Pfreemode")
        mode = OSC::RemoteParam.new($remote, ext + "Envmode")
        mode.mode = :selector
        xpts.callback = lambda { |x|
            env.xpoints = x
            env.damage_self
            whenTime.call if whenTime
        }
        ypts.callback = lambda { |x|
            env.ypoints = x.map {|xx| 2*xx-1}
            env.damage_self
        }
        pts.callback = lambda { |x|
            env.points = x
            env.damage_self
            whenTime.call if whenTime
        }
        sus.callback = lambda { |x|
            env.sustain_point = x
            env.damage_self
        }
        free.callback = lambda { |x|
            env.mouse_enable = x
        }
        mode.callback = lambda { |x|
            env.emode = x
        }
        env.valueRef = [xpts, ypts, pts, sus, free, mode]

        run_view.extern = env.extern+"out"
    ;end
  def refresh();
        return if self.valueRef.nil?
        self.valueRef.each do |v|
            v.refresh
        end
    ;end
  def get_x_points();
        tmp = Draw::DSP::norm_0_1(Draw::DSP::cumsum(xpoints[0...points]))
        Draw::DSP::pad_norm(tmp, 0.01)
    ;end
  def warp(x);
        wp = get_x_points()
        y  = []
        x.each_with_index do |xx, i|
            if((i%2) == 1)
                y << xx
            else
                ps = xx.to_i
                fr = xx-ps
                ps -= 1
                y << xx if(ps >= wp.length)
                next    if(ps >= wp.length)
                aa = wp[ps]
                bb = wp[ps]
                bb = wp[ps+1] if(ps+1 < wp.length)
                y << aa+fr*(bb-aa)
            end
        end
        y
    ;end
  def onMousePress(ev);
        #return if !self.mouse_enable
        #//Try to identify the location  of the nearest grabbable point
        #valuator.prev = ev.pos
        xdat = get_x_points()
        ydat = env.ypoints
        n = [xdat.length, ydat.length].min
        next_sel = 0
        best_dist = 1e10

        mx = ev.pos.x-global_x
        my = ev.pos.y-global_y
        (0...n).each do |i|
            xx = w*xdat[i];
            yy = h/2-h/2*ydat[i];

            dst = (mx-xx)**2 + (my-yy)**2
            if(dst < best_dist)
                best_dist = dst
                next_sel  = i
            end
        end
        if(env.selected != next_sel)
            env.selected = next_sel
            env.root.damage_item env
        end
        env.prev = ev.pos
    ;end
  def bound_points(array, low, high);
        n = array.length
        (0...n).each do |i|
            if(array[i] < low)
                array[i] = low
            elsif(array[i] > high)
                array[i] = high
            end
        end
    ;end
  def onMouseMove(ev);
        #return if !self.mouse_enable

        if(env.selected)
            scalex = 4*(env.xpoints[env.selected]+10)
            dy = 2*(ev.pos.y - env.prev.y)/env.h
            dx = scalex*(ev.pos.x - env.prev.x)/env.w
            n  = [env.xpoints.length, env.ypoints.length].min
            if(env.selected == 0 || env.selected == n-1)
                env.ypoints[env.selected] -= dy
            else
                env.xpoints[env.selected] += dx
                env.ypoints[env.selected] -= dy
            end

            bound_points(env.xpoints,  0.0, 40950.0)
            bound_points(env.ypoints, -1.0, 1.0)

            send_points() if mouse_enable
            update_nonfree_x(env.xpoints) if !mouse_enable
            update_nonfree_y(env.ypoints) if !mouse_enable

            env.prev = ev.pos
            env.root.damage_item env
        end
    ;end
  def cvt_x(x);
        fval = Math::log2(x/10.0 + 1.0) * 127.0/12.0
        [0, [127, fval.round].min].max.to_i
    ;end
  def update_nonfree_x(pts);
        if(emode == 1)
            $remote.seti(extern + "PA_dt", cvt_x(pts[1]))
            $remote.seti(extern + "PD_dt", cvt_x(pts[2]))
            $remote.seti(extern + "PR_dt", cvt_x(pts[3]))
        elsif(emode == 2)
            $remote.seti(extern + "PA_dt", cvt_x(pts[1]))
            $remote.seti(extern + "PD_dt", cvt_x(pts[2]))
            $remote.seti(extern + "PR_dt", cvt_x(pts[3]))
        elsif(emode == 3)
            $remote.seti(extern + "PA_dt", cvt_x(pts[1]))
            $remote.seti(extern + "PR_dt", cvt_x(pts[2]))
        elsif(emode == 4)
            $remote.seti(extern + "PA_dt", cvt_x(pts[1]))
            $remote.seti(extern + "PD_dt", cvt_x(pts[2]))
            $remote.seti(extern + "PR_dt", cvt_x(pts[3]))
        elsif(emode == 5)
            $remote.seti(extern + "PA_dt", cvt_x(pts[1]))
            $remote.seti(extern + "PR_dt", cvt_x(pts[2]))
        end
        whenTime.call if whenTime
    ;end
  def cvt_y(x);
        [0, [127, 127*(x+1)/2].min].max.to_i
    ;end
  def update_nonfree_y(pts);
        if(emode == 1)
            pts[0]      = -1.0
            pts[1]      = +1.0
            $remote.seti(extern + "PS_val", cvt_y(pts[2]))
            pts[3]      = -1.0
        elsif(emode == 2)
            pts[0]      = -1.0
            pts[1]      = +1.0
            $remote.seti(extern + "PS_val", cvt_y(pts[2]))
            pts[3]      = -1.0
        elsif(emode == 3)
            $remote.seti(extern + "PA_val", cvt_y(pts[0]))
            pts[1]      = 0.0
            $remote.seti(extern + "PR_val", cvt_y(pts[2]))
        elsif(emode == 4)
            $remote.seti(extern + "PA_val", cvt_y(pts[0]))
            $remote.seti(extern + "PD_val", cvt_y(pts[1]))
            pts[2]      = 0.0
            $remote.seti(extern + "PR_val", cvt_y(pts[3]))
        elsif(emode == 5)
            $remote.seti(extern + "PA_val", cvt_y(pts[0]))
            pts[1]      = 0.0
            $remote.seti(extern + "PR_val", cvt_y(pts[2]))
        end
    ;end
  def send_points();
        return if self.extern.nil?
        ry = ypoints.map {|xx| (xx+1)/2}
        valueRef[0].value = env.xpoints
        valueRef[1].value = ry
    ;end
  def class_name();
        "Envelope"
    ;end
  def draw(vg);
        xdat = get_x_points()
        ydat = env.ypoints

        fill_color   = Theme::VisualBackground
        stroke_color = Theme::VisualStroke
        light_fill   = Theme::VisualLightFill
        bright       = Theme::VisualBright
        dim          = Theme::VisualDim
        sel_color    = Theme::VisualSelect

        padfactor = 20
        bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)

        pts = Draw::zipToPos(xdat, ydat)

        background(fill_color)

        #Draw Highlights
        Draw::WaveForm::under_highlight(vg, bb, pts, light_fill)
        Draw::WaveForm::over_highlight(vg,  bb, pts, light_fill)

        #Draw Zero Line
        Draw::WaveForm::zero_line(vg, bb, dim)

        #Indicate Sustain Point
        Draw::WaveForm::env_sel_line(vg, bb, self.sustain_point, pts, dim)

        #Draw Actual Line
        Draw::WaveForm::env_plot(vg, bb, pts, bright, selected)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @env = self
@run_view = Qml::Widget.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["env"] = self
context["run_view"] = @run_view
@run_view.instance_eval do
def run_view=(k);@run_view=k;end
def run_view;    @run_view;  end
end
Qml::context_apply(context, @run_view)
Qml::prop_add(self, "xpoints")
Qml::prop_add(self, "ypoints")
Qml::prop_add(self, "points")
Qml::prop_add(self, "sustain_point")
Qml::prop_add(self, "mouse_enable")
Qml::prop_add(self, "whenTime")
Qml::prop_add(self, "emode")

        @run_view.instance_eval("def class_name();\n            \"EnvVisAnimation\"\n        ;end", "src/mruby-zest/qml/Envelope.qml", 280)

        @run_view.instance_eval("def valueRef=(value_ref);\n            @value_ref = value_ref\n        ;end", "src/mruby-zest/qml/Envelope.qml", 286)

        @run_view.instance_eval("def valueRef();\n            @value_ref\n        ;end", "src/mruby-zest/qml/Envelope.qml", 291)

        @run_view.instance_eval("def runtime_points=(pts);\n            @runtime_points = pts\n        ;end", "src/mruby-zest/qml/Envelope.qml", 296)

        @run_view.instance_eval("def update_points(xx);\n            self.runtime_points = xx\n            damage_self\n\n            @last = Time.new\n        ;end", "src/mruby-zest/qml/Envelope.qml", 312)

        @run_view.instance_eval("def animate();\n            run_view.valueRef.watch run_view.extern\n            now     = Time.new\n            @last ||= now\n            update_points([]) if((now-@last)>0.1)\n        ;end", "src/mruby-zest/qml/Envelope.qml", 320)

        @run_view.instance_eval("def draw(vg);\n            sel_color    = Theme::VisualSelect\n            dim_color    = Theme::VisualDimTrans\n            #Draw the data\n            pts   = @runtime_points\n            pts ||= []\n            return if pts.class != Array\n\n            padfactor = 20\n            bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)\n\n            Draw::WaveForm::overlay(vg, bb, pts)\n        ;end", "src/mruby-zest/qml/Envelope.qml", 328)
@run_view.layer = 1
@db.connect_property(@run_view.properties["extern"], lambda {run_view.parent.extern})
@db.connect_watcher(@run_view.properties["extern"], lambda {begin

            return if run_view.extern.nil?
            run_view.valueRef = OSC::RemoteParam.new($remote, run_view.extern)
            run_view.valueRef.set_watch
            run_view.valueRef.callback = Proc.new {|x|
                run_view.update_points(env.warp(x))
                run_view.runtime_points = env.warp(x);
                run_view.root.damage_item run_view
            }
        
end})
Qml::add_child(self, @run_view)
self.prev = nil
self.selected = nil
@db.connect_property(self.properties["xpoints"], lambda {[0.0, 0.2, 0.7, 0.8, 1.0]})
@db.connect_property(self.properties["ypoints"], lambda {[0.0, 0.5, 0.3, -0.9, 0.0]})
self.points = 5
self.sustain_point = 3
self.valueRef = nil
self.mouse_enable = true
self.whenTime = nil
self.emode = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        env.setup_valuerefs
    
end})

    #puts "Envelope, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::FlowLayout < Object
  attr_reader :this
  def bound_child(l, c);
        p  = this
        ch = c
        l.let(:pw, p.w, :ph, p.h,
              :x, c.x, :y, c.y, :w, c.w, :h, c.h) do
            l.c(:x, :w, :<, :pw)
            l.c(:y, :h, :<, :ph)
        end
    ;end
  def layout(l);
        if(l.parent)
            bound_parent(l)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super()
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @this = self

    
context = Hash.new
context["this"] = self

    #puts "FlowLayout, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Group < Qml::Widget
  attr_reader :mod
  attr_reader :titleW
  attr_reader :title
  attr_reader :content
  attr_accessor :copyable
  attr_property :toggleable
  attr_accessor :editable
  attr_accessor :whenClick
  attr_accessor :topSize
  attr_property :pxpad
  def class_name(); "Group" ;end
  def onMousePress(ev);
        whenClick.call if whenClick
    ;end
  def onSetup(old=nil);
        mch = mod.children
        cch = content.children
        if(mch.length > 2)
            cch = cch+mch[2..-1]
            mch = mch[0..1]
            content.children = cch
            mod.children     = mch
        else
        end
        content.children.each do |ch|
            ch.parent = content
        end
    ;end
  def layout(l, selfBox);
        titleW.fixed(l,   selfBox, 0, 0,       1, topSize)
        content.fixed_long(l, selfBox, 0, topSize, 1, 1-topSize, pxpad, pxpad, -2*pxpad, -2*pxpad)

        selfBox
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0, 0, mod.w, mod.h)
            v.fill_color(color("000000"))
            v.fill
        end
        #//paint the top half
        pos = [0, 0, mod.w, topSize*mod.h]
        vg.path do |v|
            border(pos[3]*0.01, pos);
            vg.rect(*pos);
            v.fill_color Theme::ModuleGrad2
            vg.fill
        end

        #upperspace = [pos[0]+pos[2]*2.0f/3.0f, pos[1], pos[2]/3.0f, pos[3]];
        #//upper.draw(upperspace);

        pos2 = [*pos]
        pos2[2] /= 3;
        pad((1-topSize), pos2);
        #drawLeftLabel(vg, str, SPLAT(pos2));



        innerspace = [0, mod.h*topSize, mod.w, mod.h*(1-topSize)];
        border(w*0.003, innerspace);
        #//paint the inner panel
        vg.path do |v|
            v.rect(*innerspace)
            paint = v.linear_gradient(0,innerspace[1],0,innerspace[3],
            Theme::ModuleGrad1, Theme::ModuleGrad2)
            v.fill_paint paint
            v.fill
        end

    ;end
end
class Qml::Text < Qml::Widget
  attr_reader :text
  attr_accessor :textColor
  attr_property :height
  attr_property :align
  def class_name();
        "Text"
    ;end
  def draw(vg);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, label.upcase)


        vg.font_face("bold")
        if(w/(self.height*h) < bb)
            vg.font_size self.height*h
        else
            vg.font_size self.height*h
        end
        vg.fill_color(self.textColor)
        if(align == :center)
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h/2,label.upcase)
        else
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(0,h/2,label.upcase)
        end
    ;end
  def layout(l, selfBox);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, label.upcase)
        if(bb != 0 && !self.layoutOpts.include?(:ignoreAspect) &&
           !self.layoutOpts.include?(:no_constraint))
            #Width cannot be so small that letters overflow
            if(height == 0.5)
                l.aspect(selfBox, bb, 1.5*scale)
            else
                l.aspect(selfBox, bb, scale)
            end
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text = self

    
context = Hash.new
context["text"] = self
Qml::prop_add(self, "height")
Qml::prop_add(self, "align")
self.textColor = Theme::TextColor
self.height = 0.5
self.align = :center

    #puts "Text, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Group
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @mod = self
@titleW = Qml::Widget.new(database, ui_path+"1"+'/')
@title = Qml::Text.new(database, ui_path+"2"+'/')
@content = Qml::Widget.new(database, ui_path+"3"+'/')

    
context = Hash.new
context["mod"] = self
context["titleW"] = @titleW
@titleW.instance_eval do
def mod=(k);@mod=k;end
def mod;    @mod;  end
end
context["title"] = @title
context["content"] = @content
Qml::context_apply(context, @titleW)
Qml::prop_add(self, "toggleable")
Qml::prop_add(self, "pxpad")

        @titleW.instance_eval("def onMousePress(ev);\n            mod.whenClick.call if mod.whenClick\n        ;end", "src/mruby-zest/qml/Group.qml", 99)

        @titleW.instance_eval("def class_name(); \"TitleBox\" ;end", "src/mruby-zest/qml/Group.qml", 103)

        @titleW.instance_eval("def pack_misc(l, selfBox, ch);\n            #Create A list of boxes\n            bbs = []\n            n = children.length\n            children.each_with_index do |ch, i|\n                box = l.genConstBox(selfBox.w*i/n, 0,\n                selfBox.w/n, selfBox.h, ch)\n                bb = ch.layout(l, box)\n                bbs << bb\n            end\n\n            bbs[0].x = 15 if bbs[0].x > 15\n\n            #put stuff on backwards\n            bx = selfBox.w\n            bbs[1..-1].reverse.each do |b|\n                b.x  = bx-b.w\n                bx  -= b.w\n            end\n\n            return selfBox\n\n            minheight = 99999\n            widths = Hash.new\n            bbs.each do |b|\n                tp = b.info.class == Qml::Knob ? :knob : :other\n                minheight     = b.h if b.h < minheight\n                widths[tp]  ||= b.w\n                widths[tp]    = b.w if b.w < widths[tp]\n            end\n            bbs.each do |b|\n                tp = b.info.class == Qml::Knob ? :knob : :other\n                #b.y += (b.h-widths[tp])/2\n                b.y  += (b.h-minheight)/2\n                b.h   = minheight\n                b.x   = 0\n                b.w   = selfBox.w\n            end\n            selfBox\n            return\n            prev = ch[0]\n            ch[1..-1].each_with_index do |c,i|\n                l.contains(selfBox, c)\n                l.rightOf(prev,c)\n                l.aspect(c, 1, 1) if children[i+1].class == Qml::PowButton\n                prev = c\n            end\n            l.weak(ch[1].x) if ch.length > 1\n            selfBox\n        ;end", "src/mruby-zest/qml/Group.qml", 105)

        @titleW.instance_eval("def layout(l, selfBox);\n            #return selfBox if children.length == 1\n            pack_misc(l, selfBox, children)\n        ;end", "src/mruby-zest/qml/Group.qml", 157)

        @titleW.instance_eval("def onSetup(old=nil);\n            return if children.length != 1\n            if(mod.toggleable)\n                pb = Qml::PowButton.new(db)\n                pb.extern = mod.toggleable\n                Qml::add_child(self, pb)\n            end\n            if(mod.copyable)\n                cb = Qml::CopyButton.new(db)\n                pb = Qml::PasteButton.new(db)\n                cb.extern = mod.extern\n                pb.extern = mod.extern\n                Qml::add_child(self, cb)\n                Qml::add_child(self, pb)\n            end\n        ;end", "src/mruby-zest/qml/Group.qml", 171)
@db.connect_property(@title.properties["label"], lambda {mod.label})
@title.align = :left
@title.textColor = Theme::TextModColor
@title.height = 0.8
Qml::add_child(@titleW, @title)
Qml::add_child(self, @titleW)

        @content.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vpack(l, selfBox, children)\n        ;end", "src/mruby-zest/qml/Group.qml", 192)

        @content.instance_eval("def class_name(); \"ContentBox\" ;end", "src/mruby-zest/qml/Group.qml", 196)
Qml::add_child(self, @content)
self.copyable = true
self.toggleable = false
self.editable = false
self.whenClick = nil
self.topSize = 0.1
self.pxpad = 8.0
self.layoutOpts = [:horizontal]
@db.connect_watcher(self.properties["toggleable"], lambda {begin

        return if(mod.copyable && titleW.children.length < 2)
        if(mod.toggleable && titleW.children[1].class != Qml::PowButton)
            pb = Qml::PowButton.new(mod.db)
            pb.extern = mod.toggleable
            pb.parent = titleW
            ch = titleW.children
            ch.insert(1, pb)
            titleW.children = ch
        end
    
end})

    #puts "Group, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Keyboard < Qml::Widget
  attr_reader :keyboard
  attr_property :data
  attr_accessor :valueRef
  attr_property :whiteKeys
  attr_property :fixpad
  attr_property :select_id
  attr_property :prev_note
  attr_property :velocity
  attr_property :velrnd
  attr_property :octave
  attr_property :priority
  def set_data(data_);
        if(self.data != data_)
            self.data = data_
            self.damage_self
        end
    ;end
  def onSetup(old=nil);
        self.valueRef = OSC::RemoteParam.new($remote, "/active_keys")
        self.valueRef.callback = lambda { |x| keyboard.set_data x }
    ;end
  def animate();
        self.valueRef.refresh
    ;end
  def class_name();"Keyboard";end
  def white_key_id(ind);
        @@key_lookup ||= Hash.new
        return @@key_lookup[ind] if @@key_lookup.include?(ind)

        base = 9
        black_pattern = [1,0,1,1,0,1,1];

        off = 0

        black_pos = 0
        ind.times do
            black_pos = black_pos % black_pattern.length
            off += 1
            off += black_pattern[black_pos]
            black_pos += 1
        end

        @@key_lookup[ind] = base + off
        base + off
    ;end
  def get_note(pos);
        rel = Pos.new(pos.x-global_x, pos.y-global_y)
        (0..whiteKeys).each do |i|
            bid = black_key_id(i)
            wid = white_key_id(i)
            bb  = black_bb(i, whiteKeys, false)
            wb  = white_bb(i, whiteKeys, false)
            if(Rect.new(*bb).include? rel)
                return bid
            elsif(Rect.new(*wb).include? rel)
                return wid
            end
        end
        return nil
    ;end
  def onMousePress(ev);
        note = get_note(ev.pos)
        self.prev_note = note

        noteOn(note)
    ;end
  def noteOn(note);
        vel = velocity+(rand-0.5)*velrnd;
        vel = [0, [127, vel].min].max.to_i

        if(note && data && data[note])
            return
        end


        if(note && $remote)
            $remote.action("/noteOn", 0, note, vel)
        end
    ;end
  def onMouseMove(ev);
        note = get_note(ev.pos)
        if(note != self.prev_note)
            $remote.action("/noteOff", 0, self.prev_note)
            noteOn(note)
            self.prev_note = note
        end
    ;end
  def onMouseRelease(ev);
        note = get_note(ev.pos)
        if(note && $remote)
            $remote.action("/noteOff", 0, note)
        end
    ;end
  def onKey(k, act);
        qwerty_high = "q2w3er5t6y7ui9o0p[=]\\"
        qwerty_low  = "zsxdcvgbhnjm,l.;/"

        note = nil
        off = 0
        qwerty_low.each_char do |i|
            if(k==i)
                note = 60 - 12 + off + self.octave*12
                break
            end
            off += 1
        end

        off = 0
        qwerty_high.each_char do |i|
            if(k==i)
                note = 60 + off
                break
            end
            off += 1
        end

        return if note.nil?

        if(act == "press")
            noteOn(note)
        else
            $remote.action("/noteOff", 0, note)
        end

    ;end
  def black_key_id(ind);
        white_key_id(ind) + 1
    ;end
  def no_black(i);
        black_pattern = [1,0,1,1,0,1,1];
        black_pattern[i%7] == 0
    ;end
  def white_bb(i, white_keys, pad=true);
        padh = fixpad/2
        box  = [i*w*1.0/(white_keys+1), padh, w*1.0/(white_keys+1), h-padh];
        if(pad)
            fixedpad(box, fixpad)
        else
            box
        end
    ;end
  def black_bb(i, white_keys, pad=true);
        width = 0.65
        box = [(i+1.0-width/2)*w*1.0/(white_keys+1), 0,
               w*width/(white_keys+1), h*0.7];
        if(pad)
            fixedpad(box, fixpad)
        else
            box
        end
    ;end
  def draw_white(vg, i, white_keys);
        white1       = Theme::KeyWhiteGrad1
        white2       = Theme::KeyWhiteGrad2
        white_accent = Theme::KeyWhiteAccent
        enable_color = Theme::KeyEnable

        box = white_bb(i, white_keys)
        wid = white_key_id i
        vg.path do |vg|
            vg.rect(*box)
            paint = vg.linear_gradient(0,0,0,h,white1, white2)
            vg.fill_paint paint
            vg.fill_color enable_color if data && data[wid]
            vg.fill_color enable_color if wid == select_id
            vg.stroke_color white_accent

            vg.fill
            vg.stroke_width fixpad
            vg.stroke
        end
    ;end
  def draw_black(vg, i, white_keys);
        black_color  = Theme::KeyBlack
        black_accent = Theme::KeyBlackAccent
        enable_color = Theme::KeyEnable
        bg_color     = Theme::KeyBackground

        id = black_key_id i
        return if no_black i
        box = black_bb(i, white_keys)
        vg.path do |vg|
            vg.rect(*box)
            vg.fill_color black_color
            vg.fill_color enable_color if data && data[id]
            vg.fill_color enable_color if id == select_id
            vg.stroke_color bg_color
            vg.fill
            vg.stroke
        end
        id = black_key_id i
        return if no_black i

        box = black_bb(i, white_keys)
        vg.path do |vg|
            p = fixpad/2
            vg.move_to(box[0]       -p,p+box[1]+box[3])
            vg.line_to(box[0]+box[2]+p,p+box[1]+box[3])
            vg.stroke_color bg_color
            vg.stroke_width fixpad*2
            vg.stroke
        end
        vg.path do |vg|
            p = fixpad/2
            o = fixpad/4
            vg.move_to(box[0]       +p,box[1]+box[3]-o)
            vg.line_to(box[0]+box[2]-p,box[1]+box[3]-o)
            vg.stroke_color black_accent
            vg.stroke_width fixpad
            vg.stroke
        end
    ;end
  def draw(vg);
        white_keys = 8*7-3;
        white_keys = self.whiteKeys;
        black_pattern = [1,0,1,1,0,1,1];

        bg_color     = Theme::KeyBackground
        background bg_color

        #draw the white keys 7 octaves + 2
        (0..white_keys).each do |i|
            draw_white(vg, i, white_keys)
        end

        #draw the black keys at the joints
        (0..white_keys-2).each do |i|
            draw_black(vg, i, white_keys)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @keyboard = self

    
context = Hash.new
context["keyboard"] = self
Qml::prop_add(self, "data")
Qml::prop_add(self, "whiteKeys")
Qml::prop_add(self, "fixpad")
Qml::prop_add(self, "select_id")
Qml::prop_add(self, "prev_note")
Qml::prop_add(self, "velocity")
Qml::prop_add(self, "velrnd")
Qml::prop_add(self, "octave")
Qml::prop_add(self, "priority")
self.data = nil
self.valueRef = nil
@db.connect_property(self.properties["whiteKeys"], lambda {8*7-5})
self.fixpad = 1.5
self.select_id = nil
self.prev_note = nil
self.velocity = 100
self.velrnd = 0
self.octave = 0
self.priority = 10

    #puts "Keyboard, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Valuator < Qml::Widget
  attr_reader :valuator
  attr_accessor :valueRef
  attr_accessor :prev
  attr_property :value
  attr_accessor :dragScale
  attr_accessor :vertical
  attr_accessor :whenValue
  attr_property :active
  attr_property :type
  attr_property :units
  def setValue(v);
        self.value = lim(v, 0.0, 1.0)
        damage_self
    ;end
  def refresh(); valueRef.refresh if valueRef ;end
  def create_radial();
        return
        gbl_cx = valuator.global_x + 0.5*valuator.w
        gbl_cy = valuator.global_y + 0.5*valuator.h
        gbl_w  = window.w
        gbl_h  = window.h

        ropt = [gbl_cx, gbl_cy, gbl_w-gbl_cx, gbl_h-gbl_cy].min

        diameter = [2.0*ropt, 3.0*0.5*(valuator.w+valuator.h)].min

        widget = RadialMenu.new(valuator.db)
        widget.w = diameter
        widget.h = diameter
        widget.x = valuator.w/2-diameter/2
        widget.y = valuator.h/2-diameter/2
        widget.layer = 2
        widget.callback = lambda {|sel|
            case sel
            when 0
                valueRef.midi_learn if valueRef
            when 1
                valueRef.rand if valueRef
            when 2
                valueRef.midi_unlearn if valueRef
            when 3
                valueRef.default if valueRef
            end
        }

        Qml::add_child(valuator, widget)
        valuator.root.smash_draw_seq
        valuator.root.damage_item widget
    ;end
  def onScroll(ev);
        return if !self.active
        fine = root.fine_mode ? 0.2 : 1.0
        updatePos(-fine*ev.dy/50.0)
    ;end
  def reset();
        reset_value = 64.01/127.0
        if(valueRef)
            old_dsp = self.valueRef.display_value

            self.valueRef.value = reset_value
            self.value = reset_value
            new_dsp = self.valueRef.display_value
            whenValue.call if whenValue && (new_dsp.nil? || old_dsp != new_dsp)
            valuator.root.log(:user_value, valuator.valueRef.display_value, src=valuator.label)
        else
            self.value = reset_value
        end
        damage_self
    ;end
  def onMousePress(ev);
        mouse_handle(ev)
    ;end
  def mouse_handle(ev);
        return if !self.active
        if(root.learn_mode && extern)
            $remote.automate(extern)
            root.log(:tooltip, "Learning #{extern}")
        end
        if(ev.buttons.include? :leftButton)
            valuator.prev = ev.pos
            now = Time.new
            #400 ms delta for double click
            if(@click_time && (now-@click_time) < 0.400)
                $remote.default(extern) if extern
                whenValue.call if whenValue
            end
            @click_time = now
        elsif(ev.buttons.include? :rightButton)
            if(children.empty?)
                create_radial
            end
        elsif(ev.buttons.include? :middleButton)
            reset
        elsif(ev.buttons.include? :drag_and_drop)
            puts "sending dnd event..."
            $remote.action("/last_dnd", self.extern)
        else
            valuator.prev = nil
        end
    ;end
  def onMouseMove(ev);
        return if !self.active
        fine = root.fine_mode ? 0.05 : 1.0
        if(prev)
            delta = if(vertical)
                +(ev.pos.y - self.prev.y)
            else
                -(ev.pos.x - self.prev.x)
            end
            updatePos(fine*delta/dragScale)
            self.prev = ev.pos
        end
    ;end
  def onMouseEnter(ev);
        self.root.log(:tooltip, self.tooltip)
    ;end
  def lim(x, low, high);
        [low, [x, high].min].max
    ;end
  def updatePosAbs(tmp);
        nvalue = lim(tmp, 0.0, 1.0)
        if(valuator.valueRef)
            old_dsp = valuator.valueRef.display_value
            valuator.valueRef.value = nvalue
            valuator.value = nvalue
            new_dsp = valuator.valueRef.display_value
            whenValue.call if whenValue && (new_dsp.nil? || old_dsp != new_dsp)
            out_value = valuator.valueRef.display_value
            if(valuator.units && valuator.units != "none")
                out_value = out_value.round(2) if out_value.class == Float
                out_value = out_value.to_s
                out_value += " "
                out_value += valuator.units
            end
            valuator.root.log(:user_value, out_value, src=valuator.label)
        else
            valuator.value = nvalue
            whenValue.call if whenValue
        end
        damage_self
    ;end
  def updatePos(delta);
        updatePosAbs(valuator.value - delta)
    ;end
  def onMerge(val);
        return if self.class != val.class
        valuator.value = val.value if(val.respond_to?(:value))
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @valuator = self

    
context = Hash.new
context["valuator"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "active")
Qml::prop_add(self, "type")
Qml::prop_add(self, "units")
self.valueRef = nil
self.prev = nil
self.value = 1.0
self.dragScale = 200.0
self.vertical = true
self.whenValue = nil
self.active = true
self.type = nil
self.units = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, valuator.extern)
        valuator.label   = meta.short_name
        valuator.tooltip = meta.tooltip
        valuator.units   = meta.units

        valuator.valueRef = OSC::RemoteParam.new($remote, valuator.extern)
        valuator.valueRef.set_min(meta.min)
        valuator.valueRef.set_max(meta.max)
        valuator.valueRef.set_scale(meta.scale)
        valuator.valueRef.type     = "f" if valuator.type
        valuator.valueRef.callback = Proc.new {|x| valuator.setValue(x)}

    
end})

    #puts "Valuator, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Knob < Qml::Valuator
  attr_reader :knob
  def draw_classic(vg);
        pi = 3.14159
        start = pi/4;
        end_   = pi*3.0/4.0;
        inner  = 0.2*[h,w].min
        outer = 0.4*[h,w].min
        cy = h/2;
        cx = w/2;
        kr = h/4;
        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color color("114575")
            v.fill
        end

        len = (3.0/2.0*pi)*knob.value;
        startt = end_ + len;

        vg.path do |v|
            v.arc(cx, cy, inner, end_, startt, 2);
            v.arc(cx, cy, outer, startt, end_, 1);
            v.close_path
            v.fill_color color("3AC5EC")
            v.fill
        end
    ;end
  def strike_through(vg);
        strike_color = Theme::TextColor
        vg.path do
            vg.move_to(w*0, h*0)
            vg.line_to(w*(1-2*0), h*(1-2*0))
            vg.stroke_width 1.0
            vg.stroke_color strike_color
            vg.stroke
        end
    ;end
  def draw(vg);
        #background color("123456")
        pi = 3.14159
        start = pi/4;
        end_   = pi*3.0/4.0;
        radius  = 0.25*[h*1.0,w].min
        inner  = 0.35*[h*1.0,w].min
        outer = 0.48*[h*1.0,w].min

        #lowest point is 0.707 cy-outer*0.707
        #shift by 0.293/2 to compinsate
        cy = h/2 + h*0.293/4;
        cx = w/2;
        kr = h/4;
        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color color("114575")
            v.fill
        end

        len = (3.0/2.0*pi)*knob.value;
        startt = end_ + len;

        vg.path do |v|
            v.arc(cx, cy, inner, end_, startt, 2);
            v.arc(cx, cy, outer, startt, end_, 1);
            v.close_path
            v.fill_color color("3AC5EC")
            v.fill
            v.stroke
        end

        knob_dash = color("E7E5E2")

        vg.path do |v|
            v.circle(cx, cy, radius)
            paint = v.linear_gradient(0,0,0,h,
            Theme::KnobGrad1, Theme::KnobGrad2)
            v.fill_paint paint
            v.fill
            v.stroke
        end

        vg.path do |v|
            angle = 3*pi/4+3*pi/2*knob.value
            v.move_to(cx, cy)
            v.line_to(cx+radius*Math.cos(angle), cy+radius*Math.sin(angle))
            v.stroke_color knob_dash
            v.stroke_width 0.2*radius
            v.stroke
        end
        
        strike_through(vg) if !self.active
        
    ;end
  def class_name();
        "Knob"
    ;end
  def layout(l, selfBox);
        l.aspect(selfBox, 0.9, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @knob = self

    
context = Hash.new
context["knob"] = self

    #puts "Knob, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::LogWidget < Qml::Widget
  attr_property :lines
  def display_log(type, message, src);
        if(self.label != message)
            self.label = message
            damage_self
        end
    ;end
  def onSetup(old);
        self.root.log_widget = self
    ;end
  def draw(vg);
        textColor  = color("3ac5ec")
        splitColor = color("133A4C")

        vg.path do |vg|
            vg.move_to(0,0.5*h)
            vg.line_to(w,0.5*h)
            vg.stroke_width 2.0
            vg.stroke_color splitColor
            vg.stroke
        end

        vg.font_face("bold")
        vg.font_size h/self.lines*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color(textColor)
        vg.text_box(0,h/4,w,self.label.upcase)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous13 = self

    
context = Hash.new
Qml::prop_add(self, "lines")
self.lines = 2

    #puts "LogWidget, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ModWheel < Qml::Valuator
  def draw(vg);
        c1 = color("0b0b0b")
        c2 = color("525252")
        bright_green = color("00AE9C")
        grad1 = vg.linear_gradient(0,0,0,h/2,c1,c2)
        grad2 = vg.linear_gradient(0,h/2,0,h,c2,c1)
        vg.path do |v|
            v.rect(0, 0, w, h/2)
            vg.fill_paint(grad1)
            v.fill
        end
        vg.path do |v|
            v.rect(0, h/2, w, h/2)
            vg.fill_paint(grad2)
            v.fill
        end
        vg.path do |v|
            v.rect(0, h*7/8*(1-value), w, h/8)
            scale = 2*(0.5-[0.5-value, value-0.5].max)
            scale = scale+0.4
            bright_green = NVG.rgba(0x00*scale,0xAe*scale,0x9c*scale, 255)
            v.fill_color(bright_green)
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous14 = self

    
context = Hash.new

    #puts "ModWheel, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::NumEntry < Qml::Widget
  attr_reader :numentry
  attr_accessor :whenValue
  attr_property :value
  attr_accessor :valueRef
  attr_property :format
  attr_property :offset
  attr_property :minimum
  attr_property :maximum
  attr_property :active
  def class_name();"num_entry";end
  def layout(l, selfBox);
        return selfBox if layoutOpts.include?(:free)

        #Assume all digit bounding boxes are roughly the same
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, "- 9999 +"+format)
        l.aspect(selfBox, bb, scale)

        selfBox
    ;end
  def refresh();
        self.valueRef.refresh if self.valueRef
    ;end
  def draw(vg);
        textHeight = 0.8
        vg.path do
            vg.rect(0, 0, w, h)
            paint = vg.linear_gradient(0,0,0,h,
                Theme::ButtonGrad1, Theme::ButtonGrad2)
            vg.fill_paint paint
            vg.fill
            vg.stroke_width 1
            vg.stroke
        end

        vg.path do
            vg.move_to(0.2*w, 0*h)
            vg.line_to(0.2*w, 1*h)
            vg.move_to(0.8*w, 0*h)
            vg.line_to(0.8*w, 1*h)
            vg.stroke
        end

        vg.font_face("bold")
        vg.font_size h*textHeight
        vg.fill_color Theme::TextColor
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.text(0.1*w,h/2,"-")
        vg.text(0.9*w,h/2,"+")
        vg.text(0.5*w,h/2,self.format+self.value.to_s)

        if(!self.active)
            pad = 1
            vg.path do
                vg.move_to(pad, pad)
                vg.line_to(w-2*pad, h-2*pad)
                vg.stroke_color Theme::TextColor
                vg.stroke
            end
        end
    ;end
  def setValue(val);
        self.value = val + offset
        damage_self
    ;end
  def updatePos(delta);
        return if !self.active
        if self.value + delta > maximum
            self.value = maximum
            return
        end
        if self.value + delta < minimum
            self.value = minimum
            return
        end
        self.value += delta
        if(self.valueRef)
            self.valueRef.value = self.value - offset
        end
        whenValue.call if whenValue
        damage_self
    ;end
  def onScroll(ev);
        updatePos(+1) if ev.dy > 0
        updatePos(-1) if ev.dy < 0
    ;end
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def onMousePress(ev);
        return if !ev.buttons.include? :leftButton
        if(ev.pos.x-global_x > 0.5*w)
            updatePos(+1)
        else
            updatePos(-1)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @numentry = self

    
context = Hash.new
context["numentry"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "format")
Qml::prop_add(self, "offset")
Qml::prop_add(self, "minimum")
Qml::prop_add(self, "maximum")
Qml::prop_add(self, "active")
self.whenValue = nil
self.value = 0
self.valueRef = nil
self.format = ""
self.offset = 0
self.minimum = 0
self.maximum = 10
self.active = true
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, numentry.extern)
        numentry.label   = meta.short_name
        numentry.tooltip = meta.tooltip
        numentry.minimum = meta.min.to_i if numentry.minimum == 0
        numentry.maximum = meta.max.to_i if numentry.maximum == 10

        numentry.valueRef = OSC::RemoteParam.new($remote, numentry.extern)
        numentry.valueRef.mode     = :full
        numentry.valueRef.callback = Proc.new {|x| numentry.setValue(x)}

    
end})

    #puts "NumEntry, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TriggerButton < Qml::Button
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        whenValue.call if whenValue
        damage_self
    ;end
  def animate();
        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.9
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous40 = self

    
context = Hash.new

    #puts "TriggerButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::PanicButton < Qml::TriggerButton
  def layout(l, selfBox);
        selfBox
    ;end
  def draw(vg);
        text_color1   = color("505050")
        text_color2   = color("B9CADE")
        pad = 1/64
        vg.path do |v|
            v.rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad))
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end
        vg.path do |v|
            angle = 30*3.14/180
            delta = 0.4*[w,h].min
            dy = delta*Math.sin(angle)
            dx = delta*Math.cos(angle)
            cy = 0.6*h
            v.move_to(0.5*w,    cy-delta)
            v.line_to(0.5*w+dx, cy+dy)
            v.line_to(0.5*w-dx, cy+dy)
            v.close_path
            v.fill_color(text_color2)
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*0.75
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.fill_color(text_color1)
        vg.text(w/2,h/2,"!")

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous16 = self

    
context = Hash.new
self.whenValue = lambda { $remote.action("/Panic") }

    #puts "PanicButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ParModule < Qml::Widget
  attr_reader :mod
  attr_reader :titleW
  attr_reader :title
  attr_reader :button_c
  attr_reader :button_p
  attr_reader :content
  attr_reader :labels
  attr_accessor :copyable
  attr_accessor :editable
  attr_property :extern
  def class_name(); "ParModule" ;end
  def onSetup(old=nil);
        n = content.children.length
        (0...n).each do |i|
            label = createInstance("Text", labels, mod.db)
            if(content.children[i].class_name != "Button")
                label.label = content.children[i].label
            end
        end
    ;end
  def layout(l);
        selfBox = l.genBox :parmodule, mod
        contentBox = content.layout(l)
        labelsBox  = labels.layout(l)
        titleBox   = titleW.layout(l)

        #Constrain Content Box
        l.contains(selfBox,titleBox)
        l.contains(selfBox,contentBox)
        l.contains(selfBox,labelsBox)
        l.topOf(titleBox,contentBox)
        l.topOf(contentBox,labelsBox)
        l.sheq([contentBox.h, selfBox.h], [1, -0.5], 0)
        l.sheq([contentBox.w, selfBox.w], [1, -1], 0)
        l.sheq([labelsBox.h, selfBox.h],  [1, -0.3], 0)
        l.sheq([labelsBox.w, selfBox.w],  [1, -1], 0)

        l.contains(selfBox, titleBox)
        l.sheq([titleBox.x],[1],0)
        l.sheq([titleBox.h, selfBox.h], [1, -0.2], 0)


        prev = nil
        blist = []
        mod.content.children.each do |ch|
            bb = ch.layout(l)
            if(bb)
                blist << bb
                l.contains(contentBox, bb)
                #l.le([bb.y, selfBox.h], [-1, 0.2])
                if(prev)
                    l.rightOf(prev,bb, true)
                end
                prev = bb
            end
        end

        #Punish Different Sizes
        blist.each do |x|
            l.punish2([selfBox.w], [1.0/blist.length], x.w)
        end
        
        #Punish floating in the y dir
        blist.each do |x|
            l.weak(x.y)
        end

        #Center Stuff Horizontally
        if(!blist.empty?)
            l.sheq([blist[0].x, selfBox.w, blist[-1].x, blist[-1].w],
            [1,       -1,       1,     1], 0)
        end
        breal = blist

        content_items = blist

        prev = nil
        blist = []
        lh    = l.gensym :modlabelHeight
        mod.labels.children.each do |ch|
            bb = ch.layout(l)
            if(bb)
                blist << bb
                l.contains(labelsBox, bb)
                l.sheq([lh, bb.h], [1, -1], 0)
                if(prev)
                    l.rightOf(prev,bb, true)
                end
                prev = bb
            end
        end

        #Punish Different Sizes
        blist.each do |x|
            l.punish2([selfBox.w], [1.0/blist.length], x.w)
        end

        #Center Stuff Horizontally
        if(!blist.empty?)
            l.sheq([blist[0].x, selfBox.w, blist[-1].x, blist[-1].w],
            [1,       -1,       1,     1], 0)
        end

        #Stitch labels
        n = blist.length
        (1...n).each do |i|
            l.rightOf(breal[i-1],blist[i])
        end

        label_items = blist

        #Center labels
        if(label_items.length == content_items.length && true)
            (1...n).each do |i|
                li = label_items[i]
                ci = content_items[i]
                l.sheq([li.x, li.w, ci.x, ci.w],
                       [1, 0.5, -1, -0.5], 0)
            end
        end



        #Layout Custom Stuff
        #buttonBox = button.layout(l)
        #l.contains(titleBox,buttonBox)

        #blist.each_with_index do |x, i|
        #    blist.each_with_index do |y, j|
        #        if(i!=j)
        #            l.punish_difference(x.w,y.w)
        #        end
        #    end
        #end
        selfBox
    ;end
  def draw(vg);
        #//paint the top half
        pos = [0, 0, mod.w, mod.h]
        pos[3] *= 0.2;
        vg.path do |v|
            border(pos[3]*0.01, pos);
            vg.rect(*pos);
            vg.fill_color(color("0A2E4C"))
            vg.fill
        end

        #upperspace = [pos[0]+pos[2]*2.0f/3.0f, pos[1], pos[2]/3.0f, pos[3]];
        #//upper.draw(upperspace);

        pos2 = [*pos]
        pos2[2] /= 3;
        pad(0.9, pos2);
        #drawLeftLabel(vg, str, SPLAT(pos2));



        innerspace = [0, mod.h*0.2, mod.w, mod.h*0.8];
        border(h*0.01, innerspace);
        #//paint the inner panel
        vg.path do |v|
            v.rect(*innerspace)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ModuleGrad1, Theme::ModuleGrad2)
            v.fill_paint paint
            v.fill
        end

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @mod = self
@titleW = Qml::Widget.new(database, ui_path+"1"+'/')
@title = Qml::Text.new(database, ui_path+"2"+'/')
@button_c = Qml::Button.new(database, ui_path+"3"+'/')
@button_p = Qml::Button.new(database, ui_path+"4"+'/')
@content = Qml::Widget.new(database, ui_path+"5"+'/')
@labels = Qml::Widget.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["mod"] = self
context["titleW"] = @titleW
@titleW.instance_eval do
def titleW=(k);@titleW=k;end
def titleW;    @titleW;  end
def title=(k);@title=k;end
def title;    @title;  end
def button_c=(k);@button_c=k;end
def button_c;    @button_c;  end
def button_p=(k);@button_p=k;end
def button_p;    @button_p;  end
end
context["title"] = @title
context["button_c"] = @button_c
@button_c.instance_eval do
def button_c=(k);@button_c=k;end
def button_c;    @button_c;  end
end
context["button_p"] = @button_p
@button_p.instance_eval do
def button_p=(k);@button_p=k;end
def button_p;    @button_p;  end
end
context["content"] = @content
context["labels"] = @labels
Qml::context_apply(context, @titleW)
Qml::context_apply(context, @button_c)
Qml::context_apply(context, @button_p)
Qml::prop_add(self, "extern")

        @titleW.instance_eval("def class_name(); \"TitleBox\" ;end", "src/mruby-zest/qml/ParModule.qml", 193)

        @titleW.instance_eval("def layout(l);\n            selfBox = l.genBox :titleBox, titleW\n            t  = title.layout(l)\n            l.contains(selfBox,t)\n\n            bc = button_c.layout(l)\n            l.contains(selfBox,bc)\n\n            bp = button_p.layout(l)\n            l.contains(selfBox,bp)\n            l.rightOf(t,bc)\n            l.rightOf(bc,bp)\n            l.weak(bc.x)\n            selfBox\n        ;end", "src/mruby-zest/qml/ParModule.qml", 195)
@db.connect_property(@title.properties["label"], lambda {mod.label})
Qml::add_child(@titleW, @title)

        @button_c.instance_eval("def layout(l);\n                selfBox = l.genBox :copyButton, button_c\n                l.aspect(selfBox, 1, 1)\n                selfBox\n            ;end", "src/mruby-zest/qml/ParModule.qml", 219)
@button_c.label = "C"
Qml::add_child(@titleW, @button_c)

        @button_p.instance_eval("def layout(l);\n                selfBox = l.genBox :copyButton, button_p\n                l.aspect(selfBox, 1, 1)\n                selfBox\n            ;end", "src/mruby-zest/qml/ParModule.qml", 229)
@button_p.label = "P"
Qml::add_child(@titleW, @button_p)
Qml::add_child(self, @titleW)

        @content.instance_eval("def class_name(); \"ContentBox\" ;end", "src/mruby-zest/qml/ParModule.qml", 241)
Qml::add_child(self, @content)

        @labels.instance_eval("def class_name(); \"LabelBox\" ;end", "src/mruby-zest/qml/ParModule.qml", 245)
Qml::add_child(self, @labels)
self.copyable = false
self.editable = false
self.extern = ""
self.layoutOpts = [:horizontal]
@db.connect_watcher(self.properties["children"], lambda {begin

        mch = mod.children
        cch = content.children
        if(mch.length > 3)
            cch = cch+mch[3..-1]
            mch = mch[0..2]
            content.children = cch
            mod.children     = mch
        else
        end
    
end})

    #puts "ParModule, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ParModuleRow < Qml::Widget
  attr_reader :row
  attr_reader :content
  attr_reader :labels
  attr_property :lsize
  attr_property :whitespace
  attr_property :outer
  def onSetup(old=nil);
        mch = row.children
        cch = content.children
        if(mch.length > 2)
            cch = cch+mch[2..-1]
            mch = mch[0..1]
            content.children = cch
            row.children     = mch
        else
        end

        n = content.children.length
        labels.children = []
        (0...n).each do |i|
            label = createInstance("Text", labels, row.db)
            label.height = 1.0
            label.layoutOpts = [:ignoreAspect]
            if(content.children[i].class_name != "Button")
                label.label = content.children[i].label
            end
        end
    ;end
  def layout_hpack(l, selfBox, children, mode, breal=nil);
        #Create A list of boxes
        blist = []
        return blist if children.empty?
        weight_list = []
        bbs = []
        begin
            prev = nil
            n = children.length
            children.each_with_index do |ch, i|
                box = l.genConstBox(selfBox.w*i/n, 0,
                selfBox.w/n, selfBox.h*1, ch)
                bb = ch.layout(l, box)
                bbs << bb
                if(ch.layoutOpts.include? :weight)
                    weight_list << ch.layoutOpts[:weight]
                else
                    weight_list << 1.0
                end
            end
        end


        heights = Hash.new
        bbs.each do |b|
            tp = b.info.class == Qml::Knob ? :knob : :other
            heights[tp]  ||= b.h
            heights[tp]    = b.h if b.h < heights[tp]
        end

        bbs.each do |b|
            tp = b.info.class == Qml::Knob ? :knob : :other
            b.y += (b.h-heights[tp])/2
            b.h  = heights[tp]
        end

        content_items = blist
    ;end
  def layout(l, selfBox);
        contentBox = content.fixed(l, selfBox, 0, 0, 1, (1-lsize))
        labelBox   = labels.fixed(l,  selfBox, 0, (1-lsize), 1, lsize)

        content_items = layout_hpack(l, contentBox, row.content.children, :normal)
        label_items   = layout_hpack(l, labelBox,   row.labels.children,  :labels,
                                     content_items)


        if(label_items.length == content_items.length)
            n = label_items.length
            (0...n).each do |i|
                li = label_items[i]
                ci = content_items[i]
                l.sheq([li.x, li.w, ci.x, ci.w],
                       [1, 0.5, -1, -0.5], 0)
            end
        end
        selfBox
    ;end
  def class_name(); "ParModuleRow" ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @row = self
@content = Qml::Widget.new(database, ui_path+"1"+'/')
@labels = Qml::Widget.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["row"] = self
context["content"] = @content
context["labels"] = @labels
Qml::prop_add(self, "lsize")
Qml::prop_add(self, "whitespace")
Qml::prop_add(self, "outer")

        @content.instance_eval("def class_name(); \"ContentBox\" ;end", "src/mruby-zest/qml/ParModuleRow.qml", 97)
Qml::add_child(self, @content)

        @labels.instance_eval("def class_name(); \"LabelBox\" ;end", "src/mruby-zest/qml/ParModuleRow.qml", 101)
Qml::add_child(self, @labels)
self.lsize = 0.3
self.whitespace = nil
self.outer = nil
self.layoutOpts = [:fixed_width]

    #puts "ParModuleRow, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ToggleButton < Qml::Button
  attr_reader :toggle
  attr_accessor :valueRef
  def setValue(x);
        self.value = x
        whenValue.call if whenValue
        damage_self
    ;end
  def refresh();
        self.valueRef.refresh if self.valueRef
    ;end
  def onMousePress(ev);
        if(root.learn_mode && extern)
            $remote.automate(extern)
            root.log(:tooltip, "Learning #{extern}")
            return
        end
        self.value = !self.value
        self.valueRef.value = self.value if self.valueRef
        damage_self
        whenValue.call if whenValue
    ;end
  def onMouseEnter(ev);
        if(self.tooltip.nil? || self.tooltip.empty?)
            self.root.log(:tooltip, "extern = <" + self.extern + ">")
        else
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @toggle = self

    
context = Hash.new
context["toggle"] = self
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, toggle.extern)
        toggle.label   = meta.short_name if toggle.label.empty?
        toggle.tooltip = meta.tooltip
        toggle.valueRef.clean if toggle.valueRef
        toggle.valueRef = OSC::RemoteParam.new($remote, toggle.extern)
        toggle.valueRef.callback = lambda {|x| toggle.setValue(x)}
    
end})

    #puts "ToggleButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::PowButton < Qml::ToggleButton
  def class_name();"powbutton";end
  def draw(vg);
        #Draw background
        grad = if(self.value)
            bright_green = color("00Ae9c")
            dark_green   = color("007368")
            vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        else
            grey1 = Theme::ButtonGrad1
            grey2 = Theme::ButtonGrad2
            vg.linear_gradient(0,0,0,h, grey1, grey2)
        end
        vg.stroke_color color("000000", 0xa0)

        grey1 = Theme::ButtonGrad1
        grey2 = Theme::ButtonGrad2
        grad2 = vg.linear_gradient(0,0,0,h, grey1, grey2)
        vg.path do |v|
            pad = 0
            v.rounded_rect(pad,0,w,h,3)
            if(self.value)
                v.fill_paint grad
            else
                v.fill_paint grad2
            end
            v.fill
            v.stroke
        end




        #Draw power button
        pi     = 3.141592653
        cx     = w/2.0;
        cy     = h/2.0;
        center = pi/2.0;
        sw     = [h,w].min/15.0;
        r      = 0.18*[h,w].min
        super_green = color("72fced")
        vg.line_cap(NVG::SQUARE);
        vg.path do |vg|
            vg.arc(cx, cy, r, center+pi*0.75, center-pi*0.75, 1);
            vg.move_to(cx, 0.45*h);
            vg.line_to(cx, 0.30*h);
            vg.stroke_width(w/20.0);
            vg.stroke_color(super_green)
            vg.stroke
        end
    ;end
  def layout(l, selfBox);
        l.aspect(selfBox, 1, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous27 = self

    
context = Hash.new

    #puts "PowButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::RadialMenu < Qml::Widget
  attr_reader :radial
  attr_property :fields
  attr_property :flavor
  attr_property :callback
  def draw(vg);
        inner  = 0.1*[h,w].min
        outer = 0.5*[h,w].min
        cx = radial.w/2
        cy = radial.h/2
        pi = 3.14159
        start = pi/4;
        end_   = pi*3.0/4.0;
        start = 0
        end_  = -2.0*pi

        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color color("114575",205)
            v.fill
        end

        if(self.flavor)
            start = self.flavor*2*pi/4 - pi/4
            end_  = start-pi/2
            vg.path do |v|
                v.arc(cx, cy, outer, start, end_, 1);
                v.arc(cx, cy, inner, end_, start, 2);
                v.close_path
                v.fill_color color("2185F5", 205)
                v.fill
            end
        end

        vg.path do |v|
            outer = 0.5*0.707*[h,w].min
            v.move_to(cx-outer,cy-outer)
            v.line_to(cx+outer,cy+outer)
            v.move_to(cx+outer,cy-outer)
            v.line_to(cx-outer,cy+outer)
            v.stroke_color color("ffffff")
            v.stroke
        end

        textColor = color("3AC5EC")

        #Draw North
        vg.font_face("bold")
        vg.font_size h/8
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.fill_color(textColor)
        vg.text(w*0.5,h*0.2,fields[0].upcase)

        #Draw East
        vg.text(w*0.8,h*0.5,fields[1].upcase)

        #Draw South
        vg.text(w*0.5,h*0.8,fields[2].upcase)

        #Draw West
        vg.text(w*0.2,h*0.5,fields[3].upcase)
    ;end
  def abs(x);
        if(x < 0)
            return -x
        else
            return x
        end
    ;end
  def onMouseMove(ev);
        dx = ev.pos.x-self.global_x-radial.w/2
        dy = ev.pos.y-self.global_y-radial.h/2
        slice = find_slice(dx, dy)
        if(slice != self.flavor)
            self.flavor = slice
            self.root.damage_item(self) if(self.root)
        end
    ;end
  def onMouseRelease(ev);
        onMousePress(ev)
    ;end
  def find_slice(dx, dy);
        dist     = (dx**2 + dy**2)**0.5
        min_dist = 0.05*[self.w, self.h].min
        return nil if dist < min_dist
        if(abs(dx) > abs(dy))
            #east vs west
            if(dx > 0)
                return 1
            else
                return 3
            end
        else
            #north vs south
            if(dy < 0)
                return 0
            else
                return 2
            end
        end
    ;end
  def onMousePress(ev);
        dx = ev.pos.x-radial.w/2
        dy = ev.pos.y-radial.h/2
        slice = find_slice(dx, dy)
        callback.call(self.flavor) if callback
        rt = radial.root
        rt.ego_death radial
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @radial = self

    
context = Hash.new
context["radial"] = self
Qml::prop_add(self, "fields")
Qml::prop_add(self, "flavor")
Qml::prop_add(self, "callback")
@db.connect_property(self.properties["fields"], lambda {["Learn", "Rand", "Unlearn", "default"]})
self.flavor = nil
self.callback = nil

    #puts "RadialMenu, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Selector < Qml::Widget
  attr_reader :selector
  attr_accessor :valueRef
  attr_property :extern
  attr_accessor :options
  attr_accessor :opt_vals
  attr_accessor :selected
  attr_property :active
  attr_accessor :whenValue
  attr_property :max_text
  def refresh();
        self.valueRef.refresh if self.valueRef
    ;end
  def set_value(x);
        sel = 0
        opt_vals.each_with_index do |i, ind|
            sel = ind if i == x
        end

        self.selected = sel
        rt = self.root
        rt.damage_item self if rt
        whenValue.call if whenValue
    ;end
  def aspect2(box, ww, hh);
        provided    = box.w/box.h
        recommended = ww/hh
        if(recommended > provided)
            #Decrease height
            new_height = box.h*provided/recommended
            box.y += (box.h-new_height)/2
            box.h = new_height
        else
            new_width = box.w*recommended/provided
            box.x += (box.w-new_width)/2
            box.w  = new_width
        end
    ;end
  def layout(l, selfBox);
        if(!layoutOpts.include?(:no_constraint))
            scale = 100
            $vg.font_size scale
            bb = 1
            long_mode = layoutOpts.include?(:long_mode)
            selector.options.each do |x|
                x = x[0..self.max_text] if x.length > self.max_text && !long_mode
                bbl  = $vg.text_bounds(0, 0, (x+"    ").upcase)
                bb   = [bb, bbl].max
            end
            scale *= layoutOpts[0] if layoutOpts.include?(:rescale)
            if(bb != 0)
                #Width cannot be so small that letters overflow
                self.aspect2(selfBox, bb, scale)
            end
        end
        selfBox
    ;end
  def class_name();
        "Selector"
    ;end
  def set_value_user(val);
        if(val.nil?)
            damage_self
            return
        end
        return if(val == self.selected)
        self.selected = val
        if(self.valueRef && self.selected <= opt_vals.length)
            self.valueRef.value = self.opt_vals[self.selected]
        end
        if(self.valueRef && self.root)
            self.root.log(:user_value, self.valueRef.display_value, src=self.label)
        end
        whenValue.call if whenValue

        damage_self
    ;end
  def onMousePress(ev);
        create_menu if self.active && options.length != 1
    ;end
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def draw_strike(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do
            vg.move_to(w*pad, h*pad)
            vg.line_to(w*pad2, h*pad2)
            vg.stroke_color Theme::TextColor
            vg.stroke_width 1
            vg.stroke
        end
    ;end
  def draw_arrow(vg);
        pad  = 1/64
        pad2 = (1-2*pad)

        vg.path do |v|
            tx = w*pad2-h*pad2
            tw = h*pad2
            ty = h*pad
            th = h*pad2

            v.move_to(tx+0.25*tw, ty+0.3*th)
            v.line_to(tx+0.50*tw, ty+0.7*th)
            v.line_to(tx+0.75*tw, ty+0.3*th)
            v.close_path

            v.fill_color Theme::TextColor
            v.fill
        end
    ;end
  def draw_button(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 2)
            paint = v.linear_gradient(0,0,0,h, Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            hh = h/20
            v.move_to(w*pad+2,       h*pad+hh)
            v.line_to(w*(1-2*pad)+1, h*pad+hh)
            v.stroke_color color("5c5c5c")
            v.stroke_width hh
            v.stroke
        end
    ;end
  def draw(vg);
        draw_button(vg)

        draw_arrow(vg) if options.length != 1

        draw_strike(vg) if !self.active

        return if options[selected].nil?

        pad  = 1/64
        text_color = Theme::TextColor

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color text_color

        long_mode = layoutOpts.include?(:long_mode)
        str = options[selected].upcase
        str = str[0..self.max_text] if str.length > self.max_text && !long_mode

        vg.text(3+w*pad*2,h/2, str)
    ;end
  def create_menu();
        n = options.length
        return if n < 1
        widget = DropDown.new(self.db)
        widget.w = self.w
        widget.h = self.h*n
        widget.x = 0
        widget.y = 0
        widget.layer = 2
        widget.options = options
        widget.callback = lambda { |v|
            set_value_user(v)
        }
        #Check if it's too big
        if(widget.h+global_y > window.h)
            if(global_y - self.h*(n-1) < 0)
                if(global_y > window.h*0.5)#upper
                    widget.h = (global_y+self.h)
                    widget.y = -global_y
                else#lower
                    widget.h = (window.h-global_y)
                end
            else
                widget.y = -self.h*(n-1)
            end
        end
        widget.prime(root())

        Qml::add_child(self, widget)
        root.smash_draw_seq
        root.damage_item widget
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @selector = self

    
context = Hash.new
context["selector"] = self
Qml::prop_add(self, "extern")
Qml::prop_add(self, "active")
Qml::prop_add(self, "max_text")
self.valueRef = nil
self.extern = ""
self.options = ["text", "test", "asdf"]
self.opt_vals = []
self.selected = 0
self.active = true
self.whenValue = nil
self.max_text = 8
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote,
                                       selector.extern)
        selector.label   = meta.short_name
        selector.tooltip = meta.tooltip
        if(meta.options)
            nopts = []
            vals  = []
            meta.options.each do |x|
                vals  << x[0]
                nopts << x[1]
            end
            selector.options = nopts
            selector.opt_vals = vals
        end
        selector.valueRef = OSC::RemoteParam.new($remote, selector.extern)
        selector.valueRef.mode = :options
        selector.valueRef.callback = Proc.new {|x| selector.set_value(x)}
    
end})

    #puts "Selector, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Slider < Qml::Valuator
  attr_reader :slider
  attr_accessor :pad
  attr_property :centered
  attr_property :visual
  def class_name();
        "Slider"
    ;end
  def draw_centered(vg);
        pad2 = (1-2*pad)
        if(value > 0.5)
            src = (h/2-h*pad2*(value-0.5))
            dst = h/2
            vg.path do |v|
                v.rect(pad*w, src, pad2*w, (src-dst).abs)
                v.fill_color Theme::SliderActive
                v.fill
            end
        else
            vg.path do |v|
                v.rect(pad*w, h/2, pad2*w, pad2*h*(0.5-value))
                v.fill_color Theme::SliderActive
                v.fill
            end
        end
    ;end
  def draw_normal(vg);
        col  = Theme::SliderActive
        col  = Theme::SliderVisActive if visual
        pad2 = (1-2*pad)
        src  = (h-h*pad2*value)
        dst  = h
        vg.path do |v|
            v.rect(pad*w, src, pad2*w, (src-dst).abs)
            v.fill_color col
            v.fill
        end
    ;end
  def draw(vg);
        self.dragScale = h
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(pad*w, pad*h, pad2*w, pad2*h)
            v.fill_color Theme::SliderBackground
            v.fill
        end

        vg.path do |v|
            v.move_to(pad* w, pad* h)
            v.line_to(pad* w, pad2*h)
            v.move_to(pad2*w, pad* h)
            v.line_to(pad2*w, pad2*h)
            v.stroke_color color(:black)
            v.stroke
        end

        return if value.class != Float

        draw_centered(vg) if  centered
        draw_normal(vg)   if !centered

        
        vg.path do |v|
            yloc = (h-h*pad2*value)
            v.move_to(w*pad,  yloc)
            v.line_to(w*pad2, yloc)
            v.stroke_color Theme::SliderStroke
            v.stroke_width 2.0
            v.stroke
        end
    ;end
  def onMousePress(ev);
        mouse_handle(ev) if !visual
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @slider = self

    
context = Hash.new
context["slider"] = self
Qml::prop_add(self, "centered")
Qml::prop_add(self, "visual")
self.pad = 0.1
self.centered = true
self.visual = false

    #puts "Slider, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Swappable < Qml::Widget
  attr_reader :swappable
  attr_property :content
  attr_accessor :whenSwapped
  def rec_del_props(widget, plist=[]);
        return plist if widget.nil?
        #Add all properties to the delete list
        widget.properties.each do |k,v|
            plist << v
        end
        #Add all children properties to the list
        widget.children.each do |ch|
            rec_del_props(ch, plist)
        end
        plist
    ;end
  def rec_clean_cbs(widget);
        if(widget.respond_to? :valueRef)
            v = widget.valueRef
            if(v.class == Array)
                v.each do |vv|
                    vv.clean
                end
            else
                v.clean if v.respond_to? :clean
            end
        end
        widget.children.each do |ch|
            rec_clean_cbs(ch)
        end
    ;end
  def dump_plist();
        plist = swappable.db.instance_eval{@plist}
        plist.each_with_index do |x, idx|
            if(idx<50)
                print idx
                print " "
                puts x.to_s
            end
        end
        puts plist.length
    ;end
  def remove_old_child();
        #it should only be possible that there is a single child
        if(!self.children.empty?)
            rec_clean_cbs(children[0])
            del_list = rec_del_props(swappable.children[0])
            @db.remove_properties del_list

            root.damage_item(children[0], :all) if root
            self.children = []
        end
    ;end
  def setup_widget(w);
        w.onSetup if w.respond_to?(:onSetup)
        w.children.each do |c|
            setup_widget c
        end
    ;end
  def onSetup(old=nil);
        create_new_child
    ;end
  def create_new_child();
        return if self.content.nil?
        return if !self.children.empty?

        t1=Time.new
        widget = swappable.content.new(swappable.db)
        t2=Time.new
        widget.x = 0
        widget.y = 0
        widget.w = self.w
        widget.h = self.h
        widget.extern = self.extern if !self.extern.empty?
        Qml::add_child(self, widget)
        t3=Time.new
        self.db.update_values
        t4=Time.new
        setup_widget widget
        self.db.update_values
        t5=Time.new

        if(root)
            root.smash_layout
            root.damage_item widget
        end
        self.whenSwapped.call if self.whenSwapped
        [t2-t1, t4-t3, t5-t4]
    ;end
  def onMerge(val);
        self.content = val.content if(val.respond_to?(:content))
    ;end
  def force_update();
        cls = "nil"
        cls = children[0].class if !children.empty?
        ext = ""
        ext = children[0].extern if !children.empty?
        srt = Time.new
        swappable.remove_old_child
        #mid = Time.new
        d2 = swappable.create_new_child
        dne = Time.new
        tot = dne-srt
        #scl = 100/tot
        #puts("[INFO] Content chagned to #{swappable.content} from #{cls} in #{1000*tot}ms")
        #puts("[INFO] <#{ext}> => <#{extern}>") if ext != extern
        #puts("[INFO] Content swap took #{1000*tot}ms (#{(scl*(mid-srt)).to_i}% remove) (#{(scl*(dne-mid)).to_i}% add)")
        #puts("[INFO]                                 (#{(scl*d2[0]).to_i}% init) (#{(scl*d2[1]).to_i}% update) (#{(scl*d2[2]).to_i}% setup)")
    ;end
  def layout(l,selfBox);
        if(selfBox.nil?)
            t = self.class_name.to_sym
            selfBox = l.genBox t, widget
        end

        return if children.empty?

        child = children[0]
        if(selfBox.class == LayoutConstBox)
            box = l.genConstBox(0, 0, selfBox.w, selfBox.h,
                        child)
            child.layout(l, box)
        else
            box = child.layout(l, nil)
            l.contains(selfBox, box)
        end

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @swappable = self

    
context = Hash.new
context["swappable"] = self
Qml::prop_add(self, "content")
self.content = nil
self.whenSwapped = nil
@db.connect_watcher(self.properties["content"], lambda {begin

        return if(!swappable.children.empty? &&
            swappable.children[0].class == swappable.content &&
            (swappable.extern.empty? || swappable.extern == swappable.children[0].extern))
        swappable.force_update
    
end})

    #puts "Swappable, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TabButton < Qml::Widget
  attr_reader :button
  attr_accessor :whenClick
  attr_accessor :highlight_pos
  attr_property :value
  attr_property :tabbox
  def onMerge(val);
        button.value = val.value if(val.respond_to?(:value))
    ;end
  def onMousePress(ev);
        button.value = !button.value
        damage_self
        whenClick.call if whenClick
        parent.set_tab(self) if parent.respond_to?(:set_tab)
    ;end
  def draw(vg);
        bg_color      = color("424B56")
        bg_on_color   = color("00818E")
        outline_color = color("000000")
        dark_off      = color("325A6E")
        text_color1   = color("52FAFE")
        text_color2   = Theme::TextColor
        vg.path do |v|
            v.rect(0, 0, w, h)
            if(button.value)
                v.fill_color(bg_on_color)
            else
                paint = v.linear_gradient(0,0,0,h,Theme::ButtonGrad1, Theme::ButtonGrad2)
                v.fill_paint paint
            end
            v.fill
        end
        vg.path do |v|
            v.rect(0,0,w,h)
            v.stroke_color(outline_color)
            v.stroke_width 1
            v.stroke
        end

        begin
            pos = 0
            if(button.highlight_pos == :bottom)
                pos = h*7/8
            end
            vg.path do |v|
                v.rect(0,pos,w,h/8)
                v.fill_color(text_color1) if  button.value
                v.fill_color(dark_off)    if !button.value
                v.fill
            end
        end


        vg.font_face("bold")
        vg.font_size h*0.65
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        if(button.value)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text(w/2,h/2,button.label.upcase)
    ;end
  def layout(l, selfBox);
        return selfBox if layoutOpts.include? :no_constraint
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, label.upcase)
        if(bb != 0)
            #Width cannot be so small that letters overflow
            l.aspect(selfBox, bb, scale)
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @button = self

    
context = Hash.new
context["button"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "tabbox")
self.whenClick = nil
self.highlight_pos = :bottom
self.value = false
self.tabbox = nil

    #puts "TabButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Text < Qml::Widget
  attr_reader :text
  attr_accessor :textColor
  attr_property :height
  attr_property :align
  def class_name();
        "Text"
    ;end
  def draw(vg);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, label.upcase)


        vg.font_face("bold")
        if(w/(self.height*h) < bb)
            vg.font_size self.height*h
        else
            vg.font_size self.height*h
        end
        vg.fill_color(self.textColor)
        if(align == :center)
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h/2,label.upcase)
        else
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(0,h/2,label.upcase)
        end
    ;end
  def layout(l, selfBox);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, label.upcase)
        if(bb != 0 && !self.layoutOpts.include?(:ignoreAspect) &&
           !self.layoutOpts.include?(:no_constraint))
            #Width cannot be so small that letters overflow
            if(height == 0.5)
                l.aspect(selfBox, bb, 1.5*scale)
            else
                l.aspect(selfBox, bb, scale)
            end
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text = self

    
context = Hash.new
context["text"] = self
Qml::prop_add(self, "height")
Qml::prop_add(self, "align")
self.textColor = Theme::TextColor
self.height = 0.5
self.align = :center

    #puts "Text, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextEdit < Qml::Widget
  attr_reader :text
  attr_accessor :action
  attr_accessor :textColor
  attr_property :height
  attr_accessor :valueRef
  def animate();
        if(root.key_widget != self)
            @next = nil
            if(@state)
                @state = nil
                damage_self
            end
            return
        end
        now = Time.new
        if(@next.nil?)
            @next = now + 0.1
            return
        elsif(@next < now)
            @state = !@state
            @next = now + 0.7
            damage_self
        end
    ;end
  def class_name();
        "Text"
    ;end
  def onSpecial(k, mode);
        return if @edit.nil?
        return if mode != :press
        if(k == :up)
            @edit.up
        elsif(k == :down)
            @edit.down
        elsif(k == :left)
            @edit.left
        elsif(k == :right)
            @edit.right
        end
        @state = true
        now = Time.new
        @next = now + 0.7
        damage_self
    ;end
  def onKey(k, mode);
        return if mode != "press"
        pos = self.label.length
        pos = @edit.pos if @edit
        ll = self.label
        if(k.ord == 8)
            pos -= 1
            if(pos >= ll.length)
                self.label = ll.slice(0, ll.length-1)
            elsif(pos >= 0)
                self.label = ll.slice(0, pos) + ll.slice(pos+1, ll.length)
            end
        else
            self.label = ll.insert(pos, k)
        end
        ll = self.label
        self.valueRef.value = ll if self.valueRef
        @edit     = EditRegion.new($vg, ll, w-20, height*h)
        if(k.ord == 8)
            @edit.pos = pos
        else
            @edit.pos = pos+1
        end
        damage_self
    ;end
  def draw(vg);
        background Theme::VisualBackground
        input  = self.label
        if(input.nil?)
            input  = "This is some random text which is hopefully "
            input += "is long enough to span a few different lines"
            input += ". I need it to in order to test multi-line "
            input += "rendering."
        end

        vg.font_face("bold")
        vg.font_size height*h
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color(textColor)
        ypos = height*h/2
        lasty = ypos
        lastx = 0

        #Break into lines
        @edit     ||= EditRegion.new(vg, input, w-20, height*h)

        @edit.each_string do |x, y, str, cursor|
            if(cursor == false)
                vg.text(x+10, y, str)
            else
                if(@state)
                vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
                    vg.text(x+10, y, str)
                end
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text = self

    
context = Hash.new
context["text"] = self
Qml::prop_add(self, "height")
self.action = nil
self.textColor = Theme::TextColor
self.height = 0.1
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        text.valueRef = OSC::RemoteParam.new($remote, text.extern)
        text.valueRef.callback = Proc.new {|x|
            text.label = x;
            text.damage_self}
    
end})

    #puts "TextEdit, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextScroll < Qml::Widget
  attr_reader :text
  attr_reader :scroll
  attr_accessor :action
  attr_accessor :textColor
  attr_property :lines
  attr_accessor :valueRef
  attr_property :off
  def class_name(); "TextScroll" ;end
  def onScroll(ev);
        scroll.onScroll(ev)
    ;end
  def num_lines();
        ind = 0
        self.label.each_line do |l|
            ind += 1
        end
        ind
    ;end
  def tryScroll();
        ln = num_lines()
        start = (1-scroll.value)*(ln-lines)
        start = start.to_i
        if(start != self.off)
            self.off = start
            damage_self
        end
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.00, 0, 0.02, 1.0)
        selfBox
    ;end
  def draw(vg);
        background Theme::VisualBackground
        input  = self.label

        hh = h/lines
        vg.font_face("bold")
        vg.font_size hh
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color(textColor)

        ind = 0
        input.each_line do |l|
            ii = ind-self.off
            break if ii >= self.lines
            vg.text(10+0.02*w, hh/2+hh*ii, l) if ii >= 0
            ind += 1
        end

    ;end
end
class Qml::ScrollBar < Qml::Valuator
  attr_property :bar_size
  attr_property :style
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end
    ;end
  def draw_overlay(vg);
        pad = 1
        domain = 1.0-bar_size
        if(vertical)
            center = 0.5*bar_size + (1-value)*domain
            self.dragScale = h*(1-bar_size)
            vg.path do |v|
                v.rect(pad,pad+@h*(center-0.5*bar_size),
                    @w-2*pad,@h*bar_size-2*pad)
                v.fill_color color("56c0a5")
                v.fill
            end
        else
            center = 0.5*bar_size + value*domain
            self.dragScale = w*(1-bar_size)
            vg.path do |v|
                v.rect(pad+w*(center-0.5*bar_size),pad,
                w*bar_size-2*pad,h-2*pad)
                v.fill_color color("56c0a5")
                v.fill
            end
        end
    ;end
  def draw_normal(vg);
        pad = 1
        vg.path do |v|
            v.rect(pad,pad,w-2*pad,h-2*pad)
            v.fill_color Theme::ScrollInactive
            v.fill
        end
        domain = 1.0-bar_size
        if(vertical)
            center = 0.5*bar_size + (1-value)*domain
            self.dragScale = h*(1-bar_size)
            vg.path do |v|
                v.rect(pad,pad+@h*(center-0.5*bar_size),
                    @w-2*pad,@h*bar_size-2*pad)
                v.fill_color Theme::ScrollActive
                v.fill
            end
        else
            center = 0.5*bar_size + value*domain
            self.dragScale = w*(1-bar_size)
            vg.path do |v|
                v.rect(pad+w*(center-0.5*bar_size),pad,
                w*bar_size-2*pad,h-2*pad)
                v.fill_color Theme::ScrollActive
                v.fill
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous174 = self

    
context = Hash.new
Qml::prop_add(self, "bar_size")
Qml::prop_add(self, "style")
self.bar_size = 0.1
self.style = :normal

    #puts "ScrollBar, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextScroll
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text = self
@scroll = Qml::ScrollBar.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["text"] = self
context["scroll"] = @scroll
Qml::prop_add(self, "lines")
Qml::prop_add(self, "off")
@scroll.whenValue = lambda { text.tryScroll }
Qml::add_child(self, @scroll)
self.action = nil
self.textColor = Theme::TextColor
self.lines = 30
self.valueRef = nil
self.off = 0

    #puts "TextScroll, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextSel < Qml::Widget
  attr_reader :textsel
  attr_accessor :valueRef
  attr_property :extern
  attr_accessor :options
  attr_accessor :opt_vals
  attr_accessor :selected
  attr_accessor :whenValue
  def set_value(x);
        sel = 0
        opt_vals.each_with_index do |i, ind|
            sel = ind if i == x
        end

        self.selected = sel
        rt = self.root
        rt.damage_item self if rt
    ;end
  def layout(l, selfBox);
        if(!layoutOpts.include?(:no_constraint))
            scale = 100
            $vg.font_size scale
            bb = 1
            textsel.options.each do |x|
                x = x[0..8] if x.length > 8
                bbl  = $vg.text_bounds(0, 0, (x+"    ").upcase)
                bb   = [bb, bbl].max
            end
            scale *= layoutOpts[0] if layoutOpts.include?(:rescale)
            if(bb != 0)
                #Width cannot be so small that letters overflow
                l.aspect(selfBox, scale, bb)
            end
        end
        selfBox
    ;end
  def class_name();
        "TextSel"
    ;end
  def draw(vg);
        text_color = Theme::TextColor
        pad  = 1/64
        pad2 = (1-2*pad)
        return if options[selected].nil?

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color text_color
        vg.text(3+w*pad*2,h/2,(options[selected]).upcase)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @textsel = self

    
context = Hash.new
context["textsel"] = self
Qml::prop_add(self, "extern")
self.valueRef = nil
self.extern = ""
self.options = ["text", "test", "asdf"]
self.opt_vals = []
self.selected = 0
self.whenValue = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote,
                                       textsel.extern)
        textsel.label   = meta.short_name
        textsel.tooltip = meta.tooltip
        if(meta.options)
            nopts = []
            vals  = []
            meta.options.each do |x|
                vals  << x[0]
                nopts << x[1]
            end
            textsel.options = nopts
            textsel.opt_vals = vals
        end
        textsel.valueRef = OSC::RemoteParam.new($remote, textsel.extern)
        textsel.valueRef.mode = :options
        textsel.valueRef.callback = Proc.new {|x| textsel.set_value(x)}
    
end})

    #puts "TextSel, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ThemeOld < Object
  attr_reader :theme
  def color(x, alpha=255);
        nil
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super()
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @theme = self

    
context = Hash.new
context["theme"] = self
@db.connect_property(self.properties["const"], lambda {Color global_dim:       color("#114575")})
@db.connect_property(self.properties["const"], lambda {Color global_bright:    color("#3ac5ec")})
@db.connect_property(self.properties["const"], lambda {Color env_background:   color("#808080")})
@db.connect_property(self.properties["const"], lambda {Color env_fill_color:   color("#0d0d0d")})
@db.connect_property(self.properties["const"], lambda {Color env_stroke_color: color("#014767")})
@db.connect_property(self.properties["const"], lambda {Color env_light_fill:   color("#114575", 55)})
@db.connect_property(self.properties["const"], lambda {Color env_bright:       theme.global_bright})
@db.connect_property(self.properties["const"], lambda {Color env_dim:          theme.global_dim})
@db.connect_property(self.properties["const"], lambda {Color env_sel:          color("#00ff00")})
@db.connect_property(self.properties["const"], lambda {Color knob_bright:      theme.global_bright})
@db.connect_property(self.properties["const"], lambda {Color knob_dim:         theme.global_dim})
@db.connect_property(self.properties["const"], lambda {Color button_on:        color("#00ff00")})
@db.connect_property(self.properties["const"], lambda {Color button_off:       color("#04375e")})
@db.connect_property(self.properties["const"], lambda {Color button_outline:   color("#0089b9")})
@db.connect_property(self.properties["const"], lambda {Color button_text1:     color("#00c2ea")})
@db.connect_property(self.properties["const"], lambda {Color button_text2:     color("#00c2ea")})
@db.connect_property(self.properties["const"], lambda {Color dropdown_color:   color("#000080")})

    #puts "ThemeOld, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ToggleButton < Qml::Button
  attr_reader :toggle
  attr_accessor :valueRef
  def setValue(x);
        self.value = x
        whenValue.call if whenValue
        damage_self
    ;end
  def refresh();
        self.valueRef.refresh if self.valueRef
    ;end
  def onMousePress(ev);
        if(root.learn_mode && extern)
            $remote.automate(extern)
            root.log(:tooltip, "Learning #{extern}")
            return
        end
        self.value = !self.value
        self.valueRef.value = self.value if self.valueRef
        damage_self
        whenValue.call if whenValue
    ;end
  def onMouseEnter(ev);
        if(self.tooltip.nil? || self.tooltip.empty?)
            self.root.log(:tooltip, "extern = <" + self.extern + ">")
        else
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @toggle = self

    
context = Hash.new
context["toggle"] = self
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, toggle.extern)
        toggle.label   = meta.short_name if toggle.label.empty?
        toggle.tooltip = meta.tooltip
        toggle.valueRef.clean if toggle.valueRef
        toggle.valueRef = OSC::RemoteParam.new($remote, toggle.extern)
        toggle.valueRef.callback = lambda {|x| toggle.setValue(x)}
    
end})

    #puts "ToggleButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TriggerButton < Qml::Button
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        whenValue.call if whenValue
        damage_self
    ;end
  def animate();
        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.9
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous40 = self

    
context = Hash.new

    #puts "TriggerButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VSlider < Qml::Valuator
  attr_reader :slider
  def draw(vg);
        vg.drawVZSlider(slider.t, *slider.bbox)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @slider = self

    
context = Hash.new
context["slider"] = self

    #puts "VSlider, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Valuator < Qml::Widget
  attr_reader :valuator
  attr_accessor :valueRef
  attr_accessor :prev
  attr_property :value
  attr_accessor :dragScale
  attr_accessor :vertical
  attr_accessor :whenValue
  attr_property :active
  attr_property :type
  attr_property :units
  def setValue(v);
        self.value = lim(v, 0.0, 1.0)
        damage_self
    ;end
  def refresh(); valueRef.refresh if valueRef ;end
  def create_radial();
        return
        gbl_cx = valuator.global_x + 0.5*valuator.w
        gbl_cy = valuator.global_y + 0.5*valuator.h
        gbl_w  = window.w
        gbl_h  = window.h

        ropt = [gbl_cx, gbl_cy, gbl_w-gbl_cx, gbl_h-gbl_cy].min

        diameter = [2.0*ropt, 3.0*0.5*(valuator.w+valuator.h)].min

        widget = RadialMenu.new(valuator.db)
        widget.w = diameter
        widget.h = diameter
        widget.x = valuator.w/2-diameter/2
        widget.y = valuator.h/2-diameter/2
        widget.layer = 2
        widget.callback = lambda {|sel|
            case sel
            when 0
                valueRef.midi_learn if valueRef
            when 1
                valueRef.rand if valueRef
            when 2
                valueRef.midi_unlearn if valueRef
            when 3
                valueRef.default if valueRef
            end
        }

        Qml::add_child(valuator, widget)
        valuator.root.smash_draw_seq
        valuator.root.damage_item widget
    ;end
  def onScroll(ev);
        return if !self.active
        fine = root.fine_mode ? 0.2 : 1.0
        updatePos(-fine*ev.dy/50.0)
    ;end
  def reset();
        reset_value = 64.01/127.0
        if(valueRef)
            old_dsp = self.valueRef.display_value

            self.valueRef.value = reset_value
            self.value = reset_value
            new_dsp = self.valueRef.display_value
            whenValue.call if whenValue && (new_dsp.nil? || old_dsp != new_dsp)
            valuator.root.log(:user_value, valuator.valueRef.display_value, src=valuator.label)
        else
            self.value = reset_value
        end
        damage_self
    ;end
  def onMousePress(ev);
        mouse_handle(ev)
    ;end
  def mouse_handle(ev);
        return if !self.active
        if(root.learn_mode && extern)
            $remote.automate(extern)
            root.log(:tooltip, "Learning #{extern}")
        end
        if(ev.buttons.include? :leftButton)
            valuator.prev = ev.pos
            now = Time.new
            #400 ms delta for double click
            if(@click_time && (now-@click_time) < 0.400)
                $remote.default(extern) if extern
                whenValue.call if whenValue
            end
            @click_time = now
        elsif(ev.buttons.include? :rightButton)
            if(children.empty?)
                create_radial
            end
        elsif(ev.buttons.include? :middleButton)
            reset
        elsif(ev.buttons.include? :drag_and_drop)
            puts "sending dnd event..."
            $remote.action("/last_dnd", self.extern)
        else
            valuator.prev = nil
        end
    ;end
  def onMouseMove(ev);
        return if !self.active
        fine = root.fine_mode ? 0.05 : 1.0
        if(prev)
            delta = if(vertical)
                +(ev.pos.y - self.prev.y)
            else
                -(ev.pos.x - self.prev.x)
            end
            updatePos(fine*delta/dragScale)
            self.prev = ev.pos
        end
    ;end
  def onMouseEnter(ev);
        self.root.log(:tooltip, self.tooltip)
    ;end
  def lim(x, low, high);
        [low, [x, high].min].max
    ;end
  def updatePosAbs(tmp);
        nvalue = lim(tmp, 0.0, 1.0)
        if(valuator.valueRef)
            old_dsp = valuator.valueRef.display_value
            valuator.valueRef.value = nvalue
            valuator.value = nvalue
            new_dsp = valuator.valueRef.display_value
            whenValue.call if whenValue && (new_dsp.nil? || old_dsp != new_dsp)
            out_value = valuator.valueRef.display_value
            if(valuator.units && valuator.units != "none")
                out_value = out_value.round(2) if out_value.class == Float
                out_value = out_value.to_s
                out_value += " "
                out_value += valuator.units
            end
            valuator.root.log(:user_value, out_value, src=valuator.label)
        else
            valuator.value = nvalue
            whenValue.call if whenValue
        end
        damage_self
    ;end
  def updatePos(delta);
        updatePosAbs(valuator.value - delta)
    ;end
  def onMerge(val);
        return if self.class != val.class
        valuator.value = val.value if(val.respond_to?(:value))
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @valuator = self

    
context = Hash.new
context["valuator"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "active")
Qml::prop_add(self, "type")
Qml::prop_add(self, "units")
self.valueRef = nil
self.prev = nil
self.value = 1.0
self.dragScale = 200.0
self.vertical = true
self.whenValue = nil
self.active = true
self.type = nil
self.units = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, valuator.extern)
        valuator.label   = meta.short_name
        valuator.tooltip = meta.tooltip
        valuator.units   = meta.units

        valuator.valueRef = OSC::RemoteParam.new($remote, valuator.extern)
        valuator.valueRef.set_min(meta.min)
        valuator.valueRef.set_max(meta.max)
        valuator.valueRef.set_scale(meta.scale)
        valuator.valueRef.type     = "f" if valuator.type
        valuator.valueRef.callback = Proc.new {|x| valuator.setValue(x)}

    
end})

    #puts "Valuator, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Widget < Object
  attr_reader :widget
  attr_accessor :parent
  attr_accessor :children
  attr_property :label
  attr_accessor :tooltip
  attr_accessor :layoutOpts
  attr_property :extern
  attr_accessor :layer
  attr_accessor :x
  attr_accessor :y
  attr_accessor :w
  attr_accessor :h
  def draw(vg);;end
  def onSetup(old=nil);
    ;end
  def layout(l, selfBox);
        widget.children.each do |child|
            if(child.respond_to?(:layout))
                nbox = l.genConstBox(0, 0,
                    selfBox.w, selfBox.h, child)
                child.layout(l, nbox)
            end
        end

        selfBox
    ;end
  def fixed(l, pbox, xc, yc, wc, hc);
        box = nil
        selfBox = l.genConstBox(pbox.w*xc, pbox.h*yc, pbox.w*wc, pbox.h*hc, self)
        box = self.layout(l, selfBox)
        box
    ;end
  def fixed_long(l, pbox, xc, yc, wc, hc, xf, yf, wf, hf);
        l.fixed_long(self, pbox, xc, yc, wc, hc, xf, yf, wf, hf)
    ;end
  def damage_self(all=nil);
        rt = self.root
        rt.damage_item(self, all) if rt
    ;end
  def background(c);
        $vg.path do |v|
            v.rect(0,0,w,h)
            v.fill_color c
            v.fill
        end
    ;end
  def class_name();
        "Widget"
    ;end
  def window();
        if(parent.respond_to?(:root))
            parent.window
        else
            self
        end
    ;end
  def root();
        if(parent.respond_to?(:root))
            parent.root
        else
            parent
        end
    ;end
  def global_x();
        par = 0
        if(parent.respond_to?(:global_x))
            par = parent.global_x
        end
        par + widget.x
    ;end
  def global_y();
        par = 0
        if(parent.respond_to?(:global_y))
            par = parent.global_y
        end
        par + widget.y
    ;end
  def to_s(i=0);
        out = ""
        i.times do
            out += " |"
        end

        out += "- <"+widget.class_name()+"##{widget.ui_path}:" + widget.label
        out += ">\n"
        widget.children.each do |x|
            out += x.to_s(i+1)
        end

        out
    ;end
  def abs_x();
        p = parent
        if(p.respond_to?(:root))
            widget.x + p.abs_x
        else
            widget.x
        end
    ;end
  def abs_y();
        p = parent
        if(p.respond_to?(:root))
            widget.y + p.abs_y
        else
            widget.y
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super()
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @widget = self

    
context = Hash.new
context["widget"] = self
Qml::prop_add(self, "label")
Qml::prop_add(self, "extern")
self.parent = nil
self.children = []
self.label = ""
self.tooltip = ""
self.layoutOpts = []
self.extern = ""
self.layer = 0
self.x = nil
self.y = nil
self.w = nil
self.h = nil

    #puts "Widget, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Window < Qml::Widget
  def class_name();
        "Window"
    ;end
  def draw(vg);
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous44 = self

    
context = Hash.new

    #puts "Window, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::IndentTest < Qml::Widget
  attr_reader :ident
  def class_name(); "Indent" ;end
  def layout(l);
        selfBox = l.genBox :indent, ident
        if(!indent.children.empty?)
            ch = indent.children[0].layout l
            l.contains selfBox, ch
        end
        selfBox
    ;end
  def draw(vg); 
        vg.path do |v|
            v.rounded_rect(0, 0, w, h, 10)
            v.fill_color(NVG.rgba(9, 0, 0, 255))
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @ident = self

    
context = Hash.new
context["ident"] = self

    #puts "IndentTest, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::OverlayBox < Qml::Widget
  attr_reader :overlaybox
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            v.fill_color(NVG.rgba(0xff,0xff,0xff,0x80))
            v.fill
        end
        vg.path do |v|
            v.move_to(0,0)
            v.line_to(w,h)
            v.move_to(w,0)
            v.line_to(0,h)
            v.stroke_color(NVG.rgba(0xff,0xff,0xff,0xff))
            v.stroke
        end
    ;end
  def onMousePress(ev);
        puts "I got a mouse press (overlaybox)"
        rt = overlaybox.root
        rt.ego_death overlaybox
        #puts test.root
        #if(children.empty?)
        #    widget = OverlayBox.new(test.db)
        #    widget.w = test.w
        #    widget.h = test.h
        #    widget.x = 0
        #    widget.y = 0
        #    Qml::add_child(test, widget)
        #end
        #test.root.damage_item(test)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @overlaybox = self

    
context = Hash.new
context["overlaybox"] = self

    #puts "OverlayBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::OverlayTest < Qml::Widget
  attr_reader :test
  def draw(vg);
        pi = 3.14159
        start = pi/4;
        end_   = pi*3.0/4.0;
        inner  = 0.2*[h,w].min
        outer = 0.4*[h,w].min
        cy = h/2;
        cx = w/2;
        kr = h/4;
        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color(NVG.rgba(0x11,0x45,0x75,255));
            v.fill
        end

        len = (3.0/2.0*pi)*0.5;
        startt = end_ + len;

        vg.path do |v|
            v.arc(cx, cy, inner, end_, startt, 2);
            v.arc(cx, cy, outer, startt, end_, 1);
            v.close_path
            v.fill_color(NVG.rgba(0x3a,0xc5,0xec,255));
            v.fill
        end
    ;end
  def onMousePress(ev);
        puts "I got a mouse press (overlaytest)"
        puts test.root
        if(children.empty?)
            widget = OverlayBox.new(test.db)
            widget.w = test.w
            widget.h = test.h
            widget.x = 0
            widget.y = 0
            Qml::add_child(test, widget)
            test.root.smash_draw_seq
        end
        test.root.damage_item(test)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @test = self

    
context = Hash.new
context["test"] = self

    #puts "OverlayTest, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::LfoVis < Qml::Widget
  attr_reader :lfo_vis
  attr_reader :run_view
  attr_property :points
  attr_property :type
  attr_property :drag_type
  attr_property :drag_prev
  attr_property :phase
  attr_property :depth
  attr_property :delay_time
  attr_property :period
  attr_property :time
  attr_accessor :valueRef
  def onSetup(old=nil);
        refs = []
        base = lfo_vis.extern
        type_var = OSC::RemoteParam.new($remote, base+"PLFOtype")
        type_var.mode = :options
        type_var.callback = lambda {|x|
            ntype = [:sine, :triangle, :square, :rampup,
                :rampdown, :exp1, :exp2][x]
            return if(ntype == lfo_vis.type)
            lfo_vis.type = [:sine, :triangle, :square, :rampup,
                :rampdown, :exp1, :exp2][x]}

        depth_var = OSC::RemoteParam.new($remote, base+"Pintensity")
        depth_var.callback = lambda {|x|
            lfo_vis.depth = x
            lfo_vis.damage_self}

        delay_var = OSC::RemoteParam.new($remote, base+"Pdelay")
        delay_var.callback = lambda {|x|
            lfo_vis.delay_time = Math.exp(Math.log(2000)*x)
            lfo_vis.damage_self}

        freq_var  = OSC::RemoteParam.new($remote, base+"Pfreq")
        freq_var.type = "f"
        freq_var.set_min(0)
        freq_var.set_max(1)
        freq_var.callback = lambda {|x|
            lfo_vis.period = 20*Math.exp(Math.log(1500/20)*(1-x))
            lfo_vis.damage_self}

        phase_var = OSC::RemoteParam.new($remote, base+"Pstartphase")
        phase_var.callback = lambda {|x|
            lfo_vis.phase = x
            lfo_vis.updateType
            lfo_vis.damage_self}


        refs << type_var
        refs << depth_var
        refs << freq_var
        refs << delay_var
        refs << phase_var
        self.valueRef = refs

    ;end
  def class_name();
        "LfoVis"
    ;end
  def onMousePress(ev);
        loc_x = ev.pos.x - lfo_vis.global_x
        if(loc_x < 0.2*lfo_vis.w)
            lfo_vis.drag_type = :delay
        else
            lfo_vis.drag_type = :lfo
        end
        lfo_vis.drag_prev = ev.pos
    ;end
  def onMouseMove(ev);
        if(lfo_vis.drag_prev)
            dx = ev.pos.x - lfo_vis.drag_prev.x
            dy = ev.pos.y - lfo_vis.drag_prev.y

            if(lfo_vis.drag_type == :delay)
                #dv = lfo_vis.phase + dy/100.0
                #dv_ = [1, [0, dv].max].min
                #lfo_vis.phase = dv_
                #lfo_vis.drag_prev.y = ev.pos.y if(dv == dv_)
                #lfo_vis.updateType

                dt = Math.exp((Math.log(lfo_vis.delay_time)/Math.log(2000) + dx/200.0)*Math.log(2000))
                dt_ = [2000, [1, dt].max].min
                lfo_vis.delay_time = dt_
                lfo_vis.drag_prev.x = ev.pos.x if(dt == dt_)
                self.valueRef[3].value = Math.log(dt_)/Math.log(2000)
                damage_self
            elsif(lfo_vis.drag_type == :lfo)
                lfo_vis.drag_prev = ev.pos
                dv = lfo_vis.depth + dy/300.0
                dv = [1, [0, dv].max].min
                lfo_vis.depth = dv
                self.valueRef[1].value = dv

                dt = 100*Math.exp(Math.log(0.01*lfo_vis.period) + dx/200.0)
                dt = [1500, [20, dt].max].min
                lfo_vis.period = dt

                nval = 1-Math.log(lfo_vis.period/20)/Math.log(1500/20)
                self.valueRef[2].value = nval
                damage_self
            end
        end
    ;end
  def updateType();
        shape = case lfo_vis.type
        when :triangle
            Proc.new {|phase|
                if(phase >= 0 && phase < 0.25)
                    4.0 * phase
                elsif(phase >= 0.25 && phase < 0.75)
                    2.0 - 4 * phase
                else
                    4.0 * phase - 4;
                end
            }
        when :square
            Proc.new {|phase|
                if(phase < 0.5)
                    -1.0
                else
                    1.0
                end
            }
        when :rampup
            Proc.new {|phase| (phase - 0.5) * 2.0}
        when :rampdown
            Proc.new {|phase| (0.5 - phase) * 2.0};
        when :exp1
            Proc.new {|phase| (0.05 ** phase) * 2.0 - 1.0}
        when :exp2
            Proc.new {|phase| (0.001 ** phase) * 2.0 - 1.0}
        else
            Proc.new {|x| Math.sin(2*3.14*x) }
        end

        # root point
        p = [0,0, 0, 0.2]
        # func points
        resolution = 128
        (0..resolution).each do |i|
            x = 0.2+0.8*i/resolution
            phase = i/resolution + lfo_vis.phase
            phase -= 1 if phase > 1

            y = shape.call(phase)
            p << y
            p << x
        end
        lfo_vis.points = p
        damage_self
    ;end
  def draw(vg);
        padfactor = 10
        bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)

        background Theme::VisualBackground

        dat = lfo_vis.points
        updateType if dat.empty?
        dat = lfo_vis.points
        if(dat.empty?)
            puts "[Zyn-Zest:ERROR] LfoVis - um, it's empty..."
            return
        end

        fill_color   = Theme::VisualBackground
        stroke_color = Theme::VisualStroke
        light_fill   = Theme::VisualLightFill
        light_filll  = color("911515", 20)
        bright       = Theme::VisualBright
        dim          = Theme::VisualDim
        sel_color    = Theme::VisualSelect

        background fill_color

        draw_grid(vg, lfo_vis.depth*16, lfo_vis.period/10,
                  bb.x+0.2*bb.w, bb.y, 0.8*bb.w, bb.h)
        draw_grid(vg, lfo_vis.depth*16, lfo_vis.delay_time/100,
                  bb.x,          bb.y, 0.2*bb.w, bb.h)

        #weak highlight
        vg.path do |vg|
            vg.rect(0.2*w, 0, 0.8*w, h)
            vg.fill_color light_filll
            vg.fill
        end

        pts = Draw::toPos(dat)

        #Draw Highlights
        Draw::WaveForm::under_highlight(vg, bb, pts, light_fill)
        Draw::WaveForm::over_highlight(vg,  bb, pts, light_fill)

        #Draw Zero Line
        Draw::WaveForm::zero_line(vg, bb, dim)

        Draw::WaveForm::env_sel_line(vg, bb, 1, pts, dim)

        #Draw Actual Line
        Draw::WaveForm::lfo_plot(vg, bb, pts, bright)

        vg.fill_color   color(:black)
        vg.stroke_color bright
        (0...(dat.length/2)).each do |i|
            xx = bb.x + bb.w*dat[2*i+1];
            yy = bb.y + bb.h/2*(1-dat[2*i]);
            next if(i >= 2)
            Draw::WaveForm::env_marker(vg, xx, yy, 3)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @lfo_vis = self
@run_view = Qml::Widget.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["lfo_vis"] = self
context["run_view"] = @run_view
@run_view.instance_eval do
def run_view=(k);@run_view=k;end
def run_view;    @run_view;  end
end
Qml::context_apply(context, @run_view)
Qml::prop_add(self, "points")
Qml::prop_add(self, "type")
Qml::prop_add(self, "drag_type")
Qml::prop_add(self, "drag_prev")
Qml::prop_add(self, "phase")
Qml::prop_add(self, "depth")
Qml::prop_add(self, "delay_time")
Qml::prop_add(self, "period")
Qml::prop_add(self, "time")

        @run_view.instance_eval("def class_name();\n            \"LfoVisAnimation\"\n        ;end", "src/mruby-zest/example/LfoVis.qml", 244)

        @run_view.instance_eval("def valueRef=(value_ref);\n            @value_ref = value_ref\n        ;end", "src/mruby-zest/example/LfoVis.qml", 250)

        @run_view.instance_eval("def valueRef();\n            @value_ref\n        ;end", "src/mruby-zest/example/LfoVis.qml", 255)

        @run_view.instance_eval("def runtime_points();\n            @runtime_points\n        ;end", "src/mruby-zest/example/LfoVis.qml", 260)

        @run_view.instance_eval("def runtime_points=(pts);\n            @runtime_points = pts\n        ;end", "src/mruby-zest/example/LfoVis.qml", 264)

        @run_view.instance_eval("def animate();\n            return if run_view.valueRef.nil?\n            run_view.valueRef.watch run_view.extern\n        ;end", "src/mruby-zest/example/LfoVis.qml", 282)

        @run_view.instance_eval("def draw(vg);\n            #Draw the data\n            pts   = @runtime_points\n            pts ||= []\n\n            padfactor = 10\n            bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)\n\n            Draw::WaveForm::overlay_lfo(vg, bb, pts)\n        ;end", "src/mruby-zest/example/LfoVis.qml", 287)
@run_view.layer = 1
@db.connect_property(@run_view.properties["extern"], lambda {lfo_vis.extern + "out"})
@db.connect_watcher(@run_view.properties["extern"], lambda {begin

            return if run_view.extern.nil?
            
            run_view.valueRef = OSC::RemoteParam.new($remote, run_view.extern)
            run_view.valueRef.set_watch
            run_view.valueRef.callback = Proc.new {|x|
                if(run_view.runtime_points != x)
                    run_view.runtime_points = x;
                    run_view.damage_self
                end
            }
        
end})
Qml::add_child(self, @run_view)
self.points = []
self.type = :triangle
self.drag_type = nil
self.drag_prev = nil
self.phase = 0.0
self.depth = 0.5
self.delay_time = 100
self.period = 1000
self.time = nil
self.valueRef = nil
@db.connect_watcher(self.properties["type"], lambda {begin

        lfo_vis.updateType
    
end})

    #puts "LfoVis, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TestLfoVis < Qml::LfoVis
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous48 = self

    
context = Hash.new
self.extern = "/part0/kit0/adpars/GlobalPar/FreqLfo/out"

    #puts "TestLfoVis, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TestRows < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :extern
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous49 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous50 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous51 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous52 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous54 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous55 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous58 = Qml::Col.new(database, ui_path+"9"+'/')
@anonymous56 = Qml::Selector.new(database, ui_path+"10"+'/')
@anonymous57 = Qml::Button.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "extern")
@db.connect_property(@anonymous49.properties["extern"], lambda {box.extern+"Pfreq" })
Qml::add_child(@top, @anonymous49)
@db.connect_property(@anonymous50.properties["extern"], lambda {box.extern+"Pintensity"})
Qml::add_child(@top, @anonymous50)
@db.connect_property(@anonymous51.properties["extern"], lambda {box.extern+"Pstartphase"})
Qml::add_child(@top, @anonymous51)
@db.connect_property(@anonymous52.properties["extern"], lambda {box.extern+"Pstretch"})
Qml::add_child(@top, @anonymous52)
Qml::add_child(self, @top)
@db.connect_property(@anonymous54.properties["extern"], lambda {box.extern+"Prandomness"})
Qml::add_child(@bot, @anonymous54)
@db.connect_property(@anonymous55.properties["extern"], lambda {box.extern+"Pfreqrand"})
Qml::add_child(@bot, @anonymous55)
@db.connect_property(@anonymous56.properties["extern"], lambda {box.extern+"PLFOtype"})
Qml::add_child(@anonymous58, @anonymous56)
@db.connect_property(@anonymous57.properties["extern"], lambda {box.extern+"Pcontinous"})
@anonymous57.label = "C"
Qml::add_child(@anonymous58, @anonymous57)
Qml::add_child(@bot, @anonymous58)
Qml::add_child(self, @bot)
self.label = "LFO"
self.extern = "/part0/kit0/adpars/GlobalPar/FreqLfo/"

    #puts "TestRows, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TestSwap < Qml::Widget
  attr_reader :container
  attr_reader :swap
  attr_reader :but
  def layout(l);
        selfBox = l.genBox :testRoot, self
        swapBox = swap.layout(l)
        buttBox = but.layout(l)

        l.fixed(swapBox, selfBox, 0,0,0.5,1)
        l.fixed(buttBox, selfBox, 0.5,0,0.5,1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @container = self
@swap = Qml::Swappable.new(database, ui_path+"1"+'/')
@but = Qml::Button.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["container"] = self
context["swap"] = @swap
context["but"] = @but
@db.connect_property(@swap.properties["content"], lambda {Qml::Knob})
Qml::add_child(self, @swap)
@but.layoutOpts = [:no_constraint]
@db.connect_watcher(@but.properties["value"], lambda {begin

            puts "value changed..."
            if(swap.content == Qml::Knob)
                swap.content = Qml::DropDown
            else
                swap.content = Qml::Knob
            end
        
end})
@but.label = "swap"
Qml::add_child(self, @but)

    #puts "TestSwap, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VolumeKnob < Qml::Knob
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous64 = self

    
context = Hash.new
self.extern = "/Pvolume"

    #puts "VolumeKnob, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TestBase < Object
  attr_reader :tb
  attr_accessor :parent
  attr_property :p1
  attr_property :p2
  attr_property :p3
  attr_accessor :children
  def to_s();
        "{#{p1},#{p2},#{p3}}"
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super()
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @tb = self

    
context = Hash.new
context["tb"] = self
Qml::prop_add(self, "p1")
Qml::prop_add(self, "p2")
Qml::prop_add(self, "p3")
self.parent = nil
self.p1 = nil
@db.connect_property(self.properties["p2"], lambda {tb.p1})
@db.connect_property(self.properties["p3"], lambda {tb.p2})
self.children = []

    #puts "TestBase, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TestExt < Qml::TestBase
  attr_reader :te1
  attr_reader :te2
  attr_reader :te3
  attr_reader :te4
  def to_s();
        "["+self.inspect+","+te2.to_s+","+te3.to_s+"]"
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @te1 = self
@te2 = Qml::TestBase.new(database, ui_path+"1"+'/')
@te3 = Qml::TestBase.new(database, ui_path+"2"+'/')
@te4 = Qml::TestBase.new(database, ui_path+"3"+'/')

    
context = Hash.new
context["te1"] = self
context["te2"] = @te2
context["te3"] = @te3
context["te4"] = @te4
@te4.instance_eval do
def te1=(k);@te1=k;end
def te1;    @te1;  end
end
Qml::context_apply(context, @te4)

        @te2.instance_eval("def to_s();\n            \"TE2\"\n        ;end", "src/mruby-qml-spawn/test/TestExt.qml", 6)
Qml::add_child(self, @te2)
@te3.p2 = 5
Qml::add_child(self, @te3)

        @te4.instance_eval("def tramp();\n            te1.p1\n        ;end", "src/mruby-qml-spawn/test/TestExt.qml", 21)
@db.connect_property(@te4.properties["p1"], lambda {te4.tramp})
Qml::add_child(self, @te4)

    #puts "TestExt, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ApplyButton < Qml::Button
  attr_reader :app
  attr_reader :lp
  attr_property :state
  attr_accessor :valueRef
  def onMousePress(ev);
        if(ev.buttons.include? :leftButton)
            self.state = :applying
            self.value = 1.0
            whenValue.call if whenValue
            lp.spin = true
            damage_self
        elsif(ev.buttons.include? :rightButton)
            if(children.length == 1)
                create_radial
            end
        end
    ;end
  def create_radial();
        gbl_cx = self.global_x + 0.5*self.w
        gbl_cy = self.global_y + 0.5*self.h
        gbl_w  = window.w
        gbl_h  = window.h

        ropt = [gbl_cx, gbl_cy, gbl_w-gbl_cx, gbl_h-gbl_cy].min

        diameter = [2.0*ropt, 3.0*0.5*(self.w+self.h)].min

        widget = RadialMenu.new(self.db)
        widget.w = diameter
        widget.h = diameter
        widget.x = self.w/2-diameter/2
        widget.y = self.h/2-diameter/2
        widget.layer = 2
        widget.fields = ["off", "1 sec", "off", "10 sec"]
        widget.callback = lambda {|sel|
            case sel
            when 0
                $auto_apply_dt = 1.0/0.0
            when 1
                $auto_apply_dt = 1.0
            when 2
                $auto_apply_dt = 1.0/0.0
            when 3
                $auto_apply_dt = 10.0
            end
            puts "auto apply dt is:"
            puts $auto_apply_dt
            puts sel
        }

        Qml::add_child(self, widget)
        self.root.smash_draw_seq
        self.root.damage_item widget
    ;end
  def setValue(x);
        self.value = x
        lp.spin = false if x == false
        lp.damage_self
        damage_self
    ;end
  def animate();
        now = Time.new
        $auto_apply_dt   ||= 1.0/0.0
        @auto_apply_last ||= now
        if(self.state == :waiting && (now-@auto_apply_last) > $auto_apply_dt)
            self.state = :applying
            self.value = 1.0
            whenValue.call if whenValue
            lp.spin = true
            damage_self
            @auto_apply_last = now
        end

        if(self.state == :waiting)
            @phase ||= 0.0
            self.value = Math.sin(3.14*@phase)
            @phase += 0.01
            @phase -= 1.0 if @phase > 1.0
            damage_self
            return
        end

        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.7
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  def layout(l, selfBox);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, "000 " + label.upcase)
        if(bb != 0)
            #Width cannot be so small that letters overflow
            l.aspect(selfBox, bb, scale)
        end

        lp.fixed(l, selfBox, 0.1, 0.1, 0.2, 0.8)
        selfBox
    ;end
end
class Qml::BusyLoop < Qml::Widget
  attr_property :value
  attr_property :spin
  def animate();
        return if !self.spin
        self.value += 0.01
        self.value -= 1.0 if self.value > 1.0
        damage_self
    ;end
  def draw(vg);
        pi = 3.14159
        start = 2*pi
        end_   = 0
        radius  = 0.25*[h*1.0,w].min
        inner  = 0.30*[h*1.0,w].min
        outer = 0.50*[h*1.0,w].min

        #lowest point is 0.707 cy-outer*0.707
        #shift by 0.293/2 to compinsate
        cy = h/2
        cx = w/2;
        kr = h/4;
        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color(color("114575"))
            v.fill
        end

        return if !self.spin
        self.value = 0.0 if(self.value.class != Float)
        cang  = 2*pi*value

        width = 0.4

        len = (3.0/2.0*pi)*width;
        startt = 2*pi*value + len;

        vg.path do |v|
            v.arc(cx, cy, inner, cang, startt, 2);
            v.arc(cx, cy, outer, startt, cang, 1);
            v.close_path
            v.fill_color(color("3AC5EC"))
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous107 = self

    
context = Hash.new
Qml::prop_add(self, "value")
Qml::prop_add(self, "spin")
self.layer = 2
self.value = 0.0
self.spin = false

    #puts "BusyLoop, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ApplyButton
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @app = self
@lp = Qml::BusyLoop.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["app"] = self
context["lp"] = @lp
Qml::prop_add(self, "state")
Qml::add_child(self, @lp)
self.state = :off
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        app.valueRef = OSC::RemoteParam.new($remote, app.extern)
        app.valueRef.callback = lambda {|x|
            app.setValue(x);
            if(!x)
                app.state = :off
            else
                app.state = :waiting
            end}
    
end})

    #puts "ApplyButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ArrowBox < Qml::Widget
  attr_reader :box
  attr_property :active
  attr_accessor :whenValue
  def onMousePress(ev);
        return if box.active

        box.active = true
        damage_self
        whenValue.call if whenValue
    ;end
  def draw(vg);
        scale = 0.5
        vg.path do
            vg.move_to((0.5-scale/2)*w,(0.5+scale/2)*h)
            vg.line_to((0.5-scale/2)*w,(0.5-scale/2)*h)
            vg.line_to((0.5+scale/2)*w,0.5*h)
            vg.close_path
            vg.fill_color(color("bbbbbb")) if !self.active
            vg.fill_color(color("55ff55")) if  self.active
            vg.fill
        end
    ;end
  def layout(l, selfBox);
        l.aspect(selfBox, 1, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "active")
self.active = false
self.whenValue = nil
@db.connect_watcher(self.properties["active"], lambda {begin

        box.damage_self
    
end})

    #puts "ArrowBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::BusyLoop < Qml::Widget
  attr_property :value
  attr_property :spin
  def animate();
        return if !self.spin
        self.value += 0.01
        self.value -= 1.0 if self.value > 1.0
        damage_self
    ;end
  def draw(vg);
        pi = 3.14159
        start = 2*pi
        end_   = 0
        radius  = 0.25*[h*1.0,w].min
        inner  = 0.30*[h*1.0,w].min
        outer = 0.50*[h*1.0,w].min

        #lowest point is 0.707 cy-outer*0.707
        #shift by 0.293/2 to compinsate
        cy = h/2
        cx = w/2;
        kr = h/4;
        vg.path do |v|
            v.arc(cx, cy, outer, start, end_, 1);
            v.arc(cx, cy, inner, end_, start, 2);
            v.close_path
            v.fill_color(color("114575"))
            v.fill
        end

        return if !self.spin
        self.value = 0.0 if(self.value.class != Float)
        cang  = 2*pi*value

        width = 0.4

        len = (3.0/2.0*pi)*width;
        startt = 2*pi*value + len;

        vg.path do |v|
            v.arc(cx, cy, inner, cang, startt, 2);
            v.arc(cx, cy, outer, startt, cang, 1);
            v.close_path
            v.fill_color(color("3AC5EC"))
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous107 = self

    
context = Hash.new
Qml::prop_add(self, "value")
Qml::prop_add(self, "spin")
self.layer = 2
self.value = 0.0
self.spin = false

    #puts "BusyLoop, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::BypassButton < Qml::ToggleButton
  def draw_text(vg);
        text_color1   = Theme::TextActiveColor
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*self.textScale/2
        if(value == true)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        if(layoutOpts.include? :left_text)
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(8,h/2,button.label.upcase)
        else
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h*0.666,"GL FILTER")
            vg.text(w/2,h*0.333,"BYPASS")
        end
    ;end
  def layout(l, selfBox);
        if(!self.layoutOpts.include?(:no_constraint))
            if(label.length == 1)
                l.aspect(selfBox, 1, 1)
            else
                scale = 100
                $vg.font_size scale
                bb1 = $vg.text_bounds(0, 0, "BYPASS")
                bb2 = $vg.text_bounds(0, 0, "GL FILTER")
                bb  = [bb1, bb2].max/2
                if(bb != 0)
                    #Width cannot be so small that letters overflow
                    self.aspect2(selfBox, bb, scale)
                end
            end
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous108 = self

    
context = Hash.new

    #puts "BypassButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ClearBox < Qml::Widget
  attr_reader :box
  attr_property :active
  attr_property :value
  attr_accessor :whenValue
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        whenValue.call if whenValue
        damage_self
    ;end
  def animate();
        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.9
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  def draw(vg);
        off_color     = Theme::ButtonInactive
        on_color      = Theme::ButtonActive
        cs = 0
        pad = 1/64
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 2)
            if(self.value == true)
                cs = 1
                v.fill_color on_color
            elsif(self.value.class == Float && self.value != 0)
                cs = 2
                t = self.value
                on = on_color
                of = Theme::ButtonGrad1
                v.fill_color(color_rgb(on.r*t + of.r*(1-t),
                                       on.g*t + of.g*(1-t),
                                       on.b*t + of.b*(1-t)))
            else
                paint = v.linear_gradient(0,0,0,h,
                Theme::ButtonGrad1, Theme::ButtonGrad2)
                v.fill_paint paint
            end
            v.fill
            v.stroke_width 1
            v.stroke

        end

        pad = 1/4

        vg.stroke_color(color("bbbbbb")) #if !self.active
        #vg.stroke_color(color("55ff55")) if  self.active
        vg.path do
            vg.move_to(1+pad*w,     1+pad*h)
            vg.line_to((1-pad)*w-1, (1-pad)*h-1)
            vg.stroke
        end
        vg.path do
            vg.move_to((1-pad)*w-1,  1+pad*h)
            vg.line_to(1+pad*w,     (1-pad)*h-1)
            vg.stroke
        end
    ;end
  def layout(l, selfBox);
        l.aspect(selfBox, 1, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "active")
Qml::prop_add(self, "value")
self.active = true
self.value = false
self.whenValue = nil

    #puts "ClearBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ColorBox < Qml::Widget
  attr_accessor :bg
  attr_accessor :pad
  def draw(vg);
        return if bg.nil?
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(pad*w,pad*h,pad2*w,pad2*h)
            v.fill_color self.bg
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous110 = self

    
context = Hash.new
self.bg = color("8f600f")
self.pad = 0.0

    #puts "ColorBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ColorButton < Qml::Button
  attr_property :bg_color
  def draw_button(vg);
        cs = 0
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 2)
            v.fill_color(self.bg_color)
            v.fill
            v.stroke_width 1
            v.stroke
        end
    ;end
  def draw_text(vg);
        text_color1   = color("111111")
        text_color2   = color("eeeeee")
        vg.font_face("bold")
        vg.font_size h*self.textScale
        if((self.bg_color.r + self.bg_color.b + self.bg_color.g) > 0.9)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end

        if(layoutOpts.include? :left_text)
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(8,h/2,self.label)
        else
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h/2,self.label)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous111 = self

    
context = Hash.new
Qml::prop_add(self, "bg_color")
@db.connect_property(self.properties["bg_color"], lambda {color("000000")})

    #puts "ColorButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ColorScheme < Qml::Widget
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.60, 0.40])
    ;end
end
class Qml::ColorSel < Qml::Widget
  def draw(vg);
        @hue ||= 0.2
        @sx  ||= 0.1
        @sy  ||= 0.4
        draw_color_wheel(vg, x, y, w, h,
            @hue / (2*Math::PI), @sx, @sy)
    ;end
  def abs(x); [x,-x].max ;end
  def onMousePress(ev);
        cx = x + w/2
        cy = y + h/2

        r1 = min(w,h)/2 - 5
        r0 = r1 - 20

        xx = ev.pos.x-global_x
        yy = ev.pos.y-global_y

        dx = (xx-cx)
        dy = (yy-cy)

        rr = Math::sqrt(dx*dx + dy*dy)

        #puts "#{r1} -- #{rr} -- #{r0}"

        if(r0 < rr && rr < r1)
            @hue = Math::atan2(dy,dx)
            #puts @hue
            damage_self
        end

        rt = r0 - 6

        tri_upper_x     = cx + rt*Math::cos(@hue)
        tri_upper_y     = cy + rt*Math::sin(@hue)

        tri_low_left_x  = cx + rt*Math::cos(@hue - (2*Math::PI)/3)
        tri_low_left_y  = cy + rt*Math::sin(@hue - (2*Math::PI)/3)

        tri_low_right_x = cx + rt*Math::cos(@hue + (2*Math::PI)/3)
        tri_low_right_y = cy + rt*Math::sin(@hue + (2*Math::PI)/3)

        _ux = tri_upper_x
        _uy = tri_upper_y
        _lx = tri_low_left_x
        _ly = tri_low_left_y
        _rx = tri_low_right_x
        _ry = tri_low_right_y

        _bx = (_lx+_rx)/2
        _by = (_ly+_ry)/2

        ul = dist(xx, yy, _ux, _uy, _lx, _ly)
        ur = dist(xx, yy, _rx, _ry, _ux, _uy)
        bt = dist(xx, yy, _lx, _ly, _rx, _ry)
        md = dist(xx, yy, _ux, _uy, _bx, _by)

        #puts "%%%  #{ul} -- #{ur} -- #{bt}"
        #puts "inside" if ul > 0 && ur > 0 && bt > 0

        range_y = Math::sqrt((_bx-_ux)*(_bx-_ux)+(_by-_uy)*(_by-_uy))
        range_x = Math::sqrt((_lx-_rx)*(_lx-_rx)+(_ly-_ry)*(_ly-_ry))/2



        #puts "y = #{bt/range_y} x = #{md/range_x}"
        if(ul > 0 && ur > 0 && bt > 0)
            #Yes, this is intentionally reversed
            @sy = md/range_x

            @sx = bt/range_y
            extra = Math::sin(Math::PI/6.0)
            @sx = @sx*(1+extra) - extra
            #puts "@sx=#{@sx}  @sy=#{@sy}"
            damage_self
        end
    ;end
  def dist(px, py, x1, y1, x2, y2);
        ((y2-y1)*px - (x2-x1)*py + x2*y1 - y2*x1)/Math::sqrt((y2-y1)*(y2-y1) + (x2-x1)*(x2-x1))
    ;end
  def onMouseMove(ev);
        onMousePress(ev)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous116 = self

    
context = Hash.new

    #puts "ColorSel, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ColorScheme
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous115 = self
@anonymous112 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous114 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous113 = Qml::ColorSel.new(database, ui_path+"3"+'/')

    
context = Hash.new

        @anonymous112.instance_eval("def onSetup(old=nil);\n            Theme.constants.each do |c|\n                ch = Qml::ColorButton.new(db)\n                ch.label = c.to_s\n                ch.bg_color = Theme.const_get(c)\n                Qml::add_child(self, ch)\n            end\n        ;end", "src/mruby-zest/example/ColorScheme.qml", 3)

        @anonymous112.instance_eval("def layout(l, selfBox);\n            Draw::Layout::grid(l, selfBox, children, 8, 6)\n        ;end", "src/mruby-zest/example/ColorScheme.qml", 12)
Qml::add_child(self, @anonymous112)
Qml::add_child(@anonymous114, @anonymous113)
Qml::add_child(self, @anonymous114)

    #puts "ColorScheme, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ColorSel < Qml::Widget
  def draw(vg);
        @hue ||= 0.2
        @sx  ||= 0.1
        @sy  ||= 0.4
        draw_color_wheel(vg, x, y, w, h,
            @hue / (2*Math::PI), @sx, @sy)
    ;end
  def abs(x); [x,-x].max ;end
  def onMousePress(ev);
        cx = x + w/2
        cy = y + h/2

        r1 = min(w,h)/2 - 5
        r0 = r1 - 20

        xx = ev.pos.x-global_x
        yy = ev.pos.y-global_y

        dx = (xx-cx)
        dy = (yy-cy)

        rr = Math::sqrt(dx*dx + dy*dy)

        #puts "#{r1} -- #{rr} -- #{r0}"

        if(r0 < rr && rr < r1)
            @hue = Math::atan2(dy,dx)
            #puts @hue
            damage_self
        end

        rt = r0 - 6

        tri_upper_x     = cx + rt*Math::cos(@hue)
        tri_upper_y     = cy + rt*Math::sin(@hue)

        tri_low_left_x  = cx + rt*Math::cos(@hue - (2*Math::PI)/3)
        tri_low_left_y  = cy + rt*Math::sin(@hue - (2*Math::PI)/3)

        tri_low_right_x = cx + rt*Math::cos(@hue + (2*Math::PI)/3)
        tri_low_right_y = cy + rt*Math::sin(@hue + (2*Math::PI)/3)

        _ux = tri_upper_x
        _uy = tri_upper_y
        _lx = tri_low_left_x
        _ly = tri_low_left_y
        _rx = tri_low_right_x
        _ry = tri_low_right_y

        _bx = (_lx+_rx)/2
        _by = (_ly+_ry)/2

        ul = dist(xx, yy, _ux, _uy, _lx, _ly)
        ur = dist(xx, yy, _rx, _ry, _ux, _uy)
        bt = dist(xx, yy, _lx, _ly, _rx, _ry)
        md = dist(xx, yy, _ux, _uy, _bx, _by)

        #puts "%%%  #{ul} -- #{ur} -- #{bt}"
        #puts "inside" if ul > 0 && ur > 0 && bt > 0

        range_y = Math::sqrt((_bx-_ux)*(_bx-_ux)+(_by-_uy)*(_by-_uy))
        range_x = Math::sqrt((_lx-_rx)*(_lx-_rx)+(_ly-_ry)*(_ly-_ry))/2



        #puts "y = #{bt/range_y} x = #{md/range_x}"
        if(ul > 0 && ur > 0 && bt > 0)
            #Yes, this is intentionally reversed
            @sy = md/range_x

            @sx = bt/range_y
            extra = Math::sin(Math::PI/6.0)
            @sx = @sx*(1+extra) - extra
            #puts "@sx=#{@sx}  @sy=#{@sy}"
            damage_self
        end
    ;end
  def dist(px, py, x1, y1, x2, y2);
        ((y2-y1)*px - (x2-x1)*py + x2*y1 - y2*x1)/Math::sqrt((y2-y1)*(y2-y1) + (x2-x1)*(x2-x1))
    ;end
  def onMouseMove(ev);
        onMousePress(ev)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous116 = self

    
context = Hash.new

    #puts "ColorSel, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ConfirmButton < Qml::Button
  attr_reader :toggle
  attr_accessor :valueRef
  def setValue(x);
        self.value = x
        whenValue.call if whenValue
        damage_self
    ;end
  def onMousePress(ev);
        if(self.label != "confirm" && self.value)
            @true_label = self.label
            self.label = "confirm"
            damage_self
            return
        elsif(self.label == "confirm")
            self.label = @true_label
            damage_self
        end
        self.value = !self.value
        self.valueRef.value = self.value if self.valueRef
        damage_self
        whenValue.call if whenValue
    ;end
  def onMouseEnter(ev);
        if(self.tooltip.nil? || self.tooltip.empty?)
            self.root.log(:tooltip, "extern = <" + self.extern + ">")
        else
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @toggle = self

    
context = Hash.new
context["toggle"] = self
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, toggle.extern)
        toggle.label   = meta.short_name if toggle.label.empty?
        toggle.tooltip = meta.tooltip
        toggle.valueRef = OSC::RemoteParam.new($remote, toggle.extern)
        toggle.valueRef.callback = lambda {|x| toggle.setValue(x)}
    
end})

    #puts "ConfirmButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::CopyButton < Qml::Button
  attr_reader :copy_button
  attr_property :index
  def layout(l, selfBox);
        l.aspect(selfBox, 1, 1)
        selfBox
    ;end
  def animate();
        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.9
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        if($remote && extern && !extern.empty?)
            if(self.index)
                $remote.action("/presets/copy", extern, self.index)
            else
                $remote.action("/presets/copy", extern)
            end
        end
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @copy_button = self

    
context = Hash.new
context["copy_button"] = self
Qml::prop_add(self, "index")
self.index = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        copy_button.tooltip = "copy from " + copy_button.extern.to_s
    
end})
self.label = "C"

    #puts "CopyButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::DataView < Qml::Widget
  attr_property :data
  attr_property :normal
  attr_accessor :pad
  attr_property :fixedpad
  attr_property :phase
  attr_property :ignore_phase
  def class_name(); "DataView" ;end
  def draw(vg);
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)

        vphase = ignore_phase ? 0 : phase
        if(data.class == Array && data[0].class == Float)
            Draw::WaveForm::plot(vg, self.data, box, normal, vphase)
        elsif(data.class == Array && data[0] == -40)
            Draw::WaveForm::plot(vg, self.data, box, normal, vphase)
        elsif(data.class == Array && data[0] == 0 && data[5] == 0)
            Draw::WaveForm::plot(vg, self.data, box, false, vphase)
        elsif(data.class == Array && data[0].class == Array)
            Draw::WaveForm::plot(vg, self.data[0], box, normal, vphase)
            Draw::WaveForm::plot(vg, self.data[1], box, normal, vphase)
        else
            Draw::WaveForm::sin(vg, box)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous119 = self

    
context = Hash.new
Qml::prop_add(self, "data")
Qml::prop_add(self, "normal")
Qml::prop_add(self, "fixedpad")
Qml::prop_add(self, "phase")
Qml::prop_add(self, "ignore_phase")
self.layer = 1
self.data = nil
self.normal = true
self.pad = 1/32
self.fixedpad = 0
self.phase = 0
self.ignore_phase = false

    #puts "DataView, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::DataViewAlt < Qml::Widget
  attr_reader :draw_alt
  attr_property :points
  def mk_points();
        pts = []
        while(pts.length < 256)
            pts << 0
        end
        pts
    ;end
  def draw(vg);
        pad = 5
        box = Rect.new(pad, pad, w-2*pad, h-2*pad)
        Draw::WaveForm::plot(vg, self.points, box, false)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @draw_alt = self

    
context = Hash.new
context["draw_alt"] = self
Qml::prop_add(self, "points")
self.layer = 1
@db.connect_property(self.properties["points"], lambda {draw_alt.mk_points})

    #puts "DataViewAlt, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::DemoLayers < Qml::Widget
  attr_reader :demo
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            grad = v.linear_gradient(0,0,512,512,NVG.rgba(0,0,0,255),
            NVG.rgba(rand(255),rand(255),rand(255),255))
            v.fill_paint = grad
            v.fill
        end
    ;end
  def onMousePress(ev); ;end
  def onMouseHover(ev);
        #pos = ev.pos
        #drect = Rect.new(pos.x-2, pos.y-2, 16, 16)
        #root.draw_damage(drect, demo.layer)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @demo = self
@anonymous121 = Qml::Knob.new(database, ui_path+"1"+'/')
@anonymous122 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous123 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous124 = Qml::Knob.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["demo"] = self
@anonymous121.x = 100
@anonymous121.y = 200
@anonymous121.w = 100
@anonymous121.h = 100
Qml::add_child(self, @anonymous121)
@anonymous122.x = 300
@anonymous122.y = 200
@anonymous122.w = 100
@anonymous122.h = 100
Qml::add_child(self, @anonymous122)
@anonymous123.x = 400
@anonymous123.y = 200
@anonymous123.w = 100
@anonymous123.h = 100
Qml::add_child(self, @anonymous123)
@anonymous124.x = 400
@anonymous124.y = 300
@anonymous124.w = 100
@anonymous124.h = 100
Qml::add_child(self, @anonymous124)

    #puts "DemoLayers, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::DummyBox < Qml::Widget
  def draw(vg);
        bb = Rect.new(0, 0, w, h)
        Draw::GradBox(vg, bb)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous126 = self

    
context = Hash.new

    #puts "DummyBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::FancyButton < Qml::Widget
  attr_reader :fancy
  attr_reader :pow
  attr_accessor :whenClick
  attr_property :value
  attr_accessor :valueRef
  def class_name(); "FancyButton" ;end
  def set_value(x);
        self.value = x
        damage_self
    ;end
  def set_sub_value(x);
        pow.value = x
        damage_self
    ;end
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def draw(vg);
        grad = if(value)
            bright_green = color("00AE9C")
            dark_green   = color("007368")
            vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        else
            grey1 = Theme::ButtonGrad1
            grey2 = Theme::ButtonGrad2
            vg.linear_gradient(0,0,0,h, grey1, grey2)
        end
        vg.stroke_color color("000000", 0xa0)
        vg.path do |v|
            v.rounded_rect(0,0,w/4,h, 5)
            v.fill_paint(grad)
            v.fill
            v.stroke_width 1
            v.stroke
        end

        grey1 = Theme::ButtonGrad1
        grey2 = Theme::ButtonGrad2
        grad2 = vg.linear_gradient(0,0,0,h, grey1, grey2)
        vg.path do |v|
            pad = 0
            v.rounded_rect(w/4+pad,0,w*3/4,h, 3)
            if(value)
                v.fill_paint grad
            else
                v.fill_paint grad2
            end
            v.fill
            v.stroke
        end

        text_color1   = Theme::TextActiveColor
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*0.6
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        if(value)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text(w*1.25/2,h/2,label.upcase)
    ;end
  def onMousePress(ev);
        #self.value = !self.value
        #damage_self
        valueRef.value = true if(value == false && valueRef)
        whenClick.call if whenClick
    ;end
  def layout(l,selfBox);
        box = pow.fixed(l, selfBox, 0, 0, 0.25, 1)
        if(layoutOpts.include? :no_constraint)
            box.w = 0
            box.y = 0
            box.w = 0.25*selfBox.w
            box.h = selfBox.h
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @fancy = self
@pow = Qml::PowButton.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["fancy"] = self
context["pow"] = @pow
Qml::prop_add(self, "value")

        @pow.instance_eval("def class_name(); \"PowButton\" ;end", "src/mruby-zest/example/FancyButton.qml", 105)
Qml::add_child(self, @pow)
self.whenClick = nil
self.value = false
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, fancy.extern)
        fancy.tooltip = meta.tooltip

        fancy.valueRef = OSC::RemoteParam.new($remote, fancy.extern)
        fancy.valueRef.callback = Proc.new {|x| fancy.set_sub_value(x)}
        pow.extern = fancy.extern
        pow.extern()
    
end})

    #puts "FancyButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::FileButton < Qml::TriggerButton
  attr_reader :filebutton
  attr_property :extval
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        whenValue.call if whenValue
        damage_self
    ;end
  def setup_widget(w);
        w.onSetup()
        w.children.each do |c|
            setup_widget(c)
        end
    ;end
  def file_select();
        win = window()
        wid = Qml::FileSelector.new(db)
        wid.whenValue = lambda { |x| filebutton.file_value(x)}
        wid.x = 0#-global_x
        wid.y = 0#-global_y
        wid.w = win.w
        wid.h = win.h
        wid.ext = self.extval
        Qml::add_child(win, wid)
        self.db.update_values
        setup_widget wid
        self.db.update_values

        if(root)
            root.smash_layout
            root.damage_item(win, :all)
        end
    ;end
  def file_value(val);
        return if val == :cancel
        $remote.action(self.extern, val)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @filebutton = self

    
context = Hash.new
context["filebutton"] = self
Qml::prop_add(self, "extval")
@db.connect_property(self.properties["extval"], lambda {".scl"})
self.whenValue = lambda { filebutton.file_select() }

    #puts "FileButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::FileSelector < Qml::Widget
  attr_reader :file
  attr_reader :folders
  attr_reader :files
  attr_reader :line
  attr_reader :ebutton
  attr_reader :favs
  attr_reader :add
  attr_accessor :valueRef
  attr_accessor :whenValue
  attr_property :ext
  attr_property :pat
  def draw_text(vg, w, h, text, textScale=0.8);
        vg.font_face("bold")
        vg.font_size h*textScale
        vg.fill_color color("56c0a5")
        if(layoutOpts.include? :left_text)
            vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
            vg.text(8,h/2,text.upcase)
        else
            vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            vg.text(w/2,h/2,text.upcase)
        end
    ;end
  def draw_button(vg, w, h, pad);
        vg.path do |v|
            v.rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad))
            v.stroke_color color("56c0a5")
            v.stroke_width 1
            v.stroke
        end
    ;end
  def layout(l, selfBox);
        mockx   = 1181
        mocky   = 659
        xpad    = 280/mockx
        xskip   = 20/mockx

        top_y  = 30/mocky
        top_h  = 34/mocky
        top_w  = (813-280)/mockx
        top_x  = xpad
        top_x2 = top_x+top_w+xskip
        top_w2 = 1-xpad-top_x2

        ff_y    = 85/mocky
        ff_h    = (573-85)/mocky
        ff_w    = 0.5-xskip/2-xpad
        ff_x    = xpad
        ff_x2   = 0.5+0.5*xskip

        fv_y  = 594/mocky
        fv_h  = (628-594)/mocky
        fv_x  = xpad
        fv_x2 = 2*xpad
        #values from mockup
        #fv_w  = 0.1
        #fv_w2 = 0.1
        #values to not make the current revision look dumb
        fv_w  = 0.2
        fv_w2 = 0.2

        sel_folder = children[0]
        sel_column = children[1]
        line       = children[2]
        select     = children[3]
        favs       = children[4]
        add_favs   = children[5]
        sel_folder.fixed(l, selfBox, ff_x,   ff_y,  ff_w,   ff_h)
        sel_column.fixed(l, selfBox, ff_x2,  ff_y,  ff_w,   ff_h)
        line.fixed(l,       selfBox, top_x,  top_y, top_w,  top_h)
        select.fixed(l,     selfBox, top_x2, top_y, top_w2, top_h)
        favs.fixed(l,       selfBox, fv_x,   fv_y,  fv_w,   fv_h)
        add_favs.fixed(l,   selfBox, fv_x2,  fv_y,  fv_w2,  fv_h)
        selfBox
    ;end
  def browser(); @browser ;end
  def onSetup(v=nil);
        @browser = FileBrowser.new

        #Get files when home dir (or any subsequent dir) is setup
        dirs  = OSC::RemoteParam.new($remote, "/file_list_dirs")
        dirs.callback   = lambda { |x|
            file.browser.set_dirs(x)
            file.check
        }
        files = OSC::RemoteParam.new($remote, "/file_list_files")
        files.callback = lambda { |x|
            file.browser.set_files(x)
            file.check
        }


        #Get the starting path i.e. the HOME dir
        home = nil
        if($current_dir.nil?)
            home  = OSC::RemoteParam.new($remote, "/file_home_dir")
            home.callback = lambda { |x|
                file.browser.set_home_dir(x)
                file.check
            }
        else
            @browser.set_home_dir($current_dir)
        end

        fav   = OSC::RemoteParam.new($remote, "/config/favorites")
        fav.callback  = lambda { |x| set_favs(x) }

        self.valueRef = [dirs, files, home]
    ;end
  def set_favs(x);
        return if x.nil? || x.empty?

        favs.options =  x  if x.class == Array
        favs.options = [x] if x.class != Array
        favs.damage_self
    ;end
  def addFav();
        #TODO
        $remote.action("/config/add-favorite", line.label)
        $remote.action("/config/favorites")
    ;end
  def clrFav();
        $remote.action("/config/clear-favorites")
        $remote.action("/config/favorites")
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 27) #esc
            whenCancel
            return
        elsif(k.ord == 9) #tab
            if(ebutton.value != true)
                ebutton.value = true
            else
                ebutton.value = 0.0
            end
            ebutton.damage_self
            line.label = line.label[0...-1]
            return
        elsif(k.ord == 13) #enter
            line.label = line.label[0...-1]
            whenEnter
            return
        elsif(k.ord == 8)
            @browser.del_char
        else
            @browser.add_char k
        end

        check
    ;end
  def check();
        return if !@browser.needs_refresh
        @browser.clear_flags

        path = @browser.path
        if self.ext && !path.end_with?(self.ext)
            line.label = [@browser.path, self.ext]
        else
            line.label = @browser.path
        end
        line.damage_self
        line.instance_eval("undef :onKey") if line.respond_to?(:onKey)

        search = @browser.search_path

        $remote.action("/file_list_dirs",  search)
        $remote.action("/file_list_files", search)
    ;end
  def draw(vg);
        background color("000000", 220)
    ;end
  def whenEnter();
        ll = @browser.path
        ll = ll + self.ext if self.ext && !ll.end_with?(self.ext)
        whenValue.call(ll) if whenValue
        root.ego_death self
    ;end
  def whenCancel();
        whenValue.call(:cancel) if whenValue
        root.ego_death self
    ;end
  def onMousePress(m);
        whenCancel
    ;end
end
class Qml::SelColumn < Qml::Widget
  attr_reader :col
  attr_reader :scroll
  attr_accessor :valueRef
  attr_property :number
  attr_property :skip
  attr_property :rows
  attr_property :nrows
  attr_accessor :whenValue
  attr_property :oldOff
  attr_property :value_sel
  attr_property :value_lab
  attr_property :clear_on_extern
  attr_property :doupcase
  attr_property :pattern
  attr_property :style
  def clear();
        clear_sel
        children.each do |c|
            c.value = false if c.class != Qml::ScrollBar
            c.damage_self
        end
    ;end
  def clear_sel();
        self.value_sel = nil
        self.value_lab = nil
    ;end
  def onScroll(ev);
        scroll.onScroll(ev)
    ;end
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end

    ;end
  def draw_overlay(vg);
        vg.path do
            vg.rect(0, 0, w, h)
            vg.stroke_color  color("56c0a5")
            vg.stroke_width 1.0
            vg.stroke
        end
        titleH = self.h*1.0/(children.length + 1)
        vg.fill_color Theme::TextModColor
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.font_size titleH*0.9
        vg.text(0.9*w/2,titleH/2,label.upcase)
    ;end
  def draw_normal(vg);
        titleH = self.h*1.0/(children.length + 1)
        vg.path do
            vg.rect(0, 0, 0.9*w, titleH)
            vg.fill_color Theme::BankOdd
            vg.fill
        end
        vg.fill_color Theme::TextModColor
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.font_size titleH*0.9
        vg.text(0.9*w/2,titleH/2,label.upcase)
    ;end
  def onSetup(old=nil);
        return if children.length > 10
        scroll.layer = self.layer
        (self.nrows).times do |x|
            ch = if(number)
                Qml::NumberButton.new(db)
            else
                Qml::SelButton.new(db)
            end
            if(rows.nil? || rows.empty?)
                ch.label = ""
            else
                stride = 1
                stride = 2 if skip
                ch.label = if(x>=rows.length/stride)
                    ""
                else
                    rows[x*stride]
                end
                ch.tooltip = if(x*stride+1>=rows.length)
                    ""
                else
                    rows[x*stride+1]
                end

            end
            ch.layoutOpts = [:no_constraint]
            ch.whenValue  = lambda {col.cb ch}

            ch.layer      = col.layer
            ch.number = x if number
            ch.style    = self.style    if ch.respond_to? :style
            ch.doupcase = self.doupcase if ch.respond_to? :doupcase
            if(!number)
                ch.pad = 0
                ch.bg = Theme::BankEven if x%2 == 0
                ch.bg = Theme::BankOdd if x%2 == 1
            end
            Qml::add_child(self, ch)
        end

    ;end
  def tryScroll();
        #24 items are visible at a time
        #center is at item nrows/2     at value 0
        #center is at item len-nrows/2 at value 1
        stride = 1
        stride = 2 if skip
        return if rows.nil?
        n = rows.length/stride
        return if n<nrows
        center = (n-nrows)*(1-scroll.value)+nrows/2
        off    = (center-nrows/2).to_i
        if(off != self.oldOff)
            setValue(rows, off)
        end
    ;end
  def cb(ch);
        if(ch.value)
            self.value_sel = ch.tooltip
            self.value_lab = ch.label
        else
            self.value_sel = nil
            self.value_lab = nil
        end

        children[1,99].each do |child|
            if(ch != child)
                child.value = false
            end
        end
        whenValue.call if whenValue
        damage_self
    ;end
  def filter(x);
        y = []
        x.each do |x|
            next if(x != "." && x != ".." && x[0] == ".")
            next if(x[-1] == "~")
            next if self.pattern && !self.pattern.match(x)
            y << x
        end
        y
    ;end
  def setValue(x,offset=0);
        return if(self.rows == x && offset == oldOff)
        self.oldOff = offset
        self.rows = x
        stride = 1
        stride = 2 if skip
        nn = x.length/stride
        nsize = [0.05, [1.0, self.nrows/nn].min].max
        if(nsize != scroll.bar_size)
            scroll.bar_size = nsize
            scroll.damage_self
        end
        n = [x.length/stride, children.length-1].min
        (1..n).each do |i|
            nv = (children[i].label == self.value_lab &&
                  children[i].tooltip == self.value_sel)
            children[i].label   = x[(i-1+offset)*stride]
            children[i].tooltip = x[(i-1+offset)*stride+1] if stride == 2
            children[i].value   = nv
        end
        ((n+1)...children.length).each do |i|
            children[i].label = ""
            children[i].value = false
        end
        damage_self
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.9, 0.0, 0.1, 1.0)
        miniBox = nil
        if(selfBox.class != LayoutConstBox)
            miniBox = l.genBox :selminibox, nil
            l.fixed(miniBox, selfBox, 0.0, 0.0, 0.9, 1.0)
        else
            miniBox = l.genConstBox(0.0, 0.0, 0.9*selfBox.w, selfBox.h)
        end
        titleH  = 1.0/(children.length + 1)
        Draw::Layout::vpack(l, miniBox, children[1,99], 0, 1, 0, titleH, 1.0-titleH)
        selfBox
    ;end
  def selected();
        return "" if(value_lab.nil?)
        return value_lab
    ;end
  def selected_val();
        return "" if(value_sel.nil?)
        return value_sel
    ;end
  def selected_id();
        self.children.each_with_index do |c,i|
            next       if c.class == Qml::ScrollBar
            return i-1 if c.value
        end
        nil
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @col = self
@scroll = Qml::ScrollBar.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["col"] = self
context["scroll"] = @scroll
Qml::prop_add(self, "number")
Qml::prop_add(self, "skip")
Qml::prop_add(self, "rows")
Qml::prop_add(self, "nrows")
Qml::prop_add(self, "oldOff")
Qml::prop_add(self, "value_sel")
Qml::prop_add(self, "value_lab")
Qml::prop_add(self, "clear_on_extern")
Qml::prop_add(self, "doupcase")
Qml::prop_add(self, "pattern")
Qml::prop_add(self, "style")
@db.connect_property(@scroll.properties["style"], lambda {col.style})
@scroll.whenValue = lambda { col.tryScroll }
Qml::add_child(self, @scroll)
self.valueRef = nil
self.number = false
self.skip = false
self.rows = nil
self.nrows = 24
self.whenValue = nil
self.oldOff = 0
self.value_sel = nil
self.value_lab = nil
self.clear_on_extern = nil
self.doupcase = true
self.pattern = nil
self.style = :normal
@db.connect_watcher(self.properties["extern"], lambda {begin

        col.valueRef = OSC::RemoteParam.new($remote, col.extern)
        col.valueRef.callback = Proc.new {|x|
            x = [x] if x.class != Array
            col.clear_sel if col.clear_on_extern
            col.setValue(col.filter(x)) if !col.skip
            col.setValue(x)             if  col.skip
        }
    
end})

    #puts "SelColumn, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextField < Qml::Widget
  attr_reader :text_field
  attr_accessor :valueRef
  attr_property :style
  attr_accessor :whenValue
  def animate();
        if(root.key_widget != self)
            @next = nil
            if(@state)
                @state = nil
                damage_self
            end
            return
        end
        now = Time.new
        if(@next.nil?)
            @next = now + 0.1
            return
        elsif(@next < now)
            @state = !@state
            @next = now + 0.7
            damage_self
        end
    ;end
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end

    ;end
  def draw_overlay(vg);
        background color("1b1c1c")
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = color("56c0a5")
        l = "..."
        if(label.class == String && !label.empty?)
            l = label.clone
        elsif(label.class == Array)
            l = label[0].clone
        end

        (0...l.length).each do |i|
            l[i] = "?" if l.getbyte(i) > 127
        end

        bnd = vg.text_bounds(0,0,l+"|")
        if(bnd+8 > self.w)
            vg.font_size self.h*self.w/(bnd+8)*0.8
            bnd = vg.text_bounds(0,0,l)
        end

        vg.text(8,h/2,l)
        if(@state)
            vg.text(8+bnd,h/2,"|")
        end
    ;end
  def draw_normal(vg);
        background Theme::GeneralBackground
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = Theme::TextColor
        l = label.empty? ? "..." : label.clone

        (0...l.length).each do |i|
            l[i] = "?" if l.getbyte(i) > 127
        end

        vg.text(8,h/2,l)
        bnd = vg.text_bounds(0,0,l)
        if(@state)
            vg.text(8+bnd,h/2,"|")
        end
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 8)
            self.label = self.label[0...-1]
        else
            self.label += k
        end
        whenValue.call if whenValue
        valueRef.value = self.label if valueRef
        damage_self
    ;end
  def onMousePress(ev);
    ;end
  def onMerge(val);
        self.label = val.label
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text_field = self

    
context = Hash.new
context["text_field"] = self
Qml::prop_add(self, "style")
self.valueRef = nil
self.style = :normal
self.whenValue = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        ref = OSC::RemoteParam.new($remote, text_field.extern)
        ref.callback = lambda {|x|
            text_field.label = x;
            text_field.damage_self
        }
        text_field.valueRef = ref
    
end})

    #puts "TextField, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Menu < Qml::Widget
  attr_property :extern
  attr_accessor :options
  attr_accessor :selected
  attr_accessor :whenValue
  def set_value_user(val);
        self.selected = val
        whenValue.call if whenValue
    ;end
  def layout(l,selfBox);
        if(!layoutOpts.include?(:no_constraint))
            scale = 100
            $vg.font_size scale
            bb = 1
            self.options.each do |x|
                x = x[0..8] if x.length > 8
                bbl  = $vg.text_bounds(0, 0, (x+"    ").upcase)
                bb   = [bb, bbl].max
            end
            scale *= layoutOpts[0] if layoutOpts.include?(:rescale)
            if(bb != 0)
                #Width cannot be so small that letters overflow
                l.sh([selfBox.w, selfBox.h], [-1.0, bb/scale], 0)
            end
        end
        selfBox
    ;end
  def class_name();
        "Menu"
    ;end
  def onMousePress(ev);
        create_menu
    ;end
  def draw(vg);
        text_color = Theme::TextColor
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            tx = w*pad2-h*pad2
            tw = h*pad2
            ty = h*pad
            th = h*pad2

            v.move_to(tx+0.25*tw, ty+0.3*th)
            v.line_to(tx+0.50*tw, ty+0.7*th)
            v.line_to(tx+0.75*tw, ty+0.3*th)
            v.close_path

            v.fill_color Theme::TextColor
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color text_color
        vg.text(3+w*pad*2,h/2,label.upcase)
    ;end
  def create_menu();
        n = options.length
        n = 1 if options.class != Array
        return if n < 1
        widget = DropDown.new(self.db)
        widget.w = 3*self.w
        widget.h = self.h*n
        widget.x = 0
        widget.y = 0
        widget.layer = 2
        widget.options = options
        widget.callback = lambda { |v|
            set_value_user(v)
        }
        widget.y = -self.h*(n-1) if(widget.h+global_y > window.h)
        widget.prime(root())

        Qml::add_child(self, widget)
        root.smash_draw_seq
        root.damage_item widget
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous165 = self

    
context = Hash.new
Qml::prop_add(self, "extern")
self.extern = ""
self.options = ["text", "test", "asdf"]
self.selected = 0
self.whenValue = nil

    #puts "Menu, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::FileSelector
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @file = self
@folders = Qml::SelColumn.new(database, ui_path+"1"+'/')
@files = Qml::SelColumn.new(database, ui_path+"2"+'/')
@line = Qml::TextField.new(database, ui_path+"3"+'/')
@ebutton = Qml::TriggerButton.new(database, ui_path+"4"+'/')
@favs = Qml::Menu.new(database, ui_path+"5"+'/')
@add = Qml::TriggerButton.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["file"] = self
context["folders"] = @folders
context["files"] = @files
context["line"] = @line
context["ebutton"] = @ebutton
context["favs"] = @favs
context["add"] = @add
@add.instance_eval do
def file=(k);@file=k;end
def file;    @file;  end
def favs=(k);@favs=k;end
def favs;    @favs;  end
def add=(k);@add=k;end
def add;    @add;  end
end
Qml::context_apply(context, @add)
Qml::prop_add(self, "ext")
Qml::prop_add(self, "pat")
@folders.layer = 2
@folders.layoutOpts = [:no_constraint]
@db.connect_property(@folders.properties["extern"], lambda {"/file_list_dirs"})
@folders.style = :overlay
@folders.label = "Folder"
@folders.doupcase = false
@folders.nrows = 20
@folders.clear_on_extern = true
@folders.whenValue = lambda {
            return if folders.selected.empty?
            file.browser.change_dir_rel(folders.selected)
            file.check
        }
Qml::add_child(self, @folders)
@files.layer = 2
@files.layoutOpts = [:no_constraint]
@db.connect_property(@files.properties["extern"], lambda {"/file_list_files"})
@files.style = :overlay
@files.label = "Files"
@files.nrows = 20
@files.doupcase = false
@files.clear_on_extern = true
@db.connect_property(@files.properties["pattern"], lambda {Regexp.new(file.pat) if file.pat})
@files.whenValue = lambda {
            return if files.selected.empty?
            file.browser.change_file(files.selected)
            file.check
        }
Qml::add_child(self, @files)
@line.style = :overlay
@line.layoutOpts = [:no_constraint]
@line.layer = 2
Qml::add_child(self, @line)

        @ebutton.instance_eval("def draw(vg);\n            parent.draw_button(vg, w, h, self.pad)\n            parent.draw_text(vg, w, h, self.label, self.textScale)\n        ;end", "src/mruby-zest/example/FileSelector.qml", 55)
@ebutton.layoutOpts = [:no_constraint]
@ebutton.layer = 2
@ebutton.label = "Enter"
@ebutton.whenValue = lambda {file.whenEnter}
Qml::add_child(self, @ebutton)
@favs.options = []
@favs.layer = 2
@favs.layoutOpts = [:no_constraint]
@favs.label = "favorites"
@favs.whenValue = lambda {
            file.browser.change_dir_abs(favs.options[favs.selected]) if favs.selected
            file.check
        }
Qml::add_child(self, @favs)

        @add.instance_eval("def onMousePress(ev);\n            add = \"add favorite\"\n            clr = \"clear favs?\"\n            if(ev.buttons.include? :leftButton)\n                file.addFav\n                self.label = add\n                self.whenValue.call if self.whenValue\n            elsif(ev.buttons.include? :rightButton)\n                if(self.label == clr)\n                    file.clrFav\n                    self.label = add\n                else\n                    self.label = clr\n                end\n            end\n            damage_self\n        ;end", "src/mruby-zest/example/FileSelector.qml", 79)

        @add.instance_eval("def draw(vg);\n            parent.draw_button(vg, w, h, self.pad)\n            parent.draw_text(vg, w, h, self.label, self.textScale)\n        ;end", "src/mruby-zest/example/FileSelector.qml", 97)
@add.label = "add favorite"
@add.layer = 2
@add.layoutOpts = [:no_constraint]
Qml::add_child(self, @add)
self.layer = 2
self.valueRef = nil
self.whenValue = nil
self.ext = nil
self.pat = nil

    #puts "FileSelector, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Forms < Qml::Widget
  attr_reader :forms
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ModuleGrad1, Theme::ModuleGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color(color("000000"))
            v.stroke
        end
    ;end
  def onSetup(old=nil);
        return if @setup_done
        children[1..-1].each do |c|
            ch = Qml::TextBox.new(db)
            ch.label = c.tooltip
            ch.align = :left
            Qml::add_child(self, ch)
        end
        @setup_done = true
    ;end
  def layout(l, selfBox);
        n = (children.length + 1)/2

        l.fixed_long(children[0], selfBox,
            0, 0, 1, 1/n,
            0, 0, 0, 0)
        (1...n).each do |i|
            l.fixed_long(children[i+n-1], selfBox,
                0, i/n, 0.75, 0.5/n, 
                10, 0, 0, 0)
            l.fixed_long(children[i], selfBox,
                0.25, (i+0.5)/n, 0.75, 0.5/n, 
                0, 0, -10, 0)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @forms = self
@anonymous137 = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["forms"] = self
@db.connect_property(@anonymous137.properties["label"], lambda {forms.label})
Qml::add_child(self, @anonymous137)

    #puts "Forms, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::GroupHeader < Qml::Widget
  attr_reader :titleW
  attr_reader :title
  attr_accessor :whenClick
  attr_property :toggleable
  attr_accessor :copyable
  def onMousePress(ev);
        whenClick.call if whenClick
    ;end
  def class_name(); "TitleBox" ;end
  def pack_misc(l, selfBox, ch);
        #Create A list of boxes
        bbs = []
        n = children.length
        children.each_with_index do |ch, i|
            box = l.genConstBox(selfBox.w*i/n, 0,
            selfBox.w/n, selfBox.h, ch)
            bb = ch.layout(l, box)
            bbs << bb
        end

        bbs[0].x = 15 if bbs[0].x > 15

        #put stuff on backwards
        bx = selfBox.w
        bbs[1..-1].reverse.each do |b|
            b.x  = bx-b.w
            bx  -= b.w
        end

        return selfBox
    ;end
  def layout(l, selfBox);
        pack_misc(l, selfBox, children)
    ;end
  def onSetup(old=nil);
        return if children.length != 1
        if(self.toggleable)
            pb = Qml::PowButton.new(db)
            pb.extern = self.toggleable
            Qml::add_child(self, pb)
        end
        if(self.copyable)
            cb = Qml::CopyButton.new(db)
            pb = Qml::PasteButton.new(db)
            cb.extern = self.extern
            pb.extern = self.extern
            Qml::add_child(self, cb)
            Qml::add_child(self, pb)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @titleW = self
@title = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["titleW"] = self
context["title"] = @title
Qml::prop_add(self, "toggleable")
@db.connect_property(@title.properties["label"], lambda {self.label})
@title.align = :left
@title.textColor = Theme::TextModColor
@title.height = 0.8
Qml::add_child(self, @title)
self.whenClick = nil
self.toggleable = false
self.copyable = false

    #puts "GroupHeader, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::HSlider < Qml::Valuator
  attr_property :centered
  attr_accessor :pad
  attr_property :height
  attr_property :active
  def draw_centered(vg, pad);
        pad2 = (1-2*pad)
        va = 1-value
        if(va >= 0.5)
            vv   = [0.01, va-0.5+0.01].max
            src = (w*0.5-w*pad2*vv)
            dst = (w*0.51)
            vg.path do |v|
                v.rect(src, pad*h, (src-dst).abs, pad2*h)
                v.fill_color Theme::SliderActive
                v.fill
            end
        else
            vv   = [0.01, 0.5-va+0.01].max
            vg.path do |v|
                v.rect(w*0.49, pad*h, pad2*w*vv, pad2*h)
                v.fill_color Theme::SliderActive
                v.fill
            end
        end
    ;end
  def draw(vg);
        self.dragScale = w
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(pad*w, pad*h, pad2*w, pad2*h)
            v.fill_color Theme::SliderBackground
            v.stroke_color color(:black)
            v.fill
            v.stroke
        end

        if(centered)
            draw_centered(vg, pad)
        else
            vg.path do |v|
                v.rect(pad*w, pad*h, value*w*pad2, pad2*h)
                v.fill_color Theme::SliderActive
                v.fill
            end
        end

        vg.path do |v|
            v.move_to(w*pad+w*pad2*value, pad*h)
            v.line_to(w*pad+w*pad2*value, pad2*h)
            v.stroke_color Theme::SliderStroke
            v.stroke_width 2.0
            v.stroke
        end

        vg.font_face("bold")
        vg.font_size height*h
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.fill_color(Draw::fade(Theme::TextColor))
        vg.text(w/2,h/2,label.upcase)

        if(!self.active)
            vg.path do
                vg.move_to(pad*w, pad*h)
                vg.line_to(pad2*w, pad2*h)
                vg.stroke_color Theme::SliderStroke
                vg.stroke
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous141 = self

    
context = Hash.new
Qml::prop_add(self, "centered")
Qml::prop_add(self, "height")
Qml::prop_add(self, "active")
self.centered = false
self.pad = 0.02
self.height = 0.6
self.active = true
self.vertical = false

    #puts "HSlider, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::HarmonicEdit < Qml::Widget
  attr_reader :hedit
  attr_accessor :whenValue
  attr_property :type
  attr_property :totalHarm
  attr_property :oldOff
  attr_property :curOff
  def cb(); whenValue.call if whenValue ;end
  def clear();
        children.each_with_index do |ch, i|
            mval = 64.1/127.0 if(type == :oscil)
            mval = 0/127.0  if(type == :subsynth)
            pval = 64.1/127.0
            mval = 1.0 if i == 0
            ch.children[0].value          = mval
            ch.children[0].valueRef.value = mval
            ch.children[2].value          = pval
            ch.children[2].valueRef.value = pval
        end
    ;end
  def animate();
        return if(self.oldOff == self.curOff)
        off = self.curOff
        self.oldOff = self.curOff

        children.each_with_index do |ch, i|
            ii = (i+off).to_i.to_s
            ch.children[1].label = ii
            if(self.type == :oscil)
                ch.children[0].extern = self.extern + "magnitude" + ii
                ch.children[2].extern = self.extern + "phase"     + ii
            end
        end
        damage_self
    ;end
  def set_scroll(val);
        len  = totalHarm - 32
        off  = 32*(val*len/32).to_i

        return if(self.oldOff == off)
        self.curOff = off
    ;end
  def onSetup(old=nil);
        return if !children.empty?
        (0...32).each do |ev|
            hm           = Qml::HarmonicEditSingle.new(db)
            hm.num       = ev
            hm.extern    = self.extern
            hm.slidetype = self.type
            hm.whenValue = lambda {hedit.cb}
            Qml::add_child(self, hm)
        end
    ;end
  def layout(l, selfBox);
        n = children.length
        children.each_with_index do |ch, id|
            ch.fixed(l, selfBox, id/n, 0, 1.0/n, 1.0)
        end

        selfBox
    ;end
  def draw(vg);
        return
        vg.path do |v|
            v.rect(0,0,self.w,self.h)
            v.fill_color color("ff600f")
            v.fill
        end
    ;end
  def onMousePress(ev);
        @active_widget  = root.activeWidget(ev.pos.x, ev.pos.y)
        @active_y       = ev.pos.y
        @active_buttons = ev.buttons
        return if @active_widget.class != Qml::Slider
        $remote.automate(@active_widget.extern) if(root.learn_mode && @active_widget.extern)

        if(ev.buttons.include? :leftButton)
            @prev = ev.pos
        elsif(ev.buttons.include? :rightButton)
            if(children.empty?)
                @active_widget.create_radial
            end
        elsif(ev.buttons.include? :middleButton)
            @active_widget.reset
        else
            @prev = ev.pos
        end
    ;end
  def onMouseMove(ev);
        fine = root.fine_mode ? 0.05 : 1.0
        nactive = root.activeWidget(ev.pos.x, ev.pos.y)
        if(@prev && @active_widget == nactive && @active_buttons.include?(:leftButton))
            delta = +(ev.pos.y - @prev.y)
            return if nactive.class != Qml::Slider
            @active_widget.updatePos(fine*delta/@active_widget.dragScale)
            @prev = ev.pos
        elsif(@prev && @active_buttons.include?(:leftButton))
            @active_widget = :dummy
            nactive = root.activeWidget(ev.pos.x, @active_y)
            return if nactive.class != Qml::Slider
            val = (ev.pos.y-nactive.global_y)/nactive.h
            nactive.updatePosAbs(1-val)
        elsif(@active_buttons.include?(:middleButton))
            @active_widget = :dummy
            nactive = root.activeWidget(ev.pos.x, @active_y)
            return if nactive.class != Qml::Slider
            nactive.reset
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @hedit = self

    
context = Hash.new
context["hedit"] = self
Qml::prop_add(self, "type")
Qml::prop_add(self, "totalHarm")
Qml::prop_add(self, "oldOff")
Qml::prop_add(self, "curOff")
self.whenValue = nil
self.type = :oscil
self.totalHarm = 128
self.oldOff = 0
self.curOff = 0

    #puts "HarmonicEdit, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::HarmonicEditSingle < Qml::ColorBox
  attr_reader :hes
  attr_accessor :whenValue
  attr_accessor :num
  attr_accessor :slidetype
  def cb();
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.0, 0.00, 1.0, 0.47)
        children[1].fixed(l, selfBox, 0.0, 0.48, 1.0, 0.08)
        children[2].fixed(l, selfBox, 0.0, 0.57, 1.0, 0.43)
        selfBox
    ;end
  def onSetup(v=nil);
        children.each do |c|
            c.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @hes = self
@anonymous143 = Qml::Slider.new(database, ui_path+"1"+'/')
@anonymous144 = Qml::Text.new(database, ui_path+"2"+'/')
@anonymous145 = Qml::Slider.new(database, ui_path+"3"+'/')

    
context = Hash.new
context["hes"] = self

        @anonymous143.instance_eval("def onSetup(old=nil);\n            undef :onMousePress if self.respond_to?(:onMousePress)\n            undef :onMouseMove  if self.respond_to?(:onMouseMove)\n        ;end", "src/mruby-zest/example/HarmonicEditSingle.qml", 26)
@anonymous143.value = 0.5
@anonymous143.pad = 0
@db.connect_property(@anonymous143.properties["centered"], lambda {hes.slidetype == :oscil})
@db.connect_property(@anonymous143.properties["extern"], lambda {begin

            if(hes.slidetype == :oscil)
                hes.extern + "magnitude" + hes.num.to_s
            else
                hes.extern + "Phmag" + hes.num.to_s
            end
        
end})
@anonymous143.whenValue = lambda {hes.cb}
Qml::add_child(self, @anonymous143)
@anonymous144.layoutOpts = [:ignoreAspect]
@anonymous144.height = 1.0
@db.connect_property(@anonymous144.properties["label"], lambda {(1+hes.num).to_s})
Qml::add_child(self, @anonymous144)

        @anonymous145.instance_eval("def onSetup(old=nil);\n            undef :onMousePress if self.respond_to?(:onMousePress)\n            undef :onMouseMove  if self.respond_to?(:onMouseMove)\n        ;end", "src/mruby-zest/example/HarmonicEditSingle.qml", 50)
@anonymous145.value = 0.5
@anonymous145.pad = 0
@db.connect_property(@anonymous145.properties["centered"], lambda {hes.slidetype == :oscil})
@db.connect_property(@anonymous145.properties["extern"], lambda {begin

            if(hes.slidetype == :oscil)
                hes.extern + "phase" + hes.num.to_s
            else
                hes.extern + "Phrelbw" + hes.num.to_s
            end
        
end})
@anonymous145.whenValue = lambda {hes.cb}
Qml::add_child(self, @anonymous145)
self.whenValue = nil
self.num = 0
self.slidetype = nil
self.bg = nil

    #puts "HarmonicEditSingle, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::HarmonicView < Qml::Widget
  attr_reader :harmonic_view
  attr_property :type
  attr_accessor :valueRef
  attr_property :points
  attr_property :shift
  def refresh();
        vr = self.valueRef
        if(vr)
            vr.each do |r|
                r.refresh
            end
        end
    ;end
  def make_points();
        pts = [1.0, 0.5, 0.0, 0.0, 0.7]
        loop {
            break if pts.length == 128
            pts << 0.0
        }
        pts
    ;end
  def draw(vg);
        pad  = 1/32
        pad2 = (1-2*pad)
        background Theme::VisualBackground
        bb = Rect.new(pad*w, pad*h, pad2*w, pad2*h)
        Draw::WaveForm::bar(vg, self.points, bb,Theme::HarmonicColor, self.shift)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @harmonic_view = self

    
context = Hash.new
context["harmonic_view"] = self
Qml::prop_add(self, "type")
Qml::prop_add(self, "points")
Qml::prop_add(self, "shift")
self.type = :osc
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        rem   = harmonic_view.extern
        type  = harmonic_view.type
        rharm = rem                       if(type == :osc)
        rharm = rem + "oscilgen/spectrum" if(type == :pad)

        ref = OSC::RemoteParam.new($remote, rharm)
        ref.callback = lambda {|x|
            #x = x.map {|x| Math.log10(x)}
            x = Draw::DSP::norm_harmonics(x)
            nharmonics = 64
            x = x[0..nharmonics-1] if x.length > nharmonics
            harmonic_view.points = x
            harmonic_view.damage_self
        }
        refs = [ref]
        if(type == :pad)
            ref = OSC::RemoteParam.new($remote,rem + "nhr")
            ref.callback = lambda {|x|
                harmonic_view.shift = x
                harmonic_view.damage_self
            }
            refs << ref
        end

        harmonic_view.valueRef = refs

    
end})
@db.connect_property(self.properties["points"], lambda {harmonic_view.make_points})
self.shift = nil

    #puts "HarmonicView, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Indent < Qml::Widget
  attr_reader :indent
  attr_accessor :pad
  def class_name(); "Indent" ;end
  def layout(l, selfBox);
        return selfBox if children.empty?

        child = self.children[0]

        chBox = l.genConstBox(pad, pad,
                selfBox.w-2*pad, selfBox.h-2*pad, child)
        child.layout(l, chBox)
        selfBox
    ;end
  def draw(vg);
        indent_color = color("232C36")
        vg.path do |v|
            v.rounded_rect(0, 0, w, h, 3)
            v.fill_color Theme::VisualBackground
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @indent = self

    
context = Hash.new
context["indent"] = self
self.pad = 2

    #puts "Indent, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::KitButton < Qml::Widget
  attr_reader :button
  attr_accessor :action
  attr_property :value
  attr_property :enable
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def set_enable(v);
        if(button.enable != v)
            button.enable = v
            damage_self
        end
    ;end
  def onMousePress(ev);
        if(ev.buttons.include? :leftButton)
            action.call(:change_view) if action
        elsif(ev.buttons.include? :middleButton)
            action.call(:toggle) if action
        end
        damage_self
    ;end
  def onMerge(val);
        button.value = val.value
    ;end
  def class_name();
        "KitButton"
    ;end
  def layout(l, selfBox);
        if(selfBox.nil?)
            t = widget.class_name.to_sym
            selfBox = l.genBox t, widget
        end
        selfBox
    ;end
  def draw(vg);
        off_color     = Theme::ButtonInactive
        on_color      = Theme::ButtonActive
        outline_color = color("0089b9")
        text_color1   = Theme::TextActiveColor
        text_color2   = color("B9CADE")
        #TODO check textcolor2 and outline
        pad = 1/32
        vg.path do |v|
            v.rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad))
            if(button.value)
                v.fill_color on_color
            else
                v.fill_color off_color
            end
            v.fill
            if(self.enable)
                v.stroke_color(outline_color)
                v.stroke_width 1
                v.stroke
            end
        end

        vg.font_face("bold")
        vg.font_size (h*0.75).to_i
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        if(value)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text(w/2,h/2,button.label.upcase)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @button = self

    
context = Hash.new
context["button"] = self
Qml::prop_add(self, "value")
Qml::prop_add(self, "enable")
self.action = nil
self.value = false
self.enable = false
self.tooltip = "Use Middle Mouse To Toggle"

    #puts "KitButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::KitButtons < Qml::Widget
  attr_reader :kitbuttons
  attr_property :rows
  attr_property :sym
  attr_accessor :valueRef
  def set_view();
        prt  = root.get_view_pos(:part)
        kid  = root.get_view_pos(:kit)

        self.valueRef.each do |v|
            v.clean
        end

        v = []
        (0...rows).each do |r|
            [0,1,2,3].each do |c|
                ii         = (1+c + 4*r)
                but  = children[ii-1]
                path = ""
                path = "/part#{prt}/kit#{ii-1}/Penabled" if self.sym == :kit
                path = "/part#{ii-1}/Penabled" if self.sym == :part
                path = "/part#{prt}/kit#{kid}/adpars/VoicePar#{ii-1}/Enabled" if self.sym == :voice

                rr = OSC::RemoteParam.new($remote, path)
                rr.callback = lambda {|vv| but.set_enable(vv)}
                v << rr
            end
        end
        self.valueRef = v
    ;end
  def onSetup(old=nil);
        return if children.length > 0
        (0...rows).each do |r|
            [0,1,2,3].each do |c|
                ii         = (1+c + 4*r)
                but        = Qml::KitButton.new(db)
                but.label  = ii.to_s
                but.action = lambda {|act|
                    if(act == :change_view)
                        kitbuttons.change_active(ii-1)
                    else
                        kitbuttons.toggle(ii-1)
                    end
                }
                Qml::add_child(self, but)
            end
        end

        set_view
    ;end
  def set_kit_enable(ii, val);
        return if ii < 0
        return if ii >= children.length
        children[ii].set_enable(val)
    ;end
  def change_active(ii);
        root.set_view_pos(self.sym,ii) if self.sym
        root.set_view_pos(:view, :add_synth) if self.sym == :voice
        root.change_view               if self.sym

        children.each_with_index do |ch, i|
            n = (i == ii)
            if(n != ch.value)
                ch.value = n
                ch.damage_self
            end
        end
    ;end
  def toggle(ii);
        self.valueRef[ii].value = !children[ii].enable
    ;end
  def layout(l, selfBox);
        cols = 4
        l.aspect(selfBox, cols, rows)

        b = 0
        ch = self.children
        (0...rows).each do |r|
            (0...cols).each do |c|
                child = ch[b]
                b += 1
                child.fixed(l, selfBox, c/4, r/self.rows, 0.25, 1/self.rows)
            end
        end
        selfBox
    ;end
  def animate();
        return if !self.sym
        vv = root.get_view_pos self.sym
        return if(vv == @vv)
        @vv = vv
        children.each_with_index do |ch, i|
            n = (i == vv)
            if(n != ch.value)
                ch.value = n
                ch.damage_self
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @kitbuttons = self

    
context = Hash.new
context["kitbuttons"] = self
Qml::prop_add(self, "rows")
Qml::prop_add(self, "sym")
self.rows = 4
self.sym = nil
self.valueRef = []

    #puts "KitButtons, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::LabelButton < Qml::Button
  attr_reader :lbut
  attr_accessor :valueRef
  attr_accessor :action
  def animate();
        if(root.key_widget != self)
            @next = nil
            if(@state)
                @state = nil
                damage_self
            end
            return
        end
        now = Time.new
        if(@next.nil?)
            @next = now + 0.1
            return
        elsif(@next < now)
            @state = !@state
            @next = now + 0.7
            damage_self
        end
    ;end
  def draw_text(vg);
        text_color1   = Theme::TextActiveColor
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*self.textScale
        if(value == true)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        if(@state)
            vg.text(8,h/2,button.label.upcase + "|")
        else
            vg.text(8,h/2,button.label.upcase)
        end
    ;end
  def setValue(x);
        self.label = x
        damage_self
    ;end
  def onMousePress(ev);
        whenValue.call if whenValue
    ;end
  def onMouseEnter(ev);
        if(self.tooltip.nil? || self.tooltip.empty?)
            self.root.log(:tooltip, "extern = <" + self.extern + ">")
        else
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 8)
            self.label[-1] = "\0" if self.label.length > 0
            self.label = self.label[0...-1]
        else
            self.label += k
        end
        whenValue.call if whenValue
        valueRef.value = self.label if valueRef
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @lbut = self

    
context = Hash.new
context["lbut"] = self
self.valueRef = nil
self.action = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        meta = OSC::RemoteMetadata.new($remote, lbut.extern)
        lbut.tooltip = meta.tooltip
        lbut.valueRef = OSC::RemoteParam.new($remote, lbut.extern)
        lbut.valueRef.callback = lambda {|x| lbut.setValue(x)}
    
end})

    #puts "LabelButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::LearnButton < Qml::Button
  def onMousePress(ev);
        root.learn_mode = !root.learn_mode
    ;end
  def onMousePress(ev);
        root.learn_mode = !root.learn_mode
    ;end
  def animate();
        if(self.value != root.learn_mode)
            self.value = root.learn_mode
            damage_self
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous152 = self

    
context = Hash.new
self.label = "learn"
self.layoutOpts = [:no_constraint]

    #puts "LearnButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::LfoVis < Qml::Widget
  attr_reader :lfo_vis
  attr_reader :run_view
  attr_property :points
  attr_property :type
  attr_property :drag_type
  attr_property :drag_prev
  attr_property :phase
  attr_property :depth
  attr_property :delay_time
  attr_property :period
  attr_property :time
  attr_accessor :valueRef
  def onSetup(old=nil);
        refs = []
        base = lfo_vis.extern
        type_var = OSC::RemoteParam.new($remote, base+"PLFOtype")
        type_var.mode = :options
        type_var.callback = lambda {|x|
            ntype = [:sine, :triangle, :square, :rampup,
                :rampdown, :exp1, :exp2][x]
            return if(ntype == lfo_vis.type)
            lfo_vis.type = [:sine, :triangle, :square, :rampup,
                :rampdown, :exp1, :exp2][x]}

        depth_var = OSC::RemoteParam.new($remote, base+"Pintensity")
        depth_var.callback = lambda {|x|
            lfo_vis.depth = x
            lfo_vis.damage_self}

        delay_var = OSC::RemoteParam.new($remote, base+"Pdelay")
        delay_var.callback = lambda {|x|
            lfo_vis.delay_time = Math.exp(Math.log(2000)*x)
            lfo_vis.damage_self}

        freq_var  = OSC::RemoteParam.new($remote, base+"Pfreq")
        freq_var.type = "f"
        freq_var.set_min(0)
        freq_var.set_max(1)
        freq_var.callback = lambda {|x|
            lfo_vis.period = 20*Math.exp(Math.log(1500/20)*(1-x))
            lfo_vis.damage_self}

        phase_var = OSC::RemoteParam.new($remote, base+"Pstartphase")
        phase_var.callback = lambda {|x|
            lfo_vis.phase = x
            lfo_vis.updateType
            lfo_vis.damage_self}


        refs << type_var
        refs << depth_var
        refs << freq_var
        refs << delay_var
        refs << phase_var
        self.valueRef = refs

    ;end
  def class_name();
        "LfoVis"
    ;end
  def onMousePress(ev);
        loc_x = ev.pos.x - lfo_vis.global_x
        if(loc_x < 0.2*lfo_vis.w)
            lfo_vis.drag_type = :delay
        else
            lfo_vis.drag_type = :lfo
        end
        lfo_vis.drag_prev = ev.pos
    ;end
  def onMouseMove(ev);
        if(lfo_vis.drag_prev)
            dx = ev.pos.x - lfo_vis.drag_prev.x
            dy = ev.pos.y - lfo_vis.drag_prev.y

            if(lfo_vis.drag_type == :delay)
                #dv = lfo_vis.phase + dy/100.0
                #dv_ = [1, [0, dv].max].min
                #lfo_vis.phase = dv_
                #lfo_vis.drag_prev.y = ev.pos.y if(dv == dv_)
                #lfo_vis.updateType

                dt = Math.exp((Math.log(lfo_vis.delay_time)/Math.log(2000) + dx/200.0)*Math.log(2000))
                dt_ = [2000, [1, dt].max].min
                lfo_vis.delay_time = dt_
                lfo_vis.drag_prev.x = ev.pos.x if(dt == dt_)
                self.valueRef[3].value = Math.log(dt_)/Math.log(2000)
                damage_self
            elsif(lfo_vis.drag_type == :lfo)
                lfo_vis.drag_prev = ev.pos
                dv = lfo_vis.depth + dy/300.0
                dv = [1, [0, dv].max].min
                lfo_vis.depth = dv
                self.valueRef[1].value = dv

                dt = 100*Math.exp(Math.log(0.01*lfo_vis.period) + dx/200.0)
                dt = [1500, [20, dt].max].min
                lfo_vis.period = dt

                nval = 1-Math.log(lfo_vis.period/20)/Math.log(1500/20)
                self.valueRef[2].value = nval
                damage_self
            end
        end
    ;end
  def updateType();
        shape = case lfo_vis.type
        when :triangle
            Proc.new {|phase|
                if(phase >= 0 && phase < 0.25)
                    4.0 * phase
                elsif(phase >= 0.25 && phase < 0.75)
                    2.0 - 4 * phase
                else
                    4.0 * phase - 4;
                end
            }
        when :square
            Proc.new {|phase|
                if(phase < 0.5)
                    -1.0
                else
                    1.0
                end
            }
        when :rampup
            Proc.new {|phase| (phase - 0.5) * 2.0}
        when :rampdown
            Proc.new {|phase| (0.5 - phase) * 2.0};
        when :exp1
            Proc.new {|phase| (0.05 ** phase) * 2.0 - 1.0}
        when :exp2
            Proc.new {|phase| (0.001 ** phase) * 2.0 - 1.0}
        else
            Proc.new {|x| Math.sin(2*3.14*x) }
        end

        # root point
        p = [0,0, 0, 0.2]
        # func points
        resolution = 128
        (0..resolution).each do |i|
            x = 0.2+0.8*i/resolution
            phase = i/resolution + lfo_vis.phase
            phase -= 1 if phase > 1

            y = shape.call(phase)
            p << y
            p << x
        end
        lfo_vis.points = p
        damage_self
    ;end
  def draw(vg);
        padfactor = 10
        bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)

        background Theme::VisualBackground

        dat = lfo_vis.points
        updateType if dat.empty?
        dat = lfo_vis.points
        if(dat.empty?)
            puts "[Zyn-Zest:ERROR] LfoVis - um, it's empty..."
            return
        end

        fill_color   = Theme::VisualBackground
        stroke_color = Theme::VisualStroke
        light_fill   = Theme::VisualLightFill
        light_filll  = color("911515", 20)
        bright       = Theme::VisualBright
        dim          = Theme::VisualDim
        sel_color    = Theme::VisualSelect

        background fill_color

        draw_grid(vg, lfo_vis.depth*16, lfo_vis.period/10,
                  bb.x+0.2*bb.w, bb.y, 0.8*bb.w, bb.h)
        draw_grid(vg, lfo_vis.depth*16, lfo_vis.delay_time/100,
                  bb.x,          bb.y, 0.2*bb.w, bb.h)

        #weak highlight
        vg.path do |vg|
            vg.rect(0.2*w, 0, 0.8*w, h)
            vg.fill_color light_filll
            vg.fill
        end

        pts = Draw::toPos(dat)

        #Draw Highlights
        Draw::WaveForm::under_highlight(vg, bb, pts, light_fill)
        Draw::WaveForm::over_highlight(vg,  bb, pts, light_fill)

        #Draw Zero Line
        Draw::WaveForm::zero_line(vg, bb, dim)

        Draw::WaveForm::env_sel_line(vg, bb, 1, pts, dim)

        #Draw Actual Line
        Draw::WaveForm::lfo_plot(vg, bb, pts, bright)

        vg.fill_color   color(:black)
        vg.stroke_color bright
        (0...(dat.length/2)).each do |i|
            xx = bb.x + bb.w*dat[2*i+1];
            yy = bb.y + bb.h/2*(1-dat[2*i]);
            next if(i >= 2)
            Draw::WaveForm::env_marker(vg, xx, yy, 3)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @lfo_vis = self
@run_view = Qml::Widget.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["lfo_vis"] = self
context["run_view"] = @run_view
@run_view.instance_eval do
def run_view=(k);@run_view=k;end
def run_view;    @run_view;  end
end
Qml::context_apply(context, @run_view)
Qml::prop_add(self, "points")
Qml::prop_add(self, "type")
Qml::prop_add(self, "drag_type")
Qml::prop_add(self, "drag_prev")
Qml::prop_add(self, "phase")
Qml::prop_add(self, "depth")
Qml::prop_add(self, "delay_time")
Qml::prop_add(self, "period")
Qml::prop_add(self, "time")

        @run_view.instance_eval("def class_name();\n            \"LfoVisAnimation\"\n        ;end", "src/mruby-zest/example/LfoVis.qml", 244)

        @run_view.instance_eval("def valueRef=(value_ref);\n            @value_ref = value_ref\n        ;end", "src/mruby-zest/example/LfoVis.qml", 250)

        @run_view.instance_eval("def valueRef();\n            @value_ref\n        ;end", "src/mruby-zest/example/LfoVis.qml", 255)

        @run_view.instance_eval("def runtime_points();\n            @runtime_points\n        ;end", "src/mruby-zest/example/LfoVis.qml", 260)

        @run_view.instance_eval("def runtime_points=(pts);\n            @runtime_points = pts\n        ;end", "src/mruby-zest/example/LfoVis.qml", 264)

        @run_view.instance_eval("def animate();\n            return if run_view.valueRef.nil?\n            run_view.valueRef.watch run_view.extern\n        ;end", "src/mruby-zest/example/LfoVis.qml", 282)

        @run_view.instance_eval("def draw(vg);\n            #Draw the data\n            pts   = @runtime_points\n            pts ||= []\n\n            padfactor = 10\n            bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)\n\n            Draw::WaveForm::overlay_lfo(vg, bb, pts)\n        ;end", "src/mruby-zest/example/LfoVis.qml", 287)
@run_view.layer = 1
@db.connect_property(@run_view.properties["extern"], lambda {lfo_vis.extern + "out"})
@db.connect_watcher(@run_view.properties["extern"], lambda {begin

            return if run_view.extern.nil?
            
            run_view.valueRef = OSC::RemoteParam.new($remote, run_view.extern)
            run_view.valueRef.set_watch
            run_view.valueRef.callback = Proc.new {|x|
                if(run_view.runtime_points != x)
                    run_view.runtime_points = x;
                    run_view.damage_self
                end
            }
        
end})
Qml::add_child(self, @run_view)
self.points = []
self.type = :triangle
self.drag_type = nil
self.drag_prev = nil
self.phase = 0.0
self.depth = 0.5
self.delay_time = 100
self.period = 1000
self.time = nil
self.valueRef = nil
@db.connect_watcher(self.properties["type"], lambda {begin

        lfo_vis.updateType
    
end})

    #puts "LfoVis, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::MainMenu < Qml::Widget
  attr_reader :menu
  attr_reader :file
  attr_reader :learn
  def layout(l, selfBox);
        pad  = 1/32
        pad2 = 0.5-2*pad
        children[0].fixed(l, selfBox, 0.0+pad, 0.0+pad, pad2, pad2)
        children[1].fixed(l, selfBox, 0.5+pad, 0.0+pad, pad2, pad2)
        children[2].fixed(l, selfBox, 0.0+pad, 0.5+pad, pad2, pad2)
        children[3].fixed(l, selfBox, 0.5+pad, 0.5+pad, pad2, pad2)

        selfBox
    ;end
  def setup_widget(w);
        w.onSetup()
        w.children.each do |c|
            setup_widget(c)
        end
    ;end
  def file_select();
        opt = file.options[file.selected]
        win = window()
        wid = Qml::FileSelector.new(db)
        wid.whenValue = lambda { |x| menu.file_value(x)}
        wid.x = 0#-global_x
        wid.y = 0#-global_y
        wid.w = win.w
        wid.h = win.h
        wid.ext = ".xiz" if opt == "save instrument"
        wid.ext = ".xmz" if opt == "save master"
        wid.ext = ".wav" if opt == "setup record"
        wid.pat = ".xiz" if opt == "load instrument"
        wid.pat = ".xmz" if opt == "load master"
        wid.pat = ".xsz" if opt == "load microtonal"
        wid.pat = ".xsz" if opt == "save microtonal"
        wid.pat = ".xlz" if opt == "load midi bindings"
        wid.pat = ".xlz" if opt == "save midi bindings"
        Qml::add_child(win, wid)
        self.db.update_values
        setup_widget wid
        self.db.update_values

        if(root)
            root.smash_layout
            root.damage_item(win, :all)
        end
    ;end
  def file_sel();
        opt = file.options[file.selected]
        if(["load instrument", "save instrument", "load master", "save master", "setup record",
            "load microtonal", "save microtonal", "load midi bindings", "save midi bindings"].include? opt)
            menu.file_select
        elsif(opt == "clear master")
            root.set_view_pos(:part, 0)
            root.set_view_pos(:kit, 0)
            root.set_view_pos(:voice, 0)
            root.set_view_pos(:view, :add_synth)
            root.set_view_pos(:subview, :global)
            root.change_view
            $remote.action("/reset_master")
        elsif(opt == "clear instrument")
            root.set_view_pos(:voice, 0)
            root.set_view_pos(:kit, 0)
            root.set_view_pos(:view, :add_synth)
            root.set_view_pos(:subview, :global)
            prt  = root.get_view_pos(:part)
            root.change_view
            $remote.action("/part#{prt}/clear")
        elsif(opt == "quit")
            $remote.action("/quit")
            root.quit
        else
            puts "[WARNING] Unhandled Option #{opt}"
        end

    ;end
  def file_value(val);
        return if val == :cancel
        opt = file.options[file.selected]
        prt  = root.get_view_pos(:part)
        if(opt == "load instrument")
            $remote.action("/load_xiz", prt, val)
        elsif(opt == "save instrument")
            $remote.action("/save_xiz", prt, val)
        elsif(opt == "load master")
            $remote.action("/load_xmz", val)
        elsif(opt == "save master")
            $remote.action("/save_xmz", val)
        elsif(opt == "save microtonal")
            $remote.action("/save_xsz", val)
        elsif(opt == "load microtonal")
            $remote.action("/load_xsz", val)
        elsif(opt == "load midi bindins")
            $remote.action("/save_xlz", val)
        elsif(opt == "save midi bindings")
            $remote.action("/load_xlz", val)
        elsif(opt == "setup record")
            $remote.action("/HDDRecorder/preparefile", val)
        end
    ;end
end
class Qml::Record < Qml::Widget
  def draw(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            v.move_to(w*1/3, 0)
            v.line_to(w*1/3, h)
            v.move_to(w*2/3, 0)
            v.line_to(w*2/3, h)
            v.stroke
        end

        text_color = Theme::TextColor

        draw_stop(vg, text_color)
        draw_pause( vg, text_color)
        draw_play(  vg, text_color)
    ;end
  def draw_record(vg, text_color);
        r = [1/6*w, 1/2*h].min*0.4
        vg.path do |v|
            v.circle(1/6*w, 1/2*h, r)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_stop(vg, text_color);
        r = [1/6*w, 1/2*h].min*0.4
        vg.path do |v|
            v.rect(1/6*w-r, 1/2*h-r, 2*r, 2*r)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_pause(vg, text_color);
        vg.path do |v|
            v.rect((3/6-1/18)*w, 0.2*h, 1/24*w, 0.6*h)
            v.rect((3/6+1/18)*w, 0.2*h, 1/24*w, 0.6*h)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_play(vg, text_color);
        vg.path do |v|
            v.move_to((5/6-1/18)*w, 0.25*h)
            v.line_to((5/6-1/18)*w, 0.75*h)
            v.line_to((5/6+1/18)*w, 0.5*h)
            v.close_path
            v.fill_color text_color
            v.fill
        end
    ;end
  def onMousePress(ev);
        px = (ev.pos.x-global_x)/w
        if(px > 0.666)
            $remote.action("/HDDRecorder/start")
        elsif(px > 0.333)
            $remote.action("/HDDRecorder/pause")
        else
            $remote.action("/HDDRecorder/stop")
        end

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous173 = self

    
context = Hash.new

    #puts "Record, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::MainMenu
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @menu = self
@file = Qml::Menu.new(database, ui_path+"1"+'/')
@learn = Qml::Button.new(database, ui_path+"2"+'/')
@anonymous157 = Qml::Record.new(database, ui_path+"3"+'/')
@anonymous158 = Qml::LearnButton.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["menu"] = self
context["file"] = @file
context["learn"] = @learn
@file.label = "file"
@file.layoutOpts = [:no_constraint]
@file.options = ["load instrument", "save instrument",
                  "load master", "save master",
                  "load microtonal", "save microtonal",
                  "load midi bindings", "save midi bindings",
                  "clear master", "clear instrument",
                  "setup record", "quit"]
@file.whenValue = lambda {
            menu.file_sel if file.selected
        }
Qml::add_child(self, @file)
@learn.label = "midi"
@learn.tooltip = "use ctrl to enable midi learn"
@learn.layoutOpts = [:no_constraint]
Qml::add_child(self, @learn)
Qml::add_child(self, @anonymous157)
@anonymous158.tooltip = "just click on a widget while in learning mode to start mapping"
Qml::add_child(self, @anonymous158)

    #puts "MainMenu, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::MainWindow < Qml::Widget
  attr_reader :window
  attr_reader :side
  attr_reader :head1
  attr_reader :footerbox
  attr_reader :main_widget
  def class_name(); "MainWindow" ;end
  def layout(l, selfBox);
        side.fixed(l,        selfBox, 0, 0.1, 0.1, 0.8)
        head1.fixed(l,       selfBox, 0, 0,   1.0, 0.1)
        footerbox.fixed(l,        selfBox, 0, 0.9, 1.0, 0.1)
        main_widget.fixed(l, selfBox, 0.102, 0.1, 0.89, 0.80)

        #Layout overlay panel (e.g. file browser)
        if(children.length == 5)
            children[4].fixed(l, selfBox, 0, 0, 1, 1)
        end

        selfBox
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0, 0, w, h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::WindowGrad1, Theme::WindowGrad2)
            v.fill_paint paint
            v.fill
        end
    ;end
  def set_content(type);
        root.set_view_pos(:view, type)
        root.change_view
    ;end
  def valueRef();
        return @refs
    ;end
  def onSetup(old=nil);
        return if main_widget.content
        @info = Hash.new

        #Assume parts are active for testing purposes
        (0...16).each do |prt|
            (0...16).each do |kit|
                @info[[prt,kit,:add]] = true
                @info[[prt,kit,:sub]] = true
                @info[[prt,kit,:pad]] = true
            end
        end

        @refs = []
        (0...16).each do |prt|
            (0...16).each do |kit|
                add = OSC::RemoteParam.new($remote,
                    "/part#{prt}/kit#{kit}/Padenabled")
                sub = OSC::RemoteParam.new($remote,
                    "/part#{prt}/kit#{kit}/Psubenabled")
                pad = OSC::RemoteParam.new($remote,
                    "/part#{prt}/kit#{kit}/Ppadenabled")
                add.callback =
                    lambda {|x| window.info(prt,kit,:add,x)}
                sub.callback =
                    lambda {|x| window.info(prt,kit,:sub,x)}
                pad.callback =
                    lambda {|x| window.info(prt,kit,:pad,x)}
                @refs << add
                @refs << sub
                @refs << pad
            end
        end


        set_view()
    ;end
  def info(a,b,c,d);
        @info[[a,b,c]] = d
    ;end
  def set_view();
        prt  = root.get_view_pos(:part)
        kit  = root.get_view_pos(:kit)
        type = root.get_view_pos(:view)

        if(type == :add_synth && !@info[[prt,kit,:add]])
            root.set_view_pos(:view, :kits)
        elsif(type == :pad_synth && !@info[[prt,kit,:pad]])
            root.set_view_pos(:view, :kits)
        elsif(type == :sub_synth && !@info[[prt,kit,:sub]])
            root.set_view_pos(:view, :kits)
        end

        type = root.get_view_pos(:view)

        if(type == :add_synth)
            main_widget.extern  = "/part#{prt}/kit#{kit}/adpars/"
            main_widget.content = Qml::ZynAddSynth
        elsif(type == :pad_synth)
            main_widget.extern  = "/part#{prt}/kit#{kit}/padpars/"
            main_widget.content = Qml::ZynPadSynth
        elsif(type == :sub_synth)
            main_widget.extern  = "/part#{prt}/kit#{kit}/subpars/"
            main_widget.content = Qml::ZynSubSynth
        elsif(type == :part)
            main_widget.extern  = "/part#{prt}/"
            main_widget.content = Qml::ZynPart
        elsif(type == :kits)
            main_widget.extern  = "/part#{prt}/"
            main_widget.content = Qml::ZynKit
        elsif(type == :effects)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynEffects
        elsif(type == :midi_learn)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynMidiLearn
        elsif(type == :mixer)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynMixer
        elsif(type == :banks)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynBank
        elsif(type == :about)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynAbout
        elsif(type == :automate)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynAutomation
        elsif(type == :colors)
            main_widget.extern  = "/"
            main_widget.content = Qml::ColorScheme
        elsif(type == :settings)
            main_widget.extern  = "/"
            main_widget.content = Qml::ZynSettings
        else
            main_widget.content = Qml::Widget
        end
    ;end
end
class Qml::ZynSidebar < Qml::Widget
  attr_reader :side
  attr_reader :partsetting
  attr_reader :part
  attr_reader :browser
  attr_reader :mixer
  attr_reader :kitedit
  attr_reader :kit
  attr_reader :arp
  attr_reader :eff
  attr_reader :add
  attr_reader :voices
  attr_reader :subButton
  attr_reader :padButton
  def class_name(); "ZynSidebar" ;end
  def set_content(type, flag=nil);
        tabs = [:part, :banks, :mixer, :kits, :automate, :effects,
                :add_synth, :sub_synth, :pad_synth]
        ind = 0
        self.children.each do |ch|
            if([Qml::FancyButton, Qml::SidebarButton].include? ch.class)
                n = (type == tabs[ind])
                if(n != ch.value)
                    ch.value = n
                    root.damage_item ch if root
                end
                ind += 1
            end
        end

        parent.set_content(type) if flag.nil? && parent.respond_to?(:set_content)
    ;end
  def set_view();
        prt  = root.get_view_pos(:part)
        kid  = root.get_view_pos(:kit)
        typ  = root.get_view_pos(:view)
        padButton.extern = "/part#{prt}/kit#{kid}/Ppadenabled"
        subButton.extern = "/part#{prt}/kit#{kid}/Psubenabled"
        add.extern       = "/part#{prt}/kit#{kid}/Padenabled"
        set_content(typ, true)
    ;end
  def onSetup(old=nil);
        set_view
    ;end
  def layout(l, selfBox);
        partsetting.fixed(l, selfBox, 0.1, 0,     0.8, 0.045)
        part.fixed(l,        selfBox, 0.1, 0.05,  0.8, 0.19)
        browser.fixed(l,     selfBox, 0.1, 0.248, 0.8, 0.045)
        mixer.fixed(l,       selfBox, 0.1, 0.3,   0.8, 0.045)
        kitedit.fixed(l,     selfBox, 0.1, 0.35,  0.8, 0.045)
        kit.fixed(l,         selfBox, 0.1, 0.405, 0.8, 0.187)
        arp.fixed(l,         selfBox, 0.1, 0.6,   0.8, 0.045)
        eff.fixed(l,         selfBox, 0.1, 0.65,  0.8, 0.045)
        add.fixed(l,         selfBox, 0.1, 0.7,   0.8, 0.045)
        voices.fixed(l,      selfBox, 0.1, 0.76,  0.8, 0.10)
        subButton.fixed(l,   selfBox, 0.1, 0.88,  0.8, 0.045)
        padButton.fixed(l,   selfBox, 0.1, 0.94,  0.8, 0.045)

        selfBox
    ;end
end
class Qml::SidebarButton < Qml::Widget
  attr_reader :button
  attr_accessor :whenClick
  attr_property :value
  def onMousePress(ev);
        self.value = !self.value
        damage_self
        self.whenClick.call if whenClick
    ;end
  def draw_button(vg);
        pad = 0
        off_color     = Theme::ButtonInactive
        on_color      = Theme::ButtonActive
        cs = 0
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 3)
            if(button.value == true)
                cs = 1
                v.fill_color on_color
            elsif(button.value.class == Float && button.value != 0)
                cs = 2
                t = button.value
                on = on_color
                of = Theme::ButtonGrad1
                v.fill_color(color_rgb(on.r*t + of.r*(1-t),
                                       on.g*t + of.g*(1-t),
                                       on.b*t + of.b*(1-t)))
            else
                paint = v.linear_gradient(0,0,0,h,
                Theme::ButtonGrad1, Theme::ButtonGrad2)
                v.fill_paint paint
            end
            v.fill
            v.stroke_width 1
            v.stroke

        end

        vg.path do |v|
            hh = h/20
            v.move_to(w*pad+3,       h*pad+hh)
            v.line_to(w*(1-2*pad)-2, h*pad+hh)
            if(cs == 0)
                v.stroke_color color("5c5c5c")
            elsif(cs == 1)
                v.stroke_color color("16a39c")
            end
            if([0,1].include?(cs))
                v.stroke_width hh 
                v.stroke
            end
        end
    ;end
  def draw(vg);
        #h = button.h
        #w = button.w
        ##puts("drawing a button...")
        #bright_green = color("00AE9C")
        #dark_green   = color("007368")
        #grad = vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        #vg.stroke_color(NVG.rgba(0,0,0,0x80))

        #grey1 = Theme::ButtonGrad1
        #grey2 = Theme::ButtonGrad2
        #grad2 = vg.linear_gradient(0,0,0,h, grey1, grey2)
        #vg.path do |v|
        #    v.rounded_rect(0,0,w,h,2)
        #    if(value)
        #        bright_green = color("00AE9C")
        #        dark_green   = color("007368")
        #        grad = vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        #        v.fill_paint grad
        #    else
        #        v.fill_paint grad2
        #    end
        #    v.stroke_width 1
        #    v.fill
        #    v.stroke
        #end
        draw_button(vg)

        text_color1   = color("52FAFE")
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*0.55
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        if(value)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text(w/2,h/2,label.upcase)

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @button = self

    
context = Hash.new
context["button"] = self
Qml::prop_add(self, "value")
self.whenClick = nil
self.value = false

    #puts "SidebarButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSidebar
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @side = self
@partsetting = Qml::SidebarButton.new(database, ui_path+"1"+'/')
@part = Qml::Indent.new(database, ui_path+"2"+'/')
@anonymous1084 = Qml::KitButtons.new(database, ui_path+"3"+'/')
@browser = Qml::SidebarButton.new(database, ui_path+"4"+'/')
@mixer = Qml::SidebarButton.new(database, ui_path+"5"+'/')
@kitedit = Qml::SidebarButton.new(database, ui_path+"6"+'/')
@kit = Qml::Indent.new(database, ui_path+"7"+'/')
@anonymous1089 = Qml::KitButtons.new(database, ui_path+"8"+'/')
@arp = Qml::SidebarButton.new(database, ui_path+"9"+'/')
@eff = Qml::SidebarButton.new(database, ui_path+"10"+'/')
@add = Qml::FancyButton.new(database, ui_path+"11"+'/')
@voices = Qml::Indent.new(database, ui_path+"12"+'/')
@anonymous1094 = Qml::KitButtons.new(database, ui_path+"13"+'/')
@subButton = Qml::FancyButton.new(database, ui_path+"14"+'/')
@padButton = Qml::FancyButton.new(database, ui_path+"15"+'/')

    
context = Hash.new
context["side"] = self
context["partsetting"] = @partsetting
context["part"] = @part
context["browser"] = @browser
context["mixer"] = @mixer
context["kitedit"] = @kitedit
context["kit"] = @kit
context["arp"] = @arp
context["eff"] = @eff
context["add"] = @add
context["voices"] = @voices
context["subButton"] = @subButton
context["padButton"] = @padButton
@partsetting.label = "part settings"
@partsetting.whenClick = lambda { side.set_content :part }
Qml::add_child(self, @partsetting)
@anonymous1084.sym = :part
Qml::add_child(@part, @anonymous1084)
Qml::add_child(self, @part)
@browser.label = "browser"
@browser.whenClick = lambda { side.set_content :banks }
Qml::add_child(self, @browser)
@mixer.label = "mixer"
@mixer.whenClick = lambda { side.set_content :mixer }
Qml::add_child(self, @mixer)
@kitedit.label = "kits"
@kitedit.whenClick = lambda { side.set_content :kits }
Qml::add_child(self, @kitedit)
@anonymous1089.sym = :kit
Qml::add_child(@kit, @anonymous1089)
Qml::add_child(self, @kit)
@arp.label = "macro learn"
@arp.whenClick = lambda { side.set_content :automate }
Qml::add_child(self, @arp)
@eff.label = "effects"
@eff.whenClick = lambda { side.set_content :effects }
Qml::add_child(self, @eff)
@add.label = "add"
@add.value = true
@add.whenClick = lambda { side.set_content :add_synth }
Qml::add_child(self, @add)
@anonymous1094.sym = :voice
@anonymous1094.rows = 2
Qml::add_child(@voices, @anonymous1094)
Qml::add_child(self, @voices)
@subButton.label = "sub"
@subButton.whenClick = lambda { side.set_content :sub_synth }
Qml::add_child(self, @subButton)
@padButton.label = "pad"
@padButton.whenClick = lambda { side.set_content :pad_synth }
Qml::add_child(self, @padButton)

    #puts "ZynSidebar, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynHeader < Qml::Widget
  attr_reader :head
  attr_reader :logo
  attr_reader :menuIndent
  attr_reader :status
  attr_reader :shortcuts
  attr_reader :vumeter
  attr_reader :panic
  attr_reader :close
  attr_reader :minimize
  def class_name();"ZynHeader";end
  def layout(l, selfBox);
        logo.fixed(l,       selfBox, 0.016, 0.22,  0.070, 0.56)
        menuIndent.fixed(l, selfBox, 0.100, 0.15, 0.112, 0.7)
        status.fixed(l,     selfBox, 0.220, 0.15, 0.520, 0.7)
        shortcuts.fixed(l,  selfBox, 0.745, 0.15, 0.160, 0.7)
        vumeter.fixed(l,    selfBox, 0.910, 0.15, 0.015, 0.7)
        panic.fixed(l,      selfBox, 0.930, 0.15, 0.040, 0.7)
        close.fixed(l,      selfBox, 0.975, 0.15, 0.015, 0.3)
        minimize.fixed(l,   selfBox, 0.975, 0.54, 0.015, 0.3)

        selfBox
    ;end
end
class Qml::ZynLogo < Qml::Widget
  def draw_placeholder(vg);
        #background color("123456")
        #vg.path do |v|
        #    v.rect(0, 0, w, h)
        #    v.fill_color(NVG.rgba(0xaa, 0xaa, 0xaa, 255))
        #    v.fill
        #end
        logo_top = color("4DC3C7")
        logo_bot = color("56C0A5")

        vg.path do |v|
            l = 0.4*w
            v.move_to(0,0)
            v.line_to(l,      0.1*h)
            v.line_to(l*0.75, 0.6*h)
            v.close_path
            v.fill_color logo_top
            v.fill
        end

        vg.path do |v|
            l = 0.4*w
            v.move_to(l,h)
            v.line_to(0,      0.9*h)
            v.line_to(l*0.25, 0.4*h)
            v.close_path
            v.fill_color logo_bot
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color logo_top
        vg.text(w*1/2,h/2,"ZYN")
    ;end
  def draw_path(vg, path, co);
        wscale = 80.0
        hscale = 35.0
        n = path.length/2
        xmin = 64
        ymin = 60
        vg.path do
            xx = (w*(path[2*0+0]-xmin)/wscale).round
            yy = (h*(path[2*0+1]-ymin)/hscale).round
            vg.move_to(xx, yy)
            (1...n).each do |i|
                xx = (w*(path[2*i+0]-xmin)/wscale).round
                yy = (h*(path[2*i+1]-ymin)/hscale).round
                vg.line_to(xx,yy)
            end
            vg.close_path
            vg.fill_color co
            vg.fill
            #vg.stroke_color color("2E3239")
            #vg.stroke_width 2.0
            #vg.stroke
        end
    ;end
  def draw_new(vg);
        st0 = color("686868")
        st1 = color("2E3239")
        st1 = color("010101")
        st2 = color("50C3C7")
        st3 = color("57C1A6")

        st1
        path_1 = [97,64,64,61,67.3,63.3,64,63,87,79,97,66,95.6,65.9]
        path_2 = [97,90,74,74,64,87,65.4,87.1,64,89,97,92,93.7,89.7]
        path_3 = [103,69,103.7,71,103,71,104,74,107.5,74,103,80,103,82.1,103,83,103,85,115,85,115,83,115,81.4,115,80,110.3,80,115,74,115,72,115,71,115,69]
        path_4 = [129,69,124,69,123,74,122,69,117,69,117.8,71,117,71,121,81,121,83,121,85,125,85,125,83,125,81,129,71,128.2,71]
        path_5 = [131,69,135,69,138.5,74,138.8,75,139,75,139,69,143,69,143,85,139,85,135.5,80,135.8,79,135,79,135,85,131,85]
        draw_path(vg, path_1, st1)
        draw_path(vg, path_2, st1)
        draw_path(vg, path_3, st1)
        draw_path(vg, path_4, st1)
        draw_path(vg, path_5, st1)

        st2 
        path_6 = [64,62,97,65,87,78]
        path_7 = [115,84,103,84,103,81,109,73,109,73,104,73,103,70,115,70,115,73,109,81,109,81,115,81,115,84]
        path_8 = [122,70,123,75,123,75,124,70,129,70,125,80,125,84,121,84,121,80,117,70]
        path_9 = [131,70,135,70,138.5,75,138.8,76,139,76,139,70,143,70,143,84,139,84,135.5,79,135.8,78,135,78,135,84,131,84]
        draw_path(vg, path_6, st2)
        draw_path(vg, path_7, st2)
        draw_path(vg, path_8, st2)
        draw_path(vg, path_9, st2)

        st3
        path_10 = [97,91,64,88,74,75]
        draw_path(vg, path_10, st3)
    ;end
  def draw(vg);
        #draw_placeholder(vg)
        draw_new(vg)
    ;end
  def onMousePress(m);
        view = root.get_view_pos(:view)
        root.set_view_pos(:view, :about)    if(view != :about)
        root.set_view_pos(:view, :settings) if(view == :about)
        root.change_view
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous726 = self

    
context = Hash.new

    #puts "ZynLogo, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynHeaderParams < Qml::Widget
  attr_reader :fn
  def layout(l, selfBox);
        padw  = 1/256
        padw2 = 0.5-2*padw
        padh = 1/32
        padh2 = 0.5-2*padh
        children[0].fixed(l, selfBox, 0.0+padw, 0.0+padh, padw2, padh2)
        children[1].fixed(l, selfBox, 0.5+padw, 0.0+padh, padw2, padh2)
        children[2].fixed(l, selfBox, 0.0+padw, 0.5+padh, padw2, padh2)
        children[3].fixed(l, selfBox, 0.5+padw, 0.5+padh, padw2, padh2)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous658 = self
@fn = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous655 = Qml::HSlider.new(database, ui_path+"2"+'/')
@anonymous656 = Qml::Button.new(database, ui_path+"3"+'/')
@anonymous657 = Qml::HSlider.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["fn"] = @fn
@fn.instance_eval do
def fn=(k);@fn=k;end
def fn;    @fn;  end
end
Qml::context_apply(context, @fn)

        @fn.instance_eval("def animate();\n            if(fn.value != root.fine_mode)\n                fn.value = root.fine_mode\n                fn.damage_self\n            end\n        ;end", "src/mruby-zest/example/ZynHeaderParams.qml", 7)
@fn.label = "fine"
@fn.layoutOpts = [:no_constraint]
Qml::add_child(self, @fn)
@anonymous655.extern = "/Pvolume"
@anonymous655.height = 0.8
@anonymous655.label = "volume"
Qml::add_child(self, @anonymous655)
@anonymous656.layoutOpts = [:no_constraint]
@anonymous656.label = "nrpn"
Qml::add_child(self, @anonymous656)
@anonymous657.extern = "/Pkeyshift"
@anonymous657.height = 0.8
@anonymous657.label = "key shift"
Qml::add_child(self, @anonymous657)

    #puts "ZynHeaderParams, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VuMeter < Qml::Widget
  attr_reader :meter
  attr_accessor :valueRef
  attr_property :data
  def animate();
        self.valueRef.refresh if(self.valueRef)
    ;end
  def rap2dB(x); 20*Math::log10(x) ;end
  def bound(x); [0.0, [1.0, x].min].max ;end
  def cv(x);min_db = -40;bound((min_db-rap2dB(x))/min_db);end
  def draw(vg);
        indent_color = Theme::VisualBackground
        background indent_color

        bar_color = Theme::VisualLine
        pad  = 3
        pad2 = (h-2*pad)
        v1 = 0.3
        v2 = 0.5
        if(!data.nil?)
            v1 = cv(data[0])
            v2 = cv(data[1])
        end

        vg.path do |v|
            v.rect(pad,(1-v1)*pad2, 0.2*w, (v1)*(h-pad))
            v.fill_color bar_color
            v.fill
        end

        vg.path do |v|
            v.rect(0.8*w-pad,(1-v2)*pad2, 0.2*w, (v2)*(h-pad))
            v.fill_color bar_color
            v.fill
        end
    ;end
  def update_data(x);
        if(self.data != x)
            self.data = x
            self.damage_self
        end
    ;end
  def onSetup(old=nil);
        self.valueRef = OSC::RemoteParam.new($remote, "/vu-meter")
        self.valueRef.callback = lambda {|x| meter.update_data(x) }
        animate() if self.data
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @meter = self

    
context = Hash.new
context["meter"] = self
Qml::prop_add(self, "data")
self.valueRef = nil
self.data = nil

    #puts "VuMeter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynHeader
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @head = self
@logo = Qml::ZynLogo.new(database, ui_path+"1"+'/')
@menuIndent = Qml::MainMenu.new(database, ui_path+"2"+'/')
@status = Qml::Indent.new(database, ui_path+"3"+'/')
@anonymous646 = Qml::LogWidget.new(database, ui_path+"4"+'/')
@shortcuts = Qml::ZynHeaderParams.new(database, ui_path+"5"+'/')
@vumeter = Qml::VuMeter.new(database, ui_path+"6"+'/')
@panic = Qml::PanicButton.new(database, ui_path+"7"+'/')
@close = Qml::Widget.new(database, ui_path+"8"+'/')
@minimize = Qml::Widget.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["head"] = self
context["logo"] = @logo
context["menuIndent"] = @menuIndent
context["status"] = @status
context["shortcuts"] = @shortcuts
context["vumeter"] = @vumeter
context["panic"] = @panic
context["close"] = @close
context["minimize"] = @minimize
Qml::add_child(self, @logo)
Qml::add_child(self, @menuIndent)
@anonymous646.label = "console log"
Qml::add_child(@status, @anonymous646)
@status.pad = 0.01
Qml::add_child(self, @status)
Qml::add_child(self, @shortcuts)
Qml::add_child(self, @vumeter)
@panic.layoutOpts = [:no_constraint]
Qml::add_child(self, @panic)
Qml::add_child(self, @close)
Qml::add_child(self, @minimize)

    #puts "ZynHeader, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFooter < Qml::Widget
  attr_reader :foot
  attr_reader :wheelwell
  attr_reader :wheel
  attr_reader :key
  attr_reader :ctrl
  attr_reader :vel
  attr_reader :rnd
  attr_reader :oct
  attr_reader :cc
  attr_reader :cctype
  def class_name();"ZynFooter";end
  def set_cc(x);
        cc_num =  cctype.options[cctype.selected].to_i
        $remote.action("/setController", 0, cc_num, x)

    ;end
  def layout(l, selfBox);
        wheelwell.fixed(l,  selfBox, 0.010, 0.15, 0.017, (1-2*0.15))
        key.fixed(l,        selfBox, 0.032, 0.10, 0.536, 0.8)
        ctrl.fixed(l,       selfBox, 0.600, 0.05, 0.400, 0.9)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @foot = self
@wheelwell = Qml::Indent.new(database, ui_path+"1"+'/')
@wheel = Qml::ModWheel.new(database, ui_path+"2"+'/')
@key = Qml::Keyboard.new(database, ui_path+"3"+'/')
@ctrl = Qml::ParModuleRow.new(database, ui_path+"4"+'/')
@vel = Qml::Knob.new(database, ui_path+"5"+'/')
@rnd = Qml::Knob.new(database, ui_path+"6"+'/')
@oct = Qml::Knob.new(database, ui_path+"7"+'/')
@cc = Qml::Knob.new(database, ui_path+"8"+'/')
@cctype = Qml::Selector.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["foot"] = self
context["wheelwell"] = @wheelwell
@wheelwell.instance_eval do
def wheel=(k);@wheel=k;end
def wheel;    @wheel;  end
end
context["wheel"] = @wheel
context["key"] = @key
context["ctrl"] = @ctrl
context["vel"] = @vel
context["rnd"] = @rnd
context["oct"] = @oct
context["cc"] = @cc
context["cctype"] = @cctype
Qml::context_apply(context, @wheelwell)

        @wheelwell.instance_eval("def layout(l, selfBox);\n            wheel.fixed(l, selfBox, 0.1, 0.1, 0.8, 0.8)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynFooter.qml", 17)
@wheel.value = 0
@wheel.label = "wheel"
@wheel.tooltip = "modulation wheel"
@wheel.whenValue = lambda {
                $remote.action("/setController", 0, 1, (wheelwell.children[0].value*127).to_i)
            }
Qml::add_child(@wheelwell, @wheel)
Qml::add_child(self, @wheelwell)
@db.connect_property(@key.properties["velocity"], lambda {0.7*127})
@db.connect_property(@key.properties["velrnd"], lambda {0.2*127})
Qml::add_child(self, @key)
@vel.whenValue = lambda { key.velocity = vel.value*127 }
@vel.value = 0.7
@vel.label = "velocity"
@vel.tooltip = "velocity of virtual keyboard notes"
Qml::add_child(@ctrl, @vel)
@rnd.whenValue = lambda { key.velrnd = rnd.value*127 }
@rnd.value = 0.2
@rnd.label = "vrnd"
@rnd.tooltip = "velocity randomness of virtual keyboard notes"
Qml::add_child(@ctrl, @rnd)
@oct.value = 0.5
@oct.label = "octave"
@oct.whenValue = lambda {key.octave = (4*(oct.value-0.5)).to_i}
Qml::add_child(@ctrl, @oct)
@db.connect_property(@cc.properties["label"], lambda {"c.val"})
@cc.whenValue = lambda { foot.set_cc((cc.value*127).to_i) }
@cc.tooltip = "midi CC control"
Qml::add_child(@ctrl, @cc)
@cctype.label = "MIDI CC"
@cctype.tooltip = "midi CC selector"
@cctype.layoutOpts = {:weight=>2.0, :long_mode=>true}
@cctype.options = [ "01: Mod.Wheel",
                       "07: Volume",
                       "10: Panning",
                       "11: Expression",
                       "64: Sustain",
                       "65: Portamento",
                       "71: Filter Q",
                       "74: Filter Freq.",
                       "75: Bandwidth",
                       "76: FM Gain",
                       "77: Res. c. freq",
                       "78: Res. bw."]
Qml::add_child(@ctrl, @cctype)
@ctrl.layoutOpts = []
Qml::add_child(self, @ctrl)

    #puts "ZynFooter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::MainWindow
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @window = self
@side = Qml::ZynSidebar.new(database, ui_path+"1"+'/')
@head1 = Qml::ZynHeader.new(database, ui_path+"2"+'/')
@footerbox = Qml::ZynFooter.new(database, ui_path+"3"+'/')
@main_widget = Qml::Swappable.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["window"] = self
context["side"] = @side
context["head1"] = @head1
context["footerbox"] = @footerbox
context["main_widget"] = @main_widget
Qml::add_child(self, @side)
Qml::add_child(self, @head1)
Qml::add_child(self, @footerbox)
Qml::add_child(self, @main_widget)

    #puts "MainWindow, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Menu < Qml::Widget
  attr_property :extern
  attr_accessor :options
  attr_accessor :selected
  attr_accessor :whenValue
  def set_value_user(val);
        self.selected = val
        whenValue.call if whenValue
    ;end
  def layout(l,selfBox);
        if(!layoutOpts.include?(:no_constraint))
            scale = 100
            $vg.font_size scale
            bb = 1
            self.options.each do |x|
                x = x[0..8] if x.length > 8
                bbl  = $vg.text_bounds(0, 0, (x+"    ").upcase)
                bb   = [bb, bbl].max
            end
            scale *= layoutOpts[0] if layoutOpts.include?(:rescale)
            if(bb != 0)
                #Width cannot be so small that letters overflow
                l.sh([selfBox.w, selfBox.h], [-1.0, bb/scale], 0)
            end
        end
        selfBox
    ;end
  def class_name();
        "Menu"
    ;end
  def onMousePress(ev);
        create_menu
    ;end
  def draw(vg);
        text_color = Theme::TextColor
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            tx = w*pad2-h*pad2
            tw = h*pad2
            ty = h*pad
            th = h*pad2

            v.move_to(tx+0.25*tw, ty+0.3*th)
            v.line_to(tx+0.50*tw, ty+0.7*th)
            v.line_to(tx+0.75*tw, ty+0.3*th)
            v.close_path

            v.fill_color Theme::TextColor
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color text_color
        vg.text(3+w*pad*2,h/2,label.upcase)
    ;end
  def create_menu();
        n = options.length
        n = 1 if options.class != Array
        return if n < 1
        widget = DropDown.new(self.db)
        widget.w = 3*self.w
        widget.h = self.h*n
        widget.x = 0
        widget.y = 0
        widget.layer = 2
        widget.options = options
        widget.callback = lambda { |v|
            set_value_user(v)
        }
        widget.y = -self.h*(n-1) if(widget.h+global_y > window.h)
        widget.prime(root())

        Qml::add_child(self, widget)
        root.smash_draw_seq
        root.damage_item widget
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous165 = self

    
context = Hash.new
Qml::prop_add(self, "extern")
self.extern = ""
self.options = ["text", "test", "asdf"]
self.selected = 0
self.whenValue = nil

    #puts "Menu, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::NumBox < Qml::ColorBox
  attr_reader :numBox
  attr_property :height
  attr_property :align
  attr_accessor :valueRef
  def layout(l, selfBox);
        return selfBox if layoutOpts.include?(:free)

        #Assume all digit bounding boxes are roughly the same
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, " 0.12 ")
        l.aspect(selfBox, bb, scale)
        children[0].fixed(l, selfBox, 0, 0, 1, 1)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @numBox = self
@anonymous166 = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["numBox"] = self
Qml::prop_add(self, "height")
Qml::prop_add(self, "align")
@anonymous166.label = ""
@db.connect_property(@anonymous166.properties["height"], lambda {numBox.height})
@anonymous166.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous166.properties["align"], lambda {numBox.align})
Qml::add_child(self, @anonymous166)
self.pad = 0.003
self.bg = nil
self.height = 0.7
self.align = :center
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        ref = OSC::RemoteParam.new($remote, numBox.extern)
        ref.mode     = :full
        ref.callback = lambda {|x|
            numBox.children[0].label = (((10*x).to_i)/10).to_s;
            numBox.damage_self
        }
        numBox.valueRef = ref
    
end})

    #puts "NumBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::NumberButton < Qml::Widget
  attr_reader :num
  attr_reader :but
  attr_reader :pow
  attr_accessor :whenValue
  attr_property :value
  attr_property :number
  def class_name(); "NumButton" ;end
  def cb();
        whenValue.call if whenValue
    ;end
  def onMousePress(ev);
        self.value = !self.value
        damage_self
        cb()
    ;end
  def layout(l);
        selfBox = l.genBox :numButton, self
        buttBox = l.genBox :numButtont, children[0]
        numbBox = l.genBox :numButtonn, children[1]
        l.contains(selfBox, buttBox)
        l.contains(selfBox, numbBox)
        l.rightOf(numbBox, buttBox)
        l.aspect(numbBox, 1.0, 1.0)
        #l.fixed(buttBox, selfBox, 0, 0, 0.25, 1)
        #l.fixed(buttBox, selfBox, 0, 0, 0.25, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @num = self
@but = Qml::Button.new(database, ui_path+"1"+'/')
@pow = Qml::Button.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["num"] = self
context["but"] = @but
context["pow"] = @pow
Qml::prop_add(self, "value")
Qml::prop_add(self, "number")
@but.pad = 0
@db.connect_property(@but.properties["label"], lambda {num.label})
@but.whenValue = lambda { num.cb }
@but.layoutOpts = [:no_constraint]
Qml::add_child(self, @but)
@db.connect_property(@pow.properties["label"], lambda {num.number.to_s})
@pow.pad = 0
@pow.whenValue = lambda { num.cb }
@pow.layoutOpts = [:no_constraint]
Qml::add_child(self, @pow)
self.whenValue = nil
self.value = false
self.number = 0
@db.connect_watcher(self.properties["label"], lambda {begin

        but.label = num.label
    
end})

    #puts "NumberButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::PasteButton < Qml::Button
  attr_reader :paste_button
  attr_property :index
  def layout(l, selfBox);
        l.aspect(selfBox, 1, 1)
        selfBox
    ;end
  def animate();
        return if self.value == 0
        return if self.value == false
        return if self.value == true
        self.value *= 0.9
        self.value  = 0 if self.value < 0.02
        damage_self
    ;end
  def onMousePress(ev);
        return if !self.active
        self.value = 1.0
        if($remote && extern && !extern.empty?)
            if(self.index)
                $remote.action("/presets/paste", extern, self.index)
            else
                $remote.action("/presets/paste", extern)
            end
        end
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @paste_button = self

    
context = Hash.new
context["paste_button"] = self
Qml::prop_add(self, "index")
self.index = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        paste_button.tooltip = "paste to " + paste_button.extern.to_s
    
end})
self.label = "P"

    #puts "PasteButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::PlotLabelY < Qml::Widget
  attr_property :labels
  def get_width(vg, text);
        scale = 100
        $vg.font_size scale
        bb = $vg.text_bounds(0, 0, text.upcase)


        vg.font_face("bold")
        return w*scale/bb
    ;end
  def draw(vg);
        return if labels.class != Array
        best_height = h/(labels.length+1)
        xx = []
        labels.each do |l|
            tmp = get_width(vg, l)
            best_height = tmp if tmp < best_height
            xx << tmp
        end

        vg.font_size best_height
        labels.each_index do |i|
            if(i == 0)
                vg.text_align NVG::ALIGN_BOTTOM | NVG::ALIGN_LEFT
                vg.text(0, h, labels[0].upcase)
            elsif(i == labels.length-1)
                vg.text_align NVG::ALIGN_TOP | NVG::ALIGN_LEFT
                vg.text(0, 0, labels[i].upcase)
            else
                vg.text_align NVG::ALIGN_MIDDLE | NVG::ALIGN_LEFT
                vg.text(0, h*i/(labels.length-1), labels[i].upcase)
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous172 = self

    
context = Hash.new
Qml::prop_add(self, "labels")
@db.connect_property(self.properties["labels"], lambda {["-1", "0", "1"]})

    #puts "PlotLabelY, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::Record < Qml::Widget
  def draw(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            v.move_to(w*1/3, 0)
            v.line_to(w*1/3, h)
            v.move_to(w*2/3, 0)
            v.line_to(w*2/3, h)
            v.stroke
        end

        text_color = Theme::TextColor

        draw_stop(vg, text_color)
        draw_pause( vg, text_color)
        draw_play(  vg, text_color)
    ;end
  def draw_record(vg, text_color);
        r = [1/6*w, 1/2*h].min*0.4
        vg.path do |v|
            v.circle(1/6*w, 1/2*h, r)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_stop(vg, text_color);
        r = [1/6*w, 1/2*h].min*0.4
        vg.path do |v|
            v.rect(1/6*w-r, 1/2*h-r, 2*r, 2*r)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_pause(vg, text_color);
        vg.path do |v|
            v.rect((3/6-1/18)*w, 0.2*h, 1/24*w, 0.6*h)
            v.rect((3/6+1/18)*w, 0.2*h, 1/24*w, 0.6*h)
            v.fill_color text_color
            v.fill
        end
    ;end
  def draw_play(vg, text_color);
        vg.path do |v|
            v.move_to((5/6-1/18)*w, 0.25*h)
            v.line_to((5/6-1/18)*w, 0.75*h)
            v.line_to((5/6+1/18)*w, 0.5*h)
            v.close_path
            v.fill_color text_color
            v.fill
        end
    ;end
  def onMousePress(ev);
        px = (ev.pos.x-global_x)/w
        if(px > 0.666)
            $remote.action("/HDDRecorder/start")
        elsif(px > 0.333)
            $remote.action("/HDDRecorder/pause")
        else
            $remote.action("/HDDRecorder/stop")
        end

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous173 = self

    
context = Hash.new

    #puts "Record, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ScrollBar < Qml::Valuator
  attr_property :bar_size
  attr_property :style
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end
    ;end
  def draw_overlay(vg);
        pad = 1
        domain = 1.0-bar_size
        if(vertical)
            center = 0.5*bar_size + (1-value)*domain
            self.dragScale = h*(1-bar_size)
            vg.path do |v|
                v.rect(pad,pad+@h*(center-0.5*bar_size),
                    @w-2*pad,@h*bar_size-2*pad)
                v.fill_color color("56c0a5")
                v.fill
            end
        else
            center = 0.5*bar_size + value*domain
            self.dragScale = w*(1-bar_size)
            vg.path do |v|
                v.rect(pad+w*(center-0.5*bar_size),pad,
                w*bar_size-2*pad,h-2*pad)
                v.fill_color color("56c0a5")
                v.fill
            end
        end
    ;end
  def draw_normal(vg);
        pad = 1
        vg.path do |v|
            v.rect(pad,pad,w-2*pad,h-2*pad)
            v.fill_color Theme::ScrollInactive
            v.fill
        end
        domain = 1.0-bar_size
        if(vertical)
            center = 0.5*bar_size + (1-value)*domain
            self.dragScale = h*(1-bar_size)
            vg.path do |v|
                v.rect(pad,pad+@h*(center-0.5*bar_size),
                    @w-2*pad,@h*bar_size-2*pad)
                v.fill_color Theme::ScrollActive
                v.fill
            end
        else
            center = 0.5*bar_size + value*domain
            self.dragScale = w*(1-bar_size)
            vg.path do |v|
                v.rect(pad+w*(center-0.5*bar_size),pad,
                w*bar_size-2*pad,h-2*pad)
                v.fill_color Theme::ScrollActive
                v.fill
            end
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous174 = self

    
context = Hash.new
Qml::prop_add(self, "bar_size")
Qml::prop_add(self, "style")
self.bar_size = 0.1
self.style = :normal

    #puts "ScrollBar, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::SearchBox < Qml::Widget
  attr_accessor :whenValue
  def animate();
        if(root.key_widget != self)
            @next = nil
            if(@state)
                @state = nil
                damage_self
            end
            return
        end
        now = Time.new
        if(@next.nil?)
            @next = now + 0.1
            return
        elsif(@next < now)
            @state = !@state
            @next = now + 0.7
            damage_self
        end
    ;end
  def draw(vg);
        background Theme::GeneralBackground
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = Theme::TextColor
        l = label.empty? ? "search..." : label
        vg.text(8,h/2,l.upcase)
        bnd = vg.text_bounds(0,0,l.upcase)
        if(@state)
            vg.text(8+bnd,h/2,"|")
        end
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 8)
            self.label = self.label[0...-1]
        else
            self.label += k
        end
        whenValue.call if whenValue
        damage_self
    ;end
  def onMerge(val);
        self.label = val.label
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous175 = self

    
context = Hash.new
self.whenValue = nil

    #puts "SearchBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::SelButton < Qml::Button
  attr_accessor :bg
  attr_property :doupcase
  attr_property :style
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end
    ;end
  def draw_overlay(vg);
        vg.font_face("bold")
        vg.font_size h*self.textScale
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = color("56c0a5")

        ll = self.label.clone
        ll = ll.upcase if self.doupcase
        (0...ll.length).each do |i|
            ll[i] = "?" if ll.getbyte(i) > 127
        end
        vg.text(8,h/2,ll)
    ;end
  def draw_normal(vg);
        on_color      = Theme::ButtonActive
        vg.path do |v|
            v.rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad))
            if(button.value)
                v.fill_color on_color
            else
                v.fill_color bg if bg
            end
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*self.textScale
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = if(value)
            Theme::TextActiveColor
        else
            Theme::TextColor
        end
        ll = self.label.clone
        ll = ll.upcase if self.doupcase
        (0...ll.length).each do |i|
            ll[i] = "?" if ll.getbyte(i) > 127
        end
        vg.text(8,h/2,ll)
    ;end
  def onMouseEnter(ev);
        if(self.tooltip != "")
            self.root.log(:tooltip, self.tooltip)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous176 = self

    
context = Hash.new
Qml::prop_add(self, "doupcase")
Qml::prop_add(self, "style")
self.bg = nil
self.doupcase = true
self.style = :normal

    #puts "SelButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::SelColumn < Qml::Widget
  attr_reader :col
  attr_reader :scroll
  attr_accessor :valueRef
  attr_property :number
  attr_property :skip
  attr_property :rows
  attr_property :nrows
  attr_accessor :whenValue
  attr_property :oldOff
  attr_property :value_sel
  attr_property :value_lab
  attr_property :clear_on_extern
  attr_property :doupcase
  attr_property :pattern
  attr_property :style
  def clear();
        clear_sel
        children.each do |c|
            c.value = false if c.class != Qml::ScrollBar
            c.damage_self
        end
    ;end
  def clear_sel();
        self.value_sel = nil
        self.value_lab = nil
    ;end
  def onScroll(ev);
        scroll.onScroll(ev)
    ;end
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end

    ;end
  def draw_overlay(vg);
        vg.path do
            vg.rect(0, 0, w, h)
            vg.stroke_color  color("56c0a5")
            vg.stroke_width 1.0
            vg.stroke
        end
        titleH = self.h*1.0/(children.length + 1)
        vg.fill_color Theme::TextModColor
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.font_size titleH*0.9
        vg.text(0.9*w/2,titleH/2,label.upcase)
    ;end
  def draw_normal(vg);
        titleH = self.h*1.0/(children.length + 1)
        vg.path do
            vg.rect(0, 0, 0.9*w, titleH)
            vg.fill_color Theme::BankOdd
            vg.fill
        end
        vg.fill_color Theme::TextModColor
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        vg.font_size titleH*0.9
        vg.text(0.9*w/2,titleH/2,label.upcase)
    ;end
  def onSetup(old=nil);
        return if children.length > 10
        scroll.layer = self.layer
        (self.nrows).times do |x|
            ch = if(number)
                Qml::NumberButton.new(db)
            else
                Qml::SelButton.new(db)
            end
            if(rows.nil? || rows.empty?)
                ch.label = ""
            else
                stride = 1
                stride = 2 if skip
                ch.label = if(x>=rows.length/stride)
                    ""
                else
                    rows[x*stride]
                end
                ch.tooltip = if(x*stride+1>=rows.length)
                    ""
                else
                    rows[x*stride+1]
                end

            end
            ch.layoutOpts = [:no_constraint]
            ch.whenValue  = lambda {col.cb ch}

            ch.layer      = col.layer
            ch.number = x if number
            ch.style    = self.style    if ch.respond_to? :style
            ch.doupcase = self.doupcase if ch.respond_to? :doupcase
            if(!number)
                ch.pad = 0
                ch.bg = Theme::BankEven if x%2 == 0
                ch.bg = Theme::BankOdd if x%2 == 1
            end
            Qml::add_child(self, ch)
        end

    ;end
  def tryScroll();
        #24 items are visible at a time
        #center is at item nrows/2     at value 0
        #center is at item len-nrows/2 at value 1
        stride = 1
        stride = 2 if skip
        return if rows.nil?
        n = rows.length/stride
        return if n<nrows
        center = (n-nrows)*(1-scroll.value)+nrows/2
        off    = (center-nrows/2).to_i
        if(off != self.oldOff)
            setValue(rows, off)
        end
    ;end
  def cb(ch);
        if(ch.value)
            self.value_sel = ch.tooltip
            self.value_lab = ch.label
        else
            self.value_sel = nil
            self.value_lab = nil
        end

        children[1,99].each do |child|
            if(ch != child)
                child.value = false
            end
        end
        whenValue.call if whenValue
        damage_self
    ;end
  def filter(x);
        y = []
        x.each do |x|
            next if(x != "." && x != ".." && x[0] == ".")
            next if(x[-1] == "~")
            next if self.pattern && !self.pattern.match(x)
            y << x
        end
        y
    ;end
  def setValue(x,offset=0);
        return if(self.rows == x && offset == oldOff)
        self.oldOff = offset
        self.rows = x
        stride = 1
        stride = 2 if skip
        nn = x.length/stride
        nsize = [0.05, [1.0, self.nrows/nn].min].max
        if(nsize != scroll.bar_size)
            scroll.bar_size = nsize
            scroll.damage_self
        end
        n = [x.length/stride, children.length-1].min
        (1..n).each do |i|
            nv = (children[i].label == self.value_lab &&
                  children[i].tooltip == self.value_sel)
            children[i].label   = x[(i-1+offset)*stride]
            children[i].tooltip = x[(i-1+offset)*stride+1] if stride == 2
            children[i].value   = nv
        end
        ((n+1)...children.length).each do |i|
            children[i].label = ""
            children[i].value = false
        end
        damage_self
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.9, 0.0, 0.1, 1.0)
        miniBox = nil
        if(selfBox.class != LayoutConstBox)
            miniBox = l.genBox :selminibox, nil
            l.fixed(miniBox, selfBox, 0.0, 0.0, 0.9, 1.0)
        else
            miniBox = l.genConstBox(0.0, 0.0, 0.9*selfBox.w, selfBox.h)
        end
        titleH  = 1.0/(children.length + 1)
        Draw::Layout::vpack(l, miniBox, children[1,99], 0, 1, 0, titleH, 1.0-titleH)
        selfBox
    ;end
  def selected();
        return "" if(value_lab.nil?)
        return value_lab
    ;end
  def selected_val();
        return "" if(value_sel.nil?)
        return value_sel
    ;end
  def selected_id();
        self.children.each_with_index do |c,i|
            next       if c.class == Qml::ScrollBar
            return i-1 if c.value
        end
        nil
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @col = self
@scroll = Qml::ScrollBar.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["col"] = self
context["scroll"] = @scroll
Qml::prop_add(self, "number")
Qml::prop_add(self, "skip")
Qml::prop_add(self, "rows")
Qml::prop_add(self, "nrows")
Qml::prop_add(self, "oldOff")
Qml::prop_add(self, "value_sel")
Qml::prop_add(self, "value_lab")
Qml::prop_add(self, "clear_on_extern")
Qml::prop_add(self, "doupcase")
Qml::prop_add(self, "pattern")
Qml::prop_add(self, "style")
@db.connect_property(@scroll.properties["style"], lambda {col.style})
@scroll.whenValue = lambda { col.tryScroll }
Qml::add_child(self, @scroll)
self.valueRef = nil
self.number = false
self.skip = false
self.rows = nil
self.nrows = 24
self.whenValue = nil
self.oldOff = 0
self.value_sel = nil
self.value_lab = nil
self.clear_on_extern = nil
self.doupcase = true
self.pattern = nil
self.style = :normal
@db.connect_watcher(self.properties["extern"], lambda {begin

        col.valueRef = OSC::RemoteParam.new($remote, col.extern)
        col.valueRef.callback = Proc.new {|x|
            x = [x] if x.class != Array
            col.clear_sel if col.clear_on_extern
            col.setValue(col.filter(x)) if !col.skip
            col.setValue(x)             if  col.skip
        }
    
end})

    #puts "SelColumn, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::SidebarButton < Qml::Widget
  attr_reader :button
  attr_accessor :whenClick
  attr_property :value
  def onMousePress(ev);
        self.value = !self.value
        damage_self
        self.whenClick.call if whenClick
    ;end
  def draw_button(vg);
        pad = 0
        off_color     = Theme::ButtonInactive
        on_color      = Theme::ButtonActive
        cs = 0
        vg.path do |v|
            v.rounded_rect(w*pad, h*pad, w*(1-2*pad), h*(1-2*pad), 3)
            if(button.value == true)
                cs = 1
                v.fill_color on_color
            elsif(button.value.class == Float && button.value != 0)
                cs = 2
                t = button.value
                on = on_color
                of = Theme::ButtonGrad1
                v.fill_color(color_rgb(on.r*t + of.r*(1-t),
                                       on.g*t + of.g*(1-t),
                                       on.b*t + of.b*(1-t)))
            else
                paint = v.linear_gradient(0,0,0,h,
                Theme::ButtonGrad1, Theme::ButtonGrad2)
                v.fill_paint paint
            end
            v.fill
            v.stroke_width 1
            v.stroke

        end

        vg.path do |v|
            hh = h/20
            v.move_to(w*pad+3,       h*pad+hh)
            v.line_to(w*(1-2*pad)-2, h*pad+hh)
            if(cs == 0)
                v.stroke_color color("5c5c5c")
            elsif(cs == 1)
                v.stroke_color color("16a39c")
            end
            if([0,1].include?(cs))
                v.stroke_width hh 
                v.stroke
            end
        end
    ;end
  def draw(vg);
        #h = button.h
        #w = button.w
        ##puts("drawing a button...")
        #bright_green = color("00AE9C")
        #dark_green   = color("007368")
        #grad = vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        #vg.stroke_color(NVG.rgba(0,0,0,0x80))

        #grey1 = Theme::ButtonGrad1
        #grey2 = Theme::ButtonGrad2
        #grad2 = vg.linear_gradient(0,0,0,h, grey1, grey2)
        #vg.path do |v|
        #    v.rounded_rect(0,0,w,h,2)
        #    if(value)
        #        bright_green = color("00AE9C")
        #        dark_green   = color("007368")
        #        grad = vg.linear_gradient(0,0,0,h, bright_green, dark_green)
        #        v.fill_paint grad
        #    else
        #        v.fill_paint grad2
        #    end
        #    v.stroke_width 1
        #    v.fill
        #    v.stroke
        #end
        draw_button(vg)

        text_color1   = color("52FAFE")
        text_color2   = Theme::TextColor
        vg.font_face("bold")
        vg.font_size h*0.55
        vg.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
        if(value)
            vg.fill_color(text_color1)
        else
            vg.fill_color(text_color2)
        end
        vg.text(w/2,h/2,label.upcase)

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @button = self

    
context = Hash.new
context["button"] = self
Qml::prop_add(self, "value")
self.whenClick = nil
self.value = false

    #puts "SidebarButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TabGroup < Qml::Widget
  def layout(l, selfBox);
        selfBox = Draw::Layout::tabpack(l, selfBox, self)
    ;end
  def get_tab(wid);
        n = children.length
        selected = 0
        tab_id = 0
        (0..n).each do |ch_id|
            child = children[ch_id]
            if(child.class == Qml::TabButton)
                if(wid == child)
                    child.value = true
                    selected = tab_id
                else
                    if(child.value)
                        child.value = false
                        child.damage_self
                    end
                end
                tab_id += 1
            end
        end
        selected
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous180 = self

    
context = Hash.new

    #puts "TabGroup, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextBox < Qml::ColorBox
  attr_reader :textBox
  attr_property :height
  attr_property :align
  attr_accessor :valueRef
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0, 0, 1, 1)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @textBox = self
@anonymous181 = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["textBox"] = self
Qml::prop_add(self, "height")
Qml::prop_add(self, "align")
@db.connect_property(@anonymous181.properties["label"], lambda {textBox.label})
@db.connect_property(@anonymous181.properties["height"], lambda {textBox.height})
@anonymous181.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous181.properties["align"], lambda {textBox.align})
Qml::add_child(self, @anonymous181)
self.pad = 0.003
self.bg = nil
self.height = 0.7
self.align = :center
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        ref = OSC::RemoteParam.new($remote, textBox.extern)
        ref.callback = lambda {|x|
            textBox.label = x;
            textBox.damage_self
        }
        textBox.valueRef = ref
    
end})

    #puts "TextBox, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextField < Qml::Widget
  attr_reader :text_field
  attr_accessor :valueRef
  attr_property :style
  attr_accessor :whenValue
  def animate();
        if(root.key_widget != self)
            @next = nil
            if(@state)
                @state = nil
                damage_self
            end
            return
        end
        now = Time.new
        if(@next.nil?)
            @next = now + 0.1
            return
        elsif(@next < now)
            @state = !@state
            @next = now + 0.7
            damage_self
        end
    ;end
  def draw(vg);
        if(self.style == :overlay)
            draw_overlay(vg)
        else
            draw_normal(vg)
        end

    ;end
  def draw_overlay(vg);
        background color("1b1c1c")
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = color("56c0a5")
        l = "..."
        if(label.class == String && !label.empty?)
            l = label.clone
        elsif(label.class == Array)
            l = label[0].clone
        end

        (0...l.length).each do |i|
            l[i] = "?" if l.getbyte(i) > 127
        end

        bnd = vg.text_bounds(0,0,l+"|")
        if(bnd+8 > self.w)
            vg.font_size self.h*self.w/(bnd+8)*0.8
            bnd = vg.text_bounds(0,0,l)
        end

        vg.text(8,h/2,l)
        if(@state)
            vg.text(8+bnd,h/2,"|")
        end
    ;end
  def draw_normal(vg);
        background Theme::GeneralBackground
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = Theme::TextColor
        l = label.empty? ? "..." : label.clone

        (0...l.length).each do |i|
            l[i] = "?" if l.getbyte(i) > 127
        end

        vg.text(8,h/2,l)
        bnd = vg.text_bounds(0,0,l)
        if(@state)
            vg.text(8+bnd,h/2,"|")
        end
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 8)
            self.label = self.label[0...-1]
        else
            self.label += k
        end
        whenValue.call if whenValue
        valueRef.value = self.label if valueRef
        damage_self
    ;end
  def onMousePress(ev);
    ;end
  def onMerge(val);
        self.label = val.label
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @text_field = self

    
context = Hash.new
context["text_field"] = self
Qml::prop_add(self, "style")
self.valueRef = nil
self.style = :normal
self.whenValue = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        ref = OSC::RemoteParam.new($remote, text_field.extern)
        ref.callback = lambda {|x|
            text_field.label = x;
            text_field.damage_self
        }
        text_field.valueRef = ref
    
end})

    #puts "TextField, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TextLine < Qml::Widget
  attr_reader :textedit
  attr_accessor :whenValue
  attr_accessor :valueRef
  attr_property :upcase
  attr_property :ext
  def draw(vg);
        background Theme::GeneralBackground
        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color = Theme::TextColor
        l = label.empty? ? "..." : label.clone
        l = l+self.ext if self.ext && !l.end_with?(self.ext)
        l = l.upcase if self.upcase
        (0...l.length).each do |i|
            l[i] = "?" if l.getbyte(i) > 127
        end
        vg.text(8,h/2,l)
    ;end
  def onKey(k, mode);
        return if mode != "press"
        if(k.ord == 8)
            self.label = self.label[0...-1]
        else
            self.label += k
        end
        whenValue.call if whenValue
        valueRef.value = self.label if valueRef
        damage_self
    ;end
  def onMerge(val);
        self.label = val.label
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @textedit = self

    
context = Hash.new
context["textedit"] = self
Qml::prop_add(self, "upcase")
Qml::prop_add(self, "ext")
self.whenValue = nil
self.valueRef = nil
self.upcase = true
self.ext = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        ref = OSC::RemoteParam.new($remote, textedit.extern)
        ref.callback = lambda {|x|
            textedit.label = x;
            textedit.damage_self
        }
        textedit.valueRef = ref
    
end})

    #puts "TextLine, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::TitleBar < Qml::Widget
  attr_reader :title_bar
  attr_reader :title
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,self.w,self.h)
            v.fill_color Theme::TitleBar
            v.fill
            v.stroke
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @title_bar = self
@title = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["title_bar"] = self
context["title"] = @title
@db.connect_property(@title.properties["label"], lambda {title_bar.label})
@title.height = 0.8
Qml::add_child(self, @title)

    #puts "TitleBar, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VGroup < Qml::ColorBox
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous187 = self

    
context = Hash.new
self.bg = nil
self.layoutOpts = [:think_of_the_children]

    #puts "VGroup, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisEq < Qml::Widget
  attr_reader :vis_eq
  attr_reader :data_view
  attr_accessor :valueRef
  attr_property :coeff
  attr_property :doUpdate
  def refresh();
        self.valueRef.refresh if valueRef
    ;end
  def update_coeff(x);
        return if self.coeff == x
        n2 = x.length/2
        self.coeff = x
        self.doUpdate = true
    ;end
  def animate();
        return if !doUpdate
        self.doUpdate = false
        xpts = Draw::DSP::logspace(1, 20000, 256)
        pts  = eq_response(xpts, coeff)
        ypts = []
        pts.each do |pt|
            ypts << Math.log10(pt)/4
        end
        data_view.data = ypts
        data_view.damage_self
    ;end
  def draw(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        vg.path do
            vg.rect(box.x, box.y, box.w, box.h)
            vg.fill_color Theme::VisualBackground
            vg.fill
        end
        Draw::Grid::log_x(vg, 1, 20000, box)
        Draw::Grid::linear_y(vg, 1, 20, box)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vis_eq = self
@data_view = Qml::DataView.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["vis_eq"] = self
context["data_view"] = @data_view
Qml::prop_add(self, "coeff")
Qml::prop_add(self, "doUpdate")
@data_view.normal = false
@data_view.pad = 0
@data_view.fixedpad = 5
Qml::add_child(self, @data_view)
self.valueRef = nil
self.coeff = nil
self.doUpdate = false
@db.connect_watcher(self.properties["extern"], lambda {begin

        vis_eq.valueRef = OSC::RemoteParam.new($remote, vis_eq.extern)
        vis_eq.valueRef.callback = lambda {|x| vis_eq.update_coeff x}
    
end})

    #puts "VisEq, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisFilter < Qml::Widget
  attr_reader :vis_filter
  attr_reader :data_view
  attr_accessor :valueRef
  attr_property :doUpdate
  attr_property :coeff
  def animate();
        return if !doUpdate
        self.doUpdate = false
        x = self.coeff
        b = nil
        a = nil
        if(x.length == 7)
            b = x[1..3]
            a = x[4..6]
        elsif(x.length == 5)
            b = x[1..2]
            a = x[3..4]
        end
        vis_filter.coeff = x;
        xpts = Draw::DSP::logspace(50, 30000, 256)
        xnorm = []
        xpts.each do |pt|
            xnorm << pt / 48000.0
        end
        yy = Draw::opt_magnitude(b, a, xnorm, x[0]+1)
        ypts = []
        yy.each do |pt|
            ypts << Math::log10(pt)/4
        end
        #vis_filter.xpoints = xpts
        data_view.data = ypts
        data_view.damage_self
    ;end
  def update_coeff(x);
       return if(x == vis_filter.coeff)
       self.coeff = x
       self.doUpdate = true
    ;end
  def refresh();
        self.valueRef.refresh if valueRef
    ;end
  def draw(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        background Theme::VisualBackground
        Draw::Grid::log_x(vg, 50, 30000, box)
        Draw::Grid::linear_y(vg, 1, 20, box)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vis_filter = self
@data_view = Qml::DataView.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["vis_filter"] = self
context["data_view"] = @data_view
Qml::prop_add(self, "doUpdate")
Qml::prop_add(self, "coeff")
@data_view.normal = false
@data_view.pad = 0
@data_view.fixedpad = 5
Qml::add_child(self, @data_view)
self.valueRef = nil
self.doUpdate = false
self.coeff = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        return if vis_filter.extern.nil?
        vis_filter.valueRef = OSC::RemoteParam.new($remote, vis_filter.extern)
        vis_filter.valueRef.callback = lambda {|x| vis_filter.update_coeff x}
    
end})

    #puts "VisFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisFormant < Qml::Widget
  attr_reader :vfor
  attr_reader :data_view
  attr_accessor :valueRef
  attr_property :numformants
  attr_property :q_value
  attr_property :stages
  attr_property :gain_value
  attr_property :vowels
  attr_property :pending_damage
  attr_property :refreshing
  attr_property :vowel_num
  def change();
        self.pending_damage = true
    ;end
  def refresh();
        self.refreshing = true
    ;end
  def onSetup(old=nil);
        base = vfor.extern
        refs = []

        #Vowel data
        val = OSC::RemoteParam.new($remote, base+"vowels")
        val.callback = lambda {|x|
            nvowels   = x[0]
            nformants = x[1]
            vs = []
            (0...nvowels).each do |vv|
                v = []
                (0...nformants).each do |f|
                    ind = 2 + 3*nformants*vv + 3*f
                    v << Formant.new(x[ind+0], x[ind+1], x[ind+2])
                end
                vs << v
            end
            vfor.vowels = vs;
            vfor.change
        }
        refs << val

        #num formants
        val = OSC::RemoteParam.new($remote, base+"Pnumformants")
        val.mode = :full
        val.callback = lambda {|x|
            vfor.numformants = x
            vfor.change
        }
        refs << val

        #q
        val = OSC::RemoteParam.new($remote, base+"q_value")
        val.callback = lambda {|x|
            vfor.q_value = x
            vfor.change
        }
        refs << val

        #stages
        val = OSC::RemoteParam.new($remote, base+"Pstages")
        val.callback = lambda {|x|
            vfor.stages = x + 1
            vfor.change
        }
        refs << val

        #gain
        val = OSC::RemoteParam.new($remote, base+"Pgain")
        val.callback = lambda {|x|
            vfor.gain_value = x
            vfor.change
        }
        refs << val
        self.valueRef = refs
    ;end
  def draw(vg);
        padfactor = 10
        bb = Draw::indent(Rect.new(0,0,w,h), padfactor, padfactor)
        background Theme::VisualBackground
        Draw::Grid::log_x(vg, 1, 20000, bb)
        Draw::Grid::linear_y(vg, 1, 20, bb)
    ;end
  def animate();
        if(self.refreshing)
            self.refreshing = false
            valueRef.each do |r|
                r.refresh()
            end
        end
        if(self.pending_damage)
            self.pending_damage = false
            response(self.vowel_num) if !self.vowels.nil?
        end
    ;end
  def response(nvowel);
        xpts = Draw::DSP::logspace(1, 20000, 256)
        vo   = self.vowels[nvowel]
        return if vo.nil?
        fo   = vo
        fo   = fo[0...numformants] if numformants <= fo.length
        ypts = formant_filter_response(xpts, fo, q_value, stages,
               gain_value)
        data_view.data = ypts
        data_view.damage_self
        damage_self
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vfor = self
@data_view = Qml::DataView.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["vfor"] = self
context["data_view"] = @data_view
Qml::prop_add(self, "numformants")
Qml::prop_add(self, "q_value")
Qml::prop_add(self, "stages")
Qml::prop_add(self, "gain_value")
Qml::prop_add(self, "vowels")
Qml::prop_add(self, "pending_damage")
Qml::prop_add(self, "refreshing")
Qml::prop_add(self, "vowel_num")
@data_view.normal = true
@data_view.pad = 0
@data_view.fixedpad = 5
Qml::add_child(self, @data_view)
self.valueRef = nil
self.numformants = 4
self.q_value = 1
self.stages = 1
self.gain_value = 1.0
self.vowels = nil
self.pending_damage = false
self.refreshing = false
self.vowel_num = 0

    #puts "VisFormant, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisHarmonic < Qml::Widget
  attr_reader :hm
  attr_accessor :pad
  attr_accessor :valueRef
  attr_property :points
  attr_property :needRefresh
  def setValue(x);
        if(x != self.points)
            self.points = x
            damage_self
        end
    ;end
  def animate();
        if(self.valueRef && self.needRefresh)
            self.needRefresh = false
            self.valueRef.refresh
        end
    ;end
  def refresh();
        self.needRefresh = true
    ;end
  def draw(vg);
        pad2 = (1-2*pad)
        box = Rect.new(w*pad, h*pad, w*pad2, h*pad2)

        background Theme::VisualBackground

        Draw::Grid::linear_x(vg,0,10,box, 0.5)
        Draw::Grid::linear_y(vg,0,10,box, 0.5)

        xpoints = Draw::DSP::linspace(-2,2,128)
        ypoints = nil
        vline   = 0.15
        if(self.points)
            ypoints = self.points[1..-1].map {|x| 2*x-1}
            vline   = self.points[0]/2
        else
            ypoints = xpoints.map {|x| 2*Math.exp(-x**2/0.1)-1 }
        end

        Draw::WaveForm::plot(vg, ypoints, box)

        vg.path do |v|
            v.move_to(0.5*w+vline*w, h*pad)
            v.line_to(0.5*w+vline*w, h*pad2)
            v.move_to(0.5*w-vline*w, h*pad)
            v.line_to(0.5*w-vline*w, h*pad2)
            v.stroke_color Theme::VisualStroke
            v.stroke
        end

        vg.path do |v|
            v.rect(0.5*w-vline*w, h*pad, 2*vline*w, h*pad2)
            v.fill_color Theme::VisualLightFill
            v.fill
        end
        vg.path do |v|
            v.move_to(pad*w, h*(pad+pad2))
            n = ypoints.length
            ypoints.each_with_index do |y, i|
                v.line_to(pad*w+pad2*w*i/n, (1-y)/2*h)
            end
            v.line_to((pad+pad2)*w,(pad+pad2)*h)
            v.close_path
            v.fill_color Theme::VisualLightFill
            v.fill
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @hm = self

    
context = Hash.new
context["hm"] = self
Qml::prop_add(self, "points")
Qml::prop_add(self, "needRefresh")
self.pad = 1/32
self.valueRef = nil
self.points = nil
self.needRefresh = false
@db.connect_watcher(self.properties["extern"], lambda {begin

        hm.valueRef = OSC::RemoteParam.new($remote, hm.extern)
        hm.valueRef.callback = Proc.new {|x| hm.setValue(x)}
    
end})

    #puts "VisHarmonic, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisResonance < Qml::Widget
  attr_reader :visreson
  attr_reader :draw_layer
  attr_accessor :prev
  attr_accessor :whenValue
  attr_accessor :valueRef
  def refresh();
        valueRef.refresh if valueRef
    ;end
  def un_rescale(x);
        o = []
        x.each do |xx|
            o << 2*xx-1
        end
        o
    ;end
  def rescale(x);
        o = []
        x.each do |xx|
            o << (xx+1)/2
        end
        o
    ;end
  def animate();
        if(@send_update)
            @send_update = false
            self.valueRef.value = rescale(draw_layer.points)
        end
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            v.fill_color   Theme::VisualBackground
            v.fill
        end

        pad = 5
        bb  = Rect.new(pad,pad,w-2*pad,h-2*pad)
        Draw::Grid::linear_x(vg, 0, 10, bb)
        Draw::Grid::linear_y(vg, 0, 10, bb)
    ;end
  def runUpdate(pos, type);
        rely  = 1 - 2*(pos.y - global_y) / h
        relx  = (pos.x - global_x) / w
        return if(relx < 0 || relx > 1)
        rely  = [1, [-1, rely].max].min

        n = draw_layer.points.length
        sel = (n*relx).to_i
        draw_layer.points[sel] = rely

        #Check to see if interpolation is needed
        if(type == :move && ![prev-1, prev, prev+1].include?(sel))
            srt = [sel,prev].min
            dst = [sel,prev].max
            a   = draw_layer.points[srt]
            b   = draw_layer.points[dst]
            (srt...dst).each do |i|
                ri = (i-srt)/(dst-srt)
                draw_layer.points[i] = a+ri*(b-a) if i >= 0 && i<draw_layer.points.length
            end
        end
        self.prev = sel
        draw_layer.damage_self
        @send_update = true
    ;end
  def onMousePress(ev);
        runUpdate(ev.pos, :press)
    ;end
  def onMouseMove(ev);
        runUpdate(ev.pos, :move)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @visreson = self
@draw_layer = Qml::DataViewAlt.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["visreson"] = self
context["draw_layer"] = @draw_layer
Qml::add_child(self, @draw_layer)
self.prev = nil
self.whenValue = nil
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        visreson.valueRef =
            OSC::RemoteParam.new($remote, visreson.extern)
        visreson.valueRef.callback = lambda {|x|
            draw_layer.points = visreson.un_rescale(x)
            draw_layer.damage_self
        }
    
end})

    #puts "VisResonance, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VisSubHarmonics < Qml::Widget
  attr_reader :vis_sub
  attr_reader :data_view
  attr_accessor :valueRef
  attr_property :next_coeff
  def refresh();
        valueRef.refresh if valueRef
    ;end
  def update_coeff(x);
        x = [x] if x.class != Array
        self.next_coeff = x
    ;end
  def do_update_coeff(x);
        now = Time.new
        xpts = Draw::DSP::logspace(400, 20000, 512)
        tmp = sub_synth_response(xpts, x)
        ypts = tmp.map { |x| Math.log10(x)/4 }

        #max  = Draw::DSP::ary_max ypts
        #ypts = ypts.map { |x| x - max + 1}

        data_view.data = ypts
        root.damage_item data_view
        t2 = Time.new
    ;end
  def draw(vg);
        fixedpad = 5
        background Theme::VisualBackground
        box = Rect.new(fixedpad, fixedpad, w-2*fixedpad, h-2*fixedpad)
        Draw::Grid::log_x(vg, 400, 20000, box)
        Draw::Grid::linear_y(vg, 1, 10, box, 1, 40)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vis_sub = self
@data_view = Qml::DataView.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["vis_sub"] = self
context["data_view"] = @data_view
@data_view.instance_eval do
def vis_sub=(k);@vis_sub=k;end
def vis_sub;    @vis_sub;  end
end
Qml::context_apply(context, @data_view)
Qml::prop_add(self, "next_coeff")

        @data_view.instance_eval("def animate();\n            if(vis_sub.next_coeff)\n                vis_sub.do_update_coeff(vis_sub.next_coeff)\n                vis_sub.next_coeff = nil\n            end\n        ;end", "src/mruby-zest/example/VisSubHarmonics.qml", 55)
@data_view.normal = false
@data_view.pad = 0.0
@data_view.fixedpad = 5
Qml::add_child(self, @data_view)
@db.connect_watcher(self.properties["extern"], lambda {begin

        return if vis_sub.extern.nil?
        #meta = OSC::RemoteMetadata.new($remote, vis_sub.extern)

        vis_sub.valueRef = OSC::RemoteParam.new($remote, vis_sub.extern)
        vis_sub.valueRef.callback = lambda {|x| vis_sub.update_coeff x}
    
end})
self.valueRef = nil
self.next_coeff = nil

    #puts "VisSubHarmonics, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::VuMeter < Qml::Widget
  attr_reader :meter
  attr_accessor :valueRef
  attr_property :data
  def animate();
        self.valueRef.refresh if(self.valueRef)
    ;end
  def rap2dB(x); 20*Math::log10(x) ;end
  def bound(x); [0.0, [1.0, x].min].max ;end
  def cv(x);min_db = -40;bound((min_db-rap2dB(x))/min_db);end
  def draw(vg);
        indent_color = Theme::VisualBackground
        background indent_color

        bar_color = Theme::VisualLine
        pad  = 3
        pad2 = (h-2*pad)
        v1 = 0.3
        v2 = 0.5
        if(!data.nil?)
            v1 = cv(data[0])
            v2 = cv(data[1])
        end

        vg.path do |v|
            v.rect(pad,(1-v1)*pad2, 0.2*w, (v1)*(h-pad))
            v.fill_color bar_color
            v.fill
        end

        vg.path do |v|
            v.rect(0.8*w-pad,(1-v2)*pad2, 0.2*w, (v2)*(h-pad))
            v.fill_color bar_color
            v.fill
        end
    ;end
  def update_data(x);
        if(self.data != x)
            self.data = x
            self.damage_self
        end
    ;end
  def onSetup(old=nil);
        self.valueRef = OSC::RemoteParam.new($remote, "/vu-meter")
        self.valueRef.callback = lambda {|x| meter.update_data(x) }
        animate() if self.data
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @meter = self

    
context = Hash.new
context["meter"] = self
Qml::prop_add(self, "data")
self.valueRef = nil
self.data = nil

    #puts "VuMeter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::WaveView < Qml::Widget
  attr_reader :wave_view
  attr_reader :data_view
  attr_property :grid
  attr_accessor :valueRef
  attr_property :phase
  attr_property :noise
  attr_property :noise_mode
  def class_name(); "WaveView" ;end
  def set_noise_mode(x);
        return if x == self.noise_mode
        self.noise_mode = x
        if(self.noise_mode)
            #Draw a N
            u = []
            20.times  { u << 0.0 }
            10.times  { u << 0.8 }
            20.times  {|x| u << (x/19)*-1.0+0.8 }
            10.times  { u << 0.8 }
            20.times  { u << 0.0 }
            l = []
            20.times  { l << 0.0 }
            10.times   { l << -0.8 }
            20.times  {|x| l << (x/19)*-1.0+0.2 }
            10.times   { l << -0.8 }
            20.times  { l << 0.0 }
            data_view.ignore_phase = true
            data_view.data = [u, l]
            data_view.damage_self
        else
            data_view.ignore_phase = false
            self.valueRef[0].refresh if valueRef
        end
    ;end
  def draw(vg);
        pad  = 1/32
        pad2 = (1-2*pad)
        box = Rect.new(w*pad, h*pad, w*pad2, h*pad2)

        background Theme::VisualBackground

        if(grid)
            Draw::Grid::linear_x(vg,0,10,box, 0.5)
            Draw::Grid::linear_y(vg,0,10,box, 0.5)
        end

        if(extern.nil? || extern.empty?)
            Draw::WaveForm::sin(vg, box, 128)
        end
    ;end
  def refresh();
        return if self.valueRef.nil?
        self.valueRef.each do |v|
            v.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @wave_view = self
@data_view = Qml::DataView.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["wave_view"] = self
context["data_view"] = @data_view
Qml::prop_add(self, "grid")
Qml::prop_add(self, "phase")
Qml::prop_add(self, "noise")
Qml::prop_add(self, "noise_mode")
@db.connect_property(@data_view.properties["phase"], lambda {wave_view.phase})
Qml::add_child(self, @data_view)
self.grid = true
self.valueRef = nil
self.phase = 0.0
self.noise = nil
self.noise_mode = false
@db.connect_watcher(self.properties["extern"], lambda {begin

        data = OSC::RemoteParam.new($remote, wave_view.extern)
        data.callback = lambda {|x|
            data_view.data = x if !wave_view.noise_mode
            data_view.damage_self
        }
        noise = nil
        if(wave_view.noise)
            noise = OSC::RemoteParam.new($remote, wave_view.noise)
            noise.callback = lambda {|x|
                wave_view.set_noise_mode(x != 0)
            }
        end
        if(noise)
            wave_view.valueRef = [data, noise]
        else
            wave_view.valueRef = [data]
        end
    
end})
@db.connect_watcher(self.properties["phase"], lambda {begin

        return if((data_view.phase*100).to_i == (wave_view.phase*100).to_i)
        data_view.phase = wave_view.phase

        data_view.damage_self
    
end})

    #puts "WaveView, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAbout < Qml::Widget
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous203 = self
@anonymous202 = Qml::TextScroll.new(database, ui_path+"1"+'/')

    
context = Hash.new
@db.connect_property(@anonymous202.properties["label"], lambda {begin

            l  = ""
            l << "ZynAddSubFX is a powerful open source software synthesizer\n"
            l << "Zyn-Fusion is the completely rewritten interface for ZynAddSubFX's 3.0.0 Release\n"
            l << "\n"
            l << "Maintainer (2009-2016):\n"
            l << "    Mark McCurry\n"
            l << "\n"
            l << "Maintainer (2002-2010):\n"
            l << "    Nasca Octavian Paul\n"
            l << "\n"
            l << "3.0.0 UI & Icon Design:\n"
            l << "    Budislav Stepanov\n"
            l << "\n"
            l << "Contributors:\n"
            l << "    Hans Petter Selasky (BSD Compat)\n"
            l << "    Christopher Oliver (Unison + presets fix, mousewheel support,\n"
            l << "                        SUBnote overtones, unison enhancements, ...)\n"
            l << "    Gerald Folcher (legato, mono notes memory)\n"
            l << "    Lars Luthman (zombie fix,jack midi, LASH support)\n"
            l << "    Daniel Clemente (with a workaround of X11 repeated key bug)\n"
            l << "    Emmanuel Saracco (fix for JACK output)\n"
            l << "    Achim Settelmeier (QUERTZ keyboard layout for virtual keyboard)\n"
            l << "    Jérémie Andréi (AZERTY keyboard layout, Array index fix, OSS failsafe)\n"
            l << "    Alexis Ballier (const char* <-> string mismatch, NULLMidi prototype fix)\n"
            l << "    Tobias Doerffel (static-instance variables fix, missing include fix)\n"
            l << "    James Morris (Memory leaks in FLTK GUI)\n"
            l << "    Alan Calvert (Portions of New IO)\n"
            l << "    Stephen Parry (DSSI rebuild)\n"
            l << "    Ryan Billing (APhaser)\n"
            l << "    Hans Petter Selasky (OSS Midi, FreeBSD support, Bank UI bug fix)\n"
            l << "    Damien Goutte-Gattat (Bank select midi support)\n"
            l << "    Lieven Moors (Spike/Circle waveform)\n"
            l << "    Olaf Schulz (MIDI Aftertouch support)\n"
            l << "    Jonathan Liles (NSM & NTK support)\n"
            l << "    Johannes Lorenz (Effect Documentation, PID Files, Testing Extrodanare)\n"
            l << "    Ilario Glasgo (Italian Doc Translation)\n"
            l << "    Filipe Coelho (Globals Cleanup, LV2/VST Support)\n"
            l << "    Andre Sklenar (UI Pixmaps)\n"
            l << "    Harald Hvaal (General Code Modernization)\n"
            l << "    Olivier Jolly (DSSI Bank Load Fix)\n"
            l << "\n"
            l << "This is the 3.0.0 version of ZynAddSubFX\n"
            l << "For more information please see http://zynaddsubfx.sourceforge.net/\n"
            l
        
end})
Qml::add_child(self, @anonymous202)

    #puts "ZynAbout, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAddGlobal < Qml::Widget
  attr_reader :addglobal
  attr_reader :row1
  attr_reader :row2
  attr_reader :amp_gen
  attr_reader :amp_env
  attr_reader :amp_lfo
  attr_reader :footer
  attr_accessor :valueRef
  attr_property :filtertype
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.55, 0.40, 0.05])
        selfBox
    ;end
  def set_view();
        subsubview = root.get_view_pos(:subsubview)
        types = [:amplitude, :frequency, :filter]
        if(!types.include?(subsubview))
            subsubview = :amplitude
            root.set_view_pos(:subsubview, subsubview)
        end
        vis = root.get_view_pos(:vis)
        types = [:envelope, :lfo, :filter]
        if(!types.include?(vis))
            vis = :envelope
            root.set_view_pos(:vis, vis)
        end

        if(subsubview == :amplitude)
            set_amp(self.extern)
        elsif(subsubview == :frequency)
            set_freq(self.extern)
        elsif(subsubview == :filter)
            set_filter(self.extern)
        end

        if(vis == :lfo)
            set_vis_lfo(self.extern, subsubview)
        elsif(vis == :envelope)
            set_vis_env(self.extern, subsubview)
        elsif(vis == :filter)
            set_vis_filter(self.extern, subsubview)
        end
        db.update_values
    ;end
  def set_vis_lfo(ext, tab);
        e_  = {:filter    => "FilterLfo/",
               :amplitude => "AmpLfo/",
               :frequency => "FreqLfo/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::LfoVis
        row1.children[0].children[0].extern = amp_lfo.extern+"out"
    ;end
  def set_vis_env(ext, tab);
        e_  = {:filter    => "FilterEnvelope/",
               :amplitude => "AmpEnvelope/",
               :frequency => "FreqEnvelope/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::ZynEnvEdit
        #self.children[0].children[0].extern = amp_env.extern+"out"
        amp_env.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_vis_filter(ext, dummy);
        row1.extern = ext + "GlobalFilter/"
        if(addglobal.filtertype == :formant)
            row1.content = Qml::ZynFormant
        else
            row1.content = Qml::VisFilter
            row1.children[0].extern = ext + "GlobalFilter/response"
        end
        amp_gen.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_amp(base);
        footer.children[0].value = true
        amp_gen.extern  = base
        amp_env.extern  = base + "AmpEnvelope/"
        amp_lfo.extern  = base + "AmpLfo/"
        amp_gen.content = Qml::ZynAmpGeneral
        amp_env.content = Qml::ZynAmpEnv
        amp_lfo.content = Qml::ZynLFO
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env, :amp)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :amp)}
    ;end
  def set_freq(base);
        footer.children[1].value = true
        amp_gen.extern  = base
        amp_env.extern  = base + "FreqEnvelope/"
        amp_lfo.extern  = base + "FreqLfo/"
        amp_gen.content = Qml::ZynFreqGeneral
        amp_env.content = Qml::ZynFreqEnv
        amp_lfo.content = Qml::ZynLFO
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env, :freq)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :freq)}
    ;end
  def set_filter(base);
        footer.children[2].value = true
        amp_gen.extern  = base + "GlobalFilter/"
        amp_env.extern  = base + "FilterEnvelope/"
        amp_lfo.extern  = base + "FilterLfo/"
        amp_gen.content = Qml::ZynAnalogFilter
        amp_env.content = Qml::ZynFilterEnv
        amp_lfo.content = Qml::ZynLFO
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env, :filter)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :filter)}
    ;end
  def onSetup(old=nil);
        return if self.valueRef
        if(self.valueRef.nil?)
            path = self.extern + "GlobalFilter/Pcategory"
            self.valueRef = OSC::RemoteParam.new($remote, path)
            self.valueRef.mode = :full
            self.valueRef.callback = lambda {|x|
                addglobal.filtertype = [:analog, :formant, :statevar][x]
            }
        end
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @addglobal = self
@row1 = Qml::Swappable.new(database, ui_path+"1"+'/')
@row2 = Qml::Widget.new(database, ui_path+"2"+'/')
@amp_gen = Qml::Swappable.new(database, ui_path+"3"+'/')
@amp_env = Qml::Swappable.new(database, ui_path+"4"+'/')
@amp_lfo = Qml::Swappable.new(database, ui_path+"5"+'/')
@footer = Qml::Widget.new(database, ui_path+"6"+'/')
@anonymous209 = Qml::TabButton.new(database, ui_path+"7"+'/')
@anonymous210 = Qml::TabButton.new(database, ui_path+"8"+'/')
@anonymous211 = Qml::TabButton.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["addglobal"] = self
context["row1"] = @row1
context["row2"] = @row2
context["amp_gen"] = @amp_gen
context["amp_env"] = @amp_env
context["amp_lfo"] = @amp_lfo
context["footer"] = @footer
Qml::prop_add(self, "filtertype")

        @row1.instance_eval("def setDataVis(type, tab);\n            root.set_view_pos(:vis, type)\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynAddGlobal.qml", 17)
Qml::add_child(self, @row1)

        @row2.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynAddGlobal.qml", 25)
@amp_gen.whenSwapped = lambda {
                if(amp_gen.content == Qml::ZynAnalogFilter)
                    ch = amp_gen.children[0]
                    ch.whenClick = lambda {row1.setDataVis(:filter, :filter)}
                end
            }
Qml::add_child(@row2, @amp_gen)
Qml::add_child(@row2, @amp_env)
Qml::add_child(@row2, @amp_lfo)
Qml::add_child(self, @row2)

        @footer.instance_eval("def layout(l, selfBox);\n            selfBox = Draw::Layout::tabpack(l, selfBox, self)\n        ;end", "src/mruby-zest/example/ZynAddGlobal.qml", 49)

        @footer.instance_eval("def setTab(id);\n            (0..2).each do |ch_id|\n                children[ch_id].value = (ch_id == id)\n                children[ch_id].damage_self\n            end\n            mapping = {0 => :amplitude,\n                       1 => :frequency,\n                       2 => :filter}\n            root.set_view_pos(:subsubview, mapping[id])\n            root.change_view\n            #db.update_values\n        ;end", "src/mruby-zest/example/ZynAddGlobal.qml", 54)
@anonymous209.label = "amplitude"
@anonymous209.whenClick = lambda {footer.setTab(0)}
@anonymous209.highlight_pos = :top
Qml::add_child(@footer, @anonymous209)
@anonymous210.label = "frequency"
@anonymous210.whenClick = lambda {footer.setTab(1)}
@anonymous210.highlight_pos = :top
Qml::add_child(@footer, @anonymous210)
@anonymous211.label = "filter"
@anonymous211.whenClick = lambda {footer.setTab(2)}
@anonymous211.highlight_pos = :top
Qml::add_child(@footer, @anonymous211)
Qml::add_child(self, @footer)
self.valueRef = nil
self.filtertype = nil

    #puts "ZynAddGlobal, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAddSynth < Qml::Widget
  attr_reader :center
  attr_reader :header
  attr_reader :vpow
  attr_reader :cpy_but
  attr_reader :pst_but
  attr_reader :swap
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.05, 0.95])
    ;end
  def onSetup(old=nil);
        return if swap.content
        set_view
    ;end
  def set_view();
        vce     = root.get_view_pos(:voice)
        subview = root.get_view_pos(:subview)
        extbase = center.extern
        ext = {:global    => "GlobalPar/",
               :voice     => "VoicePar#{vce}/",
               :oscil     => "VoicePar#{vce}/OscilSmp/",
               :modosc    => "VoicePar#{vce}/FMSmp/",
               :modulate  => "VoicePar#{vce}/",
               :vce_list  => "",
               :resonance => "GlobalPar/Reson/"}
        mapping = {:global    => Qml::ZynAddGlobal,
                   :voice     => Qml::ZynAddVoice,
                   :oscil     => Qml::ZynOscil,
                   :modosc    => Qml::ZynOscil,
                   :modulate  => Qml::ZynOscilMod,
                   :vce_list  => Qml::ZynAddVoiceList,
                   :resonance => Qml::ZynResonance}
        tabid   = {:global    => 2,
                   :voice     => 3,
                   :oscil     => 4,
                   :modosc    => 5,
                   :modulate  => 6,
                   :vce_list  => 7,
                   :resonance => 8}
        if(!mapping.include?(subview))
            subview = :oscil
            root.set_view_pos(:subview, :oscil)
        end

        swap.extern  = extbase + ext[subview]
        swap.content = mapping[subview]
        header.children.each_with_index do |ch, i|
            if(ch.class == Qml::TabButton)
                ch.value = (i == tabid[subview])
                ch.damage_self
            end
        end

        header.children[1].setValue(vce+1)

        cpy_but.index  = nil
        pst_but.index  = nil

        if([:oscil, :modosc, :resonance].include?(subview))
            cpy_but.extern = extbase + ext[subview]
            pst_but.extern = extbase + ext[subview]
        elsif([:modulate,:voice].include?(subview))
            cpy_but.extern = extbase
            pst_but.extern = extbase
            cpy_but.index  = vce
            pst_but.index  = vce
        else
            cpy_but.extern = extbase
            pst_but.extern = extbase
        end
        vpow.extern    = extbase + "VoicePar#{vce}/Enabled"

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @center = self
@header = Qml::Widget.new(database, ui_path+"1"+'/')
@vpow = Qml::PowButton.new(database, ui_path+"2"+'/')
@anonymous215 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous216 = Qml::TabButton.new(database, ui_path+"4"+'/')
@anonymous217 = Qml::TabButton.new(database, ui_path+"5"+'/')
@anonymous218 = Qml::TabButton.new(database, ui_path+"6"+'/')
@anonymous219 = Qml::TabButton.new(database, ui_path+"7"+'/')
@anonymous220 = Qml::TabButton.new(database, ui_path+"8"+'/')
@anonymous221 = Qml::TabButton.new(database, ui_path+"9"+'/')
@anonymous222 = Qml::TabButton.new(database, ui_path+"10"+'/')
@cpy_but = Qml::CopyButton.new(database, ui_path+"11"+'/')
@pst_but = Qml::PasteButton.new(database, ui_path+"12"+'/')
@swap = Qml::Swappable.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["center"] = self
context["header"] = @header
@header.instance_eval do
def cpy_but=(k);@cpy_but=k;end
def cpy_but;    @cpy_but;  end
end
context["vpow"] = @vpow
context["cpy_but"] = @cpy_but
context["pst_but"] = @pst_but
context["swap"] = @swap
Qml::context_apply(context, @header)

        @header.instance_eval("def set_voice();\n            root.set_view_pos(:voice, children[1].value-1)\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynAddSynth.qml", 31)

        @header.instance_eval("def get_voice(); root.get_view_pos(:voice) ;end", "src/mruby-zest/example/ZynAddSynth.qml", 35)

        @header.instance_eval("def layout(l, selfBox);\n            selfBox = Draw::Layout::tabpack(l, selfBox, self, cpy_but)\n        ;end", "src/mruby-zest/example/ZynAddSynth.qml", 37)

        @header.instance_eval("def setTab(id);\n            #Define a mapping from tabs to values\n            mapping = {0 => :global,\n                       1 => :voice,\n                       2 => :oscil,\n                       3 => :modosc,\n                       4 => :modulate,\n                       5 => :vce_list,\n                       6 => :resonance}\n\n            root.set_view_pos(:subview, mapping[id])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynAddSynth.qml", 42)
Qml::add_child(@header, @vpow)
@db.connect_property(@anonymous215.properties["value"], lambda {header.get_voice + 1})
@anonymous215.format = "VCE "
@anonymous215.whenValue = lambda {header.set_voice()}
@anonymous215.tooltip = "voice"
@anonymous215.maximum = 8
@anonymous215.minimum = 1
Qml::add_child(@header, @anonymous215)
@anonymous216.whenClick = lambda {header.setTab(0)}
@anonymous216.label = "global"
Qml::add_child(@header, @anonymous216)
@anonymous217.whenClick = lambda {header.setTab(1)}
@anonymous217.label = "voice"
Qml::add_child(@header, @anonymous217)
@anonymous218.whenClick = lambda {header.setTab(2)}
@anonymous218.label = "osc"
Qml::add_child(@header, @anonymous218)
@anonymous219.whenClick = lambda {header.setTab(3)}
@db.connect_property(@anonymous219.properties["label"], lambda {"mod-osc"})
Qml::add_child(@header, @anonymous219)
@anonymous220.whenClick = lambda {header.setTab(4)}
@anonymous220.label = "modulation"
Qml::add_child(@header, @anonymous220)
@anonymous221.whenClick = lambda {header.setTab(5)}
@anonymous221.label = "voice list"
Qml::add_child(@header, @anonymous221)
@anonymous222.whenClick = lambda {header.setTab(6)}
@anonymous222.label = "resonance"
Qml::add_child(@header, @anonymous222)
Qml::add_child(@header, @cpy_but)
Qml::add_child(@header, @pst_but)
Qml::add_child(self, @header)
Qml::add_child(self, @swap)

    #puts "ZynAddSynth, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAddVoice < Qml::Widget
  attr_reader :addbase
  attr_reader :row1
  attr_reader :row2
  attr_reader :gen
  attr_reader :env
  attr_reader :lfo
  attr_reader :footer
  attr_accessor :valueRef
  attr_property :filtertype
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, 
                            [0.55, 0.40, 0.05])
    ;end
  def set_amp(base);
        footer.children[0].value = true
        gen.extern  = base
        env.extern  = base + "AmpEnvelope/"
        lfo.extern  = base + "AmpLfo/"
        gen.content = Qml::ZynAmpVoiceGeneral
        env.content = Qml::ZynAmpEnv
        lfo.content = Qml::ZynLFO
        env.children[0].whenClick = lambda {row1.setDataVis(:env, :amp)}
        lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :amp)}
        env.children[0].toggleable = base + "PAmpEnvelopeEnabled"
        lfo.children[0].toggleable = base + "PAmpLfoEnabled"
    ;end
  def set_freq(base);
        footer.children[1].value = true
        gen.extern  = base
        env.extern  = base + "FreqEnvelope/"
        lfo.extern  = base + "FreqLfo/"
        gen.content = Qml::ZynFreqGeneralVoice
        env.content = Qml::ZynFreqEnv
        lfo.content = Qml::ZynLFO
        env.children[0].whenClick = lambda {row1.setDataVis(:env, :freq)}
        lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :freq)}
        env.children[0].toggleable = base + "PFreqEnvelopeEnabled"
        lfo.children[0].toggleable = base + "PFreqLfoEnabled"
    ;end
  def set_filter(base);
        footer.children[2].value = true
        gen.extern  = base + "VoiceFilter/"
        env.extern  = base + "FilterEnvelope/"
        lfo.extern  = base + "FilterLfo/"
        gen.content = Qml::ZynAnalogFilter
        env.content = Qml::ZynFilterEnv
        lfo.content = Qml::ZynLFO
        gen.children[0].whenClick = lambda {row1.setDataVis(:filter, :filter)}
        env.children[0].whenClick = lambda {row1.setDataVis(:env, :filter)}
        lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :filter)}
        env.children[0].toggleable = base + "PFilterEnvelopeEnabled"
        gen.children[0].toggleable = base + "PFilterEnabled"
        lfo.children[0].toggleable = base + "PFilterLfoEnabled"
    ;end
  def set_vis_lfo(ext, tab);
        e_  = {:filter    => "FilterLfo/",
               :amplitude => "AmpLfo/",
               :frequency => "FreqLfo/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::LfoVis
        row1.children[0].children[0].extern = lfo.extern+"out"
    ;end
  def set_vis_env(ext, tab);
        e_  = {:filter    => "FilterEnvelope/",
               :amplitude => "AmpEnvelope/",
               :frequency => "FreqEnvelope/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::ZynEnvEdit
        #self.children[0].children[0].extern = env.extern+"out"
        env.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_vis_filter(ext, dummy);
        row1.extern = ext + "VoiceFilter/"
        if(addbase.filtertype == :formant)
            row1.content = Qml::ZynFormant
        else
            row1.content = Qml::VisFilter
            row1.children[0].extern = ext + "VoiceFilter/response"
        end
        gen.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_view();
        subsubview = root.get_view_pos(:subsubview)
        types = [:amplitude, :frequency, :filter]
        if(!types.include?(subsubview))
            subsubview = :amplitude
            root.set_view_pos(:subsubview, subsubview)
        end

        vis = root.get_view_pos(:vis)
        types = [:envelope, :lfo, :filter]
        if(!types.include?(vis))
            vis = :envelope
            root.set_view_pos(:vis, vis)
        end

        base = self.extern

        if(subsubview == :amplitude)
            set_amp base
        elsif(subsubview == :frequency)
            set_freq base
        else
            set_filter base
        end

        if(vis == :lfo)
            set_vis_lfo(base, subsubview)
        elsif(vis == :envelope)
            set_vis_env(base, subsubview)
        elsif(vis == :filter)
            set_vis_filter(base, subsubview)
        end
    ;end
  def onSetup(old=nil);
        return if self.valueRef
        if(self.valueRef.nil?)
            path = self.extern + "VoiceFilter/Pcategory"
            self.valueRef = OSC::RemoteParam.new($remote, path)
            self.valueRef.mode = :full
            self.valueRef.callback = lambda {|x|
                addbase.filtertype = [:analog, :formant, :statevar][x]
            }
        end
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @addbase = self
@row1 = Qml::Swappable.new(database, ui_path+"1"+'/')
@row2 = Qml::Widget.new(database, ui_path+"2"+'/')
@gen = Qml::Swappable.new(database, ui_path+"3"+'/')
@env = Qml::Swappable.new(database, ui_path+"4"+'/')
@lfo = Qml::Swappable.new(database, ui_path+"5"+'/')
@footer = Qml::Widget.new(database, ui_path+"6"+'/')
@anonymous233 = Qml::TabButton.new(database, ui_path+"7"+'/')
@anonymous234 = Qml::TabButton.new(database, ui_path+"8"+'/')
@anonymous235 = Qml::TabButton.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["addbase"] = self
context["row1"] = @row1
context["row2"] = @row2
context["gen"] = @gen
context["env"] = @env
context["lfo"] = @lfo
context["footer"] = @footer
Qml::prop_add(self, "filtertype")

        @row1.instance_eval("def setDataVis(type, tab);\n            root.set_view_pos(:vis, type)\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynAddVoice.qml", 15)
Qml::add_child(self, @row1)

        @row2.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynAddVoice.qml", 23)
Qml::add_child(@row2, @gen)
Qml::add_child(@row2, @env)
Qml::add_child(@row2, @lfo)
Qml::add_child(self, @row2)

        @footer.instance_eval("def layout(l, selfBox);\n            Draw::Layout::tabpack(l, selfBox, self)\n        ;end", "src/mruby-zest/example/ZynAddVoice.qml", 41)

        @footer.instance_eval("def setTab(id);\n            (0..2).each do |ch_id|\n                children[ch_id].value = (ch_id == id)\n                self.root.damage_item children[ch_id]\n            end\n            mapping = {0 => :amplitude,\n                       1 => :frequency,\n                       2 => :filter}\n            root.set_view_pos(:subsubview, mapping[id])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynAddVoice.qml", 46)
@anonymous233.label = "amplitude"
@anonymous233.whenClick = lambda {footer.setTab(0)}
@anonymous233.highlight_pos = :top
Qml::add_child(@footer, @anonymous233)
@anonymous234.label = "frequency"
@anonymous234.whenClick = lambda {footer.setTab(1)}
@anonymous234.highlight_pos = :top
Qml::add_child(@footer, @anonymous234)
@anonymous235.label = "filter"
@anonymous235.whenClick = lambda {footer.setTab(2)}
@anonymous235.highlight_pos = :top
Qml::add_child(@footer, @anonymous235)
Qml::add_child(self, @footer)
self.valueRef = nil
self.filtertype = nil

    #puts "ZynAddVoice, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAddVoiceList < Qml::Widget
  attr_reader :vce_list
  attr_property :weights
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  def onSetup(old=nil);
        return if children.length > 2
        (0...8).each do |r|
            row        = Qml::ZynAddVoiceListItem.new(db)
            row.num    = r
            row.extern = vce_list.extern+"VoicePar#{r}/"
            Qml::add_child(self, row)
        end
    ;end
  def layout(l, selfBox);
        n = children.length
        off  = 0
        gap  = 0.040
        step = (1.0-(n)*gap)/n
        children.each_with_index do |ch, id|
            ch.fixed(l, selfBox, 0.01, off, 0.98, step)
            off += step + gap
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vce_list = self
@anonymous244 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous238 = Qml::ColorBox.new(database, ui_path+"2"+'/')
@anonymous239 = Qml::ColorBox.new(database, ui_path+"3"+'/')
@anonymous240 = Qml::TextBox.new(database, ui_path+"4"+'/')
@anonymous241 = Qml::TextBox.new(database, ui_path+"5"+'/')
@anonymous242 = Qml::TextBox.new(database, ui_path+"6"+'/')
@anonymous243 = Qml::TextBox.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["vce_list"] = self
@anonymous244.instance_eval do
def vce_list=(k);@vce_list=k;end
def vce_list;    @vce_list;  end
end
Qml::context_apply(context, @anonymous244)
Qml::prop_add(self, "weights")

        @anonymous244.instance_eval("def layout(l, selfBox);\n            chBox   = []\n\n            off = 0.0\n            children.each_with_index do |ch, ind|\n                weight = vce_list.weights[ind]\n                ch.fixed(l, selfBox, off, 0.0, weight, 1.0)\n                off += weight\n            end\n            selfBox\n        ;end", "src/mruby-zest/example/ZynAddVoiceList.qml", 43)
@anonymous238.bg = nil 
Qml::add_child(@anonymous244, @anonymous238)
@anonymous239.bg = nil
Qml::add_child(@anonymous244, @anonymous239)
@anonymous240.bg = nil
@anonymous240.label = "volume"
Qml::add_child(@anonymous244, @anonymous240)
@anonymous241.bg = nil
@anonymous241.label = "panning"
Qml::add_child(@anonymous244, @anonymous241)
@anonymous242.bg = nil
@anonymous242.label = "detune"
Qml::add_child(@anonymous244, @anonymous242)
@anonymous243.bg = nil
@db.connect_property(@anonymous243.properties["label"], lambda {"vib-depth"})
Qml::add_child(@anonymous244, @anonymous243)
Qml::add_child(self, @anonymous244)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.05, 0.2, 0.2, 0.3, 0.2]})

    #puts "ZynAddVoiceList, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAddVoiceListItem < Qml::Widget
  attr_reader :voice_item
  attr_accessor :num
  attr_property :weights
  def layout(l, selfBox);
        chBox   = []

        off = 0.0
        hpad = 1/128
        children.each_with_index do |ch, ind|
            weight = weights[ind]
            ch.fixed(l, selfBox, off+hpad, 0.0, weight-2*hpad, 1.0)
            off += weight
        end
        selfBox
    ;end
  def onSetup(old=nil);
        children.each do |x|
            x.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @voice_item = self
@anonymous246 = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous247 = Qml::WaveView.new(database, ui_path+"2"+'/')
@anonymous248 = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous249 = Qml::HSlider.new(database, ui_path+"4"+'/')
@anonymous250 = Qml::HSlider.new(database, ui_path+"5"+'/')
@anonymous251 = Qml::HSlider.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["voice_item"] = self
Qml::prop_add(self, "weights")
@db.connect_property(@anonymous246.properties["extern"], lambda {voice_item.extern + "Enabled"})
@db.connect_property(@anonymous246.properties["label"], lambda {(voice_item.num+1).to_s})
@anonymous246.layoutOpts = [:no_constraint]
Qml::add_child(self, @anonymous246)
@db.connect_property(@anonymous247.properties["extern"], lambda {voice_item.extern + "OscilSmp/waveform"})
@db.connect_property(@anonymous247.properties["noise"], lambda {voice_item.extern + "Type"})
@anonymous247.grid = false
Qml::add_child(self, @anonymous247)
@db.connect_property(@anonymous248.properties["extern"], lambda {voice_item.extern + "PVolume"})
@db.connect_property(@anonymous248.properties["label"], lambda {"20%"})
Qml::add_child(self, @anonymous248)
@db.connect_property(@anonymous249.properties["extern"], lambda {voice_item.extern + "PPanning"})
@anonymous249.label = "centered"
@anonymous249.centered = true
@anonymous249.value = 0.5
Qml::add_child(self, @anonymous249)
@db.connect_property(@anonymous250.properties["extern"], lambda {voice_item.extern + "PDetune"})
@db.connect_property(@anonymous250.properties["label"], lambda {"+0 cents"})
@anonymous250.centered = true
@anonymous250.value = 0.5
Qml::add_child(self, @anonymous250)
@db.connect_property(@anonymous251.properties["extern"], lambda {voice_item.extern + "FreqLfo/Pintensity"})
@db.connect_property(@anonymous251.properties["label"], lambda {"100%"})
Qml::add_child(self, @anonymous251)
self.num = 0
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.05, 0.2, 0.2, 0.3, 0.2]})

    #puts "ZynAddVoiceListItem, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAlienwah < Qml::Group
  attr_reader :wah
  attr_reader :rw
  def refresh();
        return if rw.content.nil?
        return if rw.content.children.length < 4
        rw.content.children[3..-1].each do |c|
            c.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @wah = self
@rw = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous253 = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous254 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous255 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous256 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous257 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous258 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous259 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous260 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous261 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous262 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous263 = Qml::Knob.new(database, ui_path+"12"+'/')
@anonymous264 = Qml::Knob.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["wah"] = self
context["rw"] = @rw
@db.connect_property(@anonymous253.properties["extern"], lambda {wah.extern + "Alienwah/preset"})
@anonymous253.whenValue = lambda { wah.refresh }
Qml::add_child(@rw, @anonymous253)
@db.connect_property(@anonymous254.properties["extern"], lambda {wah.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous254)
@db.connect_property(@anonymous255.properties["extern"], lambda {wah.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous255)
@db.connect_property(@anonymous256.properties["extern"], lambda {wah.extern + "Alienwah/Pfreq"   })
Qml::add_child(@rw, @anonymous256)
@db.connect_property(@anonymous257.properties["extern"], lambda {wah.extern + "Alienwah/Pfreqrnd"   })
Qml::add_child(@rw, @anonymous257)
@db.connect_property(@anonymous258.properties["extern"], lambda {wah.extern + "Alienwah/PLFOtype" })
Qml::add_child(@rw, @anonymous258)
@db.connect_property(@anonymous259.properties["extern"], lambda {wah.extern + "Alienwah/PStereo" })
Qml::add_child(@rw, @anonymous259)
@db.connect_property(@anonymous260.properties["extern"], lambda {wah.extern + "Alienwah/Pdepth" })
Qml::add_child(@rw, @anonymous260)
@db.connect_property(@anonymous261.properties["extern"], lambda {wah.extern + "Alienwah/Pfeedback" })
Qml::add_child(@rw, @anonymous261)
@db.connect_property(@anonymous262.properties["extern"], lambda {wah.extern + "Alienwah/Pdelay" })
Qml::add_child(@rw, @anonymous262)
@db.connect_property(@anonymous263.properties["extern"], lambda {wah.extern + "Alienwah/Plrcross" })
Qml::add_child(@rw, @anonymous263)
@db.connect_property(@anonymous264.properties["extern"], lambda {wah.extern + "Alienwah/Pphase" })
Qml::add_child(@rw, @anonymous264)
@rw.layoutOpts = []
Qml::add_child(self, @rw)
self.label = "alienwah"
self.topSize = 0.2

    #puts "ZynAlienwah, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAmpEnv < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  def cb();
        whenModified.call if whenModified
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous267 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous268 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous269 = Qml::Knob.new(database, ui_path+"4"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous271 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous272 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous275 = Qml::Col.new(database, ui_path+"8"+'/')
@anonymous273 = Qml::ToggleButton.new(database, ui_path+"9"+'/')
@anonymous274 = Qml::ToggleButton.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@anonymous267.whenValue = lambda { box.cb }
@db.connect_property(@anonymous267.properties["extern"], lambda {box.extern+"PA_dt" })
Qml::add_child(@top, @anonymous267)
@anonymous268.whenValue = lambda { box.cb }
@db.connect_property(@anonymous268.properties["extern"], lambda {box.extern+"PD_dt"})
Qml::add_child(@top, @anonymous268)
@anonymous269.whenValue = lambda { box.cb }
@db.connect_property(@anonymous269.properties["extern"], lambda {box.extern+"PS_val"})
Qml::add_child(@top, @anonymous269)
Qml::add_child(self, @top)
@anonymous271.whenValue = lambda { box.cb }
@db.connect_property(@anonymous271.properties["extern"], lambda {box.extern+"PR_dt"})
Qml::add_child(@bot, @anonymous271)
@anonymous272.whenValue = lambda { box.cb }
@db.connect_property(@anonymous272.properties["extern"], lambda {box.extern+"Penvstretch"})
Qml::add_child(@bot, @anonymous272)
@anonymous273.label = "FRCR"
@anonymous273.whenValue = lambda { box.cb }
@db.connect_property(@anonymous273.properties["extern"], lambda {box.extern+"Pforcedrelease"})
Qml::add_child(@anonymous275, @anonymous273)
@anonymous274.label = "lin/log"
@anonymous274.whenValue = lambda { box.cb }
@db.connect_property(@anonymous274.properties["extern"], lambda {box.extern+"Plinearenvelope"})
Qml::add_child(@anonymous275, @anonymous274)
Qml::add_child(@bot, @anonymous275)
Qml::add_child(self, @bot)
self.label = "Envelope"
self.whenModified = nil

    #puts "ZynAmpEnv, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAmpGeneral < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous278 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous279 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous280 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous281 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous283 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous284 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous287 = Qml::Col.new(database, ui_path+"9"+'/')
@anonymous285 = Qml::ToggleButton.new(database, ui_path+"10"+'/')
@anonymous286 = Qml::ToggleButton.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous278.properties["extern"], lambda {box.extern+"PVolume" })
Qml::add_child(@top, @anonymous278)
@db.connect_property(@anonymous279.properties["extern"], lambda {box.extern+"PAmpVelocityScaleFunction"})
Qml::add_child(@top, @anonymous279)
@db.connect_property(@anonymous280.properties["extern"], lambda {box.extern+"PPanning"})
Qml::add_child(@top, @anonymous280)
@db.connect_property(@anonymous281.properties["extern"], lambda {box.extern+"PPunchStretch"})
Qml::add_child(@top, @anonymous281)
Qml::add_child(self, @top)
@db.connect_property(@anonymous283.properties["extern"], lambda {box.extern+"PPunchStrength"})
Qml::add_child(@bot, @anonymous283)
@db.connect_property(@anonymous284.properties["extern"], lambda {box.extern+"PPunchTime"})
Qml::add_child(@bot, @anonymous284)
@anonymous285.label = "stereo"
@db.connect_property(@anonymous285.properties["extern"], lambda {box.extern+"PStereo"})
Qml::add_child(@anonymous287, @anonymous285)
@anonymous286.label = "rnd grp"
@db.connect_property(@anonymous286.properties["extern"], lambda {box.extern+"Hrandgrouping"})
Qml::add_child(@anonymous287, @anonymous286)
Qml::add_child(@bot, @anonymous287)
Qml::add_child(self, @bot)
self.label = "General"
self.copyable = false
self.whenModified = nil

    #puts "ZynAmpGeneral, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAmpMod < Qml::Group
  attr_reader :box
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous293 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous290 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous291 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous292 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous294 = Qml::Widget.new(database, ui_path+"5"+'/')

    
context = Hash.new
context["box"] = self
@db.connect_property(@anonymous290.properties["extern"], lambda {box.extern+"PFMVolume" })
Qml::add_child(@anonymous293, @anonymous290)
@db.connect_property(@anonymous291.properties["extern"], lambda {box.extern+"PFMVolumeDamp"})
Qml::add_child(@anonymous293, @anonymous291)
@db.connect_property(@anonymous292.properties["extern"], lambda {box.extern+"PFMVelocityScaleFunction"})
Qml::add_child(@anonymous293, @anonymous292)
Qml::add_child(self, @anonymous293)
Qml::add_child(self, @anonymous294)
self.label = "General"
self.copyable = false

    #puts "ZynAmpMod, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAmpVoiceGeneral < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous296 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous297 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous298 = Qml::BypassButton.new(database, ui_path+"4"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous300 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous301 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous302 = Qml::ToggleButton.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous296.properties["extern"], lambda {box.extern+"PVolume" })
Qml::add_child(@top, @anonymous296)
@db.connect_property(@anonymous297.properties["extern"], lambda {box.extern+"PAmpVelocityScaleFunction"})
Qml::add_child(@top, @anonymous297)
@db.connect_property(@anonymous298.properties["extern"], lambda {box.extern+"Pfilterbypass" })
Qml::add_child(@top, @anonymous298)
Qml::add_child(self, @top)
@db.connect_property(@anonymous300.properties["extern"], lambda {box.extern+"PPanning"})
Qml::add_child(@bot, @anonymous300)
@db.connect_property(@anonymous301.properties["extern"], lambda {box.extern+"PDelay"})
Qml::add_child(@bot, @anonymous301)
@anonymous302.label = "reson"
@db.connect_property(@anonymous302.properties["extern"], lambda {box.extern+"Presonance"})
Qml::add_child(@bot, @anonymous302)
Qml::add_child(self, @bot)
self.label = "General"
self.copyable = false
self.whenModified = nil

    #puts "ZynAmpVoiceGeneral, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAnalogFilter < Qml::Group
  attr_reader :box
  attr_reader :snsa
  attr_reader :snsb
  attr_reader :cat
  attr_reader :typ
  attr_property :whenModified
  attr_property :type
  attr_property :subsynth
  def cb();
        whenModified.call if whenModified
    ;end
  def change_cat();
        root.change_view()
        if(cat.selected == 1)
            typ.active = false
            typ.damage_self
            return
        end
        dest = self.extern + "Ptype"    if cat.selected == 0
        dest = self.extern + "type-svf" if cat.selected == 2
        if(typ.extern != dest)
            typ.extern = dest
            typ.extern()
        end
        typ.active = true
        typ.damage_self
    ;end
  def remove_sense();
        root.ego_death snsa
        root.ego_death snsb
    ;end
  def move_sense();
        snsa.extern = path_simp(box.extern + "../PGlobalFilterVelocityScale")
        snsb.extern = path_simp(box.extern + "../PGlobalFilterVelocityScaleFunction")
    ;end
  def class_name();
        "ZynAnalogFilter"
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous310 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous305 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous306 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous307 = Qml::Knob.new(database, ui_path+"4"+'/')
@snsa = Qml::Knob.new(database, ui_path+"5"+'/')
@snsb = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous315 = Qml::ParModuleRow.new(database, ui_path+"7"+'/')
@anonymous311 = Qml::NumEntry.new(database, ui_path+"8"+'/')
@cat = Qml::Selector.new(database, ui_path+"9"+'/')
@typ = Qml::Selector.new(database, ui_path+"10"+'/')
@anonymous314 = Qml::Knob.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["box"] = self
context["snsa"] = @snsa
context["snsb"] = @snsb
context["cat"] = @cat
context["typ"] = @typ
Qml::prop_add(self, "whenModified")
Qml::prop_add(self, "type")
Qml::prop_add(self, "subsynth")
@anonymous305.type = :float
@anonymous305.whenValue = lambda { box.cb}
@db.connect_property(@anonymous305.properties["extern"], lambda {box.extern + "basefreq"})
Qml::add_child(@anonymous310, @anonymous305)
@anonymous306.type = :float
@anonymous306.whenValue = lambda { box.cb}
@db.connect_property(@anonymous306.properties["extern"], lambda {box.extern + "baseq"})
Qml::add_child(@anonymous310, @anonymous306)
@anonymous307.type = :float
@anonymous307.whenValue = lambda { box.cb}
@db.connect_property(@anonymous307.properties["extern"], lambda {box.extern + "freqtracking"})
Qml::add_child(@anonymous310, @anonymous307)
@db.connect_property(@snsa.properties["extern"], lambda {begin

                ext = "../PFilterVelocityScale"       if !box.subsynth
                ext = "../PGlobalFilterVelocityScale" if  box.subsynth
                path_simp(box.extern + ext)
            
end})
Qml::add_child(@anonymous310, @snsa)
@db.connect_property(@snsb.properties["extern"], lambda {begin

                ext = "../PFilterVelocityScaleFunction"       if !box.subsynth
                ext = "../PGlobalFilterVelocityScaleFunction" if  box.subsynth

                path_simp(box.extern + ext)
            
end})
Qml::add_child(@anonymous310, @snsb)
Qml::add_child(self, @anonymous310)
@anonymous311.whenValue = lambda { box.cb}
@db.connect_property(@anonymous311.properties["extern"], lambda {box.extern + "Pstages"})
@anonymous311.offset = 1
@anonymous311.minimum = 1
@anonymous311.maximum = 5
Qml::add_child(@anonymous315, @anonymous311)
@cat.whenValue = lambda { box.change_cat}
@db.connect_property(@cat.properties["extern"], lambda {box.extern + "Pcategory"})
Qml::add_child(@anonymous315, @cat)
@typ.whenValue = lambda { box.cb}
@db.connect_property(@typ.properties["extern"], lambda {box.extern + "Ptype"})
Qml::add_child(@anonymous315, @typ)
@anonymous314.type = :float
@anonymous314.whenValue = lambda { box.cb}
@db.connect_property(@anonymous314.properties["extern"], lambda {box.extern + "gain"})
Qml::add_child(@anonymous315, @anonymous314)
Qml::add_child(self, @anonymous315)
self.label = "General"
self.whenModified = nil
self.type = :analog
self.subsynth = false
@db.connect_watcher(self.properties["type"], lambda {begin

        #Changes filter type
    
end})

    #puts "ZynAnalogFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomation < Qml::Widget
  attr_reader :automation_view
  attr_reader :slots
  attr_reader :tgrp
  attr_reader :chan
  attr_reader :details
  attr_accessor :valueRef
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.3,0.7])
    ;end
  def set_view();
        slot        = root.get_view_pos(:slot)
        slot_path   = "/automate/slot#{slot}/"

        chan.extern = slot_path
        (0..3).each do |i|
            details.children[i].extern = slot_path + "param#{i}/"
        end
        (0...slots.children.length).each do |i|
            slots.children[i].set_active(slot == i)
            slots.children[i].extern = "/automate/slot#{i}/"
        end
    ;end
  def set_active(slot_id);
        if(root.get_view_pos(:slot) != slot_id)
            root.set_view_pos(:slot, slot_id)
            root.change_view
        end
    ;end
  def set_mode(mode);
        $remote.automation.mode = mode
        if($remote.automation.mode == :normal)
            tgrp.children[0].value = true
            tgrp.children[1].value = false
        else
            tgrp.children[0].value = false
            tgrp.children[1].value = true
        end
        tgrp.damage_self
    ;end
  def onSetup(old=nil);
        details.onSetup

        self.valueRef = OSC::RemoteParam.new($remote, "/automate/active-slot")
        self.valueRef.type = "i"
        self.valueRef.mode = :normal_int
        self.valueRef.callback = lambda {|x| set_active(x)}

        if($remote.automation.mode == :normal)
            tgrp.children[0].value = true
            tgrp.children[1].value = false
        else
            tgrp.children[0].value = false
            tgrp.children[1].value = true
        end

        set_view
    ;end
end
class Qml::ZynAutomationMidiChan < Qml::TextBox
  attr_reader :midi_chan
  def update_text();
        @cc = nil       if(@cc.nil?       || @cc < 0)
        @learning = nil if(@learning.nil? || @learning < 0)

        new_label = self.label;
        if(@cc && !@learning)
            new_label = "MIDI CC #{@cc}"
        elsif(!@cc && @learning)
            new_label = "Learning Queue #{@learning}"
        elsif(@cc && @learning)
            new_label = "MIDI CC #{@cc} - Relearning Queue #{@learning}"
        else
            new_label = "Unbound"
        end

        if(new_label != self.label)
            self.label = new_label
            damage_self
        end
    ;end
  def onSetup(old=nil);
        @cc       = nil
        @learning = nil
    ;end
  def cc=(x);@cc = x;end
  def learning=(x);@learning = x;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @midi_chan = self

    
context = Hash.new
context["midi_chan"] = self
@db.connect_watcher(self.properties["extern"], lambda {begin

        ext  = midi_chan.extern

        learn_ref = OSC::RemoteParam.new($remote, ext + "learning")
        midi_ref  = OSC::RemoteParam.new($remote, ext + "midi-cc")

        midi_ref.mode  = true
        learn_ref.mode = true

        learn_ref.callback = lambda {|x| midi_chan.learning = x; update_text()}
        midi_ref.callback  = lambda {|x| midi_chan.cc = x;       update_text()}


        midi_chan.valueRef = [learn_ref, midi_ref]
    
end})

    #puts "ZynAutomationMidiChan, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomation
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @automation_view = self
@slots = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous323 = Qml::Widget.new(database, ui_path+"2"+'/')
@tgrp = Qml::TabGroup.new(database, ui_path+"3"+'/')
@anonymous318 = Qml::Button.new(database, ui_path+"4"+'/')
@anonymous319 = Qml::Button.new(database, ui_path+"5"+'/')
@chan = Qml::ZynAutomationMidiChan.new(database, ui_path+"6"+'/')
@details = Qml::Widget.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["automation_view"] = self
context["slots"] = @slots
context["tgrp"] = @tgrp
context["chan"] = @chan
context["details"] = @details

        @slots.instance_eval("def onSetup(old=nil);\n            return if children.length > 5\n            #//Fetch all parameters\n            16.times do |i|\n                slot = Qml::ZynAutomationSlot.new(db)\n                slot.slot_id = i\n                Qml::add_child(self,slot)\n                slot.set_active if(i==root.get_view_pos(:slot))\n            end\n        ;end", "src/mruby-zest/example/ZynAutomation.qml", 8)

        @slots.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynAutomation.qml", 19)

        @slots.instance_eval("def draw(vg);\n        ;end", "src/mruby-zest/example/ZynAutomation.qml", 22)
Qml::add_child(self, @slots)

        @anonymous323.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.05,0.05,0.9])\n        ;end", "src/mruby-zest/example/ZynAutomation.qml", 62)
@anonymous318.label = "normal learn"
@anonymous318.tooltip = "learn parameters to unused slot"
@anonymous318.whenValue = lambda { automation_view.set_mode(:normal) }
Qml::add_child(@tgrp, @anonymous318)
@anonymous319.label = "macro learn"
@anonymous319.tooltip = "learn parameters to the active slot"
@anonymous319.whenValue = lambda { automation_view.set_mode(:macro) }
Qml::add_child(@tgrp, @anonymous319)
Qml::add_child(@anonymous323, @tgrp)
@db.connect_property(@chan.properties["label"], lambda {"MIDI Chan 0 CC 12 - Relearning"})
Qml::add_child(@anonymous323, @chan)

        @details.instance_eval("def onSetup(old=nil);\n                return if children.length > 2\n                #//Fetch all parameters\n                4.times do |i|\n                    slot = Qml::ZynAutomationParam.new(db)\n                    Qml::add_child(self,slot)\n                    slot.active = false if i!=0\n                end\n            ;end", "src/mruby-zest/example/ZynAutomation.qml", 48)

        @details.instance_eval("def layout(l, selfBox);\n                Draw::Layout::grid(l, selfBox, children, 2, 2, 5, 5)\n            ;end", "src/mruby-zest/example/ZynAutomation.qml", 58)
Qml::add_child(@anonymous323, @details)
Qml::add_child(self, @anonymous323)
self.valueRef = nil

    #puts "ZynAutomation, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationMidiChan < Qml::TextBox
  attr_reader :midi_chan
  def update_text();
        @cc = nil       if(@cc.nil?       || @cc < 0)
        @learning = nil if(@learning.nil? || @learning < 0)

        new_label = self.label;
        if(@cc && !@learning)
            new_label = "MIDI CC #{@cc}"
        elsif(!@cc && @learning)
            new_label = "Learning Queue #{@learning}"
        elsif(@cc && @learning)
            new_label = "MIDI CC #{@cc} - Relearning Queue #{@learning}"
        else
            new_label = "Unbound"
        end

        if(new_label != self.label)
            self.label = new_label
            damage_self
        end
    ;end
  def onSetup(old=nil);
        @cc       = nil
        @learning = nil
    ;end
  def cc=(x);@cc = x;end
  def learning=(x);@learning = x;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @midi_chan = self

    
context = Hash.new
context["midi_chan"] = self
@db.connect_watcher(self.properties["extern"], lambda {begin

        ext  = midi_chan.extern

        learn_ref = OSC::RemoteParam.new($remote, ext + "learning")
        midi_ref  = OSC::RemoteParam.new($remote, ext + "midi-cc")

        midi_ref.mode  = true
        learn_ref.mode = true

        learn_ref.callback = lambda {|x| midi_chan.learning = x; update_text()}
        midi_ref.callback  = lambda {|x| midi_chan.cc = x;       update_text()}


        midi_chan.valueRef = [learn_ref, midi_ref]
    
end})

    #puts "ZynAutomationMidiChan, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationName < Qml::Widget
  attr_reader :textrender
  def update_path(path);
        textrender.label = path if path.length > 0
        textrender.label = "Unconnected" if path.length <= 0
        damage_self
    ;end
  def create_menu(options, xx);

        n = options.length
        n = 1 if options.class != Array
        widget = DropDown.new(self.db)
        widget.w = self.w*0.2
        widget.h = self.h*n
        widget.x = xx
        widget.y = 0
        widget.layer = 2
        widget.options = options
        widget.callback = lambda { |v|
            #set_value_user(v)
        }
        widget.y = -self.h*(n-1) if(widget.h+global_y > window.h)
        widget.prime(root())

        Qml::add_child(self, widget)
        root.smash_draw_seq
        root.damage_item widget
    ;end
  def onMousePress(ev);
        return
        ps = (ev.pos.x-global_x)/w
        opts = []
        if(ps<0.2)
            (1..16).each do |i|
                opts << "Ins #{i}"
            end
        elsif(ps < 0.4)
            (1..16).each do |i|
                opts << "Kit #{i}"
            end
        elsif(ps < 0.6)
            opts << "ADD"
            opts << "PAD"
            opts << "SUB"
        elsif(ps < 0.8)
            (1..16).each do |i|
                opts << "VCE #{i}"
            end
        else
            opts << "Vol"
            opts << "Freq"
            opts << "Mod. Type"
        end
        create_menu(opts, ev.pos.x-global_x)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous327 = self
@textrender = Qml::Text.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["textrender"] = @textrender
@textrender.label = ""
@textrender.height = 1.0
Qml::add_child(self, @textrender)

    #puts "ZynAutomationName, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationParam < Qml::Widget
  attr_reader :param
  attr_reader :aname
  attr_reader :yaxis
  attr_reader :plot
  attr_reader :gain
  attr_reader :offset
  attr_reader :clr
  attr_property :active
  attr_accessor :valueRef
  def draw(vg);
        background(color("ffffff", 50)) if !param.active
    ;end
  def set_gain(gain);
        plot.set_gain(gain)
    ;end
  def set_offset(offset);
        plot.set_offset(offset)
    ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.1, 0.45, 0.35])
    ;end
  def onSetup(old=nil);
    ;end
  def animate();
        plot.set_gain(4*(gain.value-0.5))
        plot.set_offset(offset.value-0.5)
    ;end
  def update_path(path);
        aname.update_path(path)
        if(path.empty?)
            mn   = -1
            mx   = 1
            yaxis.labels = [mn, ((mn+mx)/2), mx].map{|x| x.to_s}
            yaxis.damage_self
        else
            meta = OSC::RemoteMetadata.new($remote, path)
            mn   = meta.min
            mx   = meta.max
            if(mn && mx)
                yaxis.labels = [mn, ((mn+mx)/2), mx].map{|x| x.to_s}
                yaxis.damage_self
            end
        end
    ;end
  def update_value(val);
        plot.input = 2*val-1
    ;end
  def update_active(val);
        if(val != self.active)
            self.active = val
            damage_self
        end
    ;end
end
class Qml::ZynAutomationPlot < Qml::Widget
  attr_reader :au_plot
  attr_property :minimum
  attr_property :maximum
  attr_accessor :valueRef
  def abs(x);
        if x > 0
            x
        else
            -x
        end
    ;end
  def recalc();
        @gain   ||= 1.0
        @offset ||= 0.5

        #function is out = @offset + input * @gain
        #function is over -1..1
        #function maps to 0..1

        #Calculate x=-1 y value
        ym = @offset + -1*@gain
        #Calculate x=1 y value
        yp = @offset + 1*@gain
        #Calculate y=0 x value
        xm = (0-@offset)/@gain
        #Calculate y=1 x value
        xp = (1-@offset)/@gain

        @points = []

        if(ym < 0)
            @points << [0.0, 0.0, (xm+1)/2, 0.0]
        elsif(ym > 1)
            @points << [0.0, 1.0, (xp+1)/2, 1.0]
        else
            @points << [0.0, ym]
        end

        if(yp < 0)
            @points << [(xm+1)/2, 0.0, 1.0, 0.0]
        elsif(yp > 1)
            @points << [(xp+1)/2, 1.0, 1.0, 1.0]
        else
            @points << [1.0, yp]
        end
        @points = @points.flatten
        damage_self
    ;end
  def set_offset(o);
        return if o == @offset + 0.5
        @offset = o + 0.5
        recalc()
    ;end
  def set_gain(g);
        return if @gain == g
        @gain = g
        recalc()
    ;end
  def draw(vg);
        @gain   ||= 1.0
        @offset ||= 0.5
        recalc() if(@points.nil?)
        background(color("222222"))


        vg.path do
            vg.move_to(0,h/2)
            vg.line_to(w,h/2)
            vg.stroke_color(color("777777"))
            vg.stroke_width(1)
            vg.stroke
        end

        vg.path do
            vg.move_to(w/2,0)
            vg.line_to(w/2,h)
            vg.stroke_color(color("777777"))
            vg.stroke_width(1)
            vg.stroke
        end

        vg.path do
            vg.move_to(0,0)
            vg.line_to(w,0)
            vg.line_to(w,h)
            vg.line_to(0,h)
            vg.line_to(0,0)
            vg.stroke_color(color("aaaaaa"))
            vg.stroke_width(1)
            vg.stroke
        end

        #Draw the actual line
        n = @points.length/2
        vg.path do
            vg.move_to(0,h)
            vg.move_to(w*@points[0], h*(1-@points[1]))
            (1...n).each do |p|
                vg.line_to(w*@points[2*p+0], h*(1-@points[2*p+1]))
            end
            vg.stroke_color(color("aaaaaa"))
            vg.stroke_width(2)
            vg.stroke
        end

        #Draw the active point
        @input ||= 0.3
        out = @offset + @input * @gain
        if(out > 0 && out < 1)
            vg.path do
                vg.circle((@input+1)/2*w, (1-out)*h, 5)
                vg.stroke
            end
        end



    ;end
  def input=(x);if(@input != x);@input = x;damage_self;end;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @au_plot = self

    
context = Hash.new
context["au_plot"] = self
Qml::prop_add(self, "minimum")
Qml::prop_add(self, "maximum")
self.minimum = 0
self.maximum = 0
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        value_ref =  OSC::RemoteParam.new($remote, au_plot.ext + "../value")
        value_ref.callback = lambda {|x| au_plot.input = x}
        au_plot.valueRef = [value_ref]
    
end})

    #puts "ZynAutomationPlot, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationParam
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @param = self
@aname = Qml::ZynAutomationName.new(database, ui_path+"1"+'/')
@anonymous331 = Qml::Widget.new(database, ui_path+"2"+'/')
@yaxis = Qml::PlotLabelY.new(database, ui_path+"3"+'/')
@plot = Qml::ZynAutomationPlot.new(database, ui_path+"4"+'/')
@anonymous335 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@gain = Qml::Knob.new(database, ui_path+"6"+'/')
@offset = Qml::Knob.new(database, ui_path+"7"+'/')
@clr = Qml::ClearBox.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["param"] = self
context["aname"] = @aname
context["yaxis"] = @yaxis
context["plot"] = @plot
context["gain"] = @gain
context["offset"] = @offset
context["clr"] = @clr
Qml::prop_add(self, "active")
Qml::add_child(self, @aname)

        @anonymous331.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.1, 0.9])\n        ;end", "src/mruby-zest/example/ZynAutomationParam.qml", 30)
Qml::add_child(@anonymous331, @yaxis)
Qml::add_child(@anonymous331, @plot)
Qml::add_child(self, @anonymous331)
@gain.label = "gain"
@gain.value = 0.5
@gain.type = "f"
@gain.whenValue = lambda {plot.set_gain(4*(gain.value-0.5))}
Qml::add_child(@anonymous335, @gain)
@offset.label = "offset"
@offset.value = 0.5
@offset.type = "f"
@offset.whenValue = lambda {plot.set_offset(offset.value-0.5)}
Qml::add_child(@anonymous335, @offset)
@clr.whenValue = lambda {$remote.action(param.extern + "clear")}
Qml::add_child(@anonymous335, @clr)
Qml::add_child(self, @anonymous335)
self.active = true
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        if(self.valueRef)
            self.valueRef.each do |v|
                v.clean
            end
        end
        p = param
        ext = p.extern

        path_ref   =  OSC::RemoteParam.new($remote, ext + "path")
        active_ref =  OSC::RemoteParam.new($remote, ext + "active")
        value_ref  =  OSC::RemoteParam.new($remote, path_simp(ext + "../value"))

        value_ref.set_min(0.0)
        value_ref.set_max(1.0)
        value_ref.type = "f"

        path_ref.callback = lambda {|x| param.update_path(x)}
        active_ref.callback = lambda {|x| param.update_active(x)}
        value_ref.callback = lambda {|x| param.update_value(x)}

        gain.extern = ext + "mapping/gain"
        offset.extern = ext + "mapping/offset"


        p.valueRef = [path_ref, active_ref, value_ref]
    
end})

    #puts "ZynAutomationParam, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationPlot < Qml::Widget
  attr_reader :au_plot
  attr_property :minimum
  attr_property :maximum
  attr_accessor :valueRef
  def abs(x);
        if x > 0
            x
        else
            -x
        end
    ;end
  def recalc();
        @gain   ||= 1.0
        @offset ||= 0.5

        #function is out = @offset + input * @gain
        #function is over -1..1
        #function maps to 0..1

        #Calculate x=-1 y value
        ym = @offset + -1*@gain
        #Calculate x=1 y value
        yp = @offset + 1*@gain
        #Calculate y=0 x value
        xm = (0-@offset)/@gain
        #Calculate y=1 x value
        xp = (1-@offset)/@gain

        @points = []

        if(ym < 0)
            @points << [0.0, 0.0, (xm+1)/2, 0.0]
        elsif(ym > 1)
            @points << [0.0, 1.0, (xp+1)/2, 1.0]
        else
            @points << [0.0, ym]
        end

        if(yp < 0)
            @points << [(xm+1)/2, 0.0, 1.0, 0.0]
        elsif(yp > 1)
            @points << [(xp+1)/2, 1.0, 1.0, 1.0]
        else
            @points << [1.0, yp]
        end
        @points = @points.flatten
        damage_self
    ;end
  def set_offset(o);
        return if o == @offset + 0.5
        @offset = o + 0.5
        recalc()
    ;end
  def set_gain(g);
        return if @gain == g
        @gain = g
        recalc()
    ;end
  def draw(vg);
        @gain   ||= 1.0
        @offset ||= 0.5
        recalc() if(@points.nil?)
        background(color("222222"))


        vg.path do
            vg.move_to(0,h/2)
            vg.line_to(w,h/2)
            vg.stroke_color(color("777777"))
            vg.stroke_width(1)
            vg.stroke
        end

        vg.path do
            vg.move_to(w/2,0)
            vg.line_to(w/2,h)
            vg.stroke_color(color("777777"))
            vg.stroke_width(1)
            vg.stroke
        end

        vg.path do
            vg.move_to(0,0)
            vg.line_to(w,0)
            vg.line_to(w,h)
            vg.line_to(0,h)
            vg.line_to(0,0)
            vg.stroke_color(color("aaaaaa"))
            vg.stroke_width(1)
            vg.stroke
        end

        #Draw the actual line
        n = @points.length/2
        vg.path do
            vg.move_to(0,h)
            vg.move_to(w*@points[0], h*(1-@points[1]))
            (1...n).each do |p|
                vg.line_to(w*@points[2*p+0], h*(1-@points[2*p+1]))
            end
            vg.stroke_color(color("aaaaaa"))
            vg.stroke_width(2)
            vg.stroke
        end

        #Draw the active point
        @input ||= 0.3
        out = @offset + @input * @gain
        if(out > 0 && out < 1)
            vg.path do
                vg.circle((@input+1)/2*w, (1-out)*h, 5)
                vg.stroke
            end
        end



    ;end
  def input=(x);if(@input != x);@input = x;damage_self;end;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @au_plot = self

    
context = Hash.new
context["au_plot"] = self
Qml::prop_add(self, "minimum")
Qml::prop_add(self, "maximum")
self.minimum = 0
self.maximum = 0
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        value_ref =  OSC::RemoteParam.new($remote, au_plot.ext + "../value")
        value_ref.callback = lambda {|x| au_plot.input = x}
        au_plot.valueRef = [value_ref]
    
end})

    #puts "ZynAutomationPlot, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynAutomationSlot < Qml::Widget
  attr_reader :slot
  attr_reader :clr
  attr_reader :name
  attr_reader :slide
  attr_reader :active_indicator
  attr_property :slot_id
  def set_active(active=true);
        active_indicator.active = active
    ;end
  def onSetup(old=nil);
        #//Fetch parameters

        #//Reallocate total number of child elements if
    ;end
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.10, 0.8, 0.10])
    ;end
  def animate();
        cur_time   = Time.new
        @atime   ||= cur_time
        old_time   = @atime
        if((cur_time-old_time) >= 1/30.0)
            @atime = cur_time
            slide.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @slot = self
@clr = Qml::ClearBox.new(database, ui_path+"1"+'/')
@anonymous341 = Qml::Widget.new(database, ui_path+"2"+'/')
@name = Qml::TextField.new(database, ui_path+"3"+'/')
@slide = Qml::HSlider.new(database, ui_path+"4"+'/')
@active_indicator = Qml::ArrowBox.new(database, ui_path+"5"+'/')

    
context = Hash.new
context["slot"] = self
context["clr"] = @clr
context["name"] = @name
context["slide"] = @slide
context["active_indicator"] = @active_indicator
Qml::prop_add(self, "slot_id")
@clr.tooltip = "clear automation slot"
@clr.whenValue = lambda {$remote.action(slot.extern + "clear")}
Qml::add_child(self, @clr)

        @anonymous341.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.90, 0.10])\n        ;end", "src/mruby-zest/example/ZynAutomationSlot.qml", 30)
@name.label = "slot"
Qml::add_child(@anonymous341, @name)
@slide.type = "f"
@slide.value = 0.1
Qml::add_child(@anonymous341, @slide)
Qml::add_child(self, @anonymous341)
@active_indicator.whenValue = lambda {
            $remote.action("/automate/active-slot", slot.slot_id)
            root.set_view_pos(:slot, slot.slot_id)
            root.change_view
        }
Qml::add_child(self, @active_indicator)
self.slot_id = 0
@db.connect_watcher(self.properties["extern"], lambda {begin

        #pow.extern   = slot.extern + "active"
        name.extern  = slot.extern + "name"
        slide.extern = slot.extern + "value"
    
end})

    #puts "ZynAutomationSlot, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynBandwidthEnv < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  def cb();
        whenModified.call if whenModified
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous344 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous345 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous346 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous347 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous349 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous350 = Qml::ToggleButton.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@anonymous344.whenValue = lambda { box.cb }
@db.connect_property(@anonymous344.properties["extern"], lambda {box.extern+"PA_val" })
Qml::add_child(@top, @anonymous344)
@anonymous345.whenValue = lambda { box.cb }
@db.connect_property(@anonymous345.properties["extern"], lambda {box.extern+"PA_dt" })
Qml::add_child(@top, @anonymous345)
@anonymous346.whenValue = lambda { box.cb }
@db.connect_property(@anonymous346.properties["extern"], lambda {box.extern+"PR_dt"})
Qml::add_child(@top, @anonymous346)
@anonymous347.whenValue = lambda { box.cb }
@db.connect_property(@anonymous347.properties["extern"], lambda {box.extern+"PR_val"})
Qml::add_child(@top, @anonymous347)
Qml::add_child(self, @top)
@anonymous349.whenValue = lambda { box.cb }
@db.connect_property(@anonymous349.properties["extern"], lambda {box.extern+"Penvstretch"})
Qml::add_child(@bot, @anonymous349)
@anonymous350.label = "FRCR"
@anonymous350.whenValue = lambda { box.cb }
@db.connect_property(@anonymous350.properties["extern"], lambda {box.extern+"Pforcedrelease"})
Qml::add_child(@bot, @anonymous350)
Qml::add_child(self, @bot)
self.label = "Envelope"
self.whenModified = nil

    #puts "ZynBandwidthEnv, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynBank < Qml::Widget
  attr_reader :bank
  attr_reader :lhs
  attr_reader :search
  attr_reader :bank_name
  attr_reader :bank_type
  attr_reader :bank_tag
  attr_reader :rhs
  attr_reader :rwbox
  attr_reader :ins_sel
  attr_property :mode
  def doSearch();
        search_  = bank_name.selected_val
        search_ += " " + bank_type.selected
        search_ += " " + bank_tag.selected
        search_ += " " + search.label
        $remote.action("/bank/search", search_)
    ;end
  def doBank();
        doSearch if self.mode == :read
        setBank  if self.mode == :write
    ;end
  def setBank();
        $remote.action("/bank/blist", bank_name.selected_val)
    ;end
  def doType();
        doSearch if self.mode == :read
        setType  if self.mode == :write
    ;end
  def setType();
        part = root.get_view_pos(:part)
        sel = bank_type.selected_id
        $remote.action("/part#{part}/info.Ptype", sel)
    ;end
  def doLoad();
        ins = ins_sel.selected_val
        part = root.get_view_pos(:part)
        return if part.class != Fixnum
        $remote.action("/load_xiz", part, ins) if !ins.empty?
        $remote.damage("/part#{part}/");
    ;end
  def doSave();
        return if ins_sel.selected_val.nil?
        part = root.get_view_pos(:part)
        $remote.action("/bank/save_to_slot", part, ins_sel.selected_val.to_i)
    ;end
  def doInsSelect();
        doLoad if self.mode == :read
        doSave if self.mode == :write
    ;end
  def layout(l, selfBox);
        rhs.fixed(l,  selfBox, 0.5, 0, 0.5, 1)
        lhs.fixed(l, selfBox, 0.0, 0, 0.5, 1)

        selfBox
    ;end
  def setup(old=nil);
    ;end
end
class Qml::ZynPatchInfo < Qml::Widget
  attr_reader :info_box
  attr_reader :namebox
  attr_reader :authbox
  attr_reader :commbox
  def layout(l, selfBox);
        weights = Draw::DSP::norm_sum([1,1,1,4,1,8])

        Draw::Layout::vfill(l, selfBox, children, weights)
    ;end
  def set_view();
        prt = root.get_view_pos(:part)
        namebox.extern = "/part#{prt}/Pname"
        authbox.extern = "/part#{prt}/info.Pauthor"
        commbox.extern = "/part#{prt}/info.Pcomments"
    ;end
  def onSetup(old=nil);
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @info_box = self
@anonymous966 = Qml::TextBox.new(database, ui_path+"1"+'/')
@namebox = Qml::TextLine.new(database, ui_path+"2"+'/')
@anonymous968 = Qml::TextBox.new(database, ui_path+"3"+'/')
@authbox = Qml::TextEdit.new(database, ui_path+"4"+'/')
@anonymous970 = Qml::TextBox.new(database, ui_path+"5"+'/')
@commbox = Qml::TextEdit.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["info_box"] = self
context["namebox"] = @namebox
context["authbox"] = @authbox
context["commbox"] = @commbox
@anonymous966.label = "name"
Qml::add_child(self, @anonymous966)
Qml::add_child(self, @namebox)
@anonymous968.label = "name"
Qml::add_child(self, @anonymous968)
@authbox.height = 0.25
Qml::add_child(self, @authbox)
@anonymous970.label = "comments"
Qml::add_child(self, @anonymous970)
@db.connect_property(@commbox.properties["height"], lambda {1.0/8.0})
Qml::add_child(self, @commbox)

    #puts "ZynPatchInfo, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynBank
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @bank = self
@lhs = Qml::Widget.new(database, ui_path+"1"+'/')
@search = Qml::SearchBox.new(database, ui_path+"2"+'/')
@anonymous357 = Qml::Widget.new(database, ui_path+"3"+'/')
@bank_name = Qml::SelColumn.new(database, ui_path+"4"+'/')
@bank_type = Qml::SelColumn.new(database, ui_path+"5"+'/')
@bank_tag = Qml::SelColumn.new(database, ui_path+"6"+'/')
@rhs = Qml::Widget.new(database, ui_path+"7"+'/')
@rwbox = Qml::Widget.new(database, ui_path+"8"+'/')
@anonymous359 = Qml::TabButton.new(database, ui_path+"9"+'/')
@anonymous360 = Qml::TabButton.new(database, ui_path+"10"+'/')
@anonymous361 = Qml::TriggerButton.new(database, ui_path+"11"+'/')
@anonymous365 = Qml::Widget.new(database, ui_path+"12"+'/')
@ins_sel = Qml::SelColumn.new(database, ui_path+"13"+'/')
@anonymous364 = Qml::ZynPatchInfo.new(database, ui_path+"14"+'/')

    
context = Hash.new
context["bank"] = self
context["lhs"] = @lhs
context["search"] = @search
context["bank_name"] = @bank_name
context["bank_type"] = @bank_type
context["bank_tag"] = @bank_tag
context["rhs"] = @rhs
context["rwbox"] = @rwbox
@rwbox.instance_eval do
def bank=(k);@bank=k;end
def bank;    @bank;  end
def bank_type=(k);@bank_type=k;end
def bank_type;    @bank_type;  end
end
context["ins_sel"] = @ins_sel
Qml::context_apply(context, @rwbox)
Qml::prop_add(self, "mode")

        @lhs.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0, 0.00, 1, 0.05)\n            children[1].fixed(l, selfBox, 0, 0.05, 1, 0.95)\n\n            selfBox\n        ;end", "src/mruby-zest/example/ZynBank.qml", 92)
@search.whenValue = lambda { bank.doSearch}
Qml::add_child(@lhs, @search)

        @anonymous357.instance_eval("def layout(l, selfBox);\n                Draw::Layout::hfill(l, selfBox, children, [0.3, 0.3, 0.4])\n            ;end", "src/mruby-zest/example/ZynBank.qml", 87)
@bank_name.label = "bank"
@bank_name.skip = true
@db.connect_property(@bank_name.properties["extern"], lambda {"/bank/bank_list"})
@bank_name.whenValue = lambda { bank.doBank }
Qml::add_child(@anonymous357, @bank_name)
@bank_type.label = "type"
@bank_type.extern = "/bank/types"
@bank_type.whenValue = lambda { bank.doType }
Qml::add_child(@anonymous357, @bank_type)
@bank_tag.label = "tag"
@bank_tag.extern = "/bank/tags"
@bank_tag.whenValue = lambda { bank.doSearch }
Qml::add_child(@anonymous357, @bank_tag)
Qml::add_child(@lhs, @anonymous357)
Qml::add_child(self, @lhs)

        @rhs.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0, 0.00, 1, 0.05)\n            children[1].fixed(l, selfBox, 0, 0.05, 1, 0.95)\n\n            selfBox\n        ;end", "src/mruby-zest/example/ZynBank.qml", 155)

        @rwbox.instance_eval("def setrw(x);\n                if(x == 0)\n                    bank.mode = :read\n                    children[0].value = true\n                    children[1].value = false\n                else\n                    bank.mode = :write\n                    children[0].value = false\n                    children[1].value = true\n                end\n                bank_type.clear\n                children[0].damage_self\n                children[1].damage_self\n            ;end", "src/mruby-zest/example/ZynBank.qml", 117)

        @rwbox.instance_eval("def layout(l, selfBox);\n                Draw::Layout::hpack(l, selfBox, children)\n            ;end", "src/mruby-zest/example/ZynBank.qml", 133)
@anonymous359.label = "read"
@anonymous359.value = true
@anonymous359.whenClick = lambda { rwbox.setrw(0) }
@anonymous359.layoutOpts = [:no_constraint]
Qml::add_child(@rwbox, @anonymous359)
@anonymous360.label = "write"
@anonymous360.whenClick = lambda { rwbox.setrw(1) }
@anonymous360.layoutOpts = [:no_constraint]
Qml::add_child(@rwbox, @anonymous360)
@anonymous361.label = "save"
Qml::add_child(@rwbox, @anonymous361)
Qml::add_child(@rhs, @rwbox)

        @anonymous365.instance_eval("def layout(l, selfBox);\n                children[0].fixed(l, selfBox, 0.0, 0, 0.5, 1)\n                children[1].fixed(l, selfBox, 0.5, 0, 0.5, 1)\n\n                selfBox\n            ;end", "src/mruby-zest/example/ZynBank.qml", 148)
@db.connect_property(@ins_sel.properties["extern"], lambda {"/bank/search_results"})
@ins_sel.label = "preset"
@ins_sel.number = false
@ins_sel.skip = true
@ins_sel.whenValue = lambda {bank.doInsSelect}
Qml::add_child(@anonymous365, @ins_sel)
Qml::add_child(@anonymous365, @anonymous364)
Qml::add_child(@rhs, @anonymous365)
Qml::add_child(self, @rhs)
self.mode = :read

    #puts "ZynBank, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynCenter < Qml::Widget
  attr_reader :center
  attr_reader :header
  attr_reader :row1
  attr_reader :ampl
  attr_reader :freq
  attr_reader :filt
  attr_reader :row2
  attr_reader :ampl_env
  attr_reader :freq_env
  attr_reader :filt_env
  attr_reader :explore
  attr_reader :row3
  attr_reader :ampl_lfo
  attr_reader :freq_lfo
  attr_reader :filt_lfo
  def layout(l);
        #Center layout
        selfBox = l.genBox :zynCenter, center
        headBox  = header.layout(l)
        row1Box  = row1.layout(l)
        row2Box  = row2.layout(l)
        content  = explore.layout(l)
        row3Box  = row3.layout(l)
        #layout the module
        amplBox  = ampl.layout(l)
        freqBox  = freq.layout(l)
        filtBox  = filt.layout(l)
        amplEBox = ampl_env.layout(l)
        freqEBox = freq_env.layout(l)
        filtEBox = filt_env.layout(l)
        amplLBox = ampl_lfo.layout(l)
        freqLBox = freq_lfo.layout(l)
        filtLBox = filt_lfo.layout(l)
        #module layout done
        l.contains(selfBox, headBox)
        l.contains(selfBox, row1Box)
        l.contains(selfBox, row2Box)
        l.contains(selfBox, row3Box)

        #Top Row
        l.contains(row1Box,amplBox)
        l.contains(row1Box,freqBox)
        l.contains(row1Box,filtBox)

        l.rightOf(amplBox, freqBox)
        l.rightOf(freqBox, filtBox)

        #Second Row
        l.contains(row2Box,amplEBox)
        l.contains(row2Box,freqEBox)
        l.contains(row2Box,filtEBox)

        l.rightOf(amplEBox, freqEBox)
        l.rightOf(freqEBox, filtEBox)

        #Content
        l.contains(selfBox, content)

        #Third Row
        l.contains(row3Box,amplLBox)
        l.contains(row3Box,freqLBox)
        l.contains(row3Box,filtLBox)

        l.rightOf(amplLBox, freqLBox)
        l.rightOf(freqLBox, filtLBox)

        l.punish2([selfBox.w], [1/3.0], amplLBox.w)
        l.punish2([selfBox.w], [1/3.0], freqLBox.w)
        l.punish2([selfBox.w], [1/3.0], filtLBox.w)

        l.punish2([selfBox.w], [1/3.0], amplEBox.w)
        l.punish2([selfBox.w], [1/3.0], freqEBox.w)
        l.punish2([selfBox.w], [1/3.0], filtEBox.w)

        #Global Optimizatoin
        l.topOf(headBox, row1Box)
        l.topOf(row1Box, row2Box)
        l.topOf(row2Box, content)
        l.topOf(content, row3Box)

        l.sheq([headBox.h, selfBox.h], [1, -0.2*0.2], 0)
        l.sheq([row1Box.h, row2Box.h], [1, -1], 0)
        l.sheq([row1Box.h, row3Box.h], [1, -1], 0)
        l.le([row1Box.h, selfBox.h], [1, -0.2])
        l.le([content.h, selfBox.h], [1, -0.36])
        l.punish_difference(amplBox.w,amplEBox.w)
        l.punish_difference(freqBox.w,freqEBox.w)
        #Center Layout Done

        selfBox
    ;end
end
class Qml::ZynLFO < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous704 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous705 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous706 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous707 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous709 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous710 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous711 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous714 = Qml::Col.new(database, ui_path+"10"+'/')
@anonymous712 = Qml::Selector.new(database, ui_path+"11"+'/')
@anonymous713 = Qml::Button.new(database, ui_path+"12"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
@anonymous704.type = :float
@db.connect_property(@anonymous704.properties["extern"], lambda {box.extern+"Pfreq" })
Qml::add_child(@top, @anonymous704)
@db.connect_property(@anonymous705.properties["extern"], lambda {box.extern+"Pintensity"})
Qml::add_child(@top, @anonymous705)
@db.connect_property(@anonymous706.properties["extern"], lambda {box.extern+"Pstartphase"})
Qml::add_child(@top, @anonymous706)
@db.connect_property(@anonymous707.properties["extern"], lambda {box.extern+"Pstretch"})
Qml::add_child(@top, @anonymous707)
Qml::add_child(self, @top)
@db.connect_property(@anonymous709.properties["extern"], lambda {box.extern+"Pdelay"})
Qml::add_child(@bot, @anonymous709)
@db.connect_property(@anonymous710.properties["extern"], lambda {box.extern+"Prandomness"})
Qml::add_child(@bot, @anonymous710)
@db.connect_property(@anonymous711.properties["extern"], lambda {box.extern+"Pfreqrand"})
Qml::add_child(@bot, @anonymous711)
@db.connect_property(@anonymous712.properties["extern"], lambda {box.extern+"PLFOtype"})
Qml::add_child(@anonymous714, @anonymous712)
@db.connect_property(@anonymous713.properties["extern"], lambda {box.extern+"Pcontinous"})
@anonymous713.label = "sync"
Qml::add_child(@anonymous714, @anonymous713)
Qml::add_child(@bot, @anonymous714)
Qml::add_child(self, @bot)
self.label = "LFO"
self.extern = ""
self.copyable = true

    #puts "ZynLFO, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynCenter
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @center = self
@header = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous368 = Qml::Button.new(database, ui_path+"2"+'/')
@anonymous369 = Qml::Button.new(database, ui_path+"3"+'/')
@anonymous370 = Qml::Button.new(database, ui_path+"4"+'/')
@anonymous371 = Qml::Button.new(database, ui_path+"5"+'/')
@anonymous372 = Qml::Button.new(database, ui_path+"6"+'/')
@anonymous373 = Qml::Button.new(database, ui_path+"7"+'/')
@anonymous374 = Qml::Button.new(database, ui_path+"8"+'/')
@anonymous375 = Qml::Button.new(database, ui_path+"9"+'/')
@anonymous376 = Qml::Button.new(database, ui_path+"10"+'/')
@anonymous377 = Qml::Button.new(database, ui_path+"11"+'/')
@anonymous378 = Qml::Button.new(database, ui_path+"12"+'/')
@row1 = Qml::Widget.new(database, ui_path+"13"+'/')
@ampl = Qml::ParModule.new(database, ui_path+"14"+'/')
@anonymous380 = Qml::Knob.new(database, ui_path+"15"+'/')
@anonymous381 = Qml::Knob.new(database, ui_path+"16"+'/')
@anonymous382 = Qml::Knob.new(database, ui_path+"17"+'/')
@anonymous383 = Qml::Knob.new(database, ui_path+"18"+'/')
@anonymous384 = Qml::Knob.new(database, ui_path+"19"+'/')
@anonymous385 = Qml::Knob.new(database, ui_path+"20"+'/')
@anonymous386 = Qml::Knob.new(database, ui_path+"21"+'/')
@freq = Qml::ParModule.new(database, ui_path+"22"+'/')
@anonymous388 = Qml::Knob.new(database, ui_path+"23"+'/')
@anonymous389 = Qml::Knob.new(database, ui_path+"24"+'/')
@anonymous390 = Qml::Knob.new(database, ui_path+"25"+'/')
@anonymous391 = Qml::Knob.new(database, ui_path+"26"+'/')
@filt = Qml::ParModule.new(database, ui_path+"27"+'/')
@anonymous393 = Qml::Knob.new(database, ui_path+"28"+'/')
@anonymous394 = Qml::Knob.new(database, ui_path+"29"+'/')
@anonymous395 = Qml::Knob.new(database, ui_path+"30"+'/')
@anonymous396 = Qml::Knob.new(database, ui_path+"31"+'/')
@anonymous397 = Qml::Knob.new(database, ui_path+"32"+'/')
@row2 = Qml::Widget.new(database, ui_path+"33"+'/')
@ampl_env = Qml::ParModule.new(database, ui_path+"34"+'/')
@anonymous400 = Qml::Knob.new(database, ui_path+"35"+'/')
@anonymous401 = Qml::Knob.new(database, ui_path+"36"+'/')
@anonymous402 = Qml::Knob.new(database, ui_path+"37"+'/')
@anonymous403 = Qml::Knob.new(database, ui_path+"38"+'/')
@anonymous404 = Qml::Button.new(database, ui_path+"39"+'/')
@anonymous405 = Qml::Button.new(database, ui_path+"40"+'/')
@freq_env = Qml::ParModule.new(database, ui_path+"41"+'/')
@anonymous407 = Qml::Knob.new(database, ui_path+"42"+'/')
@anonymous408 = Qml::Knob.new(database, ui_path+"43"+'/')
@anonymous409 = Qml::Knob.new(database, ui_path+"44"+'/')
@anonymous410 = Qml::Knob.new(database, ui_path+"45"+'/')
@anonymous411 = Qml::Knob.new(database, ui_path+"46"+'/')
@filt_env = Qml::ParModule.new(database, ui_path+"47"+'/')
@anonymous413 = Qml::Knob.new(database, ui_path+"48"+'/')
@anonymous414 = Qml::Knob.new(database, ui_path+"49"+'/')
@anonymous415 = Qml::Knob.new(database, ui_path+"50"+'/')
@explore = Qml::Envelope.new(database, ui_path+"51"+'/')
@row3 = Qml::Widget.new(database, ui_path+"52"+'/')
@ampl_lfo = Qml::ZynLFO.new(database, ui_path+"53"+'/')
@freq_lfo = Qml::ZynLFO.new(database, ui_path+"54"+'/')
@filt_lfo = Qml::ZynLFO.new(database, ui_path+"55"+'/')

    
context = Hash.new
context["center"] = self
context["header"] = @header
@header.instance_eval do
def header=(k);@header=k;end
def header;    @header;  end
end
context["row1"] = @row1
context["ampl"] = @ampl
context["freq"] = @freq
context["filt"] = @filt
context["row2"] = @row2
context["ampl_env"] = @ampl_env
context["freq_env"] = @freq_env
context["filt_env"] = @filt_env
context["explore"] = @explore
context["row3"] = @row3
context["ampl_lfo"] = @ampl_lfo
context["freq_lfo"] = @freq_lfo
context["filt_lfo"] = @filt_lfo
Qml::context_apply(context, @header)

        @header.instance_eval("def measure(text);\n            bounds = Nanovg::Transform.new\n            $vg.text_bounds(0,0,text,bounds)\n            bw = bounds.c-bounds.a\n            bh = bounds.d-bounds.b\n            return bw,bh\n        ;end", "src/mruby-zest/example/ZynCenter.qml", 83)

        @header.instance_eval("def layout(l);\n            selfBox = l.genBox :zynCenterHeader, header\n            prev = nil\n            header.children.each do |ch|\n                box = ch.layout(l)\n                l.contains(selfBox,box)\n                #bw,bh = measure(ch.label)\n\n                #l.aspect(box, bh, bw)\n                if(prev)\n                    l.rightOf(prev, box)\n                end\n                prev = box\n            end\n            selfBox\n        ;end", "src/mruby-zest/example/ZynCenter.qml", 92)
@db.connect_property(@anonymous368.properties["label"], lambda {"<"})
Qml::add_child(@header, @anonymous368)
@anonymous369.label = "4"
Qml::add_child(@header, @anonymous369)
@db.connect_property(@anonymous370.properties["label"], lambda {">"})
Qml::add_child(@header, @anonymous370)
@anonymous371.label = "voice"
Qml::add_child(@header, @anonymous371)
@anonymous372.label = "global parameters"
Qml::add_child(@header, @anonymous372)
@anonymous373.label = "voice parameters"
Qml::add_child(@header, @anonymous373)
@db.connect_property(@anonymous374.properties["label"], lambda {"voice & modulator oscillators"})
Qml::add_child(@header, @anonymous374)
@anonymous375.label = "voice list"
Qml::add_child(@header, @anonymous375)
@anonymous376.label = "resonance"
Qml::add_child(@header, @anonymous376)
@anonymous377.label = "c"
Qml::add_child(@header, @anonymous377)
@anonymous378.label = "p"
Qml::add_child(@header, @anonymous378)
Qml::add_child(self, @header)
@anonymous380.label = "vol"
Qml::add_child(@ampl, @anonymous380)
@db.connect_property(@anonymous381.properties["label"], lambda {"v.sns"})
Qml::add_child(@ampl, @anonymous381)
@anonymous382.label = "pan"
Qml::add_child(@ampl, @anonymous382)
@db.connect_property(@anonymous383.properties["label"], lambda {"p.str"})
Qml::add_child(@ampl, @anonymous383)
@db.connect_property(@anonymous384.properties["label"], lambda {"p.t."})
Qml::add_child(@ampl, @anonymous384)
@db.connect_property(@anonymous385.properties["label"], lambda {"p.stc."})
Qml::add_child(@ampl, @anonymous385)
@db.connect_property(@anonymous386.properties["label"], lambda {"p.vel."})
Qml::add_child(@ampl, @anonymous386)
@ampl.label = "amplitude"
Qml::add_child(@row1, @ampl)
@anonymous388.label = "detune"
Qml::add_child(@freq, @anonymous388)
@anonymous389.label = "octave"
Qml::add_child(@freq, @anonymous389)
@db.connect_property(@anonymous390.properties["label"], lambda {"rel.bw"})
Qml::add_child(@freq, @anonymous390)
@db.connect_property(@anonymous391.properties["label"], lambda {"coarse det."})
Qml::add_child(@freq, @anonymous391)
@freq.label = "frequency"
Qml::add_child(@row1, @freq)
@db.connect_property(@anonymous393.properties["label"], lambda {"c.freq"})
Qml::add_child(@filt, @anonymous393)
@anonymous394.label = "q"
Qml::add_child(@filt, @anonymous394)
@db.connect_property(@anonymous395.properties["label"], lambda {"v.sns a."})
Qml::add_child(@filt, @anonymous395)
@db.connect_property(@anonymous396.properties["label"], lambda {"freq. tr"})
Qml::add_child(@filt, @anonymous396)
@anonymous397.label = "gain"
Qml::add_child(@filt, @anonymous397)
@filt.label = "filter"
Qml::add_child(@row1, @filt)
Qml::add_child(self, @row1)
@db.connect_property(@anonymous400.properties["label"], lambda {"a.dt"})
Qml::add_child(@ampl_env, @anonymous400)
@db.connect_property(@anonymous401.properties["label"], lambda {"d.dt"})
Qml::add_child(@ampl_env, @anonymous401)
@db.connect_property(@anonymous402.properties["label"], lambda {"s.val"})
Qml::add_child(@ampl_env, @anonymous402)
@anonymous403.label = "stretch"
Qml::add_child(@ampl_env, @anonymous403)
@anonymous404.label = "frcr"
Qml::add_child(@ampl_env, @anonymous404)
@anonymous405.label = "L"
Qml::add_child(@ampl_env, @anonymous405)
@ampl_env.label = "amplitude envelope"
Qml::add_child(@row2, @ampl_env)
@db.connect_property(@anonymous407.properties["label"], lambda {"a.dt"})
Qml::add_child(@freq_env, @anonymous407)
@db.connect_property(@anonymous408.properties["label"], lambda {"d.dt"})
Qml::add_child(@freq_env, @anonymous408)
@db.connect_property(@anonymous409.properties["label"], lambda {"s.val"})
Qml::add_child(@freq_env, @anonymous409)
@db.connect_property(@anonymous410.properties["label"], lambda {"r.dt"})
Qml::add_child(@freq_env, @anonymous410)
@anonymous411.label = "stretch"
Qml::add_child(@freq_env, @anonymous411)
@freq_env.label = "frequency envelope"
Qml::add_child(@row2, @freq_env)
@db.connect_property(@anonymous413.properties["label"], lambda {"a.dt"})
Qml::add_child(@filt_env, @anonymous413)
@db.connect_property(@anonymous414.properties["label"], lambda {"d.dt"})
Qml::add_child(@filt_env, @anonymous414)
@db.connect_property(@anonymous415.properties["label"], lambda {"s.val"})
Qml::add_child(@filt_env, @anonymous415)
@filt_env.label = "filter envelope"
Qml::add_child(@row2, @filt_env)
Qml::add_child(self, @row2)
Qml::add_child(self, @explore)
@ampl_lfo.label = "amplitude lfo"
Qml::add_child(@row3, @ampl_lfo)
@freq_lfo.label = "frequency lfo"
Qml::add_child(@row3, @freq_lfo)
@filt_lfo.label = "filter lfo"
Qml::add_child(@row3, @filt_lfo)
Qml::add_child(self, @row3)

    #puts "ZynCenter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynChorus < Qml::Group
  attr_reader :chorus
  attr_reader :rw
  def refresh();
        return if rw.content.nil?
        return if rw.content.children.length < 4
        rw.content.children[3..-1].each do |c|
            c.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @chorus = self
@rw = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous424 = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous425 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous426 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous427 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous428 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous429 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous430 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous431 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous432 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous433 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous434 = Qml::Knob.new(database, ui_path+"12"+'/')
@anonymous435 = Qml::ToggleButton.new(database, ui_path+"13"+'/')
@anonymous436 = Qml::ToggleButton.new(database, ui_path+"14"+'/')

    
context = Hash.new
context["chorus"] = self
context["rw"] = @rw
@db.connect_property(@anonymous424.properties["extern"], lambda {chorus.extern + "Chorus/preset"})
@anonymous424.whenValue = lambda { chorus.refresh }
Qml::add_child(@rw, @anonymous424)
@db.connect_property(@anonymous425.properties["extern"], lambda {chorus.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous425)
@db.connect_property(@anonymous426.properties["extern"], lambda {chorus.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous426)
@db.connect_property(@anonymous427.properties["extern"], lambda {chorus.extern + "Chorus/Pfreq" })
Qml::add_child(@rw, @anonymous427)
@db.connect_property(@anonymous428.properties["extern"], lambda {chorus.extern + "Chorus/Pfreqrnd" })
Qml::add_child(@rw, @anonymous428)
@db.connect_property(@anonymous429.properties["extern"], lambda {chorus.extern + "Chorus/PLFOtype" })
Qml::add_child(@rw, @anonymous429)
@db.connect_property(@anonymous430.properties["extern"], lambda {chorus.extern + "Chorus/PStereo" })
Qml::add_child(@rw, @anonymous430)
@db.connect_property(@anonymous431.properties["extern"], lambda {chorus.extern + "Chorus/Pdepth" })
Qml::add_child(@rw, @anonymous431)
@db.connect_property(@anonymous432.properties["extern"], lambda {chorus.extern + "Chorus/Pdelay" })
Qml::add_child(@rw, @anonymous432)
@db.connect_property(@anonymous433.properties["extern"], lambda {chorus.extern + "Chorus/Pfeedback" })
Qml::add_child(@rw, @anonymous433)
@db.connect_property(@anonymous434.properties["extern"], lambda {chorus.extern + "Chorus/Plrcross" })
Qml::add_child(@rw, @anonymous434)
@db.connect_property(@anonymous435.properties["extern"], lambda {chorus.extern + "Chorus/Pflangemode" })
Qml::add_child(@rw, @anonymous435)
@db.connect_property(@anonymous436.properties["extern"], lambda {chorus.extern + "Chorus/Poutsub" })
Qml::add_child(@rw, @anonymous436)
@rw.layoutOpts = []
Qml::add_child(self, @rw)
self.label = "chorus"
self.topSize = 0.2

    #puts "ZynChorus, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynControllers < Qml::Group
  attr_reader :ctrl
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @ctrl = self
@anonymous445 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous439 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous440 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous441 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous442 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous443 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous444 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous455 = Qml::ParModuleRow.new(database, ui_path+"8"+'/')
@anonymous448 = Qml::Col.new(database, ui_path+"9"+'/')
@anonymous446 = Qml::ToggleButton.new(database, ui_path+"10"+'/')
@anonymous447 = Qml::ToggleButton.new(database, ui_path+"11"+'/')
@anonymous451 = Qml::Col.new(database, ui_path+"12"+'/')
@anonymous449 = Qml::ToggleButton.new(database, ui_path+"13"+'/')
@anonymous450 = Qml::ToggleButton.new(database, ui_path+"14"+'/')
@anonymous454 = Qml::Col.new(database, ui_path+"15"+'/')
@anonymous452 = Qml::ToggleButton.new(database, ui_path+"16"+'/')
@anonymous453 = Qml::ToggleButton.new(database, ui_path+"17"+'/')

    
context = Hash.new
context["ctrl"] = self
@db.connect_property(@anonymous439.properties["extern"], lambda {ctrl.extern+"modwheel.depth"})
Qml::add_child(@anonymous445, @anonymous439)
@db.connect_property(@anonymous440.properties["extern"], lambda {ctrl.extern+"filtercutoff.depth"})
Qml::add_child(@anonymous445, @anonymous440)
@db.connect_property(@anonymous441.properties["extern"], lambda {ctrl.extern+"filterq.depth"})
Qml::add_child(@anonymous445, @anonymous441)
@db.connect_property(@anonymous442.properties["extern"], lambda {ctrl.extern+"bandwidth.depth"})
Qml::add_child(@anonymous445, @anonymous442)
@db.connect_property(@anonymous443.properties["extern"], lambda {ctrl.extern+"pitchwheel.bendrange"})
Qml::add_child(@anonymous445, @anonymous443)
@db.connect_property(@anonymous444.properties["extern"], lambda {ctrl.extern+"panning.depth"})
Qml::add_child(@anonymous445, @anonymous444)
Qml::add_child(self, @anonymous445)
@db.connect_property(@anonymous446.properties["extern"], lambda {ctrl.extern+"modwheel.exponential"})
Qml::add_child(@anonymous448, @anonymous446)
@db.connect_property(@anonymous447.properties["extern"], lambda {ctrl.extern+"bandwidth.exponential"})
Qml::add_child(@anonymous448, @anonymous447)
Qml::add_child(@anonymous455, @anonymous448)
@db.connect_property(@anonymous449.properties["extern"], lambda {ctrl.extern+"expression.receive"})
Qml::add_child(@anonymous451, @anonymous449)
@db.connect_property(@anonymous450.properties["extern"], lambda {ctrl.extern+"volume.receive"})
Qml::add_child(@anonymous451, @anonymous450)
Qml::add_child(@anonymous455, @anonymous451)
@db.connect_property(@anonymous452.properties["extern"], lambda {ctrl.extern+"fmamp.receive"})
Qml::add_child(@anonymous454, @anonymous452)
@db.connect_property(@anonymous453.properties["extern"], lambda {ctrl.extern+"sustain.receive"})
Qml::add_child(@anonymous454, @anonymous453)
Qml::add_child(@anonymous455, @anonymous454)
@anonymous455.lsize = 0.0
Qml::add_child(self, @anonymous455)
self.label = "controllers"
self.copyable = false

    #puts "ZynControllers, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDAFilter < Qml::Widget
  attr_reader :dyn
  attr_reader :vis
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.5, 0.5])
    ;end
end
class Qml::ZynDyFilter < Qml::Group
  attr_reader :box
  attr_reader :cat
  attr_reader :typ
  attr_property :whenModified
  attr_property :type
  def cb();
        whenModified.call if whenModified
    ;end
  def change_cat();
        root.change_view()
        dest = self.extern + "Ptype"    if cat.selected == 0
        return                          if cat.selected == 1
        dest = self.extern + "type-svf" if cat.selected == 2
        if(typ.extern != dest)
            typ.extern = dest
            typ.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous487 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous481 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous482 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous483 = Qml::NumEntry.new(database, ui_path+"4"+'/')
@cat = Qml::Selector.new(database, ui_path+"5"+'/')
@typ = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous486 = Qml::Knob.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["box"] = self
context["cat"] = @cat
context["typ"] = @typ
Qml::prop_add(self, "whenModified")
Qml::prop_add(self, "type")
@anonymous481.type = :float
@anonymous481.whenValue = lambda { box.cb}
@db.connect_property(@anonymous481.properties["extern"], lambda {box.extern + "basefreq"})
Qml::add_child(@anonymous487, @anonymous481)
@anonymous482.type = :float
@anonymous482.whenValue = lambda { box.cb}
@db.connect_property(@anonymous482.properties["extern"], lambda {box.extern + "baseq"})
Qml::add_child(@anonymous487, @anonymous482)
@anonymous483.whenValue = lambda { box.cb}
@db.connect_property(@anonymous483.properties["extern"], lambda {box.extern + "Pstages"})
@anonymous483.offset = 1
@anonymous483.minimum = 1
@anonymous483.maximum = 5
Qml::add_child(@anonymous487, @anonymous483)
@db.connect_property(@cat.properties["extern"], lambda {box.extern + "Pcategory"})
Qml::add_child(@anonymous487, @cat)
@typ.whenValue = lambda { box.cb}
@db.connect_property(@typ.properties["extern"], lambda {box.extern + "Ptype"})
Qml::add_child(@anonymous487, @typ)
@anonymous486.type = :float
@anonymous486.whenValue = lambda { box.cb}
@db.connect_property(@anonymous486.properties["extern"], lambda {box.extern + "gain"})
Qml::add_child(@anonymous487, @anonymous486)
Qml::add_child(self, @anonymous487)
self.label = "General"
self.topSize = 0.2
self.whenModified = nil
self.type = :analog
@db.connect_watcher(self.properties["type"], lambda {begin

        #Changes filter type
    
end})

    #puts "ZynDyFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDAFilter
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @dyn = self
@vis = Qml::VisFilter.new(database, ui_path+"1"+'/')
@anonymous458 = Qml::ZynDyFilter.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["dyn"] = self
context["vis"] = @vis
@db.connect_property(@vis.properties["extern"], lambda {dyn.extern + "filterpars/response"})
Qml::add_child(self, @vis)
@db.connect_property(@anonymous458.properties["whenModified"], lambda {lambda { vis.refresh }})
@db.connect_property(@anonymous458.properties["extern"], lambda {dyn.extern + "filterpars/"})
Qml::add_child(self, @anonymous458)

    #puts "ZynDAFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDFFilter < Qml::Widget
  attr_reader :dyn
end
class Qml::ZynFormant < Qml::Widget
  attr_reader :fbase
  attr_reader :vis
  attr_reader :total
  attr_reader :vgrp
  attr_reader :vow_num
  attr_reader :slide_freq
  attr_reader :for_num
  attr_reader :slide_q
  attr_reader :slide_amp
  attr_reader :pos_id
  attr_reader :pos_vowel
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.7, 0.3], 0, 2)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @fbase = self
@vis = Qml::VisFormant.new(database, ui_path+"1"+'/')
@anonymous608 = Qml::Widget.new(database, ui_path+"2"+'/')
@total = Qml::Widget.new(database, ui_path+"3"+'/')
@anonymous582 = Qml::Text.new(database, ui_path+"4"+'/')
@anonymous583 = Qml::NumEntry.new(database, ui_path+"5"+'/')
@anonymous584 = Qml::HSlider.new(database, ui_path+"6"+'/')
@anonymous585 = Qml::HSlider.new(database, ui_path+"7"+'/')
@anonymous586 = Qml::HSlider.new(database, ui_path+"8"+'/')
@anonymous587 = Qml::HSlider.new(database, ui_path+"9"+'/')
@anonymous607 = Qml::Widget.new(database, ui_path+"10"+'/')
@vgrp = Qml::Group.new(database, ui_path+"11"+'/')
@anonymous597 = Qml::Widget.new(database, ui_path+"12"+'/')
@vow_num = Qml::NumEntry.new(database, ui_path+"13"+'/')
@slide_freq = Qml::HSlider.new(database, ui_path+"14"+'/')
@for_num = Qml::NumEntry.new(database, ui_path+"15"+'/')
@slide_q = Qml::HSlider.new(database, ui_path+"16"+'/')
@anonymous593 = Qml::Widget.new(database, ui_path+"17"+'/')
@anonymous594 = Qml::Widget.new(database, ui_path+"18"+'/')
@anonymous595 = Qml::Widget.new(database, ui_path+"19"+'/')
@slide_amp = Qml::HSlider.new(database, ui_path+"20"+'/')
@anonymous606 = Qml::Group.new(database, ui_path+"21"+'/')
@anonymous605 = Qml::Widget.new(database, ui_path+"22"+'/')
@anonymous599 = Qml::NumEntry.new(database, ui_path+"23"+'/')
@pos_id = Qml::NumEntry.new(database, ui_path+"24"+'/')
@anonymous601 = Qml::HSlider.new(database, ui_path+"25"+'/')
@anonymous602 = Qml::Widget.new(database, ui_path+"26"+'/')
@pos_vowel = Qml::NumEntry.new(database, ui_path+"27"+'/')
@anonymous604 = Qml::Button.new(database, ui_path+"28"+'/')

    
context = Hash.new
context["fbase"] = self
context["vis"] = @vis
context["total"] = @total
@total.instance_eval do
def vis=(k);@vis=k;end
def vis;    @vis;  end
end
context["vgrp"] = @vgrp
@vgrp.instance_eval do
def fbase=(k);@fbase=k;end
def fbase;    @fbase;  end
def vis=(k);@vis=k;end
def vis;    @vis;  end
def vow_num=(k);@vow_num=k;end
def vow_num;    @vow_num;  end
def slide_freq=(k);@slide_freq=k;end
def slide_freq;    @slide_freq;  end
def for_num=(k);@for_num=k;end
def for_num;    @for_num;  end
def slide_q=(k);@slide_q=k;end
def slide_q;    @slide_q;  end
def slide_amp=(k);@slide_amp=k;end
def slide_amp;    @slide_amp;  end
end
context["vow_num"] = @vow_num
context["slide_freq"] = @slide_freq
context["for_num"] = @for_num
context["slide_q"] = @slide_q
context["slide_amp"] = @slide_amp
context["pos_id"] = @pos_id
context["pos_vowel"] = @pos_vowel
Qml::context_apply(context, @total)
Qml::context_apply(context, @vgrp)
@db.connect_property(@vis.properties["extern"], lambda {fbase.extern})
Qml::add_child(self, @vis)

        @anonymous608.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 2)\n        ;end", "src/mruby-zest/example/ZynFormant.qml", 150)

        @total.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children, 0.1, 0.8, 12)\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 20)

        @total.instance_eval("def onSetup(old=nil);\n                lm = lambda { vis.refresh() }\n                children.each do |ch|\n                    ch.whenValue = lm if ch.respond_to? :whenValue\n                end\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 23)

        @total.instance_eval("def draw(vg);\n                Draw::GradBox(vg, Rect.new(0,0,w,h))\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 29)
@anonymous582.label = "formants"
Qml::add_child(@total, @anonymous582)
@anonymous583.format = "PKs "
@db.connect_property(@anonymous583.properties["extern"], lambda {fbase.extern + "Pnumformants"})
Qml::add_child(@total, @anonymous583)
@db.connect_property(@anonymous584.properties["extern"], lambda {fbase.extern + "Pformantslowness"})
Qml::add_child(@total, @anonymous584)
@db.connect_property(@anonymous585.properties["extern"], lambda {fbase.extern + "Pvowelclearness"})
Qml::add_child(@total, @anonymous585)
@db.connect_property(@anonymous586.properties["extern"], lambda {fbase.extern + "Pcenterfreq" })
Qml::add_child(@total, @anonymous586)
@db.connect_property(@anonymous587.properties["extern"], lambda {fbase.extern + "Poctavesfreq" })
Qml::add_child(@total, @anonymous587)
Qml::add_child(@anonymous608, @total)

        @anonymous607.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children)\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 146)

        @vgrp.instance_eval("def update_pos();\n                    vis.vowel_num = vow_num.value-1\n                    vis.change()\n                    rt = fbase.extern + \"Pvowels\#{vow_num.value-1}/Pformants\#{for_num.value-1}/\"\n                    slide_freq.extern = rt + \"freq\"\n                    slide_q.extern    = rt + \"q\"\n                    slide_amp.extern  = rt + \"amp\"\n                    db.update_values\n                ;end", "src/mruby-zest/example/ZynFormant.qml", 39)

        @anonymous597.instance_eval("def layout(l, selfBox);\n                        Draw::Layout::grid(l, selfBox, children, 4, 2, 2, 2)\n                    ;end", "src/mruby-zest/example/ZynFormant.qml", 91)
@db.connect_property(@vow_num.properties["label"], lambda {"v. num"})
@vow_num.format = "VW "
@vow_num.tooltip = "vowel number - each formant filter is composed of one or more vowels"
@vow_num.minimum = 1
@vow_num.maximum = 6
@vow_num.value = 1
@vow_num.whenValue = lambda {vgrp.update_pos}
Qml::add_child(@anonymous597, @vow_num)
@db.connect_property(@slide_freq.properties["extern"], lambda {vgrp.extern + "freq"})
@slide_freq.label = "freq"
@slide_freq.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_freq)
@for_num.label = "formant"
@for_num.format = "PK "
@for_num.tooltip = "formant peak - each formant vowel is composed of one or more peaks"
@for_num.minimum = 1
@for_num.maximum = 12
@for_num.value = 1
@for_num.whenValue = lambda {vgrp.update_pos}
Qml::add_child(@anonymous597, @for_num)
@db.connect_property(@slide_q.properties["extern"], lambda {vgrp.extern + "q"})
@slide_q.label = "q"
@slide_q.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_q)
Qml::add_child(@anonymous597, @anonymous593)
Qml::add_child(@anonymous597, @anonymous594)
Qml::add_child(@anonymous597, @anonymous595)
@db.connect_property(@slide_amp.properties["extern"], lambda {vgrp.extern + "amp"})
@slide_amp.label = "amp"
@slide_amp.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_amp)
Qml::add_child(@vgrp, @anonymous597)
@vgrp.label = "vowel"
@vgrp.topSize = 0.2
@db.connect_property(@vgrp.properties["extern"], lambda {fbase.extern + "Pvowels0/Pformants0/"})
Qml::add_child(@anonymous607, @vgrp)

        @anonymous605.instance_eval("def layout(l, selfBox);\n                        Draw::Layout::grid(l, selfBox, children, 3, 2, 2, 4)\n                    ;end", "src/mruby-zest/example/ZynFormant.qml", 141)
@db.connect_property(@anonymous599.properties["extern"], lambda {fbase.extern + "Psequencesize"})
@anonymous599.format = "LEN "
@anonymous599.label = "seqsize"
@anonymous599.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous605, @anonymous599)
@pos_id.format = "POS "
@pos_id.value = 1
@pos_id.minimum = 1
@pos_id.maximum = 8
@db.connect_property(@pos_id.properties["label"], lambda {"s.pos"})
@pos_id.whenValue = lambda {
                            pos_vowel.extern = 
                            fbase.extern + "vowel_seq#{pos_id.value-1}"
                        }
Qml::add_child(@anonymous605, @pos_id)
@db.connect_property(@anonymous601.properties["extern"], lambda {fbase.extern + "Psequencestretch"})
@anonymous601.label = "strch"
@anonymous601.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous605, @anonymous601)
Qml::add_child(@anonymous605, @anonymous602)
@pos_vowel.format = "VW "
@pos_vowel.label = "vowel"
@pos_vowel.offset = 1
@pos_vowel.minimum = 1
@pos_vowel.maximum = 8
@db.connect_property(@pos_vowel.properties["extern"], lambda {fbase.extern + "vowel_seq0"})
Qml::add_child(@anonymous605, @pos_vowel)
@db.connect_property(@anonymous604.properties["extern"], lambda {fbase.extern + "Psequencereversed"})
@db.connect_property(@anonymous604.properties["label"], lambda {"neg. input"})
@anonymous604.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous605, @anonymous604)
Qml::add_child(@anonymous606, @anonymous605)
@anonymous606.label = "sequence"
@anonymous606.topSize = 0.2
@anonymous606.copyable = false
Qml::add_child(@anonymous607, @anonymous606)
Qml::add_child(@anonymous608, @anonymous607)
Qml::add_child(self, @anonymous608)

    #puts "ZynFormant, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDFFilter
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @dyn = self
@anonymous460 = Qml::ZynFormant.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["dyn"] = self
@db.connect_property(@anonymous460.properties["extern"], lambda {dyn.extern + "filterpars/"})
Qml::add_child(self, @anonymous460)

    #puts "ZynDFFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDistortion < Qml::Widget
  attr_reader :dst
  attr_reader :wave
  attr_reader :rw
  attr_reader :rw2
  def refresh_recur(x);
        #@@recur_level ||= 0
        #@@recur_level += 1
        #print " "*@@recur_level
        #puts "Distort refresh = {#{x.class}} of {#{dst.class}}"
        x.children.each do |xx|
            #print " "*(@@recur_level+1)
            #puts "child = #{xx.class}"
            xx.refresh() if xx.respond_to? :refresh
            dst.refresh_recur(xx)
        end
        #@@recur_level -= 1
    ;end
  def refresh();
        refresh_recur(self)
    ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.15, 0.85])
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @dst = self
@anonymous462 = Qml::GroupHeader.new(database, ui_path+"1"+'/')
@anonymous479 = Qml::Widget.new(database, ui_path+"2"+'/')
@wave = Qml::WaveView.new(database, ui_path+"3"+'/')
@anonymous478 = Qml::Widget.new(database, ui_path+"4"+'/')
@rw = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous464 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous465 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous466 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous467 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous468 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous469 = Qml::Knob.new(database, ui_path+"11"+'/')
@rw2 = Qml::ParModuleRow.new(database, ui_path+"12"+'/')
@anonymous471 = Qml::Selector.new(database, ui_path+"13"+'/')
@anonymous472 = Qml::Knob.new(database, ui_path+"14"+'/')
@anonymous473 = Qml::Knob.new(database, ui_path+"15"+'/')
@anonymous476 = Qml::Col.new(database, ui_path+"16"+'/')
@anonymous474 = Qml::ToggleButton.new(database, ui_path+"17"+'/')
@anonymous475 = Qml::ToggleButton.new(database, ui_path+"18"+'/')

    
context = Hash.new
context["dst"] = self
context["wave"] = @wave
context["rw"] = @rw
context["rw2"] = @rw2
@anonymous462.label = "Distorsion"
@db.connect_property(@anonymous462.properties["extern"], lambda {dst.extern})
@anonymous462.copyable = true
Qml::add_child(self, @anonymous462)

        @anonymous479.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynDistortion.qml", 71)
@db.connect_property(@wave.properties["extern"], lambda {dst.extern + "Distorsion/waveform"})
Qml::add_child(@anonymous479, @wave)

        @anonymous478.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children)\n            ;end", "src/mruby-zest/example/ZynDistortion.qml", 66)
@db.connect_property(@anonymous464.properties["extern"], lambda {dst.extern + "Distorsion/preset"})
@anonymous464.whenValue = lambda { dst.refresh }
@anonymous464.layoutOpts = [:long_mode]
Qml::add_child(@rw, @anonymous464)
@db.connect_property(@anonymous465.properties["extern"], lambda {dst.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous465)
@db.connect_property(@anonymous466.properties["extern"], lambda {dst.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous466)
@db.connect_property(@anonymous467.properties["extern"], lambda {dst.extern + "Distorsion/Plrcross"})
@db.connect_property(@anonymous467.properties["label"], lambda {"l.rc."})
Qml::add_child(@rw, @anonymous467)
@db.connect_property(@anonymous468.properties["extern"], lambda {dst.extern + "Distorsion/Plpf"})
Qml::add_child(@rw, @anonymous468)
@db.connect_property(@anonymous469.properties["extern"], lambda {dst.extern + "Distorsion/Phpf"})
Qml::add_child(@rw, @anonymous469)
@rw.layoutOpts = []
Qml::add_child(@anonymous478, @rw)
@db.connect_property(@anonymous471.properties["extern"], lambda {dst.extern + "Distorsion/Ptype"})
@anonymous471.whenValue = lambda {wave.refresh}
@anonymous471.layoutOpts = [:long_mode]
Qml::add_child(@rw2, @anonymous471)
@db.connect_property(@anonymous472.properties["extern"], lambda {dst.extern + "Distorsion/Pdrive"})
@anonymous472.label = "drive"
@anonymous472.whenValue = lambda {wave.refresh}
Qml::add_child(@rw2, @anonymous472)
@db.connect_property(@anonymous473.properties["extern"], lambda {dst.extern + "Distorsion/Plevel"})
@anonymous473.label = "level"
Qml::add_child(@rw2, @anonymous473)
@db.connect_property(@anonymous474.properties["extern"], lambda {dst.extern + "Distorsion/Pprefiltering"})
Qml::add_child(@anonymous476, @anonymous474)
@db.connect_property(@anonymous475.properties["extern"], lambda {dst.extern + "Distorsion/Pstereo"})
Qml::add_child(@anonymous476, @anonymous475)
Qml::add_child(@rw2, @anonymous476)
@rw2.layoutOpts = []
Qml::add_child(@anonymous478, @rw2)
Qml::add_child(@anonymous479, @anonymous478)
Qml::add_child(self, @anonymous479)

    #puts "ZynDistortion, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDyFilter < Qml::Group
  attr_reader :box
  attr_reader :cat
  attr_reader :typ
  attr_property :whenModified
  attr_property :type
  def cb();
        whenModified.call if whenModified
    ;end
  def change_cat();
        root.change_view()
        dest = self.extern + "Ptype"    if cat.selected == 0
        return                          if cat.selected == 1
        dest = self.extern + "type-svf" if cat.selected == 2
        if(typ.extern != dest)
            typ.extern = dest
            typ.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous487 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous481 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous482 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous483 = Qml::NumEntry.new(database, ui_path+"4"+'/')
@cat = Qml::Selector.new(database, ui_path+"5"+'/')
@typ = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous486 = Qml::Knob.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["box"] = self
context["cat"] = @cat
context["typ"] = @typ
Qml::prop_add(self, "whenModified")
Qml::prop_add(self, "type")
@anonymous481.type = :float
@anonymous481.whenValue = lambda { box.cb}
@db.connect_property(@anonymous481.properties["extern"], lambda {box.extern + "basefreq"})
Qml::add_child(@anonymous487, @anonymous481)
@anonymous482.type = :float
@anonymous482.whenValue = lambda { box.cb}
@db.connect_property(@anonymous482.properties["extern"], lambda {box.extern + "baseq"})
Qml::add_child(@anonymous487, @anonymous482)
@anonymous483.whenValue = lambda { box.cb}
@db.connect_property(@anonymous483.properties["extern"], lambda {box.extern + "Pstages"})
@anonymous483.offset = 1
@anonymous483.minimum = 1
@anonymous483.maximum = 5
Qml::add_child(@anonymous487, @anonymous483)
@db.connect_property(@cat.properties["extern"], lambda {box.extern + "Pcategory"})
Qml::add_child(@anonymous487, @cat)
@typ.whenValue = lambda { box.cb}
@db.connect_property(@typ.properties["extern"], lambda {box.extern + "Ptype"})
Qml::add_child(@anonymous487, @typ)
@anonymous486.type = :float
@anonymous486.whenValue = lambda { box.cb}
@db.connect_property(@anonymous486.properties["extern"], lambda {box.extern + "gain"})
Qml::add_child(@anonymous487, @anonymous486)
Qml::add_child(self, @anonymous487)
self.label = "General"
self.topSize = 0.2
self.whenModified = nil
self.type = :analog
@db.connect_watcher(self.properties["type"], lambda {begin

        #Changes filter type
    
end})

    #puts "ZynDyFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynDynFilter < Qml::Widget
  attr_reader :dyn
  attr_reader :swapp
  attr_reader :cat
  attr_accessor :valueRef
  attr_property :filtertype
  def add_cat();
        return if self.valueRef
        if(self.valueRef.nil?)
            path = self.extern + "filterpars/Pcategory"
            self.valueRef = OSC::RemoteParam.new($remote, path)
            self.valueRef.mode = :full
            self.valueRef.callback = lambda {|x|
                dyn.filtertype = [:analog, :formant, :statevar][x]
                swapp.content  = Qml::ZynDAFilter if dyn.filtertype != :formant
                swapp.content  = Qml::ZynDFFilter if dyn.filtertype == :formant
            }
        end
    ;end
  def draw(vg);
        Draw::GradBox(vg, Rect.new(0, 0, w, h))
    ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.75,0.25])
    ;end
  def onSetup(old=nil);;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @dyn = self
@swapp = Qml::Swappable.new(database, ui_path+"1"+'/')
@anonymous501 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous490 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous491 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous492 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous493 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous494 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous495 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous496 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous497 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous498 = Qml::ToggleButton.new(database, ui_path+"11"+'/')
@anonymous499 = Qml::Knob.new(database, ui_path+"12"+'/')
@cat = Qml::Selector.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["dyn"] = self
context["swapp"] = @swapp
context["cat"] = @cat
Qml::prop_add(self, "filtertype")
@db.connect_property(@swapp.properties["extern"], lambda {dyn.extern})
@db.connect_property(@swapp.properties["content"], lambda {Qml::ZynDAFilter})
Qml::add_child(self, @swapp)
@db.connect_property(@anonymous490.properties["extern"], lambda {dyn.extern + "Pvolume"})
Qml::add_child(@anonymous501, @anonymous490)
@db.connect_property(@anonymous491.properties["extern"], lambda {dyn.extern + "Ppanning"})
Qml::add_child(@anonymous501, @anonymous491)
@db.connect_property(@anonymous492.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pfreq" })
Qml::add_child(@anonymous501, @anonymous492)
@db.connect_property(@anonymous493.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pfreqrnd" })
Qml::add_child(@anonymous501, @anonymous493)
@db.connect_property(@anonymous494.properties["extern"], lambda {dyn.extern + "DynamicFilter/PLFOtype" })
Qml::add_child(@anonymous501, @anonymous494)
@db.connect_property(@anonymous495.properties["extern"], lambda {dyn.extern + "DynamicFilter/PStereo" })
Qml::add_child(@anonymous501, @anonymous495)
@db.connect_property(@anonymous496.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pdepth" })
Qml::add_child(@anonymous501, @anonymous496)
@db.connect_property(@anonymous497.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pampsns" })
Qml::add_child(@anonymous501, @anonymous497)
@db.connect_property(@anonymous498.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pampsnsinv" })
Qml::add_child(@anonymous501, @anonymous498)
@db.connect_property(@anonymous499.properties["extern"], lambda {dyn.extern + "DynamicFilter/Pampsmooth" })
Qml::add_child(@anonymous501, @anonymous499)
@db.connect_property(@cat.properties["extern"], lambda {dyn.extern + "filterpars/Pcategory"})
Qml::add_child(@anonymous501, @cat)
Qml::add_child(self, @anonymous501)
self.valueRef = nil
self.filtertype = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        dyn.add_cat()
    
end})

    #puts "ZynDynFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEcho < Qml::Group
  attr_reader :echo
  attr_reader :rw
  def refresh();
        return if rw.content.nil?
        return if rw.content.children.length < 4
        rw.content.children[3..-1].each do |c|
            c.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @echo = self
@rw = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous503 = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous504 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous505 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous506 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous507 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous508 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous509 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous510 = Qml::Knob.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["echo"] = self
context["rw"] = @rw
@anonymous503.layoutOpts = [:long_mode]
@db.connect_property(@anonymous503.properties["extern"], lambda {echo.extern + "Echo/preset"})
@anonymous503.whenValue = lambda { echo.refresh }
Qml::add_child(@rw, @anonymous503)
@db.connect_property(@anonymous504.properties["extern"], lambda {echo.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous504)
@db.connect_property(@anonymous505.properties["extern"], lambda {echo.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous505)
@db.connect_property(@anonymous506.properties["extern"], lambda {echo.extern + "Echo/Pdelay"   })
Qml::add_child(@rw, @anonymous506)
@db.connect_property(@anonymous507.properties["extern"], lambda {echo.extern + "Echo/Plrdelay" })
Qml::add_child(@rw, @anonymous507)
@db.connect_property(@anonymous508.properties["extern"], lambda {echo.extern + "Echo/Plrcross" })
Qml::add_child(@rw, @anonymous508)
@db.connect_property(@anonymous509.properties["extern"], lambda {echo.extern + "Echo/Pfb" })
Qml::add_child(@rw, @anonymous509)
@db.connect_property(@anonymous510.properties["extern"], lambda {echo.extern + "Echo/Phidamp" })
Qml::add_child(@rw, @anonymous510)
@rw.layoutOpts = []
Qml::add_child(self, @rw)
self.label = "echo"
self.topSize = 0.2

    #puts "ZynEcho, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectBypass < Qml::Widget
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.7, 0.3], 0, 2)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous515 = self
@anonymous513 = Qml::Selector.new(database, ui_path+"1"+'/')
@anonymous514 = Qml::ToggleButton.new(database, ui_path+"2"+'/')

    
context = Hash.new
@anonymous513.layoutOpts = [:no_constraint]
@anonymous513.max_text = 5
Qml::add_child(self, @anonymous513)
@anonymous514.label = "b"
@anonymous514.layoutOpts = [:no_constraint]
Qml::add_child(self, @anonymous514)

    #puts "ZynEffectBypass, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectGroup < Qml::Widget
  attr_reader :egrp
  attr_property :nunits
  attr_property :offset
  attr_property :effects
  attr_property :maxeffects
  attr_property :shownWeights
  attr_accessor :valueRef
  attr_property :needsRegen
  def onSetup(old=nil);
        v = []
        (0...maxeffects).each do |i|
            rf = extern + i.to_s + "/efftype"
            r  = OSC::RemoteParam.new($remote, rf)
            r.mode = :options
            r.callback = lambda {|x|
                lu = egrp.lookup(x)
                if(egrp.effects[i] != lu)
                    egrp.effects[i] = lu
                    egrp.needsRegen = true
                end
            }
            v << r
        end
        self.valueRef = v
        generate_children #if(egrp.effects.length != 0)
    ;end
  def total_len();
        total = 0
        (0...maxeffects).each do |r|
            if(effects.include?(r) && effects[r] != :none)
                lu = get_units(effects[r])
                total += lu
            else
                total += 1
            end
        end
        total
    ;end
  def max(a,b);
        if(a>b)
            a
        else
            b
        end
    ;end
  def animate();
        v = 1.0-children[0].value
        range = max(0, total_len-self.nunits)
        noff = (v*range).to_i
        if(noff != self.offset)
            self.offset = noff
            self.needsRegen = true
        end

        regen_children if self.needsRegen
    ;end
  def rec_del_props(widget, plist=[]);
        return plist if widget.nil?
        #Add all properties to the delete list
        widget.properties.each do |k,v|
            plist << v
        end
        #Add all children properties to the list
        widget.children.each do |ch|
            rec_del_props(ch, plist)
        end
        plist
    ;end
  def rec_clean_cbs(widget);
        if(widget.respond_to? :valueRef)
            v = widget.valueRef
            if(v.class == Array)
                v.each do |vv|
                    vv.clean
                end
            else
                v.clean if v.respond_to? :clean
            end
        end
        widget.children.each do |ch|
            rec_clean_cbs(ch)
        end
    ;end
  def regen_children();
        self.needsRegen = false

        #//Clear children
        children[1..-1].each do |ch|
            rec_clean_cbs(ch)
            del_list = rec_del_props(ch)
            @db.remove_properties del_list
        end
        self.children = [self.children[0]]

        #//Create new children
        generate_children()

        #//Force layout update
        if(root)
            root.smash_layout
            damage_self :all
        end
    ;end
  def setup_widget(w);
        w.onSetup if w.respond_to?(:onSetup)
        w.children.each do |c|
            setup_widget c
        end
    ;end
  def find(arr, elm);
        return if arr.nil?
        arr.each_with_index do |x, ind|
            if(elm == x)
                return ind
            end
        end
        nil
    ;end
  def generate_children();
        return if children.length > 1

        #//Calculate offsets that every field exists at
        @children_locs = []

        begin
            total = 0
            (0...maxeffects).each do |r|
                @children_locs[r] = total
                if(effects.include?(r) && effects[r] != :none)
                    lu = get_units(effects[r])
                    total += lu
                else
                    total += 1
                end
            end
            total
        end


        #Box Weight array
        w = []


        running = 0
        excl    = 0

        (0...self.nunits).each do |r|
            place = r + self.offset
            elm = find(@children_locs, place)
            if(elm && effects[elm] && effects[elm] != :none)
                uheight = get_units(effects[elm])
                if(r + uheight <= nunits)
                    col = make_child(effects[elm])
                    col.extern = extern + elm.to_s + "/"
                    Qml::add_child(self, col)
                    db.update_values
                    setup_widget(col)
                    w << uheight
                    running += uheight
                    excl     = uheight - 1
                    next
                end
            end
            if(running < nunits && excl < 1)
                col = Qml::DummyBox.new(db)
                Qml::add_child(self, col)
                w << 1
                running += 1
            end
            excl -= 1
        end
        self.shownWeights = w
    ;end
  def make_child(type);
        if(type == :reverb)
            return Qml::ZynReverb.new(db)
        elsif(type == :echo)
            return Qml::ZynEcho.new(db)
        elsif(type == :chorus)
            return Qml::ZynChorus.new(db)
        elsif(type == :distortion)
            return Qml::ZynDistortion.new(db)
        elsif(type == :alienwah)
            return Qml::ZynAlienwah.new(db)
        elsif(type == :phaser)
            return Qml::ZynPhaser.new(db)
        elsif(type == :eq)
            return Qml::ZynEqualizer.new(db)
        elsif(type == :dynamicfilter)
            return Qml::ZynDynFilter.new(db)
        else
            col = Qml::ColorBox.new(db)
            col.bg = color(:red)
            return col
        end
    ;end
  def lookup(type);
        mapper = {0=>:none,
                  1=>:reverb,
                  2=>:echo,
                  3=>:chorus,
                  4=>:phaser,
                  5=>:alienwah,
                  6=>:distortion,
                  7=>:eq,
                  8=>:dynamicfilter}
        mapper[type]
    ;end
  def get_units(type);
        mapper = {:alienwah => 1,
                  :chorus => 1,
                  :distortion => 2,
                  :dynamicfilter => 4,
                  :eq => 2,
                  :echo => 1,
                  :phaser => 2,
                  :reverb => 1}
        mapper[type]
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.98, 0, 0.02, 1)
        padw = 5
        padh = 3
        if(children.length > 1 && self.shownWeights.length > 0)
            off = 0
            children[1..-1].each_with_index do |ch, i|
                lh = shownWeights[i]
                ch.fixed_long(l, selfBox, 0, off/nunits, 0.98,
                  lh/nunits, padw, padh, -2*padw, -2*padh)
                off += lh
            end
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @egrp = self
@anonymous516 = Qml::ScrollBar.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["egrp"] = self
Qml::prop_add(self, "nunits")
Qml::prop_add(self, "offset")
Qml::prop_add(self, "effects")
Qml::prop_add(self, "maxeffects")
Qml::prop_add(self, "shownWeights")
Qml::prop_add(self, "needsRegen")
Qml::add_child(self, @anonymous516)
self.nunits = 6
self.offset = 0
@db.connect_property(self.properties["effects"], lambda {Hash.new})
self.maxeffects = 8
self.shownWeights = []
self.valueRef = nil
self.needsRegen = false

    #puts "ZynEffectGroup, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffects < Qml::Widget
  attr_reader :swap
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.05, 0.95])
    ;end
  def set_view();
        sub = root.get_view_pos(:subview)
        if(![:sysefx, :insefx, :prtefx].include?(sub))
            sub = :sysefx
            root.set_view_pos(:subview, sub)
        end

        if(sub == :sysefx)
            swap.content = Qml::ZynEffectsSystem
        elsif(sub == :insefx)
            swap.content = Qml::ZynEffectsInsert
        elsif(sub == :prtefx)
            prt = root.get_view_pos(:part)
            swap.extern = "/part#{prt}/"
            swap.content = Qml::ZynEffectsPart
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous523 = self
@anonymous521 = Qml::TabGroup.new(database, ui_path+"1"+'/')
@anonymous518 = Qml::TabButton.new(database, ui_path+"2"+'/')
@anonymous519 = Qml::TabButton.new(database, ui_path+"3"+'/')
@anonymous520 = Qml::TabButton.new(database, ui_path+"4"+'/')
@swap = Qml::Swappable.new(database, ui_path+"5"+'/')

    
context = Hash.new
context["swap"] = @swap

        @anonymous521.instance_eval("def set_tab(wid);\n            selected = get_tab wid\n            if(selected == 0)\n                root.set_view_pos(:subview, :sysefx)\n            elsif(selected == 1)\n                root.set_view_pos(:subview, :insefx)\n            elsif(selected == 2)\n                root.set_view_pos(:subview, :prtefx)\n            end\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynEffects.qml", 7)
@anonymous518.value = true
@anonymous518.label = "system"
Qml::add_child(@anonymous521, @anonymous518)
@anonymous519.label = "insertion"
Qml::add_child(@anonymous521, @anonymous519)
@anonymous520.label = "part insertion"
Qml::add_child(@anonymous521, @anonymous520)
Qml::add_child(self, @anonymous521)
@db.connect_property(@swap.properties["content"], lambda {Qml::ZynEffectsSystem})
Qml::add_child(self, @swap)

    #puts "ZynEffects, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectsInsert < Qml::Widget
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.1, 0.9], 0, 3)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous526 = self
@anonymous524 = Qml::ColorBox.new(database, ui_path+"1"+'/')
@anonymous525 = Qml::ZynEffectGroup.new(database, ui_path+"2"+'/')

    
context = Hash.new

        @anonymous524.instance_eval("def onSetup();\n            num_eff = 8\n            (0...num_eff).each do |i|\n                but = Qml::Selector.new(db)\n                but.extern = \"/insefx\#{i}/efftype\"\n                but.active = false\n                sel = Qml::Selector.new(db)\n                sel.whenValue = lambda { but.active = sel.selected != 1; but.damage_self }\n                sel.extern = \"/Pinsparts\#{i}\"\n                Qml::add_child(self, but)\n                Qml::add_child(self, sel)\n            end\n        ;end", "src/mruby-zest/example/ZynEffectsInsert.qml", 5)

        @anonymous524.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vstack(l, selfBox, children, 5)\n        ;end", "src/mruby-zest/example/ZynEffectsInsert.qml", 20)

        @anonymous524.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynEffectsInsert.qml", 23)
@anonymous524.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous524)
@anonymous525.maxeffects = 8
@anonymous525.extern = "/insefx"
Qml::add_child(self, @anonymous525)

    #puts "ZynEffectsInsert, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectsPart < Qml::Widget
  attr_reader :base
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.1, 0.9], 0, 3)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @base = self
@anonymous527 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous528 = Qml::ZynEffectGroup.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["base"] = self
@anonymous527.instance_eval do
def base=(k);@base=k;end
def base;    @base;  end
end
Qml::context_apply(context, @anonymous527)

        @anonymous527.instance_eval("def onSetup();\n            num_eff = 3\n            (0...num_eff).each do |i|\n                #TODO part id\n                sel = Qml::Selector.new(db)\n                sel.extern = base.extern+\"partefx\#{i}/efftype\"\n                but = Qml::ZynEffectBypass.new(db)\n                but.children[1].extern = base.extern+\"Pefxbypass\#{i}\"\n                but.children[0].extern = base.extern+\"Pefxroute\#{i}\"\n                Qml::add_child(self, sel)\n                Qml::add_child(self, but)\n            end\n        ;end", "src/mruby-zest/example/ZynEffectsPart.qml", 4)

        @anonymous527.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vstack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynEffectsPart.qml", 19)

        @anonymous527.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynEffectsPart.qml", 23)
Qml::add_child(self, @anonymous527)
@anonymous528.maxeffects = 3
@db.connect_property(@anonymous528.properties["extern"], lambda {base.extern + "partefx"})
Qml::add_child(self, @anonymous528)

    #puts "ZynEffectsPart, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectsSystem < Qml::Widget
  def class_name(); "eff" ;end
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.1, 0.9], 0, 3)
    ;end
  def set_view();
        prt = root.get_view_pos(:part)
        ch = children[0].children
        n  = ch.length/2
        (0...n).each do |i|
            ch[2*i+1].extern = "/Psysefxvol#{i}/part#{prt}"
        end
    ;end
end
class Qml::ZynSendToGrid < Qml::Widget
  attr_reader :title
  def draw(vg);
        Draw::GradBox(vg, Rect.new(0,0,w,h))
    ;end
  def layout(l, selfBox);
        pad = 4
        children[0].fixed_long(l, selfBox, 0, 0, 1, 1,
        pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1059 = self
@anonymous1058 = Qml::Widget.new(database, ui_path+"1"+'/')
@title = Qml::Widget.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["title"] = @title

        @anonymous1058.instance_eval("def onSetup(old=nil);\n            return if children.length > 1\n            rows = 4\n            (rows).times do |x|\n                ch = Qml::ZynSendToRow.new(db)\n                ch.row_id = x\n                ch.cols   = 3\n                Qml::add_child(self, ch)\n            end\n        ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 6)

        @anonymous1058.instance_eval("def class_name(); \"send_to_grid\" ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 18)

        @anonymous1058.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 20)
Qml::add_child(@anonymous1058, @title)
Qml::add_child(self, @anonymous1058)

    #puts "ZynSendToGrid, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEffectsSystem
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous534 = self
@anonymous530 = Qml::ColorBox.new(database, ui_path+"1"+'/')
@anonymous533 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous531 = Qml::ZynEffectGroup.new(database, ui_path+"3"+'/')
@anonymous532 = Qml::ZynSendToGrid.new(database, ui_path+"4"+'/')

    
context = Hash.new

        @anonymous530.instance_eval("def onSetup(old=nil);\n            return if children.length > 1\n            num_eff = 4\n            prt = root.get_view_pos(:part)\n            (0...num_eff).each do |i|\n                but = Qml::Selector.new(db)\n                but.extern = \"/sysefx\#{i}/efftype\"\n                sld = Qml::HSlider.new(db)\n                sld.extern = \"/Psysefxvol\#{i}/part\#{prt}\"\n                Qml::add_child(self, but)\n                Qml::add_child(self, sld)\n            end\n        ;end", "src/mruby-zest/example/ZynEffectsSystem.qml", 5)

        @anonymous530.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vstack(l, selfBox, self.children)#, 0.05, 0.9, 20)\n        ;end", "src/mruby-zest/example/ZynEffectsSystem.qml", 20)

        @anonymous530.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynEffectsSystem.qml", 23)
@anonymous530.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous530)

        @anonymous533.instance_eval("def class_name(); \"effvert\" ;end", "src/mruby-zest/example/ZynEffectsSystem.qml", 34)

        @anonymous533.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.666, 0.333])\n        ;end", "src/mruby-zest/example/ZynEffectsSystem.qml", 35)
@anonymous531.maxeffects = 4
@anonymous531.nunits = 4
@anonymous531.extern = "/sysefx"
Qml::add_child(@anonymous533, @anonymous531)
Qml::add_child(@anonymous533, @anonymous532)
Qml::add_child(self, @anonymous533)

    #puts "ZynEffectsSystem, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEnvEdit < Qml::Widget
  attr_reader :enveditor
  attr_reader :env
  attr_reader :free
  attr_reader :add
  attr_reader :del
  attr_reader :sus
  attr_reader :total_len
  def add_point();
        return if self.extern.empty?
        return if env.selected == nil
        return if !free.value
        path = self.extern + "addPoint"
        $remote.action(path, env.selected)
        env.refresh
    ;end
  def del_point();
        return if self.extern.empty?
        return if env.selected == nil
        return if !free.value
        path = self.extern + "delPoint"
        $remote.action(path, env.selected)
        env.refresh
    ;end
  def refresh();
        env.refresh()
    ;end
  def update_time();
        t = 0.0;
        env.xpoints.each_with_index do |pt, idx|
            next if idx > env.points
            t += pt
        end

        ll = (t/1000).to_s[0..4] + "  sec" if(t > 1000)
        ll = (t).to_s[0..4]      + " msec" if(t < 1000)
        total_len.label = ll
        total_len.damage_self
    ;end
  def set_free();
        add.active = free.value
        del.active = free.value
        sus.active = free.value
        add.damage_self
        del.damage_self
        sus.damage_self
        refresh
    ;end
  def layout(l, selfBox);
        main_width = 0.9
        main_width = layoutOpts[:main_width] if layoutOpts.include?(:main_width)
        Draw::Layout::hfill(l, selfBox, children,
        [main_width, 1-main_width], 0, 3)
    ;end
  def onSetup(old=nil);
        env.extern()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @enveditor = self
@env = Qml::Envelope.new(database, ui_path+"1"+'/')
@anonymous542 = Qml::Col.new(database, ui_path+"2"+'/')
@free = Qml::ConfirmButton.new(database, ui_path+"3"+'/')
@add = Qml::TriggerButton.new(database, ui_path+"4"+'/')
@del = Qml::TriggerButton.new(database, ui_path+"5"+'/')
@anonymous539 = Qml::Text.new(database, ui_path+"6"+'/')
@sus = Qml::NumEntry.new(database, ui_path+"7"+'/')
@total_len = Qml::Text.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["enveditor"] = self
context["env"] = @env
context["free"] = @free
context["add"] = @add
context["del"] = @del
context["sus"] = @sus
context["total_len"] = @total_len
@db.connect_property(@env.properties["extern"], lambda {enveditor.extern})
@db.connect_property(@env.properties["whenTime"], lambda {lambda { enveditor.update_time }})
Qml::add_child(self, @env)
@free.whenValue = lambda {enveditor.set_free()}
@db.connect_property(@free.properties["extern"], lambda {enveditor.extern + "Pfreemode"})
@free.label = "free"
Qml::add_child(@anonymous542, @free)
@add.active = false
@add.whenValue = lambda {enveditor.add_point()}
@add.label = "add"
Qml::add_child(@anonymous542, @add)
@del.active = false
@del.whenValue = lambda {enveditor.del_point()}
@del.label = "delete"
Qml::add_child(@anonymous542, @del)
@anonymous539.height = 0.9
@anonymous539.label = "sustain pt"
Qml::add_child(@anonymous542, @anonymous539)
@sus.active = false
@db.connect_property(@sus.properties["extern"], lambda {enveditor.extern + "Penvsustain"})
Qml::add_child(@anonymous542, @sus)
@total_len.height = 0.9
@db.connect_property(@total_len.properties["label"], lambda {"1.47 sec"})
Qml::add_child(@anonymous542, @total_len)
@anonymous542.spacer = 8
Qml::add_child(self, @anonymous542)

    #puts "ZynEnvEdit, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEqFilter < Qml::Widget
  attr_reader :fil
  attr_reader :selector
  attr_reader :gain
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  def onSetup(old=nil);
        children[0].children.each do |ch|
            ch.whenValue = lambda {fil.cb} if(ch.respond_to?(:whenValue) && ch.whenValue.nil?)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @fil = self
@anonymous549 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@selector = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous545 = Qml::Knob.new(database, ui_path+"3"+'/')
@gain = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous547 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous548 = Qml::NumEntry.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["fil"] = self
context["selector"] = @selector
context["gain"] = @gain
@db.connect_property(@selector.properties["extern"], lambda {fil.extern + "Ptype"})
@selector.whenValue = lambda() {
                fil.gain.active = (selector.selected >= 7)
                fil.gain.damage_self
                fil.gain.refresh()
                fil.cb # parent callback
            }
Qml::add_child(@anonymous549, @selector)
@db.connect_property(@anonymous545.properties["extern"], lambda {fil.extern + "Pfreq"})
@anonymous545.whenValue = lambda { fil.cb }
Qml::add_child(@anonymous549, @anonymous545)
@db.connect_property(@gain.properties["extern"], lambda {fil.extern + "Pgain"})
@gain.whenValue = lambda { fil.cb }
Qml::add_child(@anonymous549, @gain)
@db.connect_property(@anonymous547.properties["extern"], lambda {fil.extern + "Pq"})
@anonymous547.whenValue = lambda { fil.cb }
Qml::add_child(@anonymous549, @anonymous547)
@db.connect_property(@anonymous548.properties["extern"], lambda {fil.extern + "Pstages"})
@anonymous548.whenValue = lambda { fil.cb }
Qml::add_child(@anonymous549, @anonymous548)
Qml::add_child(self, @anonymous549)
self.whenValue = nil

    #puts "ZynEqFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynEqualizer < Qml::Widget
  attr_reader :eq
  attr_reader :vis
  attr_reader :fil_sel
  attr_reader :filter
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.15, 0.85])
    ;end
  def change_filter();
        filter.extern  = eq.extern + "EQ/filter#{fil_sel.value-1}/"
        filter.content = Qml::ZynEqFilter
        filter.children[0].whenValue = lambda { vis.refresh }
    ;end
  def onSetup(old=nil);
        change_filter
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @eq = self
@anonymous551 = Qml::GroupHeader.new(database, ui_path+"1"+'/')
@anonymous558 = Qml::Widget.new(database, ui_path+"2"+'/')
@vis = Qml::VisEq.new(database, ui_path+"3"+'/')
@anonymous557 = Qml::Widget.new(database, ui_path+"4"+'/')
@anonymous555 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous553 = Qml::Knob.new(database, ui_path+"6"+'/')
@fil_sel = Qml::NumEntry.new(database, ui_path+"7"+'/')
@filter = Qml::Swappable.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["eq"] = self
context["vis"] = @vis
context["fil_sel"] = @fil_sel
context["filter"] = @filter
@anonymous551.label = "Equalizer"
@db.connect_property(@anonymous551.properties["extern"], lambda {eq.extern})
@anonymous551.copyable = true
Qml::add_child(self, @anonymous551)

        @anonymous558.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0, 0, w, h))\n        ;end", "src/mruby-zest/example/ZynEqualizer.qml", 37)

        @anonymous558.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynEqualizer.qml", 41)
@db.connect_property(@vis.properties["extern"], lambda {eq.extern + "EQ/coeff"})
Qml::add_child(@anonymous558, @vis)

        @anonymous557.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children)\n            ;end", "src/mruby-zest/example/ZynEqualizer.qml", 32)
@db.connect_property(@anonymous553.properties["extern"], lambda {eq.extern + "Pvolume"})
@anonymous553.label = "vol"
Qml::add_child(@anonymous555, @anonymous553)
@fil_sel.whenValue = lambda {eq.change_filter}
@fil_sel.label = "filter id"
@fil_sel.minimum = 1
@fil_sel.maximum = 8
@fil_sel.offset = 1
@fil_sel.value = 1
Qml::add_child(@anonymous555, @fil_sel)
Qml::add_child(@anonymous557, @anonymous555)
Qml::add_child(@anonymous557, @filter)
Qml::add_child(@anonymous558, @anonymous557)
Qml::add_child(self, @anonymous558)

    #puts "ZynEqualizer, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFilterEnv < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  def cb();
        whenModified.call if whenModified
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous560 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous561 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous562 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous563 = Qml::ToggleButton.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous565 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous566 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous567 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous568 = Qml::Knob.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@anonymous560.whenValue = lambda { box.cb }
@db.connect_property(@anonymous560.properties["extern"], lambda {box.extern+"PA_dt" })
Qml::add_child(@top, @anonymous560)
@anonymous561.whenValue = lambda { box.cb }
@db.connect_property(@anonymous561.properties["extern"], lambda {box.extern+"PD_dt"})
Qml::add_child(@top, @anonymous561)
@anonymous562.whenValue = lambda { box.cb }
@db.connect_property(@anonymous562.properties["extern"], lambda {box.extern+"PR_dt"})
Qml::add_child(@top, @anonymous562)
@anonymous563.label = "FRCR"
@anonymous563.whenValue = lambda { box.cb }
@db.connect_property(@anonymous563.properties["extern"], lambda {box.extern+"Pforcedrelease"})
Qml::add_child(@top, @anonymous563)
Qml::add_child(self, @top)
@anonymous565.whenValue = lambda { box.cb }
@db.connect_property(@anonymous565.properties["extern"], lambda {box.extern+"PA_val"})
Qml::add_child(@bot, @anonymous565)
@anonymous566.whenValue = lambda { box.cb }
@db.connect_property(@anonymous566.properties["extern"], lambda {box.extern+"PD_val"})
Qml::add_child(@bot, @anonymous566)
@anonymous567.whenValue = lambda { box.cb }
@db.connect_property(@anonymous567.properties["extern"], lambda {box.extern+"PR_val"})
Qml::add_child(@bot, @anonymous567)
@anonymous568.whenValue = lambda { box.cb }
@db.connect_property(@anonymous568.properties["extern"], lambda {box.extern+"Penvstretch"})
Qml::add_child(@bot, @anonymous568)
Qml::add_child(self, @bot)
self.label = "Envelope"
self.whenModified = nil
self.extern = ""

    #puts "ZynFilterEnv, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFooter < Qml::Widget
  attr_reader :foot
  attr_reader :wheelwell
  attr_reader :wheel
  attr_reader :key
  attr_reader :ctrl
  attr_reader :vel
  attr_reader :rnd
  attr_reader :oct
  attr_reader :cc
  attr_reader :cctype
  def class_name();"ZynFooter";end
  def set_cc(x);
        cc_num =  cctype.options[cctype.selected].to_i
        $remote.action("/setController", 0, cc_num, x)

    ;end
  def layout(l, selfBox);
        wheelwell.fixed(l,  selfBox, 0.010, 0.15, 0.017, (1-2*0.15))
        key.fixed(l,        selfBox, 0.032, 0.10, 0.536, 0.8)
        ctrl.fixed(l,       selfBox, 0.600, 0.05, 0.400, 0.9)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @foot = self
@wheelwell = Qml::Indent.new(database, ui_path+"1"+'/')
@wheel = Qml::ModWheel.new(database, ui_path+"2"+'/')
@key = Qml::Keyboard.new(database, ui_path+"3"+'/')
@ctrl = Qml::ParModuleRow.new(database, ui_path+"4"+'/')
@vel = Qml::Knob.new(database, ui_path+"5"+'/')
@rnd = Qml::Knob.new(database, ui_path+"6"+'/')
@oct = Qml::Knob.new(database, ui_path+"7"+'/')
@cc = Qml::Knob.new(database, ui_path+"8"+'/')
@cctype = Qml::Selector.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["foot"] = self
context["wheelwell"] = @wheelwell
@wheelwell.instance_eval do
def wheel=(k);@wheel=k;end
def wheel;    @wheel;  end
end
context["wheel"] = @wheel
context["key"] = @key
context["ctrl"] = @ctrl
context["vel"] = @vel
context["rnd"] = @rnd
context["oct"] = @oct
context["cc"] = @cc
context["cctype"] = @cctype
Qml::context_apply(context, @wheelwell)

        @wheelwell.instance_eval("def layout(l, selfBox);\n            wheel.fixed(l, selfBox, 0.1, 0.1, 0.8, 0.8)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynFooter.qml", 17)
@wheel.value = 0
@wheel.label = "wheel"
@wheel.tooltip = "modulation wheel"
@wheel.whenValue = lambda {
                $remote.action("/setController", 0, 1, (wheelwell.children[0].value*127).to_i)
            }
Qml::add_child(@wheelwell, @wheel)
Qml::add_child(self, @wheelwell)
@db.connect_property(@key.properties["velocity"], lambda {0.7*127})
@db.connect_property(@key.properties["velrnd"], lambda {0.2*127})
Qml::add_child(self, @key)
@vel.whenValue = lambda { key.velocity = vel.value*127 }
@vel.value = 0.7
@vel.label = "velocity"
@vel.tooltip = "velocity of virtual keyboard notes"
Qml::add_child(@ctrl, @vel)
@rnd.whenValue = lambda { key.velrnd = rnd.value*127 }
@rnd.value = 0.2
@rnd.label = "vrnd"
@rnd.tooltip = "velocity randomness of virtual keyboard notes"
Qml::add_child(@ctrl, @rnd)
@oct.value = 0.5
@oct.label = "octave"
@oct.whenValue = lambda {key.octave = (4*(oct.value-0.5)).to_i}
Qml::add_child(@ctrl, @oct)
@db.connect_property(@cc.properties["label"], lambda {"c.val"})
@cc.whenValue = lambda { foot.set_cc((cc.value*127).to_i) }
@cc.tooltip = "midi CC control"
Qml::add_child(@ctrl, @cc)
@cctype.label = "MIDI CC"
@cctype.tooltip = "midi CC selector"
@cctype.layoutOpts = {:weight=>2.0, :long_mode=>true}
@cctype.options = [ "01: Mod.Wheel",
                       "07: Volume",
                       "10: Panning",
                       "11: Expression",
                       "64: Sustain",
                       "65: Portamento",
                       "71: Filter Q",
                       "74: Filter Freq.",
                       "75: Bandwidth",
                       "76: FM Gain",
                       "77: Res. c. freq",
                       "78: Res. bw."]
Qml::add_child(@ctrl, @cctype)
@ctrl.layoutOpts = []
Qml::add_child(self, @ctrl)

    #puts "ZynFooter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFormant < Qml::Widget
  attr_reader :fbase
  attr_reader :vis
  attr_reader :total
  attr_reader :vgrp
  attr_reader :vow_num
  attr_reader :slide_freq
  attr_reader :for_num
  attr_reader :slide_q
  attr_reader :slide_amp
  attr_reader :pos_id
  attr_reader :pos_vowel
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.7, 0.3], 0, 2)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @fbase = self
@vis = Qml::VisFormant.new(database, ui_path+"1"+'/')
@anonymous608 = Qml::Widget.new(database, ui_path+"2"+'/')
@total = Qml::Widget.new(database, ui_path+"3"+'/')
@anonymous582 = Qml::Text.new(database, ui_path+"4"+'/')
@anonymous583 = Qml::NumEntry.new(database, ui_path+"5"+'/')
@anonymous584 = Qml::HSlider.new(database, ui_path+"6"+'/')
@anonymous585 = Qml::HSlider.new(database, ui_path+"7"+'/')
@anonymous586 = Qml::HSlider.new(database, ui_path+"8"+'/')
@anonymous587 = Qml::HSlider.new(database, ui_path+"9"+'/')
@anonymous607 = Qml::Widget.new(database, ui_path+"10"+'/')
@vgrp = Qml::Group.new(database, ui_path+"11"+'/')
@anonymous597 = Qml::Widget.new(database, ui_path+"12"+'/')
@vow_num = Qml::NumEntry.new(database, ui_path+"13"+'/')
@slide_freq = Qml::HSlider.new(database, ui_path+"14"+'/')
@for_num = Qml::NumEntry.new(database, ui_path+"15"+'/')
@slide_q = Qml::HSlider.new(database, ui_path+"16"+'/')
@anonymous593 = Qml::Widget.new(database, ui_path+"17"+'/')
@anonymous594 = Qml::Widget.new(database, ui_path+"18"+'/')
@anonymous595 = Qml::Widget.new(database, ui_path+"19"+'/')
@slide_amp = Qml::HSlider.new(database, ui_path+"20"+'/')
@anonymous606 = Qml::Group.new(database, ui_path+"21"+'/')
@anonymous605 = Qml::Widget.new(database, ui_path+"22"+'/')
@anonymous599 = Qml::NumEntry.new(database, ui_path+"23"+'/')
@pos_id = Qml::NumEntry.new(database, ui_path+"24"+'/')
@anonymous601 = Qml::HSlider.new(database, ui_path+"25"+'/')
@anonymous602 = Qml::Widget.new(database, ui_path+"26"+'/')
@pos_vowel = Qml::NumEntry.new(database, ui_path+"27"+'/')
@anonymous604 = Qml::Button.new(database, ui_path+"28"+'/')

    
context = Hash.new
context["fbase"] = self
context["vis"] = @vis
context["total"] = @total
@total.instance_eval do
def vis=(k);@vis=k;end
def vis;    @vis;  end
end
context["vgrp"] = @vgrp
@vgrp.instance_eval do
def fbase=(k);@fbase=k;end
def fbase;    @fbase;  end
def vis=(k);@vis=k;end
def vis;    @vis;  end
def vow_num=(k);@vow_num=k;end
def vow_num;    @vow_num;  end
def slide_freq=(k);@slide_freq=k;end
def slide_freq;    @slide_freq;  end
def for_num=(k);@for_num=k;end
def for_num;    @for_num;  end
def slide_q=(k);@slide_q=k;end
def slide_q;    @slide_q;  end
def slide_amp=(k);@slide_amp=k;end
def slide_amp;    @slide_amp;  end
end
context["vow_num"] = @vow_num
context["slide_freq"] = @slide_freq
context["for_num"] = @for_num
context["slide_q"] = @slide_q
context["slide_amp"] = @slide_amp
context["pos_id"] = @pos_id
context["pos_vowel"] = @pos_vowel
Qml::context_apply(context, @total)
Qml::context_apply(context, @vgrp)
@db.connect_property(@vis.properties["extern"], lambda {fbase.extern})
Qml::add_child(self, @vis)

        @anonymous608.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 2)\n        ;end", "src/mruby-zest/example/ZynFormant.qml", 150)

        @total.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children, 0.1, 0.8, 12)\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 20)

        @total.instance_eval("def onSetup(old=nil);\n                lm = lambda { vis.refresh() }\n                children.each do |ch|\n                    ch.whenValue = lm if ch.respond_to? :whenValue\n                end\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 23)

        @total.instance_eval("def draw(vg);\n                Draw::GradBox(vg, Rect.new(0,0,w,h))\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 29)
@anonymous582.label = "formants"
Qml::add_child(@total, @anonymous582)
@anonymous583.format = "PKs "
@db.connect_property(@anonymous583.properties["extern"], lambda {fbase.extern + "Pnumformants"})
Qml::add_child(@total, @anonymous583)
@db.connect_property(@anonymous584.properties["extern"], lambda {fbase.extern + "Pformantslowness"})
Qml::add_child(@total, @anonymous584)
@db.connect_property(@anonymous585.properties["extern"], lambda {fbase.extern + "Pvowelclearness"})
Qml::add_child(@total, @anonymous585)
@db.connect_property(@anonymous586.properties["extern"], lambda {fbase.extern + "Pcenterfreq" })
Qml::add_child(@total, @anonymous586)
@db.connect_property(@anonymous587.properties["extern"], lambda {fbase.extern + "Poctavesfreq" })
Qml::add_child(@total, @anonymous587)
Qml::add_child(@anonymous608, @total)

        @anonymous607.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children)\n            ;end", "src/mruby-zest/example/ZynFormant.qml", 146)

        @vgrp.instance_eval("def update_pos();\n                    vis.vowel_num = vow_num.value-1\n                    vis.change()\n                    rt = fbase.extern + \"Pvowels\#{vow_num.value-1}/Pformants\#{for_num.value-1}/\"\n                    slide_freq.extern = rt + \"freq\"\n                    slide_q.extern    = rt + \"q\"\n                    slide_amp.extern  = rt + \"amp\"\n                    db.update_values\n                ;end", "src/mruby-zest/example/ZynFormant.qml", 39)

        @anonymous597.instance_eval("def layout(l, selfBox);\n                        Draw::Layout::grid(l, selfBox, children, 4, 2, 2, 2)\n                    ;end", "src/mruby-zest/example/ZynFormant.qml", 91)
@db.connect_property(@vow_num.properties["label"], lambda {"v. num"})
@vow_num.format = "VW "
@vow_num.tooltip = "vowel number - each formant filter is composed of one or more vowels"
@vow_num.minimum = 1
@vow_num.maximum = 6
@vow_num.value = 1
@vow_num.whenValue = lambda {vgrp.update_pos}
Qml::add_child(@anonymous597, @vow_num)
@db.connect_property(@slide_freq.properties["extern"], lambda {vgrp.extern + "freq"})
@slide_freq.label = "freq"
@slide_freq.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_freq)
@for_num.label = "formant"
@for_num.format = "PK "
@for_num.tooltip = "formant peak - each formant vowel is composed of one or more peaks"
@for_num.minimum = 1
@for_num.maximum = 12
@for_num.value = 1
@for_num.whenValue = lambda {vgrp.update_pos}
Qml::add_child(@anonymous597, @for_num)
@db.connect_property(@slide_q.properties["extern"], lambda {vgrp.extern + "q"})
@slide_q.label = "q"
@slide_q.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_q)
Qml::add_child(@anonymous597, @anonymous593)
Qml::add_child(@anonymous597, @anonymous594)
Qml::add_child(@anonymous597, @anonymous595)
@db.connect_property(@slide_amp.properties["extern"], lambda {vgrp.extern + "amp"})
@slide_amp.label = "amp"
@slide_amp.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous597, @slide_amp)
Qml::add_child(@vgrp, @anonymous597)
@vgrp.label = "vowel"
@vgrp.topSize = 0.2
@db.connect_property(@vgrp.properties["extern"], lambda {fbase.extern + "Pvowels0/Pformants0/"})
Qml::add_child(@anonymous607, @vgrp)

        @anonymous605.instance_eval("def layout(l, selfBox);\n                        Draw::Layout::grid(l, selfBox, children, 3, 2, 2, 4)\n                    ;end", "src/mruby-zest/example/ZynFormant.qml", 141)
@db.connect_property(@anonymous599.properties["extern"], lambda {fbase.extern + "Psequencesize"})
@anonymous599.format = "LEN "
@anonymous599.label = "seqsize"
@anonymous599.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous605, @anonymous599)
@pos_id.format = "POS "
@pos_id.value = 1
@pos_id.minimum = 1
@pos_id.maximum = 8
@db.connect_property(@pos_id.properties["label"], lambda {"s.pos"})
@pos_id.whenValue = lambda {
                            pos_vowel.extern = 
                            fbase.extern + "vowel_seq#{pos_id.value-1}"
                        }
Qml::add_child(@anonymous605, @pos_id)
@db.connect_property(@anonymous601.properties["extern"], lambda {fbase.extern + "Psequencestretch"})
@anonymous601.label = "strch"
@anonymous601.whenValue = lambda { vis.refresh() }
Qml::add_child(@anonymous605, @anonymous601)
Qml::add_child(@anonymous605, @anonymous602)
@pos_vowel.format = "VW "
@pos_vowel.label = "vowel"
@pos_vowel.offset = 1
@pos_vowel.minimum = 1
@pos_vowel.maximum = 8
@db.connect_property(@pos_vowel.properties["extern"], lambda {fbase.extern + "vowel_seq0"})
Qml::add_child(@anonymous605, @pos_vowel)
@db.connect_property(@anonymous604.properties["extern"], lambda {fbase.extern + "Psequencereversed"})
@db.connect_property(@anonymous604.properties["label"], lambda {"neg. input"})
@anonymous604.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous605, @anonymous604)
Qml::add_child(@anonymous606, @anonymous605)
@anonymous606.label = "sequence"
@anonymous606.topSize = 0.2
@anonymous606.copyable = false
Qml::add_child(@anonymous607, @anonymous606)
Qml::add_child(@anonymous608, @anonymous607)
Qml::add_child(self, @anonymous608)

    #puts "ZynFormant, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFreqEnv < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  def cb();
        whenModified.call if whenModified
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous610 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous611 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous612 = Qml::ToggleButton.new(database, ui_path+"4"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous614 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous615 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous616 = Qml::Knob.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@anonymous610.whenValue = lambda { box.cb }
@db.connect_property(@anonymous610.properties["extern"], lambda {box.extern+"PA_dt" })
Qml::add_child(@top, @anonymous610)
@anonymous611.whenValue = lambda { box.cb }
@db.connect_property(@anonymous611.properties["extern"], lambda {box.extern+"PR_dt" })
Qml::add_child(@top, @anonymous611)
@anonymous612.label = "FRCR"
@anonymous612.whenValue = lambda { box.cb }
@db.connect_property(@anonymous612.properties["extern"], lambda {box.extern+"Pforcedrelease"})
Qml::add_child(@top, @anonymous612)
Qml::add_child(self, @top)
@anonymous614.whenValue = lambda { box.cb }
@db.connect_property(@anonymous614.properties["extern"], lambda {box.extern+"PA_val"})
Qml::add_child(@bot, @anonymous614)
@anonymous615.whenValue = lambda { box.cb }
@db.connect_property(@anonymous615.properties["extern"], lambda {box.extern+"PR_val"})
Qml::add_child(@bot, @anonymous615)
@anonymous616.whenValue = lambda { box.cb }
@db.connect_property(@anonymous616.properties["extern"], lambda {box.extern+"Penvstretch" })
Qml::add_child(@bot, @anonymous616)
Qml::add_child(self, @bot)
self.label = "Envelope"
self.whenModified = nil
self.extern = ""

    #puts "ZynFreqEnv, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFreqGeneral < Qml::Group
  attr_reader :box
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous622 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous619 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous620 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous621 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous625 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous623 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous624 = Qml::Knob.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous619.properties["extern"], lambda {box.extern + "PDetune" })
Qml::add_child(@anonymous622, @anonymous619)
@db.connect_property(@anonymous620.properties["extern"], lambda {box.extern + "octave" })
Qml::add_child(@anonymous622, @anonymous620)
@db.connect_property(@anonymous621.properties["extern"], lambda {box.extern + "PBandwidth" })
Qml::add_child(@anonymous622, @anonymous621)
Qml::add_child(self, @anonymous622)
@db.connect_property(@anonymous623.properties["extern"], lambda {box.extern + "PDetuneType" })
Qml::add_child(@anonymous625, @anonymous623)
@db.connect_property(@anonymous624.properties["extern"], lambda {box.extern + "PCoarseDetune" })
Qml::add_child(@anonymous625, @anonymous624)
Qml::add_child(self, @anonymous625)
self.label = "General"
self.whenModified = nil

    #puts "ZynFreqGeneral, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFreqGeneralVoice < Qml::Group
  attr_reader :box
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous630 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous627 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous628 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous629 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous634 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous631 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous632 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous633 = Qml::ToggleButton.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous627.properties["extern"], lambda {box.extern + "PDetune" })
Qml::add_child(@anonymous630, @anonymous627)
@db.connect_property(@anonymous628.properties["extern"], lambda {box.extern + "octave" })
Qml::add_child(@anonymous630, @anonymous628)
@db.connect_property(@anonymous629.properties["extern"], lambda {box.extern + "PfixedfreqET" })
Qml::add_child(@anonymous630, @anonymous629)
Qml::add_child(self, @anonymous630)
@db.connect_property(@anonymous631.properties["extern"], lambda {box.extern + "PDetuneType" })
Qml::add_child(@anonymous634, @anonymous631)
@db.connect_property(@anonymous632.properties["extern"], lambda {box.extern + "coarsedetune" })
Qml::add_child(@anonymous634, @anonymous632)
@db.connect_property(@anonymous633.properties["extern"], lambda {box.extern + "Pfixedfreq" })
Qml::add_child(@anonymous634, @anonymous633)
Qml::add_child(self, @anonymous634)
self.label = "General"
self.whenModified = nil

    #puts "ZynFreqGeneralVoice, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynFreqMod < Qml::Group
  attr_reader :box
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous639 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous636 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous637 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous638 = Qml::ToggleButton.new(database, ui_path+"4"+'/')
@anonymous642 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous640 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous641 = Qml::Knob.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["box"] = self
@db.connect_property(@anonymous636.properties["extern"], lambda {box.extern + "PFMDetune" })
Qml::add_child(@anonymous639, @anonymous636)
@db.connect_property(@anonymous637.properties["extern"], lambda {box.extern + "FMoctave" })
Qml::add_child(@anonymous639, @anonymous637)
@db.connect_property(@anonymous638.properties["extern"], lambda {box.extern + "PFMFixedFreq" })
Qml::add_child(@anonymous639, @anonymous638)
Qml::add_child(self, @anonymous639)
@db.connect_property(@anonymous640.properties["extern"], lambda {box.extern + "PFMDetuneType" })
Qml::add_child(@anonymous642, @anonymous640)
@db.connect_property(@anonymous641.properties["extern"], lambda {box.extern + "PFMCoarseDetune" })
Qml::add_child(@anonymous642, @anonymous641)
Qml::add_child(self, @anonymous642)
self.label = "General"

    #puts "ZynFreqMod, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynHeader < Qml::Widget
  attr_reader :head
  attr_reader :logo
  attr_reader :menuIndent
  attr_reader :status
  attr_reader :shortcuts
  attr_reader :vumeter
  attr_reader :panic
  attr_reader :close
  attr_reader :minimize
  def class_name();"ZynHeader";end
  def layout(l, selfBox);
        logo.fixed(l,       selfBox, 0.016, 0.22,  0.070, 0.56)
        menuIndent.fixed(l, selfBox, 0.100, 0.15, 0.112, 0.7)
        status.fixed(l,     selfBox, 0.220, 0.15, 0.520, 0.7)
        shortcuts.fixed(l,  selfBox, 0.745, 0.15, 0.160, 0.7)
        vumeter.fixed(l,    selfBox, 0.910, 0.15, 0.015, 0.7)
        panic.fixed(l,      selfBox, 0.930, 0.15, 0.040, 0.7)
        close.fixed(l,      selfBox, 0.975, 0.15, 0.015, 0.3)
        minimize.fixed(l,   selfBox, 0.975, 0.54, 0.015, 0.3)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @head = self
@logo = Qml::ZynLogo.new(database, ui_path+"1"+'/')
@menuIndent = Qml::MainMenu.new(database, ui_path+"2"+'/')
@status = Qml::Indent.new(database, ui_path+"3"+'/')
@anonymous646 = Qml::LogWidget.new(database, ui_path+"4"+'/')
@shortcuts = Qml::ZynHeaderParams.new(database, ui_path+"5"+'/')
@vumeter = Qml::VuMeter.new(database, ui_path+"6"+'/')
@panic = Qml::PanicButton.new(database, ui_path+"7"+'/')
@close = Qml::Widget.new(database, ui_path+"8"+'/')
@minimize = Qml::Widget.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["head"] = self
context["logo"] = @logo
context["menuIndent"] = @menuIndent
context["status"] = @status
context["shortcuts"] = @shortcuts
context["vumeter"] = @vumeter
context["panic"] = @panic
context["close"] = @close
context["minimize"] = @minimize
Qml::add_child(self, @logo)
Qml::add_child(self, @menuIndent)
@anonymous646.label = "console log"
Qml::add_child(@status, @anonymous646)
@status.pad = 0.01
Qml::add_child(self, @status)
Qml::add_child(self, @shortcuts)
Qml::add_child(self, @vumeter)
@panic.layoutOpts = [:no_constraint]
Qml::add_child(self, @panic)
Qml::add_child(self, @close)
Qml::add_child(self, @minimize)

    #puts "ZynHeader, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynHeaderParams < Qml::Widget
  attr_reader :fn
  def layout(l, selfBox);
        padw  = 1/256
        padw2 = 0.5-2*padw
        padh = 1/32
        padh2 = 0.5-2*padh
        children[0].fixed(l, selfBox, 0.0+padw, 0.0+padh, padw2, padh2)
        children[1].fixed(l, selfBox, 0.5+padw, 0.0+padh, padw2, padh2)
        children[2].fixed(l, selfBox, 0.0+padw, 0.5+padh, padw2, padh2)
        children[3].fixed(l, selfBox, 0.5+padw, 0.5+padh, padw2, padh2)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous658 = self
@fn = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous655 = Qml::HSlider.new(database, ui_path+"2"+'/')
@anonymous656 = Qml::Button.new(database, ui_path+"3"+'/')
@anonymous657 = Qml::HSlider.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["fn"] = @fn
@fn.instance_eval do
def fn=(k);@fn=k;end
def fn;    @fn;  end
end
Qml::context_apply(context, @fn)

        @fn.instance_eval("def animate();\n            if(fn.value != root.fine_mode)\n                fn.value = root.fine_mode\n                fn.damage_self\n            end\n        ;end", "src/mruby-zest/example/ZynHeaderParams.qml", 7)
@fn.label = "fine"
@fn.layoutOpts = [:no_constraint]
Qml::add_child(self, @fn)
@anonymous655.extern = "/Pvolume"
@anonymous655.height = 0.8
@anonymous655.label = "volume"
Qml::add_child(self, @anonymous655)
@anonymous656.layoutOpts = [:no_constraint]
@anonymous656.label = "nrpn"
Qml::add_child(self, @anonymous656)
@anonymous657.extern = "/Pkeyshift"
@anonymous657.height = 0.8
@anonymous657.label = "key shift"
Qml::add_child(self, @anonymous657)

    #puts "ZynHeaderParams, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynInstrumentSettings < Qml::Group
  attr_reader :insset
  attr_reader :minkey
  attr_reader :maxkey
end
class Qml::ZynKitKeyButton < Qml::Widget
  attr_accessor :whenValue
  def draw(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            v.move_to(w*1/3, 0)
            v.line_to(w*1/3, h)
            v.move_to(w*2/3, 0)
            v.line_to(w*2/3, h)
            v.stroke
        end

        text_color = Theme::TextColor
        if self.layoutOpts.include? :aspect
            vg.font_size h*0.5 
        else
            vg.font_size h*0.8 
        end
        vg.fill_color text_color
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(1/6*w,h/2,"Mn")
        end
        
        #Pause Icon
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(3/6*w,h/2,"R")
        end
        
        #Play Icon
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(5/6*w,h/2,"Mx")
        end
    ;end
  def onMouseHover(ev);
        rel = (ev.pos.x-global_x)/w
        if(rel < 1/3)
            self.root.log(:tooltip, 
            "Capture minimum key")
        elsif(rel < 2/3)
            self.root.log(:tooltip, 
            "Reset key range")
        else
            self.root.log(:tooltip, 
            "Capture maximum key")
        end
    ;end
  def onMousePress(ev);
        rel = (ev.pos.x-global_x)/w
        if(rel < 1/3)
            $remote.action(extern + "captureMin")
        elsif(rel < 2/3)
            $remote.action(extern + "Pminkey", 0)
            $remote.action(extern + "Pmaxkey", 127)
        else
            $remote.action(extern + "captureMax")
        end
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        if(self.layoutOpts.include? :aspect)
            l.aspect(selfBox, 3.0, 1.0)
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous682 = self

    
context = Hash.new
self.whenValue = nil

    #puts "ZynKitKeyButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynInstrumentSettings
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @insset = self
@anonymous673 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous663 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous659 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous660 = Qml::Knob.new(database, ui_path+"4"+'/')
@minkey = Qml::Knob.new(database, ui_path+"5"+'/')
@maxkey = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous668 = Qml::ParModuleRow.new(database, ui_path+"7"+'/')
@anonymous664 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous665 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous666 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous667 = Qml::ZynKitKeyButton.new(database, ui_path+"11"+'/')
@anonymous672 = Qml::ParModuleRow.new(database, ui_path+"12"+'/')
@anonymous669 = Qml::Selector.new(database, ui_path+"13"+'/')
@anonymous670 = Qml::ToggleButton.new(database, ui_path+"14"+'/')
@anonymous671 = Qml::Selector.new(database, ui_path+"15"+'/')

    
context = Hash.new
context["insset"] = self
context["minkey"] = @minkey
context["maxkey"] = @maxkey

        @anonymous673.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.40, 0.40,0.2],0,2)\n        ;end", "src/mruby-zest/example/ZynInstrumentSettings.qml", 51)
@db.connect_property(@anonymous659.properties["extern"], lambda {insset.extern + "Pvolume"})
Qml::add_child(@anonymous663, @anonymous659)
@db.connect_property(@anonymous660.properties["extern"], lambda {insset.extern + "Ppanning"})
Qml::add_child(@anonymous663, @anonymous660)
@db.connect_property(@minkey.properties["extern"], lambda {insset.extern + "Pminkey"})
Qml::add_child(@anonymous663, @minkey)
@db.connect_property(@maxkey.properties["extern"], lambda {insset.extern + "Pmaxkey"})
Qml::add_child(@anonymous663, @maxkey)
@anonymous663.whitespace = 2
Qml::add_child(@anonymous673, @anonymous663)
@db.connect_property(@anonymous664.properties["extern"], lambda {insset.extern + "Pvelsns"})
Qml::add_child(@anonymous668, @anonymous664)
@db.connect_property(@anonymous665.properties["extern"], lambda {insset.extern + "Pveloffs"})
Qml::add_child(@anonymous668, @anonymous665)
@db.connect_property(@anonymous666.properties["extern"], lambda {insset.extern + "Pkeyshift"})
Qml::add_child(@anonymous668, @anonymous666)
@anonymous667.layoutOpts = [:aspect]
@db.connect_property(@anonymous667.properties["extern"], lambda {insset.extern})
@anonymous667.whenValue = lambda {
                    minkey.refresh
                    maxkey.refresh
                }
Qml::add_child(@anonymous668, @anonymous667)
@anonymous668.whitespace = 3
@anonymous668.outer = :none
@anonymous668.layoutOpts = []
Qml::add_child(@anonymous673, @anonymous668)
@db.connect_property(@anonymous669.properties["extern"], lambda {insset.extern + "Prcvchn"})
@anonymous669.label = "midi chan"
Qml::add_child(@anonymous672, @anonymous669)
@db.connect_property(@anonymous670.properties["extern"], lambda {insset.extern + "ctl/portamento.portamento"})
@anonymous670.label = "portamento"
Qml::add_child(@anonymous672, @anonymous670)
@db.connect_property(@anonymous671.properties["extern"], lambda {insset.extern + "polyType"})
@anonymous671.label = "mode"
@anonymous671.layoutOpts = [:no_constraint, :long_mode]
Qml::add_child(@anonymous672, @anonymous671)
@anonymous672.lsize = 0.0
Qml::add_child(@anonymous673, @anonymous672)
Qml::add_child(self, @anonymous673)
self.label = "instrument settings"
self.copyable = false

    #puts "ZynInstrumentSettings, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynKit < Qml::Widget
  attr_reader :kit
  attr_reader :drum
  attr_accessor :valueRef
  def get_mode(x);
        (0..2).each do |i|
            children[0].children[i].value = i == x
            children[0].children[i].damage_self
        end
        any_kit = (x != 0)
        if(any_kit)
            children[1].set_active_kit()
        else
            children[1].set_non_kit()
        end
    ;end
  def set_mode(x);
        kit.valueRef.value = x
        get_mode(x)
    ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.05,0.95], 1, 2)
    ;end
end
class Qml::ZynKitTable < Qml::Widget
  attr_reader :kittable
  attr_property :weights
  attr_property :active
  def onSetup(old=nil);
        return if children.length > 2
        (0...16).each do |r|
            row         = Qml::ZynKitRow.new(db)
            row.label   = (r+1).to_s
            row.kitnum  = r
            row.weights = self.weights
            row.extern  = kittable.extern + "kit#{r}/"
            row.extern()
            row.set_active(false) if !self.active && r != 0
            Qml::add_child(self, row)
        end
    ;end
  def class_name(); "kittable" ;end
  def layout(l, selfBox);
        Draw::Layout::vpack(l, selfBox, children, 0, 1, 2)
    ;end
  def set_active_kit();
        self.active = true
        return if children.length < 3
        children[2..-1].each do |c|
            c.set_active(true)
        end
    ;end
  def set_non_kit();
        self.active = false
        return if children.length < 3
        children[2..-1].each do |c|
            c.set_active(false)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @kittable = self
@anonymous702 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous693 = Qml::ColorBox.new(database, ui_path+"2"+'/')
@anonymous694 = Qml::TextBox.new(database, ui_path+"3"+'/')
@anonymous695 = Qml::TextBox.new(database, ui_path+"4"+'/')
@anonymous696 = Qml::TextBox.new(database, ui_path+"5"+'/')
@anonymous697 = Qml::TextBox.new(database, ui_path+"6"+'/')
@anonymous698 = Qml::TextBox.new(database, ui_path+"7"+'/')
@anonymous699 = Qml::TextBox.new(database, ui_path+"8"+'/')
@anonymous700 = Qml::TextBox.new(database, ui_path+"9"+'/')
@anonymous701 = Qml::TextBox.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["kittable"] = self
@anonymous702.instance_eval do
def kittable=(k);@kittable=k;end
def kittable;    @kittable;  end
end
Qml::context_apply(context, @anonymous702)
Qml::prop_add(self, "weights")
Qml::prop_add(self, "active")

        @anonymous702.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, kittable.weights, 0, 3)\n        ;end", "src/mruby-zest/example/ZynKitTable.qml", 42)
@anonymous693.bg = Theme::TitleBar
Qml::add_child(@anonymous702, @anonymous693)
@anonymous694.bg = Theme::TitleBar
@anonymous694.label = "kit name"
Qml::add_child(@anonymous702, @anonymous694)
@anonymous695.bg = Theme::TitleBar
@db.connect_property(@anonymous695.properties["label"], lambda {"min. key"})
Qml::add_child(@anonymous702, @anonymous695)
@anonymous696.bg = Theme::TitleBar
@anonymous696.label = "last note"
Qml::add_child(@anonymous702, @anonymous696)
@anonymous697.bg = Theme::TitleBar
@db.connect_property(@anonymous697.properties["label"], lambda {"max. key"})
Qml::add_child(@anonymous702, @anonymous697)
@anonymous698.bg = Theme::TitleBar
@anonymous698.label = "add"
Qml::add_child(@anonymous702, @anonymous698)
@anonymous699.bg = Theme::TitleBar
@anonymous699.label = "sub"
Qml::add_child(@anonymous702, @anonymous699)
@anonymous700.bg = Theme::TitleBar
@anonymous700.label = "pad"
Qml::add_child(@anonymous702, @anonymous700)
@anonymous701.bg = Theme::TitleBar
@anonymous701.label = "effect route"
Qml::add_child(@anonymous702, @anonymous701)
Qml::add_child(self, @anonymous702)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.19, 0.15, 0.10, 0.15, 0.07, 0.07, 0.07, 0.15]})
self.active = true

    #puts "ZynKitTable, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynKit
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @kit = self
@anonymous679 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous675 = Qml::Button.new(database, ui_path+"2"+'/')
@anonymous676 = Qml::Button.new(database, ui_path+"3"+'/')
@anonymous677 = Qml::Button.new(database, ui_path+"4"+'/')
@drum = Qml::ToggleButton.new(database, ui_path+"5"+'/')
@anonymous680 = Qml::ZynKitTable.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["kit"] = self
@anonymous679.instance_eval do
def drum=(k);@drum=k;end
def drum;    @drum;  end
end
context["drum"] = @drum
Qml::context_apply(context, @anonymous679)

        @anonymous679.instance_eval("def layout(l, selfBox);\n            Draw::Layout::tabpack(l, selfBox, self, drum)\n        ;end", "src/mruby-zest/example/ZynKit.qml", 50)
@anonymous675.label = "No Kits"
@anonymous675.whenValue = lambda { kit.set_mode(0) }
Qml::add_child(@anonymous679, @anonymous675)
@db.connect_property(@anonymous676.properties["label"], lambda {"Multi-Kit"})
@anonymous676.whenValue = lambda { kit.set_mode(1) }
Qml::add_child(@anonymous679, @anonymous676)
@db.connect_property(@anonymous677.properties["label"], lambda {"Single-Kit"})
@anonymous677.whenValue = lambda { kit.set_mode(2) }
Qml::add_child(@anonymous679, @anonymous677)
@db.connect_property(@drum.properties["extern"], lambda {kit.extern + "Pdrummode"})
@drum.label = "drum mode"
Qml::add_child(@anonymous679, @drum)
Qml::add_child(self, @anonymous679)
@db.connect_property(@anonymous680.properties["extern"], lambda {kit.extern})
Qml::add_child(self, @anonymous680)
self.valueRef = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        kit.valueRef = OSC::RemoteParam.new($remote, kit.extern + "Pkitmode")
        kit.valueRef.mode = :options
        kit.valueRef.callback = lambda {|x| kit.get_mode(x) }

    
end})

    #puts "ZynKit, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynKitKeyButton < Qml::Widget
  attr_accessor :whenValue
  def draw(vg);
        pad  = 1/64
        pad2 = (1-2*pad)
        vg.path do |v|
            v.rect(w*pad, h*pad, w*pad2, h*pad2)
            paint = v.linear_gradient(0,0,0,h,
            Theme::ButtonGrad1, Theme::ButtonGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_width 1
            v.stroke
        end

        vg.path do |v|
            v.move_to(w*1/3, 0)
            v.line_to(w*1/3, h)
            v.move_to(w*2/3, 0)
            v.line_to(w*2/3, h)
            v.stroke
        end

        text_color = Theme::TextColor
        if self.layoutOpts.include? :aspect
            vg.font_size h*0.5 
        else
            vg.font_size h*0.8 
        end
        vg.fill_color text_color
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(1/6*w,h/2,"Mn")
        end
        
        #Pause Icon
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(3/6*w,h/2,"R")
        end
        
        #Play Icon
        vg.path do |v|
            v.text_align NVG::ALIGN_CENTER | NVG::ALIGN_MIDDLE
            v.text(5/6*w,h/2,"Mx")
        end
    ;end
  def onMouseHover(ev);
        rel = (ev.pos.x-global_x)/w
        if(rel < 1/3)
            self.root.log(:tooltip, 
            "Capture minimum key")
        elsif(rel < 2/3)
            self.root.log(:tooltip, 
            "Reset key range")
        else
            self.root.log(:tooltip, 
            "Capture maximum key")
        end
    ;end
  def onMousePress(ev);
        rel = (ev.pos.x-global_x)/w
        if(rel < 1/3)
            $remote.action(extern + "captureMin")
        elsif(rel < 2/3)
            $remote.action(extern + "Pminkey", 0)
            $remote.action(extern + "Pmaxkey", 127)
        else
            $remote.action(extern + "captureMax")
        end
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        if(self.layoutOpts.include? :aspect)
            l.aspect(selfBox, 3.0, 1.0)
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous682 = self

    
context = Hash.new
self.whenValue = nil

    #puts "ZynKitKeyButton, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynKitRow < Qml::Widget
  attr_reader :kit_item
  attr_reader :minkey
  attr_reader :maxkey
  attr_reader :add
  attr_reader :sub
  attr_reader :pad_but
  attr_property :weights
  attr_property :kitnum
  def set_active(a);
        minkey.active = a
        maxkey.active = a
        damage_self
    ;end
  def class_name(); "kitrow" ;end
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, kit_item.weights, 0, 3)
    ;end
  def onSetup(old=nil);
        children.each do |ch|
            ch.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @kit_item = self
@anonymous683 = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous684 = Qml::TextLine.new(database, ui_path+"2"+'/')
@minkey = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous686 = Qml::ZynKitKeyButton.new(database, ui_path+"4"+'/')
@maxkey = Qml::HSlider.new(database, ui_path+"5"+'/')
@add = Qml::FancyButton.new(database, ui_path+"6"+'/')
@sub = Qml::FancyButton.new(database, ui_path+"7"+'/')
@pad_but = Qml::FancyButton.new(database, ui_path+"8"+'/')
@anonymous691 = Qml::Selector.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["kit_item"] = self
context["minkey"] = @minkey
context["maxkey"] = @maxkey
context["add"] = @add
context["sub"] = @sub
context["pad_but"] = @pad_but
Qml::prop_add(self, "weights")
Qml::prop_add(self, "kitnum")
@db.connect_property(@anonymous683.properties["extern"], lambda {kit_item.extern + "Penabled"})
@db.connect_property(@anonymous683.properties["label"], lambda {kit_item.label})
@anonymous683.layoutOpts = [:no_constraint]
Qml::add_child(self, @anonymous683)
@db.connect_property(@anonymous684.properties["extern"], lambda {kit_item.extern + "Pname"})
@anonymous684.label = "kit name"
Qml::add_child(self, @anonymous684)
@db.connect_property(@minkey.properties["extern"], lambda {kit_item.extern + "Pminkey"})
@minkey.label = "0"
Qml::add_child(self, @minkey)
@db.connect_property(@anonymous686.properties["extern"], lambda {kit_item.extern})
@anonymous686.whenValue = lambda {
            minkey.refresh
            maxkey.refresh
        }
Qml::add_child(self, @anonymous686)
@db.connect_property(@maxkey.properties["extern"], lambda {kit_item.extern + "Pmaxkey"})
Qml::add_child(self, @maxkey)
@db.connect_property(@add.properties["extern"], lambda {kit_item.extern + "Padenabled"})
@add.layoutOpts = [:no_constraint]
@add.label = "edit"
@add.whenClick = lambda {
            if(!add.children[0].value)
                $remote.action(add.extern, true)
            end
            rt = kit_item.root
            rt.set_view_pos(:view, :add_synth)
            rt.set_view_pos(:kit, kit_item.kitnum)
            rt.change_view
        }
Qml::add_child(self, @add)
@db.connect_property(@sub.properties["extern"], lambda {kit_item.extern + "Psubenabled"})
@sub.layoutOpts = [:no_constraint]
@sub.label = "edit"
@sub.whenClick = lambda {
            if(!sub.children[0].value)
                $remote.action(sub.extern, true)
            end
            rt = kit_item.root
            rt.set_view_pos(:view, :sub_synth)
            rt.set_view_pos(:kit, kit_item.kitnum)
            rt.change_view
        }
Qml::add_child(self, @sub)
@db.connect_property(@pad_but.properties["extern"], lambda {kit_item.extern + "Ppadenabled"})
@pad_but.layoutOpts = [:no_constraint]
@pad_but.label = "edit"
@pad_but.whenClick = lambda {
            if(!pad_but.children[0].value)
                $remote.action(pad_but.extern, true)
            end
            rt = kit_item.root
            rt.set_view_pos(:view, :pad_synth)
            rt.set_view_pos(:kit, kit_item.kitnum)
            rt.change_view
        }
Qml::add_child(self, @pad_but)
@db.connect_property(@anonymous691.properties["extern"], lambda {kit_item.extern + "Psendtoparteffect"})
@anonymous691.layoutOpts = [:no_constraint]
@anonymous691.label = "off"
Qml::add_child(self, @anonymous691)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.10, 0.10, 0.10, 0.05, 0.10, 0.10, 0.10, 0.30]})
self.kitnum = 0

    #puts "ZynKitRow, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynKitTable < Qml::Widget
  attr_reader :kittable
  attr_property :weights
  attr_property :active
  def onSetup(old=nil);
        return if children.length > 2
        (0...16).each do |r|
            row         = Qml::ZynKitRow.new(db)
            row.label   = (r+1).to_s
            row.kitnum  = r
            row.weights = self.weights
            row.extern  = kittable.extern + "kit#{r}/"
            row.extern()
            row.set_active(false) if !self.active && r != 0
            Qml::add_child(self, row)
        end
    ;end
  def class_name(); "kittable" ;end
  def layout(l, selfBox);
        Draw::Layout::vpack(l, selfBox, children, 0, 1, 2)
    ;end
  def set_active_kit();
        self.active = true
        return if children.length < 3
        children[2..-1].each do |c|
            c.set_active(true)
        end
    ;end
  def set_non_kit();
        self.active = false
        return if children.length < 3
        children[2..-1].each do |c|
            c.set_active(false)
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @kittable = self
@anonymous702 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous693 = Qml::ColorBox.new(database, ui_path+"2"+'/')
@anonymous694 = Qml::TextBox.new(database, ui_path+"3"+'/')
@anonymous695 = Qml::TextBox.new(database, ui_path+"4"+'/')
@anonymous696 = Qml::TextBox.new(database, ui_path+"5"+'/')
@anonymous697 = Qml::TextBox.new(database, ui_path+"6"+'/')
@anonymous698 = Qml::TextBox.new(database, ui_path+"7"+'/')
@anonymous699 = Qml::TextBox.new(database, ui_path+"8"+'/')
@anonymous700 = Qml::TextBox.new(database, ui_path+"9"+'/')
@anonymous701 = Qml::TextBox.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["kittable"] = self
@anonymous702.instance_eval do
def kittable=(k);@kittable=k;end
def kittable;    @kittable;  end
end
Qml::context_apply(context, @anonymous702)
Qml::prop_add(self, "weights")
Qml::prop_add(self, "active")

        @anonymous702.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, kittable.weights, 0, 3)\n        ;end", "src/mruby-zest/example/ZynKitTable.qml", 42)
@anonymous693.bg = Theme::TitleBar
Qml::add_child(@anonymous702, @anonymous693)
@anonymous694.bg = Theme::TitleBar
@anonymous694.label = "kit name"
Qml::add_child(@anonymous702, @anonymous694)
@anonymous695.bg = Theme::TitleBar
@db.connect_property(@anonymous695.properties["label"], lambda {"min. key"})
Qml::add_child(@anonymous702, @anonymous695)
@anonymous696.bg = Theme::TitleBar
@anonymous696.label = "last note"
Qml::add_child(@anonymous702, @anonymous696)
@anonymous697.bg = Theme::TitleBar
@db.connect_property(@anonymous697.properties["label"], lambda {"max. key"})
Qml::add_child(@anonymous702, @anonymous697)
@anonymous698.bg = Theme::TitleBar
@anonymous698.label = "add"
Qml::add_child(@anonymous702, @anonymous698)
@anonymous699.bg = Theme::TitleBar
@anonymous699.label = "sub"
Qml::add_child(@anonymous702, @anonymous699)
@anonymous700.bg = Theme::TitleBar
@anonymous700.label = "pad"
Qml::add_child(@anonymous702, @anonymous700)
@anonymous701.bg = Theme::TitleBar
@anonymous701.label = "effect route"
Qml::add_child(@anonymous702, @anonymous701)
Qml::add_child(self, @anonymous702)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.19, 0.15, 0.10, 0.15, 0.07, 0.07, 0.07, 0.15]})
self.active = true

    #puts "ZynKitTable, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynLFO < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous704 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous705 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous706 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous707 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous709 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous710 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous711 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous714 = Qml::Col.new(database, ui_path+"10"+'/')
@anonymous712 = Qml::Selector.new(database, ui_path+"11"+'/')
@anonymous713 = Qml::Button.new(database, ui_path+"12"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
@anonymous704.type = :float
@db.connect_property(@anonymous704.properties["extern"], lambda {box.extern+"Pfreq" })
Qml::add_child(@top, @anonymous704)
@db.connect_property(@anonymous705.properties["extern"], lambda {box.extern+"Pintensity"})
Qml::add_child(@top, @anonymous705)
@db.connect_property(@anonymous706.properties["extern"], lambda {box.extern+"Pstartphase"})
Qml::add_child(@top, @anonymous706)
@db.connect_property(@anonymous707.properties["extern"], lambda {box.extern+"Pstretch"})
Qml::add_child(@top, @anonymous707)
Qml::add_child(self, @top)
@db.connect_property(@anonymous709.properties["extern"], lambda {box.extern+"Pdelay"})
Qml::add_child(@bot, @anonymous709)
@db.connect_property(@anonymous710.properties["extern"], lambda {box.extern+"Prandomness"})
Qml::add_child(@bot, @anonymous710)
@db.connect_property(@anonymous711.properties["extern"], lambda {box.extern+"Pfreqrand"})
Qml::add_child(@bot, @anonymous711)
@db.connect_property(@anonymous712.properties["extern"], lambda {box.extern+"PLFOtype"})
Qml::add_child(@anonymous714, @anonymous712)
@db.connect_property(@anonymous713.properties["extern"], lambda {box.extern+"Pcontinous"})
@anonymous713.label = "sync"
Qml::add_child(@anonymous714, @anonymous713)
Qml::add_child(@bot, @anonymous714)
Qml::add_child(self, @bot)
self.label = "LFO"
self.extern = ""
self.copyable = true

    #puts "ZynLFO, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynLLFO < Qml::ParModule
  attr_reader :lfo_mod
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @lfo_mod = self
@anonymous717 = Qml::Knob.new(database, ui_path+"1"+'/')
@anonymous718 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous719 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous720 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous721 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous722 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous723 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous724 = Qml::Button.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["lfo_mod"] = self
@db.connect_property(@anonymous717.properties["extern"], lambda {lfo_mod.extern+"Pfreq"})
Qml::add_child(self, @anonymous717)
@db.connect_property(@anonymous718.properties["extern"], lambda {lfo_mod.extern+"Pintensity"})
Qml::add_child(self, @anonymous718)
@db.connect_property(@anonymous719.properties["extern"], lambda {lfo_mod.extern+"Pstartphase"})
Qml::add_child(self, @anonymous719)
@db.connect_property(@anonymous720.properties["extern"], lambda {lfo_mod.extern+"Pstretch"})
Qml::add_child(self, @anonymous720)
@db.connect_property(@anonymous721.properties["extern"], lambda {lfo_mod.extern+"Prandomness"})
Qml::add_child(self, @anonymous721)
@db.connect_property(@anonymous722.properties["extern"], lambda {lfo_mod.extern+"Pfreqrand"})
Qml::add_child(self, @anonymous722)
@db.connect_property(@anonymous723.properties["extern"], lambda {lfo_mod.extern+"PLFOtype"})
Qml::add_child(self, @anonymous723)
@db.connect_property(@anonymous724.properties["extern"], lambda {lfo_mod.extern+"Pcontinous"})
@anonymous724.label = "C"
Qml::add_child(self, @anonymous724)
self.w = 500*4
self.h = 300*4
self.extern = "/part0/kit0/adpars/GlobalPar/FreqLfo/"
self.label = "generic lfo"

    #puts "ZynLLFO, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynLogo < Qml::Widget
  def draw_placeholder(vg);
        #background color("123456")
        #vg.path do |v|
        #    v.rect(0, 0, w, h)
        #    v.fill_color(NVG.rgba(0xaa, 0xaa, 0xaa, 255))
        #    v.fill
        #end
        logo_top = color("4DC3C7")
        logo_bot = color("56C0A5")

        vg.path do |v|
            l = 0.4*w
            v.move_to(0,0)
            v.line_to(l,      0.1*h)
            v.line_to(l*0.75, 0.6*h)
            v.close_path
            v.fill_color logo_top
            v.fill
        end

        vg.path do |v|
            l = 0.4*w
            v.move_to(l,h)
            v.line_to(0,      0.9*h)
            v.line_to(l*0.25, 0.4*h)
            v.close_path
            v.fill_color logo_bot
            v.fill
        end

        vg.font_face("bold")
        vg.font_size h*0.8
        vg.text_align NVG::ALIGN_LEFT | NVG::ALIGN_MIDDLE
        vg.fill_color logo_top
        vg.text(w*1/2,h/2,"ZYN")
    ;end
  def draw_path(vg, path, co);
        wscale = 80.0
        hscale = 35.0
        n = path.length/2
        xmin = 64
        ymin = 60
        vg.path do
            xx = (w*(path[2*0+0]-xmin)/wscale).round
            yy = (h*(path[2*0+1]-ymin)/hscale).round
            vg.move_to(xx, yy)
            (1...n).each do |i|
                xx = (w*(path[2*i+0]-xmin)/wscale).round
                yy = (h*(path[2*i+1]-ymin)/hscale).round
                vg.line_to(xx,yy)
            end
            vg.close_path
            vg.fill_color co
            vg.fill
            #vg.stroke_color color("2E3239")
            #vg.stroke_width 2.0
            #vg.stroke
        end
    ;end
  def draw_new(vg);
        st0 = color("686868")
        st1 = color("2E3239")
        st1 = color("010101")
        st2 = color("50C3C7")
        st3 = color("57C1A6")

        st1
        path_1 = [97,64,64,61,67.3,63.3,64,63,87,79,97,66,95.6,65.9]
        path_2 = [97,90,74,74,64,87,65.4,87.1,64,89,97,92,93.7,89.7]
        path_3 = [103,69,103.7,71,103,71,104,74,107.5,74,103,80,103,82.1,103,83,103,85,115,85,115,83,115,81.4,115,80,110.3,80,115,74,115,72,115,71,115,69]
        path_4 = [129,69,124,69,123,74,122,69,117,69,117.8,71,117,71,121,81,121,83,121,85,125,85,125,83,125,81,129,71,128.2,71]
        path_5 = [131,69,135,69,138.5,74,138.8,75,139,75,139,69,143,69,143,85,139,85,135.5,80,135.8,79,135,79,135,85,131,85]
        draw_path(vg, path_1, st1)
        draw_path(vg, path_2, st1)
        draw_path(vg, path_3, st1)
        draw_path(vg, path_4, st1)
        draw_path(vg, path_5, st1)

        st2 
        path_6 = [64,62,97,65,87,78]
        path_7 = [115,84,103,84,103,81,109,73,109,73,104,73,103,70,115,70,115,73,109,81,109,81,115,81,115,84]
        path_8 = [122,70,123,75,123,75,124,70,129,70,125,80,125,84,121,84,121,80,117,70]
        path_9 = [131,70,135,70,138.5,75,138.8,76,139,76,139,70,143,70,143,84,139,84,135.5,79,135.8,78,135,78,135,84,131,84]
        draw_path(vg, path_6, st2)
        draw_path(vg, path_7, st2)
        draw_path(vg, path_8, st2)
        draw_path(vg, path_9, st2)

        st3
        path_10 = [97,91,64,88,74,75]
        draw_path(vg, path_10, st3)
    ;end
  def draw(vg);
        #draw_placeholder(vg)
        draw_new(vg)
    ;end
  def onMousePress(m);
        view = root.get_view_pos(:view)
        root.set_view_pos(:view, :about)    if(view != :about)
        root.set_view_pos(:view, :settings) if(view == :about)
        root.change_view
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous726 = self

    
context = Hash.new

    #puts "ZynLogo, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynMidiLearn < Qml::Widget
  attr_reader :vce_list
  attr_property :weights
  attr_accessor :valueRef
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  def onSetup(old=nil);
        return if children.length > 3
        (0...16).each do |r|
            row = Qml::ZynMidiLearnRow.new(db)
            row.label = r.to_s
            row.weights = self.weights
            row.whenValue = lambda { vce_list.refresh }
            Qml::add_child(self, row)
        end

        self.valueRef = OSC::RemoteParam.new($remote, "/midi-learn-values")
        self.valueRef.callback = lambda {|x| vce_list.handle(x) }
        self.valueRef.refresh
    ;end
  def handle(x);
        n = x.length/4
        m = children.length-1
        (0...n).each do |i|
            row = children[i+1]
            row.children[1].label = "*"
            row.children[2].label = x[i*4+0].to_s
            row.children[3].label = x[i*4+1].to_s
            row.children[4].label = x[i*4+2].to_s
            row.children[5].label = x[i*4+3].to_s
        end
        (n...m).each do |i|
            row = children[i+1]
            row.children[1].label = "X"
            row.children[2].label = "X"
            row.children[3].label = ""
            row.children[4].label = "X"
            row.children[5].label = "X"
        end
        damage_self
    ;end
  def layout(l, selfBox);
        n    = children.length
        off  = 0
        gap  = 0.005
        step = (1.0-(n)*gap)/n
        children.each_with_index do |ch, id|
            ch.fixed(l, selfBox, 0.01, off, 0.98, step)
            off += step + gap
        end
        selfBox
    ;end
  def refresh();
        valueRef.refresh if valueRef
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @vce_list = self
@anonymous733 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous727 = Qml::ColorBox.new(database, ui_path+"2"+'/')
@anonymous728 = Qml::TextBox.new(database, ui_path+"3"+'/')
@anonymous729 = Qml::TextBox.new(database, ui_path+"4"+'/')
@anonymous730 = Qml::TextBox.new(database, ui_path+"5"+'/')
@anonymous731 = Qml::TextBox.new(database, ui_path+"6"+'/')
@anonymous732 = Qml::TextBox.new(database, ui_path+"7"+'/')

    
context = Hash.new
context["vce_list"] = self
@anonymous733.instance_eval do
def vce_list=(k);@vce_list=k;end
def vce_list;    @vce_list;  end
end
Qml::context_apply(context, @anonymous733)
Qml::prop_add(self, "weights")

        @anonymous733.instance_eval("def class_name(); \"zavlh\" ;end", "src/mruby-zest/example/ZynMidiLearn.qml", 71)

        @anonymous733.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, vce_list.weights)\n        ;end", "src/mruby-zest/example/ZynMidiLearn.qml", 72)
@anonymous727.bg = nil
Qml::add_child(@anonymous733, @anonymous727)
@anonymous728.bg = nil
@anonymous728.label = "ch"
Qml::add_child(@anonymous733, @anonymous728)
@anonymous729.bg = nil
@anonymous729.label = "ctrl"
Qml::add_child(@anonymous733, @anonymous729)
@anonymous730.bg = nil
@anonymous730.label = "acquired controls"
Qml::add_child(@anonymous733, @anonymous730)
@anonymous731.bg = nil
@anonymous731.label = "min"
Qml::add_child(@anonymous733, @anonymous731)
@anonymous732.bg = nil
@anonymous732.label = "max"
Qml::add_child(@anonymous733, @anonymous732)
Qml::add_child(self, @anonymous733)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.05, 0.05, 0.75, 0.05, 0.05]})
self.valueRef = nil

    #puts "ZynMidiLearn, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynMidiLearnRow < Qml::Widget
  attr_reader :voice_item
  attr_reader :ctrl
  attr_property :weights
  attr_accessor :whenValue
  def unlearn();
        return if ctrl.label.empty?
        $remote.action("/unlearn", ctrl.label)
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        off = 0.0
        hpad = 1/128
        children.each_with_index do |ch, ind|
            weight = weights[ind]
            ch.fixed(l, selfBox, off+hpad, 0.0, weight-2*hpad, 1.0)
            off += weight
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @voice_item = self
@anonymous735 = Qml::TriggerButton.new(database, ui_path+"1"+'/')
@anonymous736 = Qml::TextBox.new(database, ui_path+"2"+'/')
@anonymous737 = Qml::TextBox.new(database, ui_path+"3"+'/')
@ctrl = Qml::TextBox.new(database, ui_path+"4"+'/')
@anonymous739 = Qml::TextBox.new(database, ui_path+"5"+'/')
@anonymous740 = Qml::TextBox.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["voice_item"] = self
context["ctrl"] = @ctrl
Qml::prop_add(self, "weights")
@anonymous735.tooltip = "unlearn midi control"
@db.connect_property(@anonymous735.properties["label"], lambda {voice_item.label})
@anonymous735.layoutOpts = [:no_constraint]
@anonymous735.whenValue = lambda { voice_item.unlearn }
Qml::add_child(self, @anonymous735)
@anonymous736.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous736)
@anonymous737.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous737)
@ctrl.bg = Theme::GeneralBackground
@ctrl.pad = 0
@ctrl.align = :left
Qml::add_child(self, @ctrl)
@anonymous739.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous739)
@anonymous740.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous740)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.05, 0.2, 0.2, 0.3, 0.2]})
self.whenValue = nil

    #puts "ZynMidiLearnRow, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynMixer < Qml::Widget
  attr_reader :mixer
  attr_accessor :valueRef
  attr_property :data
  def animate();
        #self.valueRef.refresh if(self.valueRef)

        if(self.data)
            children[0].set_level(data[0], data[1])
            (1..16).each do |i|
                children[i].set_level(data[6+i-1])
            end
        end
    ;end
  def update_data(x);
        self.data = x
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  def onSetup(old=nil);
        return if children.length > 3
        col = Qml::ZynMixerMasterCol.new(db)
        Qml::add_child(self, col)
        (0...16).each do |r|
            col = Qml::ZynMixerCol.new(db)
            col.label  = (1+r).to_s
            col.idx    = r
            col.extern = "/part#{r}/"
            Qml::add_child(self, col)
        end

        self.valueRef = OSC::RemoteParam.new($remote, "/vu-meter")
        self.valueRef.callback = lambda {|x| mixer.update_data(x) }
        animate() if self.data
    ;end
  def class_name(); "mixer" ;end
  def layout(l, selfBox);
        Draw::Layout::hpack(l, selfBox, children, 0.02, 0.96, 4)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @mixer = self

    
context = Hash.new
context["mixer"] = self
Qml::prop_add(self, "data")
self.valueRef = nil
self.data = nil

    #puts "ZynMixer, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynMixerCol < Qml::Widget
  attr_reader :col
  attr_property :weights
  attr_property :idx
  def rap2dB(x); 20*Math::log10(x) ;end
  def bound(x); [0.0, [1.0, x].min].max ;end
  def cv(x);min_db = -40;bound((min_db-rap2dB(x))/min_db);end
  def set_level(l);
        old = children[2].children[0].value
        if(old != cv(l))
            children[2].children[0].value = cv(l)
            children[2].children[1].value = cv(l)
            children[2].children[0].damage_self
            children[2].children[1].damage_self
        end
    ;end
  def layout(l, selfBox);
        off = 0.0
        fixed_pad = 2
        children.each_with_index do |ch, ind|
            weight = weights[ind]
            l.fixed_long(ch, selfBox, 0, off, 1.0, weight,
                         0, fixed_pad, 0, -2*fixed_pad)
            off += weight
        end
        selfBox
    ;end
  def onSetup(old=nil);
        children.each do |ch|
            ch.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @col = self
@anonymous743 = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous744 = Qml::TextEdit.new(database, ui_path+"2"+'/')
@anonymous748 = Qml::ColorBox.new(database, ui_path+"3"+'/')
@anonymous745 = Qml::Slider.new(database, ui_path+"4"+'/')
@anonymous746 = Qml::Slider.new(database, ui_path+"5"+'/')
@anonymous747 = Qml::Slider.new(database, ui_path+"6"+'/')
@anonymous749 = Qml::HSlider.new(database, ui_path+"7"+'/')
@anonymous750 = Qml::Selector.new(database, ui_path+"8"+'/')
@anonymous751 = Qml::TriggerButton.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["col"] = self
Qml::prop_add(self, "weights")
Qml::prop_add(self, "idx")
@db.connect_property(@anonymous743.properties["extern"], lambda {col.extern + "Penabled"})
@db.connect_property(@anonymous743.properties["label"], lambda {col.label})
@anonymous743.layoutOpts = [:no_constraint]
Qml::add_child(self, @anonymous743)
@db.connect_property(@anonymous744.properties["extern"], lambda {col.extern + "Pname"})
@anonymous744.height = 0.333333
@anonymous744.label = "synth"
Qml::add_child(self, @anonymous744)

        @anonymous748.instance_eval("def class_name(); \"mixbox\" ;end", "src/mruby-zest/example/ZynMixerCol.qml", 40)

        @anonymous748.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynMixerCol.qml", 41)
@anonymous745.visual = true
@anonymous745.centered = false
@anonymous745.pad = 0.01
Qml::add_child(@anonymous748, @anonymous745)
@anonymous746.visual = true
@anonymous746.centered = false
@anonymous746.pad = 0.01
Qml::add_child(@anonymous748, @anonymous746)
@db.connect_property(@anonymous747.properties["extern"], lambda {col.extern + "Pvolume"})
@anonymous747.centered = false
@anonymous747.pad = 0.01
Qml::add_child(@anonymous748, @anonymous747)
@anonymous748.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous748)
@db.connect_property(@anonymous749.properties["extern"], lambda {col.extern + "Ppanning"})
@anonymous749.label = "pan"
Qml::add_child(self, @anonymous749)
@db.connect_property(@anonymous750.properties["extern"], lambda {col.extern + "Prcvchn"})
@anonymous750.layoutOpts = [:no_constraint]
Qml::add_child(self, @anonymous750)
@anonymous751.layoutOpts = [:no_constraint]
@anonymous751.label = "edit"
@anonymous751.whenValue = lambda { col.root.set_view_pos(:part, idx) }
Qml::add_child(self, @anonymous751)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.1, 0.70, 0.05, 0.05, 0.05]})
self.idx = 0

    #puts "ZynMixerCol, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynMixerMasterCol < Qml::Widget
  attr_reader :col
  attr_property :weights
  def rap2dB(x); 20*Math::log10(x) ;end
  def bound(x); [0.0, [1.0, x].min].max ;end
  def cv(x);min_db = -40;bound((min_db-rap2dB(x))/min_db);end
  def set_level(l,r);
        old = children[1].children[1].value
        if(old != cv(l))
            children[1].children[1].value = cv(l)
            children[1].children[0].value = cv(r)
            children[1].children[0].damage_self
            children[1].children[1].damage_self
        end
    ;end
  def layout(l, selfBox);
        off = 0.0
        fixed_pad = 2
        children.each_with_index do |ch, ind|
            weight = weights[ind]
            l.fixed_long(ch, selfBox, 0, off, 1.0, weight,
                         0, fixed_pad, 0, -2*fixed_pad)
            off += weight
        end
        selfBox
    ;end
  def onSetup(old=nil);
        children.each do |ch|
            ch.extern()
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @col = self
@anonymous753 = Qml::TextBox.new(database, ui_path+"1"+'/')
@anonymous757 = Qml::ColorBox.new(database, ui_path+"2"+'/')
@anonymous754 = Qml::Slider.new(database, ui_path+"3"+'/')
@anonymous755 = Qml::Slider.new(database, ui_path+"4"+'/')
@anonymous756 = Qml::Slider.new(database, ui_path+"5"+'/')

    
context = Hash.new
context["col"] = self
Qml::prop_add(self, "weights")
@anonymous753.label = "Master"
Qml::add_child(self, @anonymous753)

        @anonymous757.instance_eval("def class_name(); \"mixbox\" ;end", "src/mruby-zest/example/ZynMixerMasterCol.qml", 33)

        @anonymous757.instance_eval("def layout(l,selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynMixerMasterCol.qml", 34)
@anonymous754.visual = true
@anonymous754.centered = false
@anonymous754.pad = 0.01
Qml::add_child(@anonymous757, @anonymous754)
@anonymous755.visual = true
@anonymous755.centered = false
@anonymous755.pad = 0.01
Qml::add_child(@anonymous757, @anonymous755)
@anonymous756.extern = "/Pvolume"
@anonymous756.centered = false
@anonymous756.pad = 0.01
Qml::add_child(@anonymous757, @anonymous756)
@anonymous757.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous757)
@db.connect_property(self.properties["weights"], lambda {[0.05, 0.95, 0.05, 0.05, 0.05]})

    #puts "ZynMixerMasterCol, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynModAmpEnv < Qml::Group
  attr_reader :box
  attr_property :whenModified
  def cb();
        whenModified.call if whenModified
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous764 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous759 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous760 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous761 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous762 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous763 = Qml::Knob.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "whenModified")
@anonymous759.whenValue = lambda { box.cb }
@db.connect_property(@anonymous759.properties["extern"], lambda {box.extern+"PA_dt" })
Qml::add_child(@anonymous764, @anonymous759)
@anonymous760.whenValue = lambda { box.cb }
@db.connect_property(@anonymous760.properties["extern"], lambda {box.extern+"PD_dt"})
Qml::add_child(@anonymous764, @anonymous760)
@anonymous761.whenValue = lambda { box.cb }
@db.connect_property(@anonymous761.properties["extern"], lambda {box.extern+"PS_val"})
Qml::add_child(@anonymous764, @anonymous761)
@anonymous762.whenValue = lambda { box.cb }
@db.connect_property(@anonymous762.properties["extern"], lambda {box.extern+"PR_dt"})
Qml::add_child(@anonymous764, @anonymous762)
@anonymous763.whenValue = lambda { box.cb }
@db.connect_property(@anonymous763.properties["extern"], lambda {box.extern+"Penvstretch"})
Qml::add_child(@anonymous764, @anonymous763)
Qml::add_child(self, @anonymous764)
self.label = "Envelope"
self.whenModified = nil
self.extern = ""

    #puts "ZynModAmpEnv, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynModFreqGeneral < Qml::Group
  attr_reader :box
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous771 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous766 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous767 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous768 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous769 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous770 = Qml::NumEntry.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["box"] = self
@db.connect_property(@anonymous766.properties["extern"], lambda {box.extern + "PFMDetune" })
Qml::add_child(@anonymous771, @anonymous766)
@db.connect_property(@anonymous767.properties["extern"], lambda {box.extern + "FMoctave" })
Qml::add_child(@anonymous771, @anonymous767)
@db.connect_property(@anonymous768.properties["extern"], lambda {box.extern + "PFMrelbw" })
Qml::add_child(@anonymous771, @anonymous768)
@db.connect_property(@anonymous769.properties["extern"], lambda {box.extern + "PFMdetuneType" })
Qml::add_child(@anonymous771, @anonymous769)
@db.connect_property(@anonymous770.properties["extern"], lambda {box.extern + "PFMcoarseDetune" })
Qml::add_child(@anonymous771, @anonymous770)
Qml::add_child(self, @anonymous771)
self.extern = ""

    #puts "ZynModFreqGeneral, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynOscil < Qml::Widget
  attr_reader :base_osc
  attr_reader :base_title
  attr_reader :base_harm
  attr_reader :base
  attr_reader :full_title
  attr_reader :full_harm
  attr_reader :full
  attr_reader :hedit
  attr_reader :scroll
  attr_reader :middle_panel
  attr_reader :modsel
  attr_property :doRefresh
  attr_property :delayRefresh
  def animate();
        if(doRefresh)
            self.doRefresh = false
            base.refresh
            full.refresh
            base_harm.refresh
            full_harm.refresh
            self.delayRefresh = Time.new + 0.1
        end

        #Super hacky workaround for data race in adsynth oscillators
        if(delayRefresh && Time.new > delayRefresh)
            base.refresh
            base_harm.refresh
            self.delayRefresh = nil
        end
    ;end
  def clear();
        (0...128).each do |i|
            mval = 64
            pval = 64
            mval = 127 if i == 0
            $remote.seti(base_osc.extern + "magnitude" + i.to_s, mval)
            $remote.seti(base_osc.extern + "phase" + i.to_s,     pval)
        end
        refresh
    ;end
  def refresh();
        self.doRefresh = true
    ;end
  def layout(l, selfBox);
        base.fixed(l,           selfBox, 0.00, 0.15, 0.30, 0.38)
        full.fixed(l,           selfBox, 0.70, 0.15, 0.30, 0.38)
        base_harm.fixed(l,      selfBox, 0.00, 0.05, 0.30, 0.12)
        full_harm.fixed(l,      selfBox, 0.70, 0.05, 0.30, 0.12)
        base_title.fixed(l,     selfBox, 0.00, 0.00, 0.30, 0.05)
        full_title.fixed(l,     selfBox, 0.70, 0.00, 0.30, 0.05)
        hedit.fixed(l,          selfBox, 0.00, 0.53, 1.00, 0.42)
        scroll.fixed(l,         selfBox, 0.00, 0.95, 1.00, 0.05)
        middle_panel.fixed(l,   selfBox, 0.30, 0.0,  0.40, 0.53)

        #voceBox = voice_button.layout(l)
        #modbBox = mod_button.layout(l)
        #l.fixed(scrlBox, selfBox, 0.10, 0.95, 0.80, 0.05)
        #l.fixed(voceBox, selfBox, 0.00, 0.95, 0.10, 0.05)
        #l.fixed(modbBox, selfBox, 0.90, 0.95, 0.10, 0.05)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @base_osc = self
@base_title = Qml::TitleBar.new(database, ui_path+"1"+'/')
@base_harm = Qml::HarmonicView.new(database, ui_path+"2"+'/')
@base = Qml::WaveView.new(database, ui_path+"3"+'/')
@full_title = Qml::TitleBar.new(database, ui_path+"4"+'/')
@full_harm = Qml::HarmonicView.new(database, ui_path+"5"+'/')
@full = Qml::WaveView.new(database, ui_path+"6"+'/')
@hedit = Qml::HarmonicEdit.new(database, ui_path+"7"+'/')
@scroll = Qml::ScrollBar.new(database, ui_path+"8"+'/')
@middle_panel = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous823 = Qml::Widget.new(database, ui_path+"10"+'/')
@anonymous781 = Qml::TextBox.new(database, ui_path+"11"+'/')
@anonymous782 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous783 = Qml::HSlider.new(database, ui_path+"13"+'/')
@anonymous784 = Qml::Widget.new(database, ui_path+"14"+'/')
@anonymous785 = Qml::TextBox.new(database, ui_path+"15"+'/')
@modsel = Qml::Selector.new(database, ui_path+"16"+'/')
@anonymous787 = Qml::HSlider.new(database, ui_path+"17"+'/')
@anonymous788 = Qml::HSlider.new(database, ui_path+"18"+'/')
@anonymous789 = Qml::HSlider.new(database, ui_path+"19"+'/')
@anonymous790 = Qml::TriggerButton.new(database, ui_path+"20"+'/')
@anonymous791 = Qml::TextBox.new(database, ui_path+"21"+'/')
@anonymous792 = Qml::Selector.new(database, ui_path+"22"+'/')
@anonymous793 = Qml::HSlider.new(database, ui_path+"23"+'/')
@anonymous794 = Qml::Widget.new(database, ui_path+"24"+'/')
@anonymous795 = Qml::TextBox.new(database, ui_path+"25"+'/')
@anonymous796 = Qml::Selector.new(database, ui_path+"26"+'/')
@anonymous797 = Qml::HSlider.new(database, ui_path+"27"+'/')
@anonymous798 = Qml::HSlider.new(database, ui_path+"28"+'/')
@anonymous799 = Qml::Widget.new(database, ui_path+"29"+'/')
@anonymous800 = Qml::ToggleButton.new(database, ui_path+"30"+'/')
@anonymous801 = Qml::TextBox.new(database, ui_path+"31"+'/')
@anonymous802 = Qml::Selector.new(database, ui_path+"32"+'/')
@anonymous803 = Qml::HSlider.new(database, ui_path+"33"+'/')
@anonymous806 = Qml::Widget.new(database, ui_path+"34"+'/')
@anonymous804 = Qml::TriggerButton.new(database, ui_path+"35"+'/')
@anonymous805 = Qml::ToggleButton.new(database, ui_path+"36"+'/')
@anonymous807 = Qml::TextBox.new(database, ui_path+"37"+'/')
@anonymous808 = Qml::Selector.new(database, ui_path+"38"+'/')
@anonymous809 = Qml::HSlider.new(database, ui_path+"39"+'/')
@anonymous810 = Qml::HSlider.new(database, ui_path+"40"+'/')
@anonymous811 = Qml::HSlider.new(database, ui_path+"41"+'/')
@anonymous812 = Qml::TriggerButton.new(database, ui_path+"42"+'/')
@anonymous813 = Qml::TextBox.new(database, ui_path+"43"+'/')
@anonymous814 = Qml::Selector.new(database, ui_path+"44"+'/')
@anonymous815 = Qml::HSlider.new(database, ui_path+"45"+'/')
@anonymous816 = Qml::HSlider.new(database, ui_path+"46"+'/')
@anonymous817 = Qml::HSlider.new(database, ui_path+"47"+'/')
@anonymous818 = Qml::Widget.new(database, ui_path+"48"+'/')
@anonymous819 = Qml::TextBox.new(database, ui_path+"49"+'/')
@anonymous820 = Qml::Selector.new(database, ui_path+"50"+'/')
@anonymous821 = Qml::HSlider.new(database, ui_path+"51"+'/')
@anonymous822 = Qml::TriggerButton.new(database, ui_path+"52"+'/')

    
context = Hash.new
context["base_osc"] = self
context["base_title"] = @base_title
context["base_harm"] = @base_harm
context["base"] = @base
context["full_title"] = @full_title
context["full_harm"] = @full_harm
context["full"] = @full
context["hedit"] = @hedit
context["scroll"] = @scroll
context["middle_panel"] = @middle_panel
@anonymous823.instance_eval do
def base_osc=(k);@base_osc=k;end
def base_osc;    @base_osc;  end
def base=(k);@base=k;end
def base;    @base;  end
end
context["modsel"] = @modsel
Qml::context_apply(context, @anonymous823)
Qml::prop_add(self, "doRefresh")
Qml::prop_add(self, "delayRefresh")
@base_title.label = "base waveform"
Qml::add_child(self, @base_title)
@db.connect_property(@base_harm.properties["extern"], lambda {base_osc.extern + "base-spectrum"})
Qml::add_child(self, @base_harm)
@db.connect_property(@base.properties["extern"], lambda {base_osc.extern + "base-waveform"})
Qml::add_child(self, @base)
@full_title.label = "full oscillator"
Qml::add_child(self, @full_title)
@db.connect_property(@full_harm.properties["extern"], lambda {base_osc.extern + "spectrum"})
Qml::add_child(self, @full_harm)
@db.connect_property(@full.properties["extern"], lambda {base_osc.extern + "waveform"})
Qml::add_child(self, @full)
@db.connect_property(@hedit.properties["extern"], lambda {base_osc.extern})
@hedit.whenValue = lambda {base_osc.refresh}
Qml::add_child(self, @hedit)
@scroll.vertical = false
@scroll.value = 0
@scroll.whenValue = lambda {hedit.set_scroll scroll.value}
Qml::add_child(self, @scroll)

        @middle_panel.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynOscil.qml", 227)

        @middle_panel.instance_eval("def layout(l, selfBox);\n            pad = 6\n            l.fixed_long(children[0], selfBox, 0, 0, 1, 1,\n                                pad, pad, -2*pad, -2*pad)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynOscil.qml", 232)

        @anonymous823.instance_eval("def onSetup(old=nil);\n                lam = lambda {base_osc.refresh}\n                children.each do |ch|\n                    ch.layoutOpts  = [:no_constraint]\n                    ch.whenValue ||= lam if(ch.respond_to? :whenValue)\n                end\n            ;end", "src/mruby-zest/example/ZynOscil.qml", 210)

        @anonymous823.instance_eval("def class_name(); \"oscgrid\" ;end", "src/mruby-zest/example/ZynOscil.qml", 219)

        @anonymous823.instance_eval("def layout(l, selfBox);\n                pad = 3\n                Draw::Layout::gridt(l, selfBox, children, 10, 4, pad, pad)\n            ;end", "src/mruby-zest/example/ZynOscil.qml", 220)
@db.connect_property(@anonymous781.properties["label"], lambda {"base func."})
Qml::add_child(@anonymous823, @anonymous781)
@db.connect_property(@anonymous782.properties["extern"], lambda {base_osc.extern + "Pcurrentbasefunc"})
Qml::add_child(@anonymous823, @anonymous782)
@db.connect_property(@anonymous783.properties["extern"], lambda {base_osc.extern + "Pbasefuncpar"})
Qml::add_child(@anonymous823, @anonymous783)
Qml::add_child(@anonymous823, @anonymous784)
@db.connect_property(@anonymous785.properties["label"], lambda {"BF mod."})
Qml::add_child(@anonymous823, @anonymous785)
@db.connect_property(@modsel.properties["extern"], lambda {base_osc.extern + "Pbasefuncmodulation"})
Qml::add_child(@anonymous823, @modsel)
@db.connect_property(@anonymous787.properties["extern"], lambda {modsel.extern + "par1"})
Qml::add_child(@anonymous823, @anonymous787)
@db.connect_property(@anonymous788.properties["extern"], lambda {modsel.extern + "par2"})
Qml::add_child(@anonymous823, @anonymous788)
@db.connect_property(@anonymous789.properties["extern"], lambda {modsel.extern + "par3"})
Qml::add_child(@anonymous823, @anonymous789)
@anonymous790.label = "as base"
@anonymous790.tooltip = "use combined waveform as base waveform"
@anonymous790.whenValue = lambda {
                    $remote.action(base_osc.extern+"use-as-base")
                }
Qml::add_child(@anonymous823, @anonymous790)
@anonymous791.label = "waveshaping"
Qml::add_child(@anonymous823, @anonymous791)
@db.connect_property(@anonymous792.properties["extern"], lambda {base_osc.extern + "Pwaveshapingfunction"})
Qml::add_child(@anonymous823, @anonymous792)
@db.connect_property(@anonymous793.properties["extern"], lambda {base_osc.extern + "Pwaveshaping"})
Qml::add_child(@anonymous823, @anonymous793)
Qml::add_child(@anonymous823, @anonymous794)
@anonymous795.label = "filter"
Qml::add_child(@anonymous823, @anonymous795)
@db.connect_property(@anonymous796.properties["extern"], lambda {base_osc.extern + "Pfiltertype"})
Qml::add_child(@anonymous823, @anonymous796)
@db.connect_property(@anonymous797.properties["extern"], lambda {base_osc.extern + "Pfilterpar1"})
Qml::add_child(@anonymous823, @anonymous797)
@db.connect_property(@anonymous798.properties["extern"], lambda {base_osc.extern + "Pfilterpar2"})
Qml::add_child(@anonymous823, @anonymous798)
Qml::add_child(@anonymous823, @anonymous799)
@db.connect_property(@anonymous800.properties["extern"], lambda {base_osc.extern + "Pfilterbeforews"})
@anonymous800.label = "pre/post"
Qml::add_child(@anonymous823, @anonymous800)
@db.connect_property(@anonymous801.properties["label"], lambda {"mag. type"})
Qml::add_child(@anonymous823, @anonymous801)
@db.connect_property(@anonymous802.properties["extern"], lambda {base_osc.extern + "Phmagtype" })
Qml::add_child(@anonymous823, @anonymous802)
@db.connect_property(@anonymous803.properties["extern"], lambda {base_osc.extern + "Pharmonicshift"})
Qml::add_child(@anonymous823, @anonymous803)

        @anonymous806.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hpack(l, selfBox, children)\n                ;end", "src/mruby-zest/example/ZynOscil.qml", 161)
@anonymous804.label = "R"
@anonymous804.whenValue = lambda {
                        $remote.action(base_osc.extern + "Pharmonicshift", 0.to_i)
                        base_osc.refresh
                    }
Qml::add_child(@anonymous806, @anonymous804)
@db.connect_property(@anonymous805.properties["extern"], lambda {base_osc.extern + "Pharmonicshiftfirst"})
@anonymous805.label = "pre/post"
Qml::add_child(@anonymous806, @anonymous805)
Qml::add_child(@anonymous823, @anonymous806)
@db.connect_property(@anonymous807.properties["label"], lambda {"adapt. harm."})
Qml::add_child(@anonymous823, @anonymous807)
@db.connect_property(@anonymous808.properties["extern"], lambda {base_osc.extern + "Padaptiveharmonics"})
Qml::add_child(@anonymous823, @anonymous808)
@db.connect_property(@anonymous809.properties["extern"], lambda {base_osc.extern + "Padaptiveharmonicspower"})
Qml::add_child(@anonymous823, @anonymous809)
@db.connect_property(@anonymous810.properties["extern"], lambda {base_osc.extern + "Padaptiveharmonicsbasefreq"})
Qml::add_child(@anonymous823, @anonymous810)
@db.connect_property(@anonymous811.properties["extern"], lambda {base_osc.extern + "Padaptiveharmonicspar"})
Qml::add_child(@anonymous823, @anonymous811)
@anonymous812.whenValue = lambda {base_osc.clear}
@anonymous812.label = "clear all"
@anonymous812.tooltip = "clear all harmonics"
Qml::add_child(@anonymous823, @anonymous812)
@anonymous813.label = "modulation"
Qml::add_child(@anonymous823, @anonymous813)
@db.connect_property(@anonymous814.properties["extern"], lambda {base_osc.extern + "Pmodulation"})
Qml::add_child(@anonymous823, @anonymous814)
@db.connect_property(@anonymous815.properties["extern"], lambda {base_osc.extern + "Pmodulationpar1"})
Qml::add_child(@anonymous823, @anonymous815)
@db.connect_property(@anonymous816.properties["extern"], lambda {base_osc.extern + "Pmodulationpar2"})
Qml::add_child(@anonymous823, @anonymous816)
@db.connect_property(@anonymous817.properties["extern"], lambda {base_osc.extern + "Pmodulationpar3"})
Qml::add_child(@anonymous823, @anonymous817)
Qml::add_child(@anonymous823, @anonymous818)
@db.connect_property(@anonymous819.properties["label"], lambda {"spectrum adj."})
Qml::add_child(@anonymous823, @anonymous819)
@db.connect_property(@anonymous820.properties["extern"], lambda {base_osc.extern + "Psatype"})
Qml::add_child(@anonymous823, @anonymous820)
@db.connect_property(@anonymous821.properties["extern"], lambda {base_osc.extern + "Psapar"})
Qml::add_child(@anonymous823, @anonymous821)
@anonymous822.label = "to sine"
@anonymous822.tooltip = "convert full waveform to sinasoidal components"
@anonymous822.whenValue = lambda {
                    $remote.action(base_osc.extern+"convert2sine")
                }
Qml::add_child(@anonymous823, @anonymous822)
Qml::add_child(@middle_panel, @anonymous823)
Qml::add_child(self, @middle_panel)
self.doRefresh = false
self.delayRefresh = nil

    #puts "ZynOscil, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynOscilMod < Qml::Widget
  attr_reader :basemod
  attr_reader :vis
  attr_reader :gen
  attr_reader :env
  attr_reader :footer
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children,
        [0.95,0.05])
    ;end
  def set_view();
        subsubview = root.get_view_pos(:subsubview)
        vs = [:amplitude, :frequency]
        if(!vs.include?(subsubview))
            subsubview = :amplitude
            root.set_view_pos(:subsubview, subsubview)
        end
        if(subsubview == :amplitude)
            footer.children[0].value = true
            gen.extern  = basemod.extern
            env.extern  = basemod.extern + "FMAmpEnvelope/"
            vis.extern  = basemod.extern + "FMAmpEnvelope/"
            gen.content = Qml::ZynAmpMod
            env.content = Qml::ZynAmpEnv
            vis.content = Qml::ZynEnvEdit
            vis.children[0].layoutOpts = Hash.new
            vis.children[0].layoutOpts[:main_width] = 0.8

            env.children[0].toggleable   = basemod.extern + "PFMAmpEnvelopeEnabled"
            env.children[0].whenModified = lambda {
                elm = vis.children[0]
                elm.refresh if elm.respond_to? :refresh
            }
        else
            footer.children[1].value = true
            gen.extern  = basemod.extern
            env.extern  = basemod.extern + "FMFreqEnvelope/"
            vis.extern  = basemod.extern + "FMFreqEnvelope/"
            gen.content = Qml::ZynFreqMod
            env.content = Qml::ZynFreqEnv
            vis.content = Qml::ZynEnvEdit
            vis.children[0].layoutOpts = Hash.new
            vis.children[0].layoutOpts[:main_width] = 0.8

            env.children[0].toggleable   = basemod.extern + "PFMFreqEnvelopeEnabled"
            env.children[0].whenModified = lambda {
                elm = vis.children[0]
                elm.refresh if elm.respond_to? :refresh
            }
        end
        db.update_values
    ;end
  def onSetup(old=nil);
        set_view
    ;end
end
class Qml::ZynOscilModRight < Qml::Widget
  attr_reader :base
  attr_reader :extmod
  attr_reader :ext
  attr_reader :phase_osc
  attr_reader :osc_wave
  attr_reader :extfm
  attr_reader :phase_mod
  attr_reader :mod_wave
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children,
            [0.15,0.2,0.65])
    ;end
  def onSetup(old=nil);
        vce     = root.get_view_pos(:voice)
        mapper  = [-1]
        names   = ["Normal"]
        names2  = ["Normal"]
        (0...vce).each do |i|
            mapper << i
            names  << "Oscil #{i+1}"
            names2 << "Mod   #{i+1}"
        end

        extfm.opt_vals = mapper
        extfm.options  = names
        extfm.extern   = base.extern + "PextFMoscil"
        ext.opt_vals   = mapper
        ext.options    = names
        ext.extern     = base.extern + "Pextoscil"
        extmod.opt_vals = mapper
        extmod.options  = names2
        extmod.extern   = base.extern + "PFMVoice"

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @base = self
@anonymous841 = Qml::Indent.new(database, ui_path+"1"+'/')
@anonymous840 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous838 = Qml::Selector.new(database, ui_path+"3"+'/')
@extmod = Qml::Selector.new(database, ui_path+"4"+'/')
@anonymous849 = Qml::Group.new(database, ui_path+"5"+'/')
@anonymous848 = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous842 = Qml::NumEntry.new(database, ui_path+"7"+'/')
@anonymous843 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous844 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous845 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous846 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous847 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous865 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous857 = Qml::Widget.new(database, ui_path+"14"+'/')
@anonymous855 = Qml::Group.new(database, ui_path+"15"+'/')
@anonymous852 = Qml::ParModuleRow.new(database, ui_path+"16"+'/')
@ext = Qml::Selector.new(database, ui_path+"17"+'/')
@phase_osc = Qml::Knob.new(database, ui_path+"18"+'/')
@anonymous854 = Qml::ParModuleRow.new(database, ui_path+"19"+'/')
@anonymous853 = Qml::Selector.new(database, ui_path+"20"+'/')
@osc_wave = Qml::WaveView.new(database, ui_path+"21"+'/')
@anonymous864 = Qml::Widget.new(database, ui_path+"22"+'/')
@anonymous862 = Qml::Group.new(database, ui_path+"23"+'/')
@anonymous860 = Qml::ParModuleRow.new(database, ui_path+"24"+'/')
@extfm = Qml::Selector.new(database, ui_path+"25"+'/')
@phase_mod = Qml::Knob.new(database, ui_path+"26"+'/')
@anonymous861 = Qml::Widget.new(database, ui_path+"27"+'/')
@mod_wave = Qml::WaveView.new(database, ui_path+"28"+'/')

    
context = Hash.new
context["base"] = self
context["extmod"] = @extmod
context["ext"] = @ext
context["phase_osc"] = @phase_osc
context["osc_wave"] = @osc_wave
context["extfm"] = @extfm
context["phase_mod"] = @phase_mod
context["mod_wave"] = @mod_wave

        @anonymous841.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 15)
@db.connect_property(@anonymous838.properties["extern"], lambda {base.extern + "PFMEnabled"})
@anonymous838.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous840, @anonymous838)
@extmod.label = "external modulator"
Qml::add_child(@anonymous840, @extmod)
Qml::add_child(@anonymous841, @anonymous840)
@anonymous841.pad = 10
Qml::add_child(self, @anonymous841)
@db.connect_property(@anonymous842.properties["extern"], lambda {base.extern + "Unison_size"})
Qml::add_child(@anonymous848, @anonymous842)
@db.connect_property(@anonymous843.properties["extern"], lambda {base.extern + "Unison_frequency_spread"})
Qml::add_child(@anonymous848, @anonymous843)
@db.connect_property(@anonymous844.properties["extern"], lambda {base.extern + "Unison_stereo_spread"})
Qml::add_child(@anonymous848, @anonymous844)
@db.connect_property(@anonymous845.properties["extern"], lambda {base.extern + "Unison_vibratto"})
Qml::add_child(@anonymous848, @anonymous845)
@db.connect_property(@anonymous846.properties["extern"], lambda {base.extern + "Unison_vibratto_speed"})
Qml::add_child(@anonymous848, @anonymous846)
@db.connect_property(@anonymous847.properties["extern"], lambda {base.extern + "Unison_invert_phase"})
Qml::add_child(@anonymous848, @anonymous847)
Qml::add_child(@anonymous849, @anonymous848)
@anonymous849.label = "unison"
@anonymous849.topSize = 0.2
@anonymous849.copyable = false
Qml::add_child(self, @anonymous849)

        @anonymous865.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 129)

        @anonymous857.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vfill(l, selfBox, children,\n                    [0.4,0.6])\n            ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 86)
@ext.whenValue = lambda {
                            off  = ext.opt_vals[ext.selected]
                            off  = root.get_view_pos(:voice) if off == -1
                            off  = root.get_view_pos(:voice) if off.nil?
                            ext1 = "/VoicePar#{off}/OscilSmp/waveform"
                            ext2 = "/VoicePar#{off}/Type"
                            osc_wave.noise  = path_simp(base.extern+"../") + ext2
                            osc_wave.extern = path_simp(base.extern+"../") + ext1
                            osc_wave.damage_self
                        }
@db.connect_property(@ext.properties["extern"], lambda {base.extern + "Pextoscil"})
Qml::add_child(@anonymous852, @ext)
@db.connect_property(@phase_osc.properties["extern"], lambda {base.extern + "Poscilphase"})
@phase_osc.whenValue = lambda {osc_wave.phase = phase_osc.value}
Qml::add_child(@anonymous852, @phase_osc)
@anonymous852.layoutOpts = []
@anonymous852.lsize = 0.4
Qml::add_child(@anonymous855, @anonymous852)
@db.connect_property(@anonymous853.properties["extern"], lambda {base.extern + "Type"})
Qml::add_child(@anonymous854, @anonymous853)
@anonymous854.lsize = 0.4
Qml::add_child(@anonymous855, @anonymous854)
@anonymous855.label = "vce osc"
@anonymous855.topSize = 0.15
@anonymous855.copyable = false
Qml::add_child(@anonymous857, @anonymous855)
@db.connect_property(@osc_wave.properties["noise"], lambda {base.extern + "Type"})
@db.connect_property(@osc_wave.properties["extern"], lambda {base.extern + "OscilSmp/waveform"})
@osc_wave.grid = false
Qml::add_child(@anonymous857, @osc_wave)
Qml::add_child(@anonymous865, @anonymous857)

        @anonymous864.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vfill(l, selfBox, children,\n                    [0.4,0.6])\n            ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 124)
@extfm.whenValue = lambda {
                            off = extfm.opt_vals[extfm.selected]
                            off = root.get_view_pos(:voice) if off == -1
                            off = root.get_view_pos(:voice) if off.nil?
                            ext = "/VoicePar#{off}/FMSmp/waveform"
                            mod_wave.extern = path_simp(base.extern+"../") + ext
                            mod_wave.damage_self
                        }
@db.connect_property(@extfm.properties["extern"], lambda {base.extern + "PextFMoscil"})
Qml::add_child(@anonymous860, @extfm)
@db.connect_property(@phase_mod.properties["extern"], lambda {base.extern+"PFMoscilphase"})
@phase_mod.whenValue = lambda {
                            mod_wave.phase = phase_mod.value
                        }
Qml::add_child(@anonymous860, @phase_mod)
Qml::add_child(@anonymous862, @anonymous860)
Qml::add_child(@anonymous862, @anonymous861)
@anonymous862.label = "mod osc"
@anonymous862.topSize = 0.15
@anonymous862.copyable = false
Qml::add_child(@anonymous864, @anonymous862)
@mod_wave.grid = false
@db.connect_property(@mod_wave.properties["extern"], lambda {base.extern + "FMSmp/waveform"})
Qml::add_child(@anonymous864, @mod_wave)
Qml::add_child(@anonymous865, @anonymous864)
Qml::add_child(self, @anonymous865)

    #puts "ZynOscilModRight, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynOscilMod
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @basemod = self
@anonymous833 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous831 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous827 = Qml::Widget.new(database, ui_path+"3"+'/')
@vis = Qml::Swappable.new(database, ui_path+"4"+'/')
@anonymous830 = Qml::Widget.new(database, ui_path+"5"+'/')
@gen = Qml::Swappable.new(database, ui_path+"6"+'/')
@env = Qml::Swappable.new(database, ui_path+"7"+'/')
@anonymous832 = Qml::ZynOscilModRight.new(database, ui_path+"8"+'/')
@footer = Qml::TabGroup.new(database, ui_path+"9"+'/')
@anonymous834 = Qml::TabButton.new(database, ui_path+"10"+'/')
@anonymous835 = Qml::TabButton.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["basemod"] = self
context["vis"] = @vis
context["gen"] = @gen
context["env"] = @env
context["footer"] = @footer

        @anonymous833.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.6,0.4])\n        ;end", "src/mruby-zest/example/ZynOscilMod.qml", 23)

        @anonymous831.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vfill(l, selfBox, children, [0.6,0.4])\n            ;end", "src/mruby-zest/example/ZynOscilMod.qml", 15)
Qml::add_child(@anonymous827, @vis)
Qml::add_child(@anonymous831, @anonymous827)

        @anonymous830.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hpack(l, selfBox, children)\n                ;end", "src/mruby-zest/example/ZynOscilMod.qml", 11)
Qml::add_child(@anonymous830, @gen)
Qml::add_child(@anonymous830, @env)
Qml::add_child(@anonymous831, @anonymous830)
Qml::add_child(@anonymous833, @anonymous831)
@db.connect_property(@anonymous832.properties["extern"], lambda {basemod.extern})
Qml::add_child(@anonymous833, @anonymous832)
Qml::add_child(self, @anonymous833)

        @footer.instance_eval("def set_tab(wid);\n            id = get_tab wid\n            vs = [:amplitude, :frequency]\n            root.set_view_pos(:subsubview, vs[id])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynOscilMod.qml", 40)
@anonymous834.label = "amplitude"
@anonymous834.highlight_pos = :top
Qml::add_child(@footer, @anonymous834)
@anonymous835.label = "frequency"
@anonymous835.highlight_pos = :top
Qml::add_child(@footer, @anonymous835)
Qml::add_child(self, @footer)

    #puts "ZynOscilMod, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynOscilModRight < Qml::Widget
  attr_reader :base
  attr_reader :extmod
  attr_reader :ext
  attr_reader :phase_osc
  attr_reader :osc_wave
  attr_reader :extfm
  attr_reader :phase_mod
  attr_reader :mod_wave
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children,
            [0.15,0.2,0.65])
    ;end
  def onSetup(old=nil);
        vce     = root.get_view_pos(:voice)
        mapper  = [-1]
        names   = ["Normal"]
        names2  = ["Normal"]
        (0...vce).each do |i|
            mapper << i
            names  << "Oscil #{i+1}"
            names2 << "Mod   #{i+1}"
        end

        extfm.opt_vals = mapper
        extfm.options  = names
        extfm.extern   = base.extern + "PextFMoscil"
        ext.opt_vals   = mapper
        ext.options    = names
        ext.extern     = base.extern + "Pextoscil"
        extmod.opt_vals = mapper
        extmod.options  = names2
        extmod.extern   = base.extern + "PFMVoice"

    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @base = self
@anonymous841 = Qml::Indent.new(database, ui_path+"1"+'/')
@anonymous840 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous838 = Qml::Selector.new(database, ui_path+"3"+'/')
@extmod = Qml::Selector.new(database, ui_path+"4"+'/')
@anonymous849 = Qml::Group.new(database, ui_path+"5"+'/')
@anonymous848 = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous842 = Qml::NumEntry.new(database, ui_path+"7"+'/')
@anonymous843 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous844 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous845 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous846 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous847 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous865 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous857 = Qml::Widget.new(database, ui_path+"14"+'/')
@anonymous855 = Qml::Group.new(database, ui_path+"15"+'/')
@anonymous852 = Qml::ParModuleRow.new(database, ui_path+"16"+'/')
@ext = Qml::Selector.new(database, ui_path+"17"+'/')
@phase_osc = Qml::Knob.new(database, ui_path+"18"+'/')
@anonymous854 = Qml::ParModuleRow.new(database, ui_path+"19"+'/')
@anonymous853 = Qml::Selector.new(database, ui_path+"20"+'/')
@osc_wave = Qml::WaveView.new(database, ui_path+"21"+'/')
@anonymous864 = Qml::Widget.new(database, ui_path+"22"+'/')
@anonymous862 = Qml::Group.new(database, ui_path+"23"+'/')
@anonymous860 = Qml::ParModuleRow.new(database, ui_path+"24"+'/')
@extfm = Qml::Selector.new(database, ui_path+"25"+'/')
@phase_mod = Qml::Knob.new(database, ui_path+"26"+'/')
@anonymous861 = Qml::Widget.new(database, ui_path+"27"+'/')
@mod_wave = Qml::WaveView.new(database, ui_path+"28"+'/')

    
context = Hash.new
context["base"] = self
context["extmod"] = @extmod
context["ext"] = @ext
context["phase_osc"] = @phase_osc
context["osc_wave"] = @osc_wave
context["extfm"] = @extfm
context["phase_mod"] = @phase_mod
context["mod_wave"] = @mod_wave

        @anonymous841.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 15)
@db.connect_property(@anonymous838.properties["extern"], lambda {base.extern + "PFMEnabled"})
@anonymous838.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous840, @anonymous838)
@extmod.label = "external modulator"
Qml::add_child(@anonymous840, @extmod)
Qml::add_child(@anonymous841, @anonymous840)
@anonymous841.pad = 10
Qml::add_child(self, @anonymous841)
@db.connect_property(@anonymous842.properties["extern"], lambda {base.extern + "Unison_size"})
Qml::add_child(@anonymous848, @anonymous842)
@db.connect_property(@anonymous843.properties["extern"], lambda {base.extern + "Unison_frequency_spread"})
Qml::add_child(@anonymous848, @anonymous843)
@db.connect_property(@anonymous844.properties["extern"], lambda {base.extern + "Unison_stereo_spread"})
Qml::add_child(@anonymous848, @anonymous844)
@db.connect_property(@anonymous845.properties["extern"], lambda {base.extern + "Unison_vibratto"})
Qml::add_child(@anonymous848, @anonymous845)
@db.connect_property(@anonymous846.properties["extern"], lambda {base.extern + "Unison_vibratto_speed"})
Qml::add_child(@anonymous848, @anonymous846)
@db.connect_property(@anonymous847.properties["extern"], lambda {base.extern + "Unison_invert_phase"})
Qml::add_child(@anonymous848, @anonymous847)
Qml::add_child(@anonymous849, @anonymous848)
@anonymous849.label = "unison"
@anonymous849.topSize = 0.2
@anonymous849.copyable = false
Qml::add_child(self, @anonymous849)

        @anonymous865.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 129)

        @anonymous857.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vfill(l, selfBox, children,\n                    [0.4,0.6])\n            ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 86)
@ext.whenValue = lambda {
                            off  = ext.opt_vals[ext.selected]
                            off  = root.get_view_pos(:voice) if off == -1
                            off  = root.get_view_pos(:voice) if off.nil?
                            ext1 = "/VoicePar#{off}/OscilSmp/waveform"
                            ext2 = "/VoicePar#{off}/Type"
                            osc_wave.noise  = path_simp(base.extern+"../") + ext2
                            osc_wave.extern = path_simp(base.extern+"../") + ext1
                            osc_wave.damage_self
                        }
@db.connect_property(@ext.properties["extern"], lambda {base.extern + "Pextoscil"})
Qml::add_child(@anonymous852, @ext)
@db.connect_property(@phase_osc.properties["extern"], lambda {base.extern + "Poscilphase"})
@phase_osc.whenValue = lambda {osc_wave.phase = phase_osc.value}
Qml::add_child(@anonymous852, @phase_osc)
@anonymous852.layoutOpts = []
@anonymous852.lsize = 0.4
Qml::add_child(@anonymous855, @anonymous852)
@db.connect_property(@anonymous853.properties["extern"], lambda {base.extern + "Type"})
Qml::add_child(@anonymous854, @anonymous853)
@anonymous854.lsize = 0.4
Qml::add_child(@anonymous855, @anonymous854)
@anonymous855.label = "vce osc"
@anonymous855.topSize = 0.15
@anonymous855.copyable = false
Qml::add_child(@anonymous857, @anonymous855)
@db.connect_property(@osc_wave.properties["noise"], lambda {base.extern + "Type"})
@db.connect_property(@osc_wave.properties["extern"], lambda {base.extern + "OscilSmp/waveform"})
@osc_wave.grid = false
Qml::add_child(@anonymous857, @osc_wave)
Qml::add_child(@anonymous865, @anonymous857)

        @anonymous864.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vfill(l, selfBox, children,\n                    [0.4,0.6])\n            ;end", "src/mruby-zest/example/ZynOscilModRight.qml", 124)
@extfm.whenValue = lambda {
                            off = extfm.opt_vals[extfm.selected]
                            off = root.get_view_pos(:voice) if off == -1
                            off = root.get_view_pos(:voice) if off.nil?
                            ext = "/VoicePar#{off}/FMSmp/waveform"
                            mod_wave.extern = path_simp(base.extern+"../") + ext
                            mod_wave.damage_self
                        }
@db.connect_property(@extfm.properties["extern"], lambda {base.extern + "PextFMoscil"})
Qml::add_child(@anonymous860, @extfm)
@db.connect_property(@phase_mod.properties["extern"], lambda {base.extern+"PFMoscilphase"})
@phase_mod.whenValue = lambda {
                            mod_wave.phase = phase_mod.value
                        }
Qml::add_child(@anonymous860, @phase_mod)
Qml::add_child(@anonymous862, @anonymous860)
Qml::add_child(@anonymous862, @anonymous861)
@anonymous862.label = "mod osc"
@anonymous862.topSize = 0.15
@anonymous862.copyable = false
Qml::add_child(@anonymous864, @anonymous862)
@mod_wave.grid = false
@db.connect_property(@mod_wave.properties["extern"], lambda {base.extern + "FMSmp/waveform"})
Qml::add_child(@anonymous864, @mod_wave)
Qml::add_child(@anonymous865, @anonymous864)
Qml::add_child(self, @anonymous865)

    #puts "ZynOscilModRight, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadAmp < Qml::Group
  attr_reader :box
  attr_reader :top
  attr_reader :bot
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@top = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous867 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous868 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous869 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous870 = Qml::Knob.new(database, ui_path+"5"+'/')
@bot = Qml::ParModuleRow.new(database, ui_path+"6"+'/')
@anonymous872 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous873 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous874 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous877 = Qml::Col.new(database, ui_path+"10"+'/')
@anonymous875 = Qml::Button.new(database, ui_path+"11"+'/')
@anonymous876 = Qml::Button.new(database, ui_path+"12"+'/')

    
context = Hash.new
context["box"] = self
context["top"] = @top
context["bot"] = @bot
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous867.properties["extern"], lambda {box.extern+"PVolume" })
Qml::add_child(@top, @anonymous867)
@db.connect_property(@anonymous868.properties["extern"], lambda {box.extern+"PAmpVelocityScaleFunction"})
Qml::add_child(@top, @anonymous868)
@db.connect_property(@anonymous869.properties["extern"], lambda {box.extern+"PPanning"})
Qml::add_child(@top, @anonymous869)
@db.connect_property(@anonymous870.properties["extern"], lambda {box.extern+"Fadein_adjustment"})
Qml::add_child(@top, @anonymous870)
Qml::add_child(self, @top)
@db.connect_property(@anonymous872.properties["extern"], lambda {box.extern+"PPunchStretch"})
Qml::add_child(@bot, @anonymous872)
@db.connect_property(@anonymous873.properties["extern"], lambda {box.extern+"PPunchStrength"})
Qml::add_child(@bot, @anonymous873)
@db.connect_property(@anonymous874.properties["extern"], lambda {box.extern+"PPunchTime"})
Qml::add_child(@bot, @anonymous874)
@anonymous875.label = "stereo"
@db.connect_property(@anonymous875.properties["extern"], lambda {box.extern+"Pstereo"})
Qml::add_child(@anonymous877, @anonymous875)
@anonymous876.label = "rnd grp"
Qml::add_child(@anonymous877, @anonymous876)
Qml::add_child(@bot, @anonymous877)
Qml::add_child(self, @bot)
self.label = "General"
self.copyable = false
self.whenModified = nil

    #puts "ZynPadAmp, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadFreq < Qml::Group
  attr_reader :box
  attr_property :whenModified
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @box = self
@anonymous883 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous880 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous881 = Qml::NumEntry.new(database, ui_path+"3"+'/')
@anonymous882 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous888 = Qml::ParModuleRow.new(database, ui_path+"5"+'/')
@anonymous884 = Qml::ToggleButton.new(database, ui_path+"6"+'/')
@anonymous885 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous886 = Qml::Selector.new(database, ui_path+"8"+'/')
@anonymous887 = Qml::NumEntry.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["box"] = self
Qml::prop_add(self, "whenModified")
@db.connect_property(@anonymous880.properties["extern"], lambda {box.extern + "PDetune" })
Qml::add_child(@anonymous883, @anonymous880)
@db.connect_property(@anonymous881.properties["extern"], lambda {box.extern + "octave" })
Qml::add_child(@anonymous883, @anonymous881)
@db.connect_property(@anonymous882.properties["extern"], lambda {box.extern + "Pbandwidth" })
Qml::add_child(@anonymous883, @anonymous882)
Qml::add_child(self, @anonymous883)
@db.connect_property(@anonymous884.properties["extern"], lambda {box.extern + "Pfixedfreq" })
Qml::add_child(@anonymous888, @anonymous884)
@db.connect_property(@anonymous885.properties["extern"], lambda {box.extern + "PfixedfreqET" })
Qml::add_child(@anonymous888, @anonymous885)
@db.connect_property(@anonymous886.properties["extern"], lambda {box.extern + "PDetuneType" })
Qml::add_child(@anonymous888, @anonymous886)
@db.connect_property(@anonymous887.properties["extern"], lambda {box.extern + "PCoarseDetune" })
Qml::add_child(@anonymous888, @anonymous887)
Qml::add_child(self, @anonymous888)
self.label = "General"
self.whenModified = nil

    #puts "ZynPadFreq, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadGlobal < Qml::Widget
  attr_reader :padglobal
  attr_reader :row1
  attr_reader :row2
  attr_reader :amp_gen
  attr_reader :amp_env
  attr_reader :amp_lfo
  attr_reader :footer
  attr_accessor :valueRef
  attr_property :filtertype
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.55, 0.4, 0.05])
    ;end
  def set_view();
        subsubview = root.get_view_pos(:subsubview)
        types = [:amplitude, :frequency, :filter]
        if(!types.include?(subsubview))
            subsubview = :amplitude
            root.set_view_pos(:subsubview, subsubview)
        end
        vis = root.get_view_pos(:vis)
        types = [:envelope, :lfo, :filter]
        if(!types.include?(vis))
            vis = :envelope
            root.set_view_pos(:vis, vis)
        end

        if(subsubview == :amplitude)
            set_amp(self.extern)
        elsif(subsubview == :frequency)
            set_freq(self.extern)
        elsif(subsubview == :filter)
            set_filter(self.extern)
        end

        if(vis == :lfo)
            set_vis_lfo(self.extern, subsubview)
        elsif(vis == :envelope)
            set_vis_env(self.extern, subsubview)
        elsif(vis == :filter)
            set_vis_filter(self.extern, subsubview)
        end
        db.update_values
    ;end
  def set_vis_lfo(ext, tab);
        e_  = {:filter    => "FilterLfo/",
               :amplitude => "AmpLfo/",
               :frequency => "FreqLfo/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::LfoVis
        row1.children[0].children[0].extern = amp_lfo.extern+"out"
    ;end
  def set_vis_env(ext, tab);
        e_  = {:filter    => "FilterEnvelope/",
               :amplitude => "AmpEnvelope/",
               :frequency => "FreqEnvelope/"}[tab]
        return if e_.nil?
        row1.extern  = ext + e_
        row1.content = Qml::ZynEnvEdit
        #self.children[0].children[0].extern = amp_env.extern+"out"
        amp_env.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_vis_filter(ext, dummy);
        row1.extern = ext + "GlobalFilter/"
        if(self.filtertype == :formant)
            row1.content = Qml::ZynFormant
        else
            row1.content = Qml::VisFilter
            row1.children[0].extern = ext + "GlobalFilter/response"
        end
        amp_gen.children[0].whenModified = lambda {
            elm = row1.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_amp(base);
        footer.children[0].value = true
        amp_gen.extern  = base
        amp_env.extern  = base + "AmpEnvelope/"
        amp_lfo.extern  = base + "AmpLfo/"
        amp_gen.content = Qml::ZynPadAmp
        amp_env.content = Qml::ZynAmpEnv
        amp_lfo.content = Qml::ZynLFO
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env, :amp)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :amp)}
    ;end
  def set_freq(base);
        footer.children[1].value = true
        amp_gen.extern  = base
        amp_env.extern  = base + "FreqEnvelope/"
        amp_lfo.extern  = base + "FreqLfo/"
        amp_gen.content = Qml::ZynPadFreq
        amp_env.content = Qml::ZynFreqEnv
        amp_lfo.content = Qml::ZynLFO
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env, :freq)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo, :freq)}
    ;end
  def set_filter(base);
        footer.children[2].value = true
        amp_gen.extern  = base + "GlobalFilter/"
        amp_env.extern  = base + "FilterEnvelope/"
        amp_lfo.extern  = base + "FilterLfo/"
        amp_gen.content = Qml::ZynAnalogFilter
        amp_env.content = Qml::ZynFilterEnv
        amp_lfo.content = Qml::ZynLFO
        amp_gen.children[0].whenClick = lambda {row1.setDataVis(:filter, :filter)}
        amp_env.children[0].whenClick = lambda {row1.setDataVis(:env,    :filter)}
        amp_lfo.children[0].whenClick = lambda {row1.setDataVis(:lfo,    :filter)}
    ;end
  def onSetup(old=nil);
        if(padglobal.valueRef.nil?)
            path = padglobal.extern + "GlobalFilter/Pcategory"
            padglobal.valueRef = OSC::RemoteParam.new($remote, path)
            padglobal.valueRef.mode = :full
            padglobal.valueRef.callback = lambda {|x|
                padglobal.filtertype = [:analog, :formant, :statevar][x]
            }
        end
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @padglobal = self
@row1 = Qml::Swappable.new(database, ui_path+"1"+'/')
@row2 = Qml::Widget.new(database, ui_path+"2"+'/')
@amp_gen = Qml::Swappable.new(database, ui_path+"3"+'/')
@amp_env = Qml::Swappable.new(database, ui_path+"4"+'/')
@amp_lfo = Qml::Swappable.new(database, ui_path+"5"+'/')
@footer = Qml::Widget.new(database, ui_path+"6"+'/')
@anonymous895 = Qml::TabButton.new(database, ui_path+"7"+'/')
@anonymous896 = Qml::TabButton.new(database, ui_path+"8"+'/')
@anonymous897 = Qml::TabButton.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["padglobal"] = self
context["row1"] = @row1
context["row2"] = @row2
context["amp_gen"] = @amp_gen
context["amp_env"] = @amp_env
context["amp_lfo"] = @amp_lfo
context["footer"] = @footer
Qml::prop_add(self, "filtertype")

        @row1.instance_eval("def setDataVis(type, tab);\n            root.set_view_pos(:vis, type)\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynPadGlobal.qml", 16)
@db.connect_property(@row1.properties["content"], lambda {Qml::LfoVis})
Qml::add_child(self, @row1)

        @row2.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynPadGlobal.qml", 24)
Qml::add_child(@row2, @amp_gen)
Qml::add_child(@row2, @amp_env)
Qml::add_child(@row2, @amp_lfo)
Qml::add_child(self, @row2)

        @footer.instance_eval("def layout(l, selfBox);\n            Draw::Layout::tabpack(l, selfBox, self)\n        ;end", "src/mruby-zest/example/ZynPadGlobal.qml", 36)

        @footer.instance_eval("def setTab(id);\n            (0..2).each do |ch_id|\n                children[ch_id].value = (ch_id == id)\n                children[ch_id].damage_self\n            end\n            mapping = {0 => :amplitude,\n                       1 => :frequency,\n                       2 => :filter}\n            root.set_view_pos(:subsubview, mapping[id])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynPadGlobal.qml", 40)
@anonymous895.label = "amplitude"
@anonymous895.whenClick = lambda {footer.setTab(0)}
@anonymous895.highlight_pos = :top
Qml::add_child(@footer, @anonymous895)
@anonymous896.label = "frequency"
@anonymous896.whenClick = lambda {footer.setTab(1)}
@anonymous896.highlight_pos = :top
Qml::add_child(@footer, @anonymous896)
@anonymous897.label = "filter"
@anonymous897.whenClick = lambda {footer.setTab(2)}
@anonymous897.highlight_pos = :top
Qml::add_child(@footer, @anonymous897)
Qml::add_child(self, @footer)
self.valueRef = nil
self.filtertype = nil

    #puts "ZynPadGlobal, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadHarmonics < Qml::Widget
  attr_reader :base
  attr_reader :nhr
  attr_reader :visharm
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0.0, 0.00, 1.0, 0.17)
        children[1].fixed(l, selfBox, 0.0, 0.17, 1.0, 0.50)
        children[2].fixed(l, selfBox, 0.0, 0.67, 1.0, 0.33)
        selfBox
    ;end
end
class Qml::ZynPadOvertone < Qml::Widget
  attr_reader :overtone
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  def layout(l, selfBox);
        pad = 4
        l.fixed_long(children[0], selfBox, 0, 0, 1, 1,
            pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @overtone = self
@anonymous930 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous918 = Qml::TextBox.new(database, ui_path+"2"+'/')
@anonymous919 = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous920 = Qml::HSlider.new(database, ui_path+"4"+'/')
@anonymous921 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous922 = Qml::Widget.new(database, ui_path+"6"+'/')
@anonymous923 = Qml::Widget.new(database, ui_path+"7"+'/')
@anonymous924 = Qml::TextBox.new(database, ui_path+"8"+'/')
@anonymous925 = Qml::TextBox.new(database, ui_path+"9"+'/')
@anonymous926 = Qml::HSlider.new(database, ui_path+"10"+'/')
@anonymous927 = Qml::Selector.new(database, ui_path+"11"+'/')
@anonymous928 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous929 = Qml::HSlider.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["overtone"] = self
@anonymous930.instance_eval do
def overtone=(k);@overtone=k;end
def overtone;    @overtone;  end
end
Qml::context_apply(context, @anonymous930)

        @anonymous930.instance_eval("def class_name(); \"overtone\" ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 57)

        @anonymous930.instance_eval("def layout(l, selfBox);\n            Draw::Layout::grid(l, selfBox, children, 4, 3, 1, 2)\n        ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 58)

        @anonymous930.instance_eval("def onSetup(old=nil);\n            cb_ = lambda {overtone.cb}\n            children.each do |ch|\n                ch.whenValue = cb_ if ch.respond_to? :whenValue\n            end\n        ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 62)
@db.connect_property(@anonymous918.properties["label"], lambda {"overtone pos."})
@anonymous918.height = 0.9
Qml::add_child(@anonymous930, @anonymous918)
@db.connect_property(@anonymous919.properties["extern"], lambda {overtone.extern + "Pbandwidth"})
@anonymous919.height = 0.8
Qml::add_child(@anonymous930, @anonymous919)
@db.connect_property(@anonymous920.properties["extern"], lambda {overtone.extern + "Phrpos.par3"})
@anonymous920.height = 0.8
Qml::add_child(@anonymous930, @anonymous920)
@db.connect_property(@anonymous921.properties["extern"], lambda {overtone.extern + "Phrpos.type"})
@anonymous921.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous930, @anonymous921)
Qml::add_child(@anonymous930, @anonymous922)
Qml::add_child(@anonymous930, @anonymous923)
@anonymous924.label = "spectral mode"
@anonymous924.height = 0.9
Qml::add_child(@anonymous930, @anonymous924)
@db.connect_property(@anonymous925.properties["label"], lambda {"bw. scale"})
@anonymous925.height = 0.9
Qml::add_child(@anonymous930, @anonymous925)
@db.connect_property(@anonymous926.properties["extern"], lambda {overtone.extern + "Phrpos.par1"})
@anonymous926.height = 0.8
Qml::add_child(@anonymous930, @anonymous926)
@anonymous927.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous927.properties["extern"], lambda {overtone.extern + "Pmode"})
Qml::add_child(@anonymous930, @anonymous927)
@anonymous928.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous928.properties["extern"], lambda {overtone.extern + "Pbwscale"})
Qml::add_child(@anonymous930, @anonymous928)
@db.connect_property(@anonymous929.properties["extern"], lambda {overtone.extern + "Phrpos.par2"})
@anonymous929.height = 0.8
Qml::add_child(@anonymous930, @anonymous929)
Qml::add_child(self, @anonymous930)
self.whenValue = nil

    #puts "ZynPadOvertone, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadProfile < Qml::Widget
  attr_reader :padprofile
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        pad = 4
        children[0].fixed_long(l, selfBox, 0, 0, 1, 1,
            pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @padprofile = self
@anonymous948 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous932 = Qml::TextBox.new(database, ui_path+"2"+'/')
@anonymous933 = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous934 = Qml::Selector.new(database, ui_path+"4"+'/')
@anonymous935 = Qml::HSlider.new(database, ui_path+"5"+'/')
@anonymous936 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous937 = Qml::HSlider.new(database, ui_path+"7"+'/')
@anonymous938 = Qml::ToggleButton.new(database, ui_path+"8"+'/')
@anonymous939 = Qml::HSlider.new(database, ui_path+"9"+'/')
@anonymous940 = Qml::TextBox.new(database, ui_path+"10"+'/')
@anonymous941 = Qml::HSlider.new(database, ui_path+"11"+'/')
@anonymous942 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous943 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous944 = Qml::TextBox.new(database, ui_path+"14"+'/')
@anonymous945 = Qml::HSlider.new(database, ui_path+"15"+'/')
@anonymous946 = Qml::Selector.new(database, ui_path+"16"+'/')
@anonymous947 = Qml::HSlider.new(database, ui_path+"17"+'/')

    
context = Hash.new
context["padprofile"] = self
@anonymous948.instance_eval do
def padprofile=(k);@padprofile=k;end
def padprofile;    @padprofile;  end
end
Qml::context_apply(context, @anonymous948)

        @anonymous948.instance_eval("def onSetup();\n            cb_ = lambda {padprofile.cb}\n            children.each do |x|\n                x.layoutOpts = [:no_constraint]\n                x.whenValue  = cb_ if x.respond_to? :whenValue\n            end\n        ;end", "src/mruby-zest/example/ZynPadProfile.qml", 9)

        @anonymous948.instance_eval("def class_name(); \"overtone\" ;end", "src/mruby-zest/example/ZynPadProfile.qml", 46)

        @anonymous948.instance_eval("def layout(l, selfBox);\n            Draw::Layout::grid(l, selfBox, children, 8, 2, 1, 2)\n        ;end", "src/mruby-zest/example/ZynPadProfile.qml", 47)
@anonymous932.label = "base type"
Qml::add_child(@anonymous948, @anonymous932)
@db.connect_property(@anonymous933.properties["extern"], lambda {padprofile.extern + "Php.modulator.par1"})
Qml::add_child(@anonymous948, @anonymous933)
@db.connect_property(@anonymous934.properties["extern"], lambda {padprofile.extern + "Php.base.type"})
Qml::add_child(@anonymous948, @anonymous934)
@db.connect_property(@anonymous935.properties["extern"], lambda {padprofile.extern + "Php.modulator.freq"})
Qml::add_child(@anonymous948, @anonymous935)
@db.connect_property(@anonymous936.properties["extern"], lambda {padprofile.extern + "Php.onehalf"})
Qml::add_child(@anonymous948, @anonymous936)
@db.connect_property(@anonymous937.properties["extern"], lambda {padprofile.extern + "Php.width"})
Qml::add_child(@anonymous948, @anonymous937)
@anonymous938.label = "autoscale"
@db.connect_property(@anonymous938.properties["extern"], lambda {padprofile.extern + "Php.autoscale"})
Qml::add_child(@anonymous948, @anonymous938)
@db.connect_property(@anonymous939.properties["extern"], lambda {padprofile.extern + "Php.base.par1"})
Qml::add_child(@anonymous948, @anonymous939)
@db.connect_property(@anonymous940.properties["label"], lambda {"amp. mlt."})
Qml::add_child(@anonymous948, @anonymous940)
@db.connect_property(@anonymous941.properties["extern"], lambda {padprofile.extern + "Php.freqmult"})
Qml::add_child(@anonymous948, @anonymous941)
@db.connect_property(@anonymous942.properties["extern"], lambda {padprofile.extern + "Php.amp.type"})
Qml::add_child(@anonymous948, @anonymous942)
Qml::add_child(@anonymous948, @anonymous943)
@anonymous944.bg = nil
@db.connect_property(@anonymous944.properties["label"], lambda {"amp. mode"})
Qml::add_child(@anonymous948, @anonymous944)
@db.connect_property(@anonymous945.properties["extern"], lambda {padprofile.extern + "Php.amp.par1"})
Qml::add_child(@anonymous948, @anonymous945)
@db.connect_property(@anonymous946.properties["extern"], lambda {padprofile.extern + "Php.amp.mode"})
Qml::add_child(@anonymous948, @anonymous946)
@db.connect_property(@anonymous947.properties["extern"], lambda {padprofile.extern + "Php.amp.par2"})
Qml::add_child(@anonymous948, @anonymous947)
Qml::add_child(self, @anonymous948)
self.whenValue = nil

    #puts "ZynPadProfile, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadHarmonics
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @base = self
@anonymous902 = Qml::Widget.new(database, ui_path+"1"+'/')
@nhr = Qml::HarmonicView.new(database, ui_path+"2"+'/')
@anonymous901 = Qml::ZynPadOvertone.new(database, ui_path+"3"+'/')
@anonymous906 = Qml::Widget.new(database, ui_path+"4"+'/')
@visharm = Qml::VisHarmonic.new(database, ui_path+"5"+'/')
@anonymous905 = Qml::Group.new(database, ui_path+"6"+'/')
@anonymous904 = Qml::ZynPadProfile.new(database, ui_path+"7"+'/')
@anonymous916 = Qml::Widget.new(database, ui_path+"8"+'/')
@anonymous907 = Qml::WaveView.new(database, ui_path+"9"+'/')
@anonymous915 = Qml::Group.new(database, ui_path+"10"+'/')
@anonymous911 = Qml::ParModuleRow.new(database, ui_path+"11"+'/')
@anonymous908 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous909 = Qml::Selector.new(database, ui_path+"13"+'/')
@anonymous910 = Qml::NumEntry.new(database, ui_path+"14"+'/')
@anonymous914 = Qml::Col.new(database, ui_path+"15"+'/')
@anonymous912 = Qml::Selector.new(database, ui_path+"16"+'/')
@anonymous913 = Qml::Text.new(database, ui_path+"17"+'/')

    
context = Hash.new
context["base"] = self
context["nhr"] = @nhr
context["visharm"] = @visharm

        @anonymous902.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0.0, 0.0, 0.6, 1.0)\n            children[1].fixed(l, selfBox, 0.6, 0.0, 0.4, 1.0)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynPadHarmonics.qml", 14)
@nhr.type = :pad
@db.connect_property(@nhr.properties["extern"], lambda {base.extern})
Qml::add_child(@anonymous902, @nhr)
@db.connect_property(@anonymous901.properties["extern"], lambda {base.extern})
@anonymous901.whenValue = lambda { nhr.refresh }
Qml::add_child(@anonymous902, @anonymous901)
Qml::add_child(self, @anonymous902)

        @anonymous906.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0.0, 0.0, 0.7, 1.0)\n            children[1].fixed(l, selfBox, 0.7, 0.0, 0.3, 1.0)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynPadHarmonics.qml", 35)
@db.connect_property(@visharm.properties["extern"], lambda {base.extern + "harmonic_profile"})
Qml::add_child(@anonymous906, @visharm)
@db.connect_property(@anonymous904.properties["extern"], lambda {base.extern})
@anonymous904.whenValue = lambda { visharm.refresh }
Qml::add_child(@anonymous905, @anonymous904)
@anonymous905.label = "harmonic profile"
@anonymous905.topSize = 0.09
@anonymous905.copyable = false
Qml::add_child(@anonymous906, @anonymous905)
Qml::add_child(self, @anonymous906)

        @anonymous916.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0.0, 0.0, 0.7, 1.0)\n            children[1].fixed(l, selfBox, 0.7, 0.0, 0.3, 1.0)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynPadHarmonics.qml", 70)
@db.connect_property(@anonymous907.properties["extern"], lambda {base.extern + "oscilgen/waveform"})
Qml::add_child(@anonymous916, @anonymous907)
@db.connect_property(@anonymous908.properties["extern"], lambda {base.extern + "Pquality.basenote"})
Qml::add_child(@anonymous911, @anonymous908)
@db.connect_property(@anonymous909.properties["extern"], lambda {base.extern + "Pquality.smpoct"})
Qml::add_child(@anonymous911, @anonymous909)
@db.connect_property(@anonymous910.properties["extern"], lambda {base.extern + "Pquality.oct"})
@anonymous910.label = "octaves"
Qml::add_child(@anonymous911, @anonymous910)
@anonymous911.lsize = 0.4
Qml::add_child(@anonymous915, @anonymous911)
@db.connect_property(@anonymous912.properties["extern"], lambda {base.extern + "Pquality.samplesize"})
@anonymous912.layoutOpts = [:long_mode]
Qml::add_child(@anonymous914, @anonymous912)
@anonymous913.label = "sample size"
Qml::add_child(@anonymous914, @anonymous913)
Qml::add_child(@anonymous915, @anonymous914)
@anonymous915.topSize = 0.15
@anonymous915.label = "harmonic content"
@anonymous915.copyable = false
Qml::add_child(@anonymous916, @anonymous915)
Qml::add_child(self, @anonymous916)
self.extern = "/part0/kit0/padpars/"

    #puts "ZynPadHarmonics, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadOvertone < Qml::Widget
  attr_reader :overtone
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  def draw(vg);
        vg.path do |v|
            v.rect(0,0,w,h)
            paint = v.linear_gradient(0,0,0,h,
            Theme::InnerGrad1, Theme::InnerGrad2)
            v.fill_paint paint
            v.fill
            v.stroke_color color(:black)
            v.stroke_width 1.0
            v.stroke
        end
    ;end
  def layout(l, selfBox);
        pad = 4
        l.fixed_long(children[0], selfBox, 0, 0, 1, 1,
            pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @overtone = self
@anonymous930 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous918 = Qml::TextBox.new(database, ui_path+"2"+'/')
@anonymous919 = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous920 = Qml::HSlider.new(database, ui_path+"4"+'/')
@anonymous921 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous922 = Qml::Widget.new(database, ui_path+"6"+'/')
@anonymous923 = Qml::Widget.new(database, ui_path+"7"+'/')
@anonymous924 = Qml::TextBox.new(database, ui_path+"8"+'/')
@anonymous925 = Qml::TextBox.new(database, ui_path+"9"+'/')
@anonymous926 = Qml::HSlider.new(database, ui_path+"10"+'/')
@anonymous927 = Qml::Selector.new(database, ui_path+"11"+'/')
@anonymous928 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous929 = Qml::HSlider.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["overtone"] = self
@anonymous930.instance_eval do
def overtone=(k);@overtone=k;end
def overtone;    @overtone;  end
end
Qml::context_apply(context, @anonymous930)

        @anonymous930.instance_eval("def class_name(); \"overtone\" ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 57)

        @anonymous930.instance_eval("def layout(l, selfBox);\n            Draw::Layout::grid(l, selfBox, children, 4, 3, 1, 2)\n        ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 58)

        @anonymous930.instance_eval("def onSetup(old=nil);\n            cb_ = lambda {overtone.cb}\n            children.each do |ch|\n                ch.whenValue = cb_ if ch.respond_to? :whenValue\n            end\n        ;end", "src/mruby-zest/example/ZynPadOvertone.qml", 62)
@db.connect_property(@anonymous918.properties["label"], lambda {"overtone pos."})
@anonymous918.height = 0.9
Qml::add_child(@anonymous930, @anonymous918)
@db.connect_property(@anonymous919.properties["extern"], lambda {overtone.extern + "Pbandwidth"})
@anonymous919.height = 0.8
Qml::add_child(@anonymous930, @anonymous919)
@db.connect_property(@anonymous920.properties["extern"], lambda {overtone.extern + "Phrpos.par3"})
@anonymous920.height = 0.8
Qml::add_child(@anonymous930, @anonymous920)
@db.connect_property(@anonymous921.properties["extern"], lambda {overtone.extern + "Phrpos.type"})
@anonymous921.layoutOpts = [:no_constraint]
Qml::add_child(@anonymous930, @anonymous921)
Qml::add_child(@anonymous930, @anonymous922)
Qml::add_child(@anonymous930, @anonymous923)
@anonymous924.label = "spectral mode"
@anonymous924.height = 0.9
Qml::add_child(@anonymous930, @anonymous924)
@db.connect_property(@anonymous925.properties["label"], lambda {"bw. scale"})
@anonymous925.height = 0.9
Qml::add_child(@anonymous930, @anonymous925)
@db.connect_property(@anonymous926.properties["extern"], lambda {overtone.extern + "Phrpos.par1"})
@anonymous926.height = 0.8
Qml::add_child(@anonymous930, @anonymous926)
@anonymous927.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous927.properties["extern"], lambda {overtone.extern + "Pmode"})
Qml::add_child(@anonymous930, @anonymous927)
@anonymous928.layoutOpts = [:no_constraint]
@db.connect_property(@anonymous928.properties["extern"], lambda {overtone.extern + "Pbwscale"})
Qml::add_child(@anonymous930, @anonymous928)
@db.connect_property(@anonymous929.properties["extern"], lambda {overtone.extern + "Phrpos.par2"})
@anonymous929.height = 0.8
Qml::add_child(@anonymous930, @anonymous929)
Qml::add_child(self, @anonymous930)
self.whenValue = nil

    #puts "ZynPadOvertone, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadProfile < Qml::Widget
  attr_reader :padprofile
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  def layout(l, selfBox);
        pad = 4
        children[0].fixed_long(l, selfBox, 0, 0, 1, 1,
            pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @padprofile = self
@anonymous948 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous932 = Qml::TextBox.new(database, ui_path+"2"+'/')
@anonymous933 = Qml::HSlider.new(database, ui_path+"3"+'/')
@anonymous934 = Qml::Selector.new(database, ui_path+"4"+'/')
@anonymous935 = Qml::HSlider.new(database, ui_path+"5"+'/')
@anonymous936 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous937 = Qml::HSlider.new(database, ui_path+"7"+'/')
@anonymous938 = Qml::ToggleButton.new(database, ui_path+"8"+'/')
@anonymous939 = Qml::HSlider.new(database, ui_path+"9"+'/')
@anonymous940 = Qml::TextBox.new(database, ui_path+"10"+'/')
@anonymous941 = Qml::HSlider.new(database, ui_path+"11"+'/')
@anonymous942 = Qml::Selector.new(database, ui_path+"12"+'/')
@anonymous943 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous944 = Qml::TextBox.new(database, ui_path+"14"+'/')
@anonymous945 = Qml::HSlider.new(database, ui_path+"15"+'/')
@anonymous946 = Qml::Selector.new(database, ui_path+"16"+'/')
@anonymous947 = Qml::HSlider.new(database, ui_path+"17"+'/')

    
context = Hash.new
context["padprofile"] = self
@anonymous948.instance_eval do
def padprofile=(k);@padprofile=k;end
def padprofile;    @padprofile;  end
end
Qml::context_apply(context, @anonymous948)

        @anonymous948.instance_eval("def onSetup();\n            cb_ = lambda {padprofile.cb}\n            children.each do |x|\n                x.layoutOpts = [:no_constraint]\n                x.whenValue  = cb_ if x.respond_to? :whenValue\n            end\n        ;end", "src/mruby-zest/example/ZynPadProfile.qml", 9)

        @anonymous948.instance_eval("def class_name(); \"overtone\" ;end", "src/mruby-zest/example/ZynPadProfile.qml", 46)

        @anonymous948.instance_eval("def layout(l, selfBox);\n            Draw::Layout::grid(l, selfBox, children, 8, 2, 1, 2)\n        ;end", "src/mruby-zest/example/ZynPadProfile.qml", 47)
@anonymous932.label = "base type"
Qml::add_child(@anonymous948, @anonymous932)
@db.connect_property(@anonymous933.properties["extern"], lambda {padprofile.extern + "Php.modulator.par1"})
Qml::add_child(@anonymous948, @anonymous933)
@db.connect_property(@anonymous934.properties["extern"], lambda {padprofile.extern + "Php.base.type"})
Qml::add_child(@anonymous948, @anonymous934)
@db.connect_property(@anonymous935.properties["extern"], lambda {padprofile.extern + "Php.modulator.freq"})
Qml::add_child(@anonymous948, @anonymous935)
@db.connect_property(@anonymous936.properties["extern"], lambda {padprofile.extern + "Php.onehalf"})
Qml::add_child(@anonymous948, @anonymous936)
@db.connect_property(@anonymous937.properties["extern"], lambda {padprofile.extern + "Php.width"})
Qml::add_child(@anonymous948, @anonymous937)
@anonymous938.label = "autoscale"
@db.connect_property(@anonymous938.properties["extern"], lambda {padprofile.extern + "Php.autoscale"})
Qml::add_child(@anonymous948, @anonymous938)
@db.connect_property(@anonymous939.properties["extern"], lambda {padprofile.extern + "Php.base.par1"})
Qml::add_child(@anonymous948, @anonymous939)
@db.connect_property(@anonymous940.properties["label"], lambda {"amp. mlt."})
Qml::add_child(@anonymous948, @anonymous940)
@db.connect_property(@anonymous941.properties["extern"], lambda {padprofile.extern + "Php.freqmult"})
Qml::add_child(@anonymous948, @anonymous941)
@db.connect_property(@anonymous942.properties["extern"], lambda {padprofile.extern + "Php.amp.type"})
Qml::add_child(@anonymous948, @anonymous942)
Qml::add_child(@anonymous948, @anonymous943)
@anonymous944.bg = nil
@db.connect_property(@anonymous944.properties["label"], lambda {"amp. mode"})
Qml::add_child(@anonymous948, @anonymous944)
@db.connect_property(@anonymous945.properties["extern"], lambda {padprofile.extern + "Php.amp.par1"})
Qml::add_child(@anonymous948, @anonymous945)
@db.connect_property(@anonymous946.properties["extern"], lambda {padprofile.extern + "Php.amp.mode"})
Qml::add_child(@anonymous948, @anonymous946)
@db.connect_property(@anonymous947.properties["extern"], lambda {padprofile.extern + "Php.amp.par2"})
Qml::add_child(@anonymous948, @anonymous947)
Qml::add_child(self, @anonymous948)
self.whenValue = nil

    #puts "ZynPadProfile, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPadSynth < Qml::Widget
  attr_reader :center
  attr_reader :header
  attr_reader :appl
  attr_reader :copy
  attr_reader :paste
  attr_reader :swap
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.05, 0.95])
    ;end
  def apply();
        $remote.action(center.extern+"prepare")
    ;end
  def get_voice(); root.get_view_pos(:voice) ;end
  def get_part(); root.get_view_pos(:part)  ;end
  def get_kit(); root.get_view_pos(:kit)   ;end
  def onSetup(old=nil);
        return if swap.content.nil?
        set_view
    ;end
  def set_view();
        subview = root.get_view_pos(:subview)

        mapping = {:harmonics   => Qml::ZynPadHarmonics,
                   :oscil       => Qml::ZynOscil,
                   :global_pad  => Qml::ZynPadGlobal}
        base = center.extern
        ext     = {:harmonics  => "",
                   :oscil      => "oscilgen/",
                   :global_pad => ""}
        tabid   = {:harmonics  => 0,
                   :oscil      => 1,
                   :global_pad => 2}
        if(!mapping.include?(subview))
            subview = :oscil
            root.set_view_pos(:subview, :oscil)
        end


        copy.extern  = base + ext[subview]
        paste.extern = base + ext[subview]
        swap.extern  = base + ext[subview]
        swap.content = mapping[subview]
        header.children[tabid[subview]].value = true
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @center = self
@header = Qml::TabGroup.new(database, ui_path+"1"+'/')
@anonymous950 = Qml::TabButton.new(database, ui_path+"2"+'/')
@anonymous951 = Qml::TabButton.new(database, ui_path+"3"+'/')
@anonymous952 = Qml::TabButton.new(database, ui_path+"4"+'/')
@appl = Qml::ApplyButton.new(database, ui_path+"5"+'/')
@copy = Qml::CopyButton.new(database, ui_path+"6"+'/')
@paste = Qml::PasteButton.new(database, ui_path+"7"+'/')
@swap = Qml::Swappable.new(database, ui_path+"8"+'/')

    
context = Hash.new
context["center"] = self
context["header"] = @header
@header.instance_eval do
def appl=(k);@appl=k;end
def appl;    @appl;  end
end
context["appl"] = @appl
context["copy"] = @copy
context["paste"] = @paste
context["swap"] = @swap
Qml::context_apply(context, @header)

        @header.instance_eval("def layout(l, selfBox);\n            selfBox = Draw::Layout::tabpack(l, selfBox, self, appl)\n        ;end", "src/mruby-zest/example/ZynPadSynth.qml", 34)

        @header.instance_eval("def set_tab(wid);\n            selected = get_tab wid\n\n            #Define a mapping from tabs to values\n            mapping = {0 => :harmonics,\n                       1 => :oscil,\n                       2 => :global_pad}\n            root.set_view_pos(:subview, mapping[selected])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynPadSynth.qml", 38)
@anonymous950.label = "harmonic structure"
Qml::add_child(@header, @anonymous950)
@anonymous951.label = "oscillator"
Qml::add_child(@header, @anonymous951)
@db.connect_property(@anonymous952.properties["label"], lambda {"envelopes & lfos"})
Qml::add_child(@header, @anonymous952)
@appl.layoutOpts = [:no_constraint]
@appl.label = "   apply"
@db.connect_property(@appl.properties["extern"], lambda {center.extern + "needPrepare"})
@appl.whenValue = lambda {center.apply() }
Qml::add_child(@header, @appl)
@db.connect_property(@copy.properties["extern"], lambda {center.extern})
Qml::add_child(@header, @copy)
@db.connect_property(@paste.properties["extern"], lambda {center.extern})
Qml::add_child(@header, @paste)
Qml::add_child(self, @header)
Qml::add_child(self, @swap)

    #puts "ZynPadSynth, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPart < Qml::Widget
  attr_reader :pset
  def class_name(); "part" ;end
  def layout(l, selfBox);
        Draw::Layout::hfill(l, selfBox, children, [0.2,0.4,0.4])
    ;end
end
class Qml::ZynPortamento < Qml::Group
  attr_reader :port
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @port = self
@anonymous1000 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous996 = Qml::Col.new(database, ui_path+"2"+'/')
@anonymous994 = Qml::ToggleButton.new(database, ui_path+"3"+'/')
@anonymous995 = Qml::ToggleButton.new(database, ui_path+"4"+'/')
@anonymous997 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous998 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous999 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1004 = Qml::ParModuleRow.new(database, ui_path+"8"+'/')
@anonymous1001 = Qml::ToggleButton.new(database, ui_path+"9"+'/')
@anonymous1002 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous1003 = Qml::Knob.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["port"] = self
@db.connect_property(@anonymous994.properties["extern"], lambda {port.extern+"portamento.receive"})
Qml::add_child(@anonymous996, @anonymous994)
@db.connect_property(@anonymous995.properties["extern"], lambda {port.extern+"portamento.pitchthreshtype"})
Qml::add_child(@anonymous996, @anonymous995)
Qml::add_child(@anonymous1000, @anonymous996)
@db.connect_property(@anonymous997.properties["extern"], lambda {port.extern+"portamento.pitchthresh"})
Qml::add_child(@anonymous1000, @anonymous997)
@db.connect_property(@anonymous998.properties["extern"], lambda {port.extern+"portamento.time"})
Qml::add_child(@anonymous1000, @anonymous998)
@db.connect_property(@anonymous999.properties["extern"], lambda {port.extern+"portamento.updowntimestretch"})
Qml::add_child(@anonymous1000, @anonymous999)
Qml::add_child(self, @anonymous1000)
@db.connect_property(@anonymous1001.properties["extern"], lambda {port.extern+"portamento.proportional"})
Qml::add_child(@anonymous1004, @anonymous1001)
@db.connect_property(@anonymous1002.properties["extern"], lambda {port.extern+"portamento.propRate"})
Qml::add_child(@anonymous1004, @anonymous1002)
@db.connect_property(@anonymous1003.properties["extern"], lambda {port.extern+"portamento.propDepth"})
Qml::add_child(@anonymous1004, @anonymous1003)
Qml::add_child(self, @anonymous1004)
self.label = "portamento"
self.copyable = false

    #puts "ZynPortamento, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynScale < Qml::Widget
  attr_reader :micro
  def class_name(); "part" ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.2,0.10,0.70])
    ;end
end
class Qml::ZynTuning < Qml::Widget
  attr_reader :scale
  attr_reader :tune
  attr_reader :mapp
  def load_scale(type);
        @opt = type
        win = window()
        wid = Qml::FileSelector.new(db)
        wid.whenValue = lambda { |x| scale.load_scale_file(x)}
        wid.x = 0
        wid.y = 0
        wid.w = win.w
        wid.h = win.h
        wid.pat = type
        Qml::add_child(win, wid)
        self.db.update_values
        setup_widget wid
        self.db.update_values

        if(root)
            root.smash_layout
            root.damage_item(win, :all)
        end
    ;end
  def setup_widget(w);
        w.onSetup()
        w.children.each do |c|
            setup_widget(c)
        end
    ;end
  def load_scale_file(val);
        $remote.action("/load_kbm", val) if @opt == ".kbm"
        $remote.action("/load_scl", val) if @opt == ".scl"
        $remote.action("/microtonal/mapping")
        $remote.action("/microtonal/tunings")
    ;end
  def class_name(); "part" ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.9, 0.1])
    ;end
  def apply();
        return false
        tune.apply
        mapp.apply
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @scale = self
@anonymous1167 = Qml::Widget.new(database, ui_path+"1"+'/')
@tune = Qml::Group.new(database, ui_path+"2"+'/')
@anonymous1163 = Qml::TextEdit.new(database, ui_path+"3"+'/')
@mapp = Qml::Group.new(database, ui_path+"4"+'/')
@anonymous1165 = Qml::TextEdit.new(database, ui_path+"5"+'/')
@anonymous1172 = Qml::ColorBox.new(database, ui_path+"6"+'/')
@anonymous1171 = Qml::ParModuleRow.new(database, ui_path+"7"+'/')
@anonymous1168 = Qml::TriggerButton.new(database, ui_path+"8"+'/')
@anonymous1169 = Qml::TriggerButton.new(database, ui_path+"9"+'/')
@anonymous1170 = Qml::TriggerButton.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["scale"] = self
context["tune"] = @tune
context["mapp"] = @mapp

        @anonymous1167.instance_eval("def class_name(); \"partsub\" ;end", "src/mruby-zest/example/ZynTuning.qml", 64)

        @anonymous1167.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.5, 0.5])\n        ;end", "src/mruby-zest/example/ZynTuning.qml", 65)
@anonymous1163.height = 0.05
@anonymous1163.extern = "/microtonal/tunings"
Qml::add_child(@tune, @anonymous1163)
@tune.label = "tunings"
@db.connect_property(@tune.properties["extern"], lambda {scale.extern + "tunings"})
@tune.copyable = false
Qml::add_child(@anonymous1167, @tune)
@anonymous1165.height = 0.05
@anonymous1165.extern = "/microtonal/mapping"
Qml::add_child(@mapp, @anonymous1165)
@mapp.label = "key mapping"
@db.connect_property(@mapp.properties["extern"], lambda {scale.extern + "mapping"})
@mapp.copyable = false
Qml::add_child(@anonymous1167, @mapp)
Qml::add_child(self, @anonymous1167)
@db.connect_property(@anonymous1168.properties["label"], lambda {"import .scl"})
@anonymous1168.tooltip = "load microtonal scale"
@anonymous1168.whenValue = lambda {scale.load_scale(".scl")}
Qml::add_child(@anonymous1171, @anonymous1168)
@db.connect_property(@anonymous1169.properties["label"], lambda {"import .kbm"})
@anonymous1169.tooltip = "load keyboard mapping"
@anonymous1169.whenValue = lambda {scale.load_scale(".kbm")}
Qml::add_child(@anonymous1171, @anonymous1169)
@anonymous1170.label = "retune"
@anonymous1170.whenValue = lambda { scale.apply }
@anonymous1170.active = false
Qml::add_child(@anonymous1171, @anonymous1170)
Qml::add_child(@anonymous1172, @anonymous1171)
@anonymous1172.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous1172)

    #puts "ZynTuning, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynScale
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @micro = self
@anonymous1044 = Qml::Group.new(database, ui_path+"1"+'/')
@anonymous1043 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous1038 = Qml::ToggleButton.new(database, ui_path+"3"+'/')
@anonymous1039 = Qml::NumBox.new(database, ui_path+"4"+'/')
@anonymous1040 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous1041 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous1042 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1054 = Qml::Widget.new(database, ui_path+"8"+'/')
@anonymous1051 = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous1047 = Qml::Widget.new(database, ui_path+"10"+'/')
@anonymous1045 = Qml::Text.new(database, ui_path+"11"+'/')
@anonymous1046 = Qml::TextLine.new(database, ui_path+"12"+'/')
@anonymous1050 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous1048 = Qml::Text.new(database, ui_path+"14"+'/')
@anonymous1049 = Qml::TextLine.new(database, ui_path+"15"+'/')
@anonymous1053 = Qml::ParModuleRow.new(database, ui_path+"16"+'/')
@anonymous1052 = Qml::Knob.new(database, ui_path+"17"+'/')
@anonymous1055 = Qml::ZynTuning.new(database, ui_path+"18"+'/')

    
context = Hash.new
context["micro"] = self
@db.connect_property(@anonymous1038.properties["extern"], lambda {micro.extern + "Penabled" })
Qml::add_child(@anonymous1043, @anonymous1038)
@db.connect_property(@anonymous1039.properties["extern"], lambda {micro.extern + "octavesize"})
@anonymous1039.label = "per/oct"
Qml::add_child(@anonymous1043, @anonymous1039)
@db.connect_property(@anonymous1040.properties["extern"], lambda {micro.extern + "PAfreq"})
@anonymous1040.type = "f"
Qml::add_child(@anonymous1043, @anonymous1040)
@db.connect_property(@anonymous1041.properties["extern"], lambda {micro.extern + "PAnote" })
Qml::add_child(@anonymous1043, @anonymous1041)
@db.connect_property(@anonymous1042.properties["extern"], lambda {micro.extern + "Pinvertupdowncenter" })
Qml::add_child(@anonymous1043, @anonymous1042)
Qml::add_child(@anonymous1044, @anonymous1043)
@anonymous1044.topSize = 0.2
@anonymous1044.copyable = false
@anonymous1044.label = "scale settings"
Qml::add_child(self, @anonymous1044)

        @anonymous1054.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynScale.qml", 59)

        @anonymous1054.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0.0, 0, 0.9, 1)\n            children[1].fixed(l, selfBox, 0.9, 0, 0.1, 1)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynScale.qml", 62)

        @anonymous1051.instance_eval("def class_name(); \"partsub\" ;end", "src/mruby-zest/example/ZynScale.qml", 51)

        @anonymous1051.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children, 0, 1, 4)\n            ;end", "src/mruby-zest/example/ZynScale.qml", 52)

        @anonymous1047.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 4)\n                ;end", "src/mruby-zest/example/ZynScale.qml", 35)
@anonymous1045.label = "name:"
@anonymous1045.height = 0.8
Qml::add_child(@anonymous1047, @anonymous1045)
@db.connect_property(@anonymous1046.properties["extern"], lambda {micro.extern + "Pname"})
Qml::add_child(@anonymous1047, @anonymous1046)
Qml::add_child(@anonymous1051, @anonymous1047)

        @anonymous1050.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 4)\n                ;end", "src/mruby-zest/example/ZynScale.qml", 47)
@anonymous1048.label = "comment:"
@anonymous1048.height = 0.8
Qml::add_child(@anonymous1050, @anonymous1048)
@db.connect_property(@anonymous1049.properties["extern"], lambda {micro.extern + "Pcomment"})
Qml::add_child(@anonymous1050, @anonymous1049)
Qml::add_child(@anonymous1051, @anonymous1050)
Qml::add_child(@anonymous1054, @anonymous1051)
@db.connect_property(@anonymous1052.properties["extern"], lambda {micro.extern + "Pscaleshift" })
Qml::add_child(@anonymous1053, @anonymous1052)
Qml::add_child(@anonymous1054, @anonymous1053)
Qml::add_child(self, @anonymous1054)
Qml::add_child(self, @anonymous1055)

    #puts "ZynScale, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPart
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @pset = self
@anonymous959 = Qml::Widget.new(database, ui_path+"1"+'/')
@anonymous963 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous960 = Qml::ZynInstrumentSettings.new(database, ui_path+"3"+'/')
@anonymous961 = Qml::ZynControllers.new(database, ui_path+"4"+'/')
@anonymous962 = Qml::ZynPortamento.new(database, ui_path+"5"+'/')
@anonymous964 = Qml::ZynScale.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["pset"] = self

        @anonymous959.instance_eval("def onSetup(old=nil);\n            return if !children.empty?\n            (1..16).each do |i|\n                but = Qml::ToggleButton.new(self.db)\n                but.label = i.to_s\n                but.pad   = 1/512\n                but.layoutOpts = [:no_constraint]\n                but.textScale  = 0.5\n                but.extern = \"/part\#{i-1}/Penabled\"\n                Qml::add_child(self, but)\n            end\n            (1..16).each do |i|\n                but = Qml::LabelButton.new(self.db)\n                but.label = \"None\"\n                but.pad   = 1/512\n                but.layoutOpts = [:no_constraint, :left_text]\n                but.textScale  = 0.5\n                but.extern = \"/part\#{i-1}/Pname\"\n                but.whenValue = lambda { but.root.set_view_pos(:part, i-1); but.root.change_view}\n                Qml::add_child(self, but)\n            end\n        ;end", "src/mruby-zest/example/ZynPart.qml", 4)

        @anonymous959.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vpack(l, selfBox, children[0..15],  0.0, 0.2)\n            Draw::Layout::vpack(l, selfBox, children[16..31], 0.2, 0.8)\n        ;end", "src/mruby-zest/example/ZynPart.qml", 28)
Qml::add_child(self, @anonymous959)

        @anonymous963.instance_eval("def class_name(); \"part\" ;end", "src/mruby-zest/example/ZynPart.qml", 34)

        @anonymous963.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.4,0.3,0.3])\n        ;end", "src/mruby-zest/example/ZynPart.qml", 35)
@db.connect_property(@anonymous960.properties["extern"], lambda {pset.extern})
Qml::add_child(@anonymous963, @anonymous960)
@db.connect_property(@anonymous961.properties["extern"], lambda {pset.extern + "ctl/"})
Qml::add_child(@anonymous963, @anonymous961)
@db.connect_property(@anonymous962.properties["extern"], lambda {pset.extern + "ctl/"})
Qml::add_child(@anonymous963, @anonymous962)
Qml::add_child(self, @anonymous963)
@anonymous964.extern = "/microtonal/"
Qml::add_child(self, @anonymous964)

    #puts "ZynPart, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPatchInfo < Qml::Widget
  attr_reader :info_box
  attr_reader :namebox
  attr_reader :authbox
  attr_reader :commbox
  def layout(l, selfBox);
        weights = Draw::DSP::norm_sum([1,1,1,4,1,8])

        Draw::Layout::vfill(l, selfBox, children, weights)
    ;end
  def set_view();
        prt = root.get_view_pos(:part)
        namebox.extern = "/part#{prt}/Pname"
        authbox.extern = "/part#{prt}/info.Pauthor"
        commbox.extern = "/part#{prt}/info.Pcomments"
    ;end
  def onSetup(old=nil);
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @info_box = self
@anonymous966 = Qml::TextBox.new(database, ui_path+"1"+'/')
@namebox = Qml::TextLine.new(database, ui_path+"2"+'/')
@anonymous968 = Qml::TextBox.new(database, ui_path+"3"+'/')
@authbox = Qml::TextEdit.new(database, ui_path+"4"+'/')
@anonymous970 = Qml::TextBox.new(database, ui_path+"5"+'/')
@commbox = Qml::TextEdit.new(database, ui_path+"6"+'/')

    
context = Hash.new
context["info_box"] = self
context["namebox"] = @namebox
context["authbox"] = @authbox
context["commbox"] = @commbox
@anonymous966.label = "name"
Qml::add_child(self, @anonymous966)
Qml::add_child(self, @namebox)
@anonymous968.label = "name"
Qml::add_child(self, @anonymous968)
@authbox.height = 0.25
Qml::add_child(self, @authbox)
@anonymous970.label = "comments"
Qml::add_child(self, @anonymous970)
@db.connect_property(@commbox.properties["height"], lambda {1.0/8.0})
Qml::add_child(self, @commbox)

    #puts "ZynPatchInfo, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPhaser < Qml::Group
  attr_reader :phaser
  attr_reader :rw
  attr_reader :rw2
  attr_reader :knobLRcross
  attr_reader :knobOffset
  attr_reader :knobPhase
  attr_reader :knobWidth
  attr_reader :analogButton
  def refresh();
        return if rw.content.nil?
        return if rw.content.children.length < 4
        rw.content.children[3..-1].each do |c|
            c.refresh
        end
        rw2.content.children.each do |c|
            c.refresh
        end
    ;end
  def changeAnalog(value);
        knobWidth.active = value
        knobPhase.active = !value
        knobOffset.active = value
        knobLRcross.active = !value
        knobWidth.damage_self
        knobPhase.damage_self
        knobOffset.damage_self
        knobLRcross.damage_self
        refresh()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @phaser = self
@rw = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous973 = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous974 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous975 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous976 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous977 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous978 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous979 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous980 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous981 = Qml::Knob.new(database, ui_path+"10"+'/')
@rw2 = Qml::ParModuleRow.new(database, ui_path+"11"+'/')
@anonymous983 = Qml::NumEntry.new(database, ui_path+"12"+'/')
@knobLRcross = Qml::Knob.new(database, ui_path+"13"+'/')
@knobOffset = Qml::Knob.new(database, ui_path+"14"+'/')
@anonymous986 = Qml::ToggleButton.new(database, ui_path+"15"+'/')
@knobPhase = Qml::Knob.new(database, ui_path+"16"+'/')
@knobWidth = Qml::Knob.new(database, ui_path+"17"+'/')
@anonymous989 = Qml::ToggleButton.new(database, ui_path+"18"+'/')
@anonymous990 = Qml::Knob.new(database, ui_path+"19"+'/')
@analogButton = Qml::ToggleButton.new(database, ui_path+"20"+'/')

    
context = Hash.new
context["phaser"] = self
context["rw"] = @rw
context["rw2"] = @rw2
context["knobLRcross"] = @knobLRcross
context["knobOffset"] = @knobOffset
context["knobPhase"] = @knobPhase
context["knobWidth"] = @knobWidth
context["analogButton"] = @analogButton
@db.connect_property(@anonymous973.properties["extern"], lambda {phaser.extern + "Phaser/preset"})
@anonymous973.whenValue = lambda { phaser.refresh }
Qml::add_child(@rw, @anonymous973)
@db.connect_property(@anonymous974.properties["extern"], lambda {phaser.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous974)
@db.connect_property(@anonymous975.properties["extern"], lambda {phaser.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous975)
@db.connect_property(@anonymous976.properties["extern"], lambda {phaser.extern + "Phaser/lfo.Pfreq" })
Qml::add_child(@rw, @anonymous976)
@db.connect_property(@anonymous977.properties["extern"], lambda {phaser.extern + "Phaser/lfo.Prandomness" })
Qml::add_child(@rw, @anonymous977)
@db.connect_property(@anonymous978.properties["extern"], lambda {phaser.extern + "Phaser/lfo.PLFOtype" })
Qml::add_child(@rw, @anonymous978)
@db.connect_property(@anonymous979.properties["extern"], lambda {phaser.extern + "Phaser/lfo.Pstereo" })
Qml::add_child(@rw, @anonymous979)
@db.connect_property(@anonymous980.properties["extern"], lambda {phaser.extern + "Phaser/Pdepth" })
Qml::add_child(@rw, @anonymous980)
@db.connect_property(@anonymous981.properties["extern"], lambda {phaser.extern + "Phaser/Pfb" })
Qml::add_child(@rw, @anonymous981)
@rw.layoutOpts = []
Qml::add_child(self, @rw)
@db.connect_property(@anonymous983.properties["extern"], lambda {phaser.extern + "Phaser/Pstages" })
Qml::add_child(@rw2, @anonymous983)
@db.connect_property(@knobLRcross.properties["extern"], lambda {phaser.extern + "Phaser/Plrcross" })
Qml::add_child(@rw2, @knobLRcross)
@db.connect_property(@knobOffset.properties["extern"], lambda {phaser.extern + "Phaser/Poffset" })
Qml::add_child(@rw2, @knobOffset)
@db.connect_property(@anonymous986.properties["extern"], lambda {phaser.extern + "Phaser/Poutsub" })
Qml::add_child(@rw2, @anonymous986)
@db.connect_property(@knobPhase.properties["extern"], lambda {phaser.extern + "Phaser/Pphase" })
Qml::add_child(@rw2, @knobPhase)
@db.connect_property(@knobWidth.properties["extern"], lambda {phaser.extern + "Phaser/Pwidth" })
Qml::add_child(@rw2, @knobWidth)
@db.connect_property(@anonymous989.properties["extern"], lambda {phaser.extern + "Phaser/Phyper" })
Qml::add_child(@rw2, @anonymous989)
@db.connect_property(@anonymous990.properties["extern"], lambda {phaser.extern + "Phaser/Pdistortion" })
Qml::add_child(@rw2, @anonymous990)
@db.connect_property(@analogButton.properties["extern"], lambda {phaser.extern + "Phaser/Panalog"})
@analogButton.whenValue = lambda() { phaser.changeAnalog(analogButton.value) }
Qml::add_child(@rw2, @analogButton)
@rw2.layoutOpts = []
Qml::add_child(self, @rw2)
self.label = "phaser"
self.topSize = 0.2

    #puts "ZynPhaser, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynPortamento < Qml::Group
  attr_reader :port
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @port = self
@anonymous1000 = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous996 = Qml::Col.new(database, ui_path+"2"+'/')
@anonymous994 = Qml::ToggleButton.new(database, ui_path+"3"+'/')
@anonymous995 = Qml::ToggleButton.new(database, ui_path+"4"+'/')
@anonymous997 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous998 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous999 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1004 = Qml::ParModuleRow.new(database, ui_path+"8"+'/')
@anonymous1001 = Qml::ToggleButton.new(database, ui_path+"9"+'/')
@anonymous1002 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous1003 = Qml::Knob.new(database, ui_path+"11"+'/')

    
context = Hash.new
context["port"] = self
@db.connect_property(@anonymous994.properties["extern"], lambda {port.extern+"portamento.receive"})
Qml::add_child(@anonymous996, @anonymous994)
@db.connect_property(@anonymous995.properties["extern"], lambda {port.extern+"portamento.pitchthreshtype"})
Qml::add_child(@anonymous996, @anonymous995)
Qml::add_child(@anonymous1000, @anonymous996)
@db.connect_property(@anonymous997.properties["extern"], lambda {port.extern+"portamento.pitchthresh"})
Qml::add_child(@anonymous1000, @anonymous997)
@db.connect_property(@anonymous998.properties["extern"], lambda {port.extern+"portamento.time"})
Qml::add_child(@anonymous1000, @anonymous998)
@db.connect_property(@anonymous999.properties["extern"], lambda {port.extern+"portamento.updowntimestretch"})
Qml::add_child(@anonymous1000, @anonymous999)
Qml::add_child(self, @anonymous1000)
@db.connect_property(@anonymous1001.properties["extern"], lambda {port.extern+"portamento.proportional"})
Qml::add_child(@anonymous1004, @anonymous1001)
@db.connect_property(@anonymous1002.properties["extern"], lambda {port.extern+"portamento.propRate"})
Qml::add_child(@anonymous1004, @anonymous1002)
@db.connect_property(@anonymous1003.properties["extern"], lambda {port.extern+"portamento.propDepth"})
Qml::add_child(@anonymous1004, @anonymous1003)
Qml::add_child(self, @anonymous1004)
self.label = "portamento"
self.copyable = false

    #puts "ZynPortamento, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynResOptions < Qml::ParModuleRow
  attr_reader :resopt
  attr_accessor :whenValue
  def cb();
        whenValue.call if whenValue
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @resopt = self
@anonymous1006 = Qml::ToggleButton.new(database, ui_path+"1"+'/')
@anonymous1007 = Qml::Knob.new(database, ui_path+"2"+'/')
@anonymous1008 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous1009 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous1010 = Qml::TriggerButton.new(database, ui_path+"5"+'/')
@anonymous1011 = Qml::ToggleButton.new(database, ui_path+"6"+'/')
@anonymous1012 = Qml::TextBox.new(database, ui_path+"7"+'/')
@anonymous1013 = Qml::TextBox.new(database, ui_path+"8"+'/')
@anonymous1014 = Qml::TriggerButton.new(database, ui_path+"9"+'/')
@anonymous1015 = Qml::TriggerButton.new(database, ui_path+"10"+'/')
@anonymous1016 = Qml::TriggerButton.new(database, ui_path+"11"+'/')
@anonymous1017 = Qml::TriggerButton.new(database, ui_path+"12"+'/')
@anonymous1018 = Qml::TriggerButton.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["resopt"] = self
@db.connect_property(@anonymous1006.properties["extern"], lambda {resopt.extern + "Penabled"})
@anonymous1006.label = "enable"
Qml::add_child(self, @anonymous1006)
@db.connect_property(@anonymous1007.properties["extern"], lambda {resopt.extern + "PmaxdB"})
@anonymous1007.label = "max"
Qml::add_child(self, @anonymous1007)
@db.connect_property(@anonymous1008.properties["extern"], lambda {resopt.extern + "Pcenterfreq"})
@db.connect_property(@anonymous1008.properties["label"], lambda {"c.f."})
Qml::add_child(self, @anonymous1008)
@db.connect_property(@anonymous1009.properties["extern"], lambda {resopt.extern + "Poctavesfreq"})
@db.connect_property(@anonymous1009.properties["label"], lambda {"oct."})
Qml::add_child(self, @anonymous1009)
@anonymous1010.whenValue = lambda {
            path = resopt.extern + "interpolatepeaks"
            $remote.action(path, 0)
            resopt.cb
        }
@anonymous1010.label = "interp"
@anonymous1010.tooltip = "Interpolate between sparsely added points"
Qml::add_child(self, @anonymous1010)
@db.connect_property(@anonymous1011.properties["extern"], lambda {resopt.extern + "Pprotectthefundamental"})
Qml::add_child(self, @anonymous1011)
Qml::add_child(self, @anonymous1012)
Qml::add_child(self, @anonymous1013)
@anonymous1014.whenValue = lambda {
            path = resopt.extern + "zero"
            $remote.action(path)
            resopt.cb
        }
@anonymous1014.label = "zero"
@anonymous1014.tooltip = "reset frequency response"
Qml::add_child(self, @anonymous1014)
@anonymous1015.whenValue = lambda {
            path = resopt.extern + "smooth"
            $remote.action(path)
            resopt.cb
        }
@anonymous1015.label = "smooth"
@anonymous1015.tooltip = "smooth out frequency response"
Qml::add_child(self, @anonymous1015)
@anonymous1016.whenValue = lambda {
            path = resopt.extern + "randomize"
            $remote.action(path, 0)
            resopt.cb
        }
@anonymous1016.label = "random 1"
@anonymous1016.tooltip = "create slow changing random response"
Qml::add_child(self, @anonymous1016)
@anonymous1017.whenValue = lambda {
            path = resopt.extern + "randomize"
            $remote.action(path, 1)
            resopt.cb
        }
@anonymous1017.label = "random 2"
@anonymous1017.tooltip = "create random response"
Qml::add_child(self, @anonymous1017)
@anonymous1018.whenValue = lambda {
            path = resopt.extern + "randomize"
            $remote.action(path, 2)
            resopt.cb
        }
@anonymous1018.label = "random 3"
@anonymous1018.tooltip = "create fast changing random response"
Qml::add_child(self, @anonymous1018)
self.whenValue = nil

    #puts "ZynResOptions, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynResonance < Qml::Widget
  attr_reader :reson
  attr_reader :vis
  def refresh();
        vis.refresh
    ;end
  def layout(l, selfBox);
        children[0].fixed(l, selfBox, 0, 0.0, 1, 0.8)
        children[1].fixed(l, selfBox, 0, 0.8, 1, 0.2)
        selfBox
    ;end
  def draw(vg);
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @reson = self
@vis = Qml::VisResonance.new(database, ui_path+"1"+'/')
@anonymous1022 = Qml::Indent.new(database, ui_path+"2"+'/')
@anonymous1021 = Qml::ZynResOptions.new(database, ui_path+"3"+'/')

    
context = Hash.new
context["reson"] = self
context["vis"] = @vis
@db.connect_property(@vis.properties["extern"], lambda {reson.extern + "respoints"})
Qml::add_child(self, @vis)

        @anonymous1022.instance_eval("def draw(vg);\n            vg.path do |v|\n                v.rect(0,0,w,h)\n                paint = v.linear_gradient(0,0,0,h,\n                Theme::InnerGrad1, Theme::InnerGrad2)\n                v.fill_paint paint\n                v.fill\n                v.stroke_color color(:black)\n                v.stroke_width 1.0\n                v.stroke\n            end\n        ;end", "src/mruby-zest/example/ZynResonance.qml", 16)
@anonymous1021.whenValue = lambda {reson.refresh}
@db.connect_property(@anonymous1021.properties["extern"], lambda {reson.extern})
Qml::add_child(@anonymous1022, @anonymous1021)
Qml::add_child(self, @anonymous1022)

    #puts "ZynResonance, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynReverb < Qml::Group
  attr_reader :reverb
  attr_reader :rw
  def refresh();
        return if rw.content.nil?
        return if rw.content.children.length < 4
        rw.content.children[3..-1].each do |c|
            c.refresh
        end
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @reverb = self
@rw = Qml::ParModuleRow.new(database, ui_path+"1"+'/')
@anonymous1024 = Qml::Selector.new(database, ui_path+"2"+'/')
@anonymous1025 = Qml::Knob.new(database, ui_path+"3"+'/')
@anonymous1026 = Qml::Knob.new(database, ui_path+"4"+'/')
@anonymous1027 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous1028 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous1029 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1030 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous1031 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous1032 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous1033 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous1034 = Qml::Knob.new(database, ui_path+"12"+'/')
@anonymous1035 = Qml::Knob.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["reverb"] = self
context["rw"] = @rw
@db.connect_property(@anonymous1024.properties["extern"], lambda {reverb.extern + "Reverb/preset"})
@anonymous1024.whenValue = lambda { reverb.refresh }
Qml::add_child(@rw, @anonymous1024)
@db.connect_property(@anonymous1025.properties["extern"], lambda {reverb.extern + "Pvolume"})
Qml::add_child(@rw, @anonymous1025)
@db.connect_property(@anonymous1026.properties["extern"], lambda {reverb.extern + "Ppanning"})
Qml::add_child(@rw, @anonymous1026)
@db.connect_property(@anonymous1027.properties["extern"], lambda {reverb.extern + "Reverb/Ptype"})
Qml::add_child(@rw, @anonymous1027)
@db.connect_property(@anonymous1028.properties["extern"], lambda {reverb.extern + "Reverb/Proomsize" })
Qml::add_child(@rw, @anonymous1028)
@db.connect_property(@anonymous1029.properties["extern"], lambda {reverb.extern + "Reverb/Ptime" })
Qml::add_child(@rw, @anonymous1029)
@db.connect_property(@anonymous1030.properties["extern"], lambda {reverb.extern + "Reverb/Pidelay"})
Qml::add_child(@rw, @anonymous1030)
@db.connect_property(@anonymous1031.properties["extern"], lambda {reverb.extern + "Reverb/Pidelayfb"})
Qml::add_child(@rw, @anonymous1031)
@db.connect_property(@anonymous1032.properties["extern"], lambda {reverb.extern + "Reverb/Plpf"})
Qml::add_child(@rw, @anonymous1032)
@db.connect_property(@anonymous1033.properties["extern"], lambda {reverb.extern + "Reverb/Phpf"})
Qml::add_child(@rw, @anonymous1033)
@db.connect_property(@anonymous1034.properties["extern"], lambda {reverb.extern + "Reverb/Plohidamp"})
Qml::add_child(@rw, @anonymous1034)
@db.connect_property(@anonymous1035.properties["extern"], lambda {reverb.extern + "Reverb/Pbandwidth" })
Qml::add_child(@rw, @anonymous1035)
@rw.layoutOpts = []
Qml::add_child(self, @rw)
self.label = "reverb"
self.topSize = 0.2

    #puts "ZynReverb, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynScale < Qml::Widget
  attr_reader :micro
  def class_name(); "part" ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.2,0.10,0.70])
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @micro = self
@anonymous1044 = Qml::Group.new(database, ui_path+"1"+'/')
@anonymous1043 = Qml::ParModuleRow.new(database, ui_path+"2"+'/')
@anonymous1038 = Qml::ToggleButton.new(database, ui_path+"3"+'/')
@anonymous1039 = Qml::NumBox.new(database, ui_path+"4"+'/')
@anonymous1040 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous1041 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous1042 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1054 = Qml::Widget.new(database, ui_path+"8"+'/')
@anonymous1051 = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous1047 = Qml::Widget.new(database, ui_path+"10"+'/')
@anonymous1045 = Qml::Text.new(database, ui_path+"11"+'/')
@anonymous1046 = Qml::TextLine.new(database, ui_path+"12"+'/')
@anonymous1050 = Qml::Widget.new(database, ui_path+"13"+'/')
@anonymous1048 = Qml::Text.new(database, ui_path+"14"+'/')
@anonymous1049 = Qml::TextLine.new(database, ui_path+"15"+'/')
@anonymous1053 = Qml::ParModuleRow.new(database, ui_path+"16"+'/')
@anonymous1052 = Qml::Knob.new(database, ui_path+"17"+'/')
@anonymous1055 = Qml::ZynTuning.new(database, ui_path+"18"+'/')

    
context = Hash.new
context["micro"] = self
@db.connect_property(@anonymous1038.properties["extern"], lambda {micro.extern + "Penabled" })
Qml::add_child(@anonymous1043, @anonymous1038)
@db.connect_property(@anonymous1039.properties["extern"], lambda {micro.extern + "octavesize"})
@anonymous1039.label = "per/oct"
Qml::add_child(@anonymous1043, @anonymous1039)
@db.connect_property(@anonymous1040.properties["extern"], lambda {micro.extern + "PAfreq"})
@anonymous1040.type = "f"
Qml::add_child(@anonymous1043, @anonymous1040)
@db.connect_property(@anonymous1041.properties["extern"], lambda {micro.extern + "PAnote" })
Qml::add_child(@anonymous1043, @anonymous1041)
@db.connect_property(@anonymous1042.properties["extern"], lambda {micro.extern + "Pinvertupdowncenter" })
Qml::add_child(@anonymous1043, @anonymous1042)
Qml::add_child(@anonymous1044, @anonymous1043)
@anonymous1044.topSize = 0.2
@anonymous1044.copyable = false
@anonymous1044.label = "scale settings"
Qml::add_child(self, @anonymous1044)

        @anonymous1054.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynScale.qml", 59)

        @anonymous1054.instance_eval("def layout(l, selfBox);\n            children[0].fixed(l, selfBox, 0.0, 0, 0.9, 1)\n            children[1].fixed(l, selfBox, 0.9, 0, 0.1, 1)\n            selfBox\n        ;end", "src/mruby-zest/example/ZynScale.qml", 62)

        @anonymous1051.instance_eval("def class_name(); \"partsub\" ;end", "src/mruby-zest/example/ZynScale.qml", 51)

        @anonymous1051.instance_eval("def layout(l, selfBox);\n                Draw::Layout::vpack(l, selfBox, children, 0, 1, 4)\n            ;end", "src/mruby-zest/example/ZynScale.qml", 52)

        @anonymous1047.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 4)\n                ;end", "src/mruby-zest/example/ZynScale.qml", 35)
@anonymous1045.label = "name:"
@anonymous1045.height = 0.8
Qml::add_child(@anonymous1047, @anonymous1045)
@db.connect_property(@anonymous1046.properties["extern"], lambda {micro.extern + "Pname"})
Qml::add_child(@anonymous1047, @anonymous1046)
Qml::add_child(@anonymous1051, @anonymous1047)

        @anonymous1050.instance_eval("def layout(l, selfBox);\n                    Draw::Layout::hfill(l, selfBox, children, [0.3, 0.7], 0, 4)\n                ;end", "src/mruby-zest/example/ZynScale.qml", 47)
@anonymous1048.label = "comment:"
@anonymous1048.height = 0.8
Qml::add_child(@anonymous1050, @anonymous1048)
@db.connect_property(@anonymous1049.properties["extern"], lambda {micro.extern + "Pcomment"})
Qml::add_child(@anonymous1050, @anonymous1049)
Qml::add_child(@anonymous1051, @anonymous1050)
Qml::add_child(@anonymous1054, @anonymous1051)
@db.connect_property(@anonymous1052.properties["extern"], lambda {micro.extern + "Pscaleshift" })
Qml::add_child(@anonymous1053, @anonymous1052)
Qml::add_child(@anonymous1054, @anonymous1053)
Qml::add_child(self, @anonymous1054)
Qml::add_child(self, @anonymous1055)

    #puts "ZynScale, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSendToGrid < Qml::Widget
  attr_reader :title
  def draw(vg);
        Draw::GradBox(vg, Rect.new(0,0,w,h))
    ;end
  def layout(l, selfBox);
        pad = 4
        children[0].fixed_long(l, selfBox, 0, 0, 1, 1,
        pad, pad, -2*pad, -2*pad)
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1059 = self
@anonymous1058 = Qml::Widget.new(database, ui_path+"1"+'/')
@title = Qml::Widget.new(database, ui_path+"2"+'/')

    
context = Hash.new
context["title"] = @title

        @anonymous1058.instance_eval("def onSetup(old=nil);\n            return if children.length > 1\n            rows = 4\n            (rows).times do |x|\n                ch = Qml::ZynSendToRow.new(db)\n                ch.row_id = x\n                ch.cols   = 3\n                Qml::add_child(self, ch)\n            end\n        ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 6)

        @anonymous1058.instance_eval("def class_name(); \"send_to_grid\" ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 18)

        @anonymous1058.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSendToGrid.qml", 20)
Qml::add_child(@anonymous1058, @title)
Qml::add_child(self, @anonymous1058)

    #puts "ZynSendToGrid, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSendToRow < Qml::Widget
  attr_reader :row
  attr_reader :name
  attr_property :row_id
  attr_property :cols
  def onSetup(old=nil);
        return if children.length > 1
        name.label = (row.row_id+1).to_s + " reverb"
        (cols-row_id).times do |x|
            ch = Qml::HSlider.new(db)
            ch.extern = "/sysefxfrom#{row_id}/to#{x+row_id+1}"
            ch.tooltip = "Route from system effect #{row_id+1} to effect #{x+row_id+2}"
            Qml::add_child(self, ch)
        end
    ;end
  def layout(l, selfBox);
        step = 1/(cols+1)
        children[0].fixed(l, selfBox, 0, 0, step, 1)
        off = step*(1+row_id)
        (1...children.length).each do |bx|
            children[bx].fixed(l, selfBox, off, 0, step, 1)
            off += step
        end
        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @row = self
@name = Qml::TextSel.new(database, ui_path+"1"+'/')

    
context = Hash.new
context["row"] = self
context["name"] = @name
Qml::prop_add(self, "row_id")
Qml::prop_add(self, "cols")
@db.connect_property(@name.properties["extern"], lambda {"/sysefx#{row.row_id}/efftype"})
Qml::add_child(self, @name)
self.row_id = 0
self.cols = 4

    #puts "ZynSendToRow, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSettings < Qml::Widget
  def layout(l, selfBox);
        Draw::Layout::hpack(l, selfBox, children, 0, 1, 10)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1082 = self
@anonymous1069 = Qml::Forms.new(database, ui_path+"1"+'/')
@anonymous1062 = Qml::TextField.new(database, ui_path+"2"+'/')
@anonymous1063 = Qml::ToggleButton.new(database, ui_path+"3"+'/')
@anonymous1064 = Qml::TextField.new(database, ui_path+"4"+'/')
@anonymous1065 = Qml::TextField.new(database, ui_path+"5"+'/')
@anonymous1066 = Qml::NumEntry.new(database, ui_path+"6"+'/')
@anonymous1067 = Qml::Selector.new(database, ui_path+"7"+'/')
@anonymous1068 = Qml::ToggleButton.new(database, ui_path+"8"+'/')
@anonymous1075 = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous1070 = Qml::Text.new(database, ui_path+"10"+'/')
@anonymous1074 = Qml::Widget.new(database, ui_path+"11"+'/')
@anonymous1071 = Qml::Button.new(database, ui_path+"12"+'/')
@anonymous1072 = Qml::Button.new(database, ui_path+"13"+'/')
@anonymous1073 = Qml::Button.new(database, ui_path+"14"+'/')
@anonymous1081 = Qml::Widget.new(database, ui_path+"15"+'/')
@anonymous1076 = Qml::Text.new(database, ui_path+"16"+'/')
@anonymous1080 = Qml::Widget.new(database, ui_path+"17"+'/')
@anonymous1077 = Qml::Button.new(database, ui_path+"18"+'/')
@anonymous1078 = Qml::Button.new(database, ui_path+"19"+'/')
@anonymous1079 = Qml::Button.new(database, ui_path+"20"+'/')

    
context = Hash.new
@anonymous1062.tooltip = "Sample Rate:"
@anonymous1062.label = "48 kHz"
Qml::add_child(@anonymous1069, @anonymous1062)
@anonymous1063.tooltip = "Swap Stereo:"
Qml::add_child(@anonymous1069, @anonymous1063)
@anonymous1064.tooltip = "Oscil Size:"
@anonymous1064.label = "256 samples"
Qml::add_child(@anonymous1069, @anonymous1064)
@anonymous1065.tooltip = "Buffer Size:"
@anonymous1065.label = "128 samples"
Qml::add_child(@anonymous1069, @anonymous1065)
@anonymous1066.tooltip = "XML Compression:"
Qml::add_child(@anonymous1069, @anonymous1066)
@anonymous1067.tooltip = "Virtual Keyboard layout:"
@anonymous1067.options = ["QWERTY", "AZERTY"]
Qml::add_child(@anonymous1069, @anonymous1067)
@anonymous1068.tooltip = "Ignore MIDI Program Change:"
Qml::add_child(@anonymous1069, @anonymous1068)
@anonymous1069.label = "Global Settings"
Qml::add_child(self, @anonymous1069)

        @anonymous1075.instance_eval("def draw(vg);\n        vg.path do |v|\n            v.rect(0,0,w,h)\n            paint = v.linear_gradient(0,0,0,h,\n            Theme::ModuleGrad1, Theme::ModuleGrad2)\n            v.fill_paint paint\n            v.fill\n            v.stroke_color(color(\"000000\"))\n            v.stroke\n        end\n    ;end", "src/mruby-zest/example/ZynSettings.qml", 15)

        @anonymous1075.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.9, 0.1])\n        ;end", "src/mruby-zest/example/ZynSettings.qml", 43)
@anonymous1070.label = "bank root"
Qml::add_child(@anonymous1075, @anonymous1070)

        @anonymous1074.instance_eval("def layout(l, selfBox);\n                Draw::Layout::hfill(l, selfBox, children, [0.7, 0.15, 0.15])\n            ;end", "src/mruby-zest/example/ZynSettings.qml", 39)
@anonymous1071.label = "prioritize"
Qml::add_child(@anonymous1074, @anonymous1071)
@db.connect_property(@anonymous1072.properties["label"], lambda {"+"})
Qml::add_child(@anonymous1074, @anonymous1072)
@db.connect_property(@anonymous1073.properties["label"], lambda {"-"})
Qml::add_child(@anonymous1074, @anonymous1073)
Qml::add_child(@anonymous1075, @anonymous1074)
Qml::add_child(self, @anonymous1075)

        @anonymous1081.instance_eval("def draw(vg);\n        vg.path do |v|\n            v.rect(0,0,w,h)\n            paint = v.linear_gradient(0,0,0,h,\n            Theme::ModuleGrad1, Theme::ModuleGrad2)\n            v.fill_paint paint\n            v.fill\n            v.stroke_color(color(\"000000\"))\n            v.stroke\n        end\n    ;end", "src/mruby-zest/example/ZynSettings.qml", 48)

        @anonymous1081.instance_eval("def layout(l, selfBox);\n            Draw::Layout::vfill(l, selfBox, children, [0.9, 0.1])\n        ;end", "src/mruby-zest/example/ZynSettings.qml", 76)
@anonymous1076.label = "preset roots"
Qml::add_child(@anonymous1081, @anonymous1076)

        @anonymous1080.instance_eval("def layout(l, selfBox);\n                Draw::Layout::hfill(l, selfBox, children, [0.7, 0.15, 0.15])\n            ;end", "src/mruby-zest/example/ZynSettings.qml", 72)
@anonymous1077.label = "prioritize"
Qml::add_child(@anonymous1080, @anonymous1077)
@db.connect_property(@anonymous1078.properties["label"], lambda {"+"})
Qml::add_child(@anonymous1080, @anonymous1078)
@db.connect_property(@anonymous1079.properties["label"], lambda {"-"})
Qml::add_child(@anonymous1080, @anonymous1079)
Qml::add_child(@anonymous1081, @anonymous1080)
Qml::add_child(self, @anonymous1081)

    #puts "ZynSettings, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSidebar < Qml::Widget
  attr_reader :side
  attr_reader :partsetting
  attr_reader :part
  attr_reader :browser
  attr_reader :mixer
  attr_reader :kitedit
  attr_reader :kit
  attr_reader :arp
  attr_reader :eff
  attr_reader :add
  attr_reader :voices
  attr_reader :subButton
  attr_reader :padButton
  def class_name(); "ZynSidebar" ;end
  def set_content(type, flag=nil);
        tabs = [:part, :banks, :mixer, :kits, :automate, :effects,
                :add_synth, :sub_synth, :pad_synth]
        ind = 0
        self.children.each do |ch|
            if([Qml::FancyButton, Qml::SidebarButton].include? ch.class)
                n = (type == tabs[ind])
                if(n != ch.value)
                    ch.value = n
                    root.damage_item ch if root
                end
                ind += 1
            end
        end

        parent.set_content(type) if flag.nil? && parent.respond_to?(:set_content)
    ;end
  def set_view();
        prt  = root.get_view_pos(:part)
        kid  = root.get_view_pos(:kit)
        typ  = root.get_view_pos(:view)
        padButton.extern = "/part#{prt}/kit#{kid}/Ppadenabled"
        subButton.extern = "/part#{prt}/kit#{kid}/Psubenabled"
        add.extern       = "/part#{prt}/kit#{kid}/Padenabled"
        set_content(typ, true)
    ;end
  def onSetup(old=nil);
        set_view
    ;end
  def layout(l, selfBox);
        partsetting.fixed(l, selfBox, 0.1, 0,     0.8, 0.045)
        part.fixed(l,        selfBox, 0.1, 0.05,  0.8, 0.19)
        browser.fixed(l,     selfBox, 0.1, 0.248, 0.8, 0.045)
        mixer.fixed(l,       selfBox, 0.1, 0.3,   0.8, 0.045)
        kitedit.fixed(l,     selfBox, 0.1, 0.35,  0.8, 0.045)
        kit.fixed(l,         selfBox, 0.1, 0.405, 0.8, 0.187)
        arp.fixed(l,         selfBox, 0.1, 0.6,   0.8, 0.045)
        eff.fixed(l,         selfBox, 0.1, 0.65,  0.8, 0.045)
        add.fixed(l,         selfBox, 0.1, 0.7,   0.8, 0.045)
        voices.fixed(l,      selfBox, 0.1, 0.76,  0.8, 0.10)
        subButton.fixed(l,   selfBox, 0.1, 0.88,  0.8, 0.045)
        padButton.fixed(l,   selfBox, 0.1, 0.94,  0.8, 0.045)

        selfBox
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @side = self
@partsetting = Qml::SidebarButton.new(database, ui_path+"1"+'/')
@part = Qml::Indent.new(database, ui_path+"2"+'/')
@anonymous1084 = Qml::KitButtons.new(database, ui_path+"3"+'/')
@browser = Qml::SidebarButton.new(database, ui_path+"4"+'/')
@mixer = Qml::SidebarButton.new(database, ui_path+"5"+'/')
@kitedit = Qml::SidebarButton.new(database, ui_path+"6"+'/')
@kit = Qml::Indent.new(database, ui_path+"7"+'/')
@anonymous1089 = Qml::KitButtons.new(database, ui_path+"8"+'/')
@arp = Qml::SidebarButton.new(database, ui_path+"9"+'/')
@eff = Qml::SidebarButton.new(database, ui_path+"10"+'/')
@add = Qml::FancyButton.new(database, ui_path+"11"+'/')
@voices = Qml::Indent.new(database, ui_path+"12"+'/')
@anonymous1094 = Qml::KitButtons.new(database, ui_path+"13"+'/')
@subButton = Qml::FancyButton.new(database, ui_path+"14"+'/')
@padButton = Qml::FancyButton.new(database, ui_path+"15"+'/')

    
context = Hash.new
context["side"] = self
context["partsetting"] = @partsetting
context["part"] = @part
context["browser"] = @browser
context["mixer"] = @mixer
context["kitedit"] = @kitedit
context["kit"] = @kit
context["arp"] = @arp
context["eff"] = @eff
context["add"] = @add
context["voices"] = @voices
context["subButton"] = @subButton
context["padButton"] = @padButton
@partsetting.label = "part settings"
@partsetting.whenClick = lambda { side.set_content :part }
Qml::add_child(self, @partsetting)
@anonymous1084.sym = :part
Qml::add_child(@part, @anonymous1084)
Qml::add_child(self, @part)
@browser.label = "browser"
@browser.whenClick = lambda { side.set_content :banks }
Qml::add_child(self, @browser)
@mixer.label = "mixer"
@mixer.whenClick = lambda { side.set_content :mixer }
Qml::add_child(self, @mixer)
@kitedit.label = "kits"
@kitedit.whenClick = lambda { side.set_content :kits }
Qml::add_child(self, @kitedit)
@anonymous1089.sym = :kit
Qml::add_child(@kit, @anonymous1089)
Qml::add_child(self, @kit)
@arp.label = "macro learn"
@arp.whenClick = lambda { side.set_content :automate }
Qml::add_child(self, @arp)
@eff.label = "effects"
@eff.whenClick = lambda { side.set_content :effects }
Qml::add_child(self, @eff)
@add.label = "add"
@add.value = true
@add.whenClick = lambda { side.set_content :add_synth }
Qml::add_child(self, @add)
@anonymous1094.sym = :voice
@anonymous1094.rows = 2
Qml::add_child(@voices, @anonymous1094)
Qml::add_child(self, @voices)
@subButton.label = "sub"
@subButton.whenClick = lambda { side.set_content :sub_synth }
Qml::add_child(self, @subButton)
@padButton.label = "pad"
@padButton.whenClick = lambda { side.set_content :pad_synth }
Qml::add_child(self, @padButton)

    #puts "ZynSidebar, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubAmp < Qml::Widget
  attr_reader :subamp
  attr_reader :edit
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.6, 0.4])
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @subamp = self
@edit = Qml::ZynEnvEdit.new(database, ui_path+"1"+'/')
@anonymous1108 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous1106 = Qml::Group.new(database, ui_path+"3"+'/')
@anonymous1104 = Qml::ParModuleRow.new(database, ui_path+"4"+'/')
@anonymous1100 = Qml::ToggleButton.new(database, ui_path+"5"+'/')
@anonymous1101 = Qml::Knob.new(database, ui_path+"6"+'/')
@anonymous1102 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1103 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous1105 = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous1107 = Qml::ZynAmpEnv.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["subamp"] = self
context["edit"] = @edit
@db.connect_property(@edit.properties["extern"], lambda {subamp.extern + "AmpEnvelope/"})
Qml::add_child(self, @edit)

        @anonymous1108.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSubAmp.qml", 25)
@db.connect_property(@anonymous1100.properties["extern"], lambda {subamp.extern + "Pstereo" })
Qml::add_child(@anonymous1104, @anonymous1100)
@db.connect_property(@anonymous1101.properties["extern"], lambda {subamp.extern + "PVolume" })
Qml::add_child(@anonymous1104, @anonymous1101)
@db.connect_property(@anonymous1102.properties["extern"], lambda {subamp.extern + "PPanning" })
Qml::add_child(@anonymous1104, @anonymous1102)
@db.connect_property(@anonymous1103.properties["extern"], lambda {subamp.extern + "PAmpVelocityScaleFunction" })
Qml::add_child(@anonymous1104, @anonymous1103)
Qml::add_child(@anonymous1106, @anonymous1104)
Qml::add_child(@anonymous1106, @anonymous1105)
@anonymous1106.label = "general"
Qml::add_child(@anonymous1108, @anonymous1106)
@db.connect_property(@anonymous1107.properties["extern"], lambda {subamp.extern+"AmpEnvelope/"})
@db.connect_property(@anonymous1107.properties["whenModified"], lambda {lambda { edit.refresh }})
Qml::add_child(@anonymous1108, @anonymous1107)
Qml::add_child(self, @anonymous1108)
self.extern = "/part0/kit0/subpars/"

    #puts "ZynSubAmp, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubBandwidth < Qml::Widget
  attr_reader :subbw
  attr_reader :edit
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.6, 0.4])
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @subbw = self
@edit = Qml::ZynEnvEdit.new(database, ui_path+"1"+'/')
@anonymous1119 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous1117 = Qml::Group.new(database, ui_path+"3"+'/')
@anonymous1115 = Qml::ParModuleRow.new(database, ui_path+"4"+'/')
@anonymous1111 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous1112 = Qml::NumEntry.new(database, ui_path+"6"+'/')
@anonymous1113 = Qml::Knob.new(database, ui_path+"7"+'/')
@anonymous1114 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous1116 = Qml::Widget.new(database, ui_path+"9"+'/')
@anonymous1118 = Qml::ZynBandwidthEnv.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["subbw"] = self
context["edit"] = @edit
@db.connect_property(@edit.properties["extern"], lambda {subbw.extern + "BandWidthEnvelope/"})
Qml::add_child(self, @edit)

        @anonymous1119.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSubBandwidth.qml", 26)
@db.connect_property(@anonymous1111.properties["extern"], lambda {subbw.extern + "Pstart" })
Qml::add_child(@anonymous1115, @anonymous1111)
@db.connect_property(@anonymous1112.properties["extern"], lambda {subbw.extern + "Pnumstages" })
Qml::add_child(@anonymous1115, @anonymous1112)
@db.connect_property(@anonymous1113.properties["extern"], lambda {subbw.extern + "Pbandwidth" })
Qml::add_child(@anonymous1115, @anonymous1113)
@db.connect_property(@anonymous1114.properties["extern"], lambda {subbw.extern + "Pbwscale" })
Qml::add_child(@anonymous1115, @anonymous1114)
Qml::add_child(@anonymous1117, @anonymous1115)
Qml::add_child(@anonymous1117, @anonymous1116)
@anonymous1117.label = "general"
Qml::add_child(@anonymous1119, @anonymous1117)
@db.connect_property(@anonymous1118.properties["toggleable"], lambda {subbw.extern + "PBandWidthEnvelopeEnabled"})
@db.connect_property(@anonymous1118.properties["whenModified"], lambda {lambda {edit.refresh}})
@db.connect_property(@anonymous1118.properties["extern"], lambda {subbw.extern + "BandWidthEnvelope/"})
Qml::add_child(@anonymous1119, @anonymous1118)
Qml::add_child(self, @anonymous1119)
self.extern = "/part0/kit0/subpars/"

    #puts "ZynSubBandwidth, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubFilter < Qml::Widget
  attr_reader :subfl
  attr_reader :edit
  attr_reader :gen
  attr_reader :env
  attr_accessor :valueRef
  attr_property :filtertype
  def change_vis(x);
        root.set_view_pos(:vis, x)
        root.change_view
    ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.6, 0.4])
    ;end
  def set_vis_env(ext);
        edit.extern  = ext + "GlobalFilterEnvelope/"
        edit.content = Qml::ZynEnvEdit
        env.whenModified = lambda {
            elm = edit.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_vis_filter(ext);
        edit.extern = ext + "GlobalFilter/"
        if(self.filtertype == :formant)
            edit.content = Qml::ZynFormant
        else
            edit.extern = ext + "GlobalFilter/response"
            edit.content = Qml::VisFilter
        end
        gen.whenModified = lambda {
            elm = edit.children[0]
            elm.refresh if elm.respond_to? :refresh
        }
    ;end
  def set_view();
        base = self.extern
        vis  = root.get_view_pos(:vis)

        types = [:envelope, :lfo, :filter]
        if(!types.include?(vis))
            vis = :envelope
            root.set_view_pos(:vis, vis)
        end

        if(vis == :envelope)
            set_vis_env(base)
        elsif(vis == :filter)
            set_vis_filter(base)
        end
    ;end
  def add_cat();
        return if self.valueRef
        if(self.valueRef.nil?)
            path = self.extern + "GlobalFilter/Pcategory"
            self.valueRef = OSC::RemoteParam.new($remote, path)
            self.valueRef.mode = :full
            self.valueRef.callback = lambda {|x|
                subfl.filtertype = [:analog, :formant, :statevar][x]
            }
        end
    ;end
  def onSetup(old=nil);

        gen.move_sense()
        set_view()
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @subfl = self
@edit = Qml::Swappable.new(database, ui_path+"1"+'/')
@anonymous1124 = Qml::Widget.new(database, ui_path+"2"+'/')
@gen = Qml::ZynAnalogFilter.new(database, ui_path+"3"+'/')
@env = Qml::ZynFilterEnv.new(database, ui_path+"4"+'/')

    
context = Hash.new
context["subfl"] = self
context["edit"] = @edit
context["gen"] = @gen
context["env"] = @env
Qml::prop_add(self, "filtertype")
Qml::add_child(self, @edit)

        @anonymous1124.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSubFilter.qml", 31)
@gen.subsynth = true
@db.connect_property(@gen.properties["toggleable"], lambda {subfl.extern + "PGlobalFilterEnabled"})
@db.connect_property(@gen.properties["extern"], lambda {subfl.extern + "GlobalFilter/"})
@gen.whenClick = lambda { subfl.change_vis(:filter) }
Qml::add_child(@anonymous1124, @gen)
@db.connect_property(@env.properties["extern"], lambda {subfl.extern+"GlobalFilterEnvelope/"})
@env.whenClick = lambda { subfl.change_vis(:envlope) }
Qml::add_child(@anonymous1124, @env)
Qml::add_child(self, @anonymous1124)
self.valueRef = nil
self.filtertype = nil
@db.connect_watcher(self.properties["extern"], lambda {begin

        subfl.add_cat()
    
end})

    #puts "ZynSubFilter, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubFreq < Qml::Widget
  attr_reader :subfq
  attr_reader :edit
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.6, 0.4])
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @subfq = self
@edit = Qml::ZynEnvEdit.new(database, ui_path+"1"+'/')
@anonymous1138 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous1136 = Qml::Group.new(database, ui_path+"3"+'/')
@anonymous1131 = Qml::ParModuleRow.new(database, ui_path+"4"+'/')
@anonymous1127 = Qml::Knob.new(database, ui_path+"5"+'/')
@anonymous1128 = Qml::Selector.new(database, ui_path+"6"+'/')
@anonymous1129 = Qml::NumEntry.new(database, ui_path+"7"+'/')
@anonymous1130 = Qml::Knob.new(database, ui_path+"8"+'/')
@anonymous1135 = Qml::ParModuleRow.new(database, ui_path+"9"+'/')
@anonymous1132 = Qml::ToggleButton.new(database, ui_path+"10"+'/')
@anonymous1133 = Qml::Knob.new(database, ui_path+"11"+'/')
@anonymous1134 = Qml::Knob.new(database, ui_path+"12"+'/')
@anonymous1137 = Qml::ZynFreqEnv.new(database, ui_path+"13"+'/')

    
context = Hash.new
context["subfq"] = self
context["edit"] = @edit
@db.connect_property(@edit.properties["extern"], lambda {subfq.extern + "FreqEnvelope/"})
Qml::add_child(self, @edit)

        @anonymous1138.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hpack(l, selfBox, children)\n        ;end", "src/mruby-zest/example/ZynSubFreq.qml", 31)
@db.connect_property(@anonymous1127.properties["extern"], lambda {subfq.extern + "PDetune" })
Qml::add_child(@anonymous1131, @anonymous1127)
@db.connect_property(@anonymous1128.properties["extern"], lambda {subfq.extern + "PDetuneType" })
Qml::add_child(@anonymous1131, @anonymous1128)
@db.connect_property(@anonymous1129.properties["extern"], lambda {subfq.extern + "octave" })
Qml::add_child(@anonymous1131, @anonymous1129)
@db.connect_property(@anonymous1130.properties["extern"], lambda {subfq.extern + "PBendAdjust"})
Qml::add_child(@anonymous1131, @anonymous1130)
Qml::add_child(@anonymous1136, @anonymous1131)
@db.connect_property(@anonymous1132.properties["extern"], lambda {subfq.extern + "Pfixedfreq" })
Qml::add_child(@anonymous1135, @anonymous1132)
@db.connect_property(@anonymous1133.properties["extern"], lambda {subfq.extern + "PfixedfreqET" })
Qml::add_child(@anonymous1135, @anonymous1133)
@db.connect_property(@anonymous1134.properties["extern"], lambda {subfq.extern + "POffsetHz"})
Qml::add_child(@anonymous1135, @anonymous1134)
Qml::add_child(@anonymous1136, @anonymous1135)
@anonymous1136.label = "general"
Qml::add_child(@anonymous1138, @anonymous1136)
@db.connect_property(@anonymous1137.properties["toggleable"], lambda {subfq.extern + "PFreqEnvelopeEnabled"})
@db.connect_property(@anonymous1137.properties["extern"], lambda {subfq.extern+"FreqEnvelope/"})
@db.connect_property(@anonymous1137.properties["whenModified"], lambda {lambda { edit.refresh }})
Qml::add_child(@anonymous1138, @anonymous1137)
Qml::add_child(self, @anonymous1138)
self.extern = "/part0/kit0/subpars/"

    #puts "ZynSubFreq, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubHarmonic < Qml::Widget
  attr_reader :harm
  attr_reader :sub_harmonics
  attr_reader :edit
  def class_name(); "subsynth" ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.50, 0.10, 0.40])
    ;end
  def clear();
        edit.clear
        sub_harmonics.refresh
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @harm = self
@sub_harmonics = Qml::VisSubHarmonics.new(database, ui_path+"1"+'/')
@anonymous1150 = Qml::Widget.new(database, ui_path+"2"+'/')
@anonymous1149 = Qml::ParModuleRow.new(database, ui_path+"3"+'/')
@anonymous1141 = Qml::TriggerButton.new(database, ui_path+"4"+'/')
@anonymous1142 = Qml::Selector.new(database, ui_path+"5"+'/')
@anonymous1143 = Qml::NumEntry.new(database, ui_path+"6"+'/')
@anonymous1144 = Qml::Widget.new(database, ui_path+"7"+'/')
@anonymous1145 = Qml::Selector.new(database, ui_path+"8"+'/')
@anonymous1146 = Qml::Knob.new(database, ui_path+"9"+'/')
@anonymous1147 = Qml::Knob.new(database, ui_path+"10"+'/')
@anonymous1148 = Qml::Knob.new(database, ui_path+"11"+'/')
@edit = Qml::HarmonicEdit.new(database, ui_path+"12"+'/')

    
context = Hash.new
context["harm"] = self
context["sub_harmonics"] = @sub_harmonics
context["edit"] = @edit
@db.connect_property(@sub_harmonics.properties["extern"], lambda {harm.extern + "response"})
Qml::add_child(self, @sub_harmonics)

        @anonymous1150.instance_eval("def draw(vg);\n            Draw::GradBox(vg, Rect.new(0,0,w,h))\n        ;end", "src/mruby-zest/example/ZynSubHarmonic.qml", 44)
@anonymous1141.label = "clear"
@anonymous1141.whenValue = lambda {harm.clear}
Qml::add_child(@anonymous1149, @anonymous1141)
@db.connect_property(@anonymous1142.properties["extern"], lambda {harm.extern + "Phmagtype"})
@anonymous1142.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1142)
@anonymous1143.label = "stages"
@db.connect_property(@anonymous1143.properties["extern"], lambda {harm.extern + "Pnumstages"})
@anonymous1143.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1143)
@anonymous1144.label = "          "
Qml::add_child(@anonymous1149, @anonymous1144)
@db.connect_property(@anonymous1145.properties["extern"], lambda {harm.extern+"POvertoneSpread.type"})
@anonymous1145.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1145)
@db.connect_property(@anonymous1146.properties["extern"], lambda {harm.extern+"POvertoneSpread.par1"})
@anonymous1146.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1146)
@db.connect_property(@anonymous1147.properties["extern"], lambda {harm.extern+"POvertoneSpread.par2"})
@anonymous1147.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1147)
@db.connect_property(@anonymous1148.properties["extern"], lambda {harm.extern+"POvertoneSpread.par3"})
@anonymous1148.whenValue = lambda {sub_harmonics.refresh}
Qml::add_child(@anonymous1149, @anonymous1148)
@anonymous1149.lsize = 0.3
Qml::add_child(@anonymous1150, @anonymous1149)
Qml::add_child(self, @anonymous1150)
@db.connect_property(@edit.properties["extern"], lambda {harm.extern})
@edit.type = :subsynth
@edit.whenValue = lambda { sub_harmonics.refresh }
Qml::add_child(self, @edit)

    #puts "ZynSubHarmonic, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynSubSynth < Qml::Widget
  attr_reader :subsynth
  attr_reader :subtabs
  attr_reader :cpy
  attr_reader :swap
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.05, 0.95])
    ;end
  def set_view();
        vw = root.get_view_pos(:subview)
        mapping = {:harmonic  => Qml::ZynSubHarmonic,
                   :amplitude => Qml::ZynSubAmp,
                   :bandwidth => Qml::ZynSubBandwidth,
                   :frequency => Qml::ZynSubFreq,
                   :filter    => Qml::ZynSubFilter}
        tabid   = {:harmonic  => 0,
                   :amplitude => 1,
                   :bandwidth => 2,
                   :frequency => 3,
                   :filter    => 4}
        if(!mapping.include?(vw))
            root.set_view_pos(:subview, :harmonic)
            vw = :harmonic
        end

        swap.extern  = subsynth.extern
        swap.content = mapping[vw]
        subtabs.children[tabid[vw]].value = true
    ;end
  def onSetup(old=nil);
        set_view
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @subsynth = self
@subtabs = Qml::TabGroup.new(database, ui_path+"1"+'/')
@anonymous1153 = Qml::TabButton.new(database, ui_path+"2"+'/')
@anonymous1154 = Qml::TabButton.new(database, ui_path+"3"+'/')
@anonymous1155 = Qml::TabButton.new(database, ui_path+"4"+'/')
@anonymous1156 = Qml::TabButton.new(database, ui_path+"5"+'/')
@anonymous1157 = Qml::TabButton.new(database, ui_path+"6"+'/')
@cpy = Qml::CopyButton.new(database, ui_path+"7"+'/')
@anonymous1159 = Qml::PasteButton.new(database, ui_path+"8"+'/')
@swap = Qml::Swappable.new(database, ui_path+"9"+'/')

    
context = Hash.new
context["subsynth"] = self
context["subtabs"] = @subtabs
@subtabs.instance_eval do
def cpy=(k);@cpy=k;end
def cpy;    @cpy;  end
end
context["cpy"] = @cpy
context["swap"] = @swap
Qml::context_apply(context, @subtabs)

        @subtabs.instance_eval("def layout(l, selfBox);\n            selfBox = Draw::Layout::tabpack(l, selfBox, self, cpy)\n        ;end", "src/mruby-zest/example/ZynSubSynth.qml", 15)

        @subtabs.instance_eval("def set_tab(wid);\n            selected = get_tab wid\n\n            #Define a mapping from tabs to values\n            mapping = {0 => :harmonic,\n                       1 => :amplitude,\n                       2 => :bandwidth,\n                       3 => :frequency,\n                       4 => :filter,\n                       }\n            return if !mapping.include?(selected)\n            root.set_view_pos(:subview, mapping[selected])\n            root.change_view\n        ;end", "src/mruby-zest/example/ZynSubSynth.qml", 19)
@anonymous1153.label = "harmonic"
Qml::add_child(@subtabs, @anonymous1153)
@anonymous1154.label = "amplitude"
Qml::add_child(@subtabs, @anonymous1154)
@anonymous1155.label = "bandwidth"
Qml::add_child(@subtabs, @anonymous1155)
@anonymous1156.label = "frequency"
Qml::add_child(@subtabs, @anonymous1156)
@anonymous1157.label = "filter"
Qml::add_child(@subtabs, @anonymous1157)
@db.connect_property(@cpy.properties["extern"], lambda {subsynth.extern})
Qml::add_child(@subtabs, @cpy)
@db.connect_property(@anonymous1159.properties["extern"], lambda {subsynth.extern})
Qml::add_child(@subtabs, @anonymous1159)
Qml::add_child(self, @subtabs)
Qml::add_child(self, @swap)

    #puts "ZynSubSynth, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynTuning < Qml::Widget
  attr_reader :scale
  attr_reader :tune
  attr_reader :mapp
  def load_scale(type);
        @opt = type
        win = window()
        wid = Qml::FileSelector.new(db)
        wid.whenValue = lambda { |x| scale.load_scale_file(x)}
        wid.x = 0
        wid.y = 0
        wid.w = win.w
        wid.h = win.h
        wid.pat = type
        Qml::add_child(win, wid)
        self.db.update_values
        setup_widget wid
        self.db.update_values

        if(root)
            root.smash_layout
            root.damage_item(win, :all)
        end
    ;end
  def setup_widget(w);
        w.onSetup()
        w.children.each do |c|
            setup_widget(c)
        end
    ;end
  def load_scale_file(val);
        $remote.action("/load_kbm", val) if @opt == ".kbm"
        $remote.action("/load_scl", val) if @opt == ".scl"
        $remote.action("/microtonal/mapping")
        $remote.action("/microtonal/tunings")
    ;end
  def class_name(); "part" ;end
  def layout(l, selfBox);
        Draw::Layout::vfill(l, selfBox, children, [0.9, 0.1])
    ;end
  def apply();
        return false
        tune.apply
        mapp.apply
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @scale = self
@anonymous1167 = Qml::Widget.new(database, ui_path+"1"+'/')
@tune = Qml::Group.new(database, ui_path+"2"+'/')
@anonymous1163 = Qml::TextEdit.new(database, ui_path+"3"+'/')
@mapp = Qml::Group.new(database, ui_path+"4"+'/')
@anonymous1165 = Qml::TextEdit.new(database, ui_path+"5"+'/')
@anonymous1172 = Qml::ColorBox.new(database, ui_path+"6"+'/')
@anonymous1171 = Qml::ParModuleRow.new(database, ui_path+"7"+'/')
@anonymous1168 = Qml::TriggerButton.new(database, ui_path+"8"+'/')
@anonymous1169 = Qml::TriggerButton.new(database, ui_path+"9"+'/')
@anonymous1170 = Qml::TriggerButton.new(database, ui_path+"10"+'/')

    
context = Hash.new
context["scale"] = self
context["tune"] = @tune
context["mapp"] = @mapp

        @anonymous1167.instance_eval("def class_name(); \"partsub\" ;end", "src/mruby-zest/example/ZynTuning.qml", 64)

        @anonymous1167.instance_eval("def layout(l, selfBox);\n            Draw::Layout::hfill(l, selfBox, children, [0.5, 0.5])\n        ;end", "src/mruby-zest/example/ZynTuning.qml", 65)
@anonymous1163.height = 0.05
@anonymous1163.extern = "/microtonal/tunings"
Qml::add_child(@tune, @anonymous1163)
@tune.label = "tunings"
@db.connect_property(@tune.properties["extern"], lambda {scale.extern + "tunings"})
@tune.copyable = false
Qml::add_child(@anonymous1167, @tune)
@anonymous1165.height = 0.05
@anonymous1165.extern = "/microtonal/mapping"
Qml::add_child(@mapp, @anonymous1165)
@mapp.label = "key mapping"
@db.connect_property(@mapp.properties["extern"], lambda {scale.extern + "mapping"})
@mapp.copyable = false
Qml::add_child(@anonymous1167, @mapp)
Qml::add_child(self, @anonymous1167)
@db.connect_property(@anonymous1168.properties["label"], lambda {"import .scl"})
@anonymous1168.tooltip = "load microtonal scale"
@anonymous1168.whenValue = lambda {scale.load_scale(".scl")}
Qml::add_child(@anonymous1171, @anonymous1168)
@db.connect_property(@anonymous1169.properties["label"], lambda {"import .kbm"})
@anonymous1169.tooltip = "load keyboard mapping"
@anonymous1169.whenValue = lambda {scale.load_scale(".kbm")}
Qml::add_child(@anonymous1171, @anonymous1169)
@anonymous1170.label = "retune"
@anonymous1170.whenValue = lambda { scale.apply }
@anonymous1170.active = false
Qml::add_child(@anonymous1171, @anonymous1170)
Qml::add_child(@anonymous1172, @anonymous1171)
@anonymous1172.bg = Theme::GeneralBackground
Qml::add_child(self, @anonymous1172)

    #puts "ZynTuning, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynWip < Qml::Widget
  def layout(l, selfBox);
        children[0].fixed(l,        selfBox, 0.05, 0.05, 0.95, 0.85)
    ;end
  def draw(vg);
        draw2(vg)
    ;end
  def draw2(vg);
        vg.text_align NVG::ALIGN_MIDDLE | NVG::ALIGN_CENTER
        vg.font_size 30.0
        vg.fill_color color("ffffff")
        vg.text(550, 10, "Format Vowel (Live?) View [radius=1/q]")
        vg.text(20, 470, "-10 dB")
        vg.text(20, 300, "-5 dB")
        vg.text(20, 100, "0 dB")
        vg.text(20, 10, "10 dB")
        
        vg.text(150, 500, "10 Hz")
        vg.text(550, 500, "100 Hz")
        vg.text(950, 500, "1 kHz")

    ;end
  def draw1(vg);
        vg.text_align NVG::ALIGN_MIDDLE | NVG::ALIGN_CENTER
        vg.font_size 30.0
        vg.fill_color color("ffffff")
        vg.text(550, 10, "Format Sequence View")
        vg.text(20, 470, "0 Hz")
        vg.text(20, 300, "100 Hz")
        vg.text(20, 100, "1k Hz")
        vg.text(20, 10, "10k Hz")
        
        vg.text(150, 500, "Seq Pos 1")
        vg.text(550, 500, "Seq Pos 2")
        vg.text(950, 500, "Seq Pos 3")

    ;end
end
class Qml::ZynWipTwo < Qml::Widget
  def circ(vg, cx, cy, r, col);
        vg.path do
            #vg.arc(cx, cy, r, 0, 3.14, 1)
            vg.circle(cx, cy, r)
            vg.fill_color col
            vg.fill
        end
    ;end
  def circ_mag(vg, cx, cy, h, col);
        vg.path do
            vg.rect(cx-5, cy-h, 10, 2*h)
            vg.fill_color col
            vg.fill
        end
    ;end
  def circ_link(vg, x1, y1, m1, x2, y2, m2, col);
        vg.path do
            vg.move_to(x1+5, y1-m1)
            vg.line_to(x2-5, y2-m2)
            vg.line_to(x2-5, y2+m2)
            vg.line_to(x1+5, y1+m1)
            vg.close_path
            vg.fill
        end
        angle = Math.atan2(y2-y1, x2-x1)
        xdiff = Math.cos(angle)*20
        ydiff = Math.sin(angle)*20
        vg.path do
            vg.move_to(x1+xdiff, y1+ydiff)
            vg.line_to(x2-xdiff, y2-ydiff)
            vg.stroke_color col
            vg.stroke
        end
    ;end
  def circ_text(vg, bx, by, label);
        vg.text_align NVG::ALIGN_MIDDLE | NVG::ALIGN_CENTER
        vg.font_size 30.0
        vg.fill_color color("ffffff")
        vg.text(bx, by, label)
    ;end
  def draw(vg);
        draw2(vg)
    ;end
  def draw1(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        vg.path do
            vg.rect(box.x, box.y, box.w, box.h)
            vg.fill_color Theme::VisualBackground
            vg.fill
        end
        Draw::Grid::linear_x(vg, 1, 10, box)
        vg.stroke_width 1.5
        Draw::Grid::log_y(vg, 1, 10000, box)

        radius = 20

        bx1 = box.x + 0.1*box.w
        by1 = box.y + 0.2*box.h
        m1  = 40

        bx2 = box.x+0.5*box.w
        by2 = box.y+0.4*box.h
        m2  = 60
        
        bx3 = box.x+0.9*box.w
        by3 = box.y+0.3*box.h
        m3  = 20

        green = color("00ff00", 100)
        lab   = "1"

        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        green = color("00ffff", 100)
        lab   = "2"
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.9*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.8*box.h
        m2  = 100
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.8*box.h
        m3  = 90
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        
        green = color("ffff00", 100)
        lab   = "3"
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.4*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.5*box.h
        m2  = 2
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.6*box.h
        m3  = 25
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        lab   = "4"
        green = color("ff0000", 100)
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.1*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.15*box.h
        m2  = 35
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.6*box.h
        m3  = 25
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        #circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        #circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        #circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        #circ_text(vg, bx3, by3, lab)

    ;end
  def circ_filter(vg, bx, by, br, lab, col, box);
        circ(vg, bx, by, br, col)

        delta = (box.h-by)/2.0
        bot_left = [bx-delta-br, box.h]
        bot_right = [bx+delta+br, box.h]

        #vg.path do
        #    vg.move_to(bx-br, by)
        #    vg.line_to(bot_left[0], bot_left[1])
        #    vg.line_to(bot_right[0], bot_right[1])
        #    vg.line_to(bx+br, by)
        #    vg.fill_color(col)
        #    vg.fill
        #end
        
        circ_text(vg, bx, by, lab)
    ;end
  def draw2(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        vg.path do
            vg.rect(box.x, box.y, box.w, box.h)
            vg.fill_color Theme::VisualBackground
            vg.fill
        end
        Draw::Grid::log_x(vg, 1, 10000, box)
        vg.stroke_width 1.5
        Draw::Grid::linear_y(vg, 1, 50, box)

        radius = 10
        green = color("00ff00", 100)
        lab   = "1"
        bx1 = box.x + 0.1*box.w
        by1 = box.y + 0.8*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 30
        green = color("00ffff", 100)
        lab   = "2"
        bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.75*box.h
        
        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 20
        green = color("ffff00", 100)
        lab   = "3"
        bx1 = box.x + 0.6*box.w
        by1 = box.y + 0.2*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 90
        green = color("ff0000", 100)
        lab   = "4"
        bx1 = box.x + 0.8*box.w
        by1 = box.y + 0.4*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1176 = self

    
context = Hash.new

    #puts "ZynWipTwo, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynWip
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1175 = self
@anonymous1174 = Qml::ZynWipTwo.new(database, ui_path+"1"+'/')

    
context = Hash.new
Qml::add_child(self, @anonymous1174)

    #puts "ZynWip, #{1000000*(Time.new-t1)}, 123456"
end
end
class Qml::ZynWipTwo < Qml::Widget
  def circ(vg, cx, cy, r, col);
        vg.path do
            #vg.arc(cx, cy, r, 0, 3.14, 1)
            vg.circle(cx, cy, r)
            vg.fill_color col
            vg.fill
        end
    ;end
  def circ_mag(vg, cx, cy, h, col);
        vg.path do
            vg.rect(cx-5, cy-h, 10, 2*h)
            vg.fill_color col
            vg.fill
        end
    ;end
  def circ_link(vg, x1, y1, m1, x2, y2, m2, col);
        vg.path do
            vg.move_to(x1+5, y1-m1)
            vg.line_to(x2-5, y2-m2)
            vg.line_to(x2-5, y2+m2)
            vg.line_to(x1+5, y1+m1)
            vg.close_path
            vg.fill
        end
        angle = Math.atan2(y2-y1, x2-x1)
        xdiff = Math.cos(angle)*20
        ydiff = Math.sin(angle)*20
        vg.path do
            vg.move_to(x1+xdiff, y1+ydiff)
            vg.line_to(x2-xdiff, y2-ydiff)
            vg.stroke_color col
            vg.stroke
        end
    ;end
  def circ_text(vg, bx, by, label);
        vg.text_align NVG::ALIGN_MIDDLE | NVG::ALIGN_CENTER
        vg.font_size 30.0
        vg.fill_color color("ffffff")
        vg.text(bx, by, label)
    ;end
  def draw(vg);
        draw2(vg)
    ;end
  def draw1(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        vg.path do
            vg.rect(box.x, box.y, box.w, box.h)
            vg.fill_color Theme::VisualBackground
            vg.fill
        end
        Draw::Grid::linear_x(vg, 1, 10, box)
        vg.stroke_width 1.5
        Draw::Grid::log_y(vg, 1, 10000, box)

        radius = 20

        bx1 = box.x + 0.1*box.w
        by1 = box.y + 0.2*box.h
        m1  = 40

        bx2 = box.x+0.5*box.w
        by2 = box.y+0.4*box.h
        m2  = 60
        
        bx3 = box.x+0.9*box.w
        by3 = box.y+0.3*box.h
        m3  = 20

        green = color("00ff00", 100)
        lab   = "1"

        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        green = color("00ffff", 100)
        lab   = "2"
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.9*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.8*box.h
        m2  = 100
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.8*box.h
        m3  = 90
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        
        green = color("ffff00", 100)
        lab   = "3"
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.4*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.5*box.h
        m2  = 2
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.6*box.h
        m3  = 25
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        circ_text(vg, bx3, by3, lab)
        
        lab   = "4"
        green = color("ff0000", 100)
        #bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.1*box.h
        m1  = 10

        #bx2 = box.x+0.4*box.w
        by2 = box.y+0.15*box.h
        m2  = 35
        
        #bx3 = box.x+0.6*box.w
        by3 = box.y+0.6*box.h
        m3  = 25
        
        circ(vg, bx1, by1, radius, green)
        circ(vg, bx2, by2, radius, green)
        #circ(vg, bx3, by3, radius, green)

        circ_mag(vg, bx1, by1, m1, green)
        circ_mag(vg, bx2, by2, m2, green)
        #circ_mag(vg, bx3, by3, m3, green)

        circ_link(vg, bx1, by1, m1, bx2, by2, m2, green)
        #circ_link(vg, bx2, by2, m2, bx3, by3, m3, green)
        circ_text(vg, bx1, by1, lab)
        circ_text(vg, bx2, by2, lab)
        #circ_text(vg, bx3, by3, lab)

    ;end
  def circ_filter(vg, bx, by, br, lab, col, box);
        circ(vg, bx, by, br, col)

        delta = (box.h-by)/2.0
        bot_left = [bx-delta-br, box.h]
        bot_right = [bx+delta+br, box.h]

        #vg.path do
        #    vg.move_to(bx-br, by)
        #    vg.line_to(bot_left[0], bot_left[1])
        #    vg.line_to(bot_right[0], bot_right[1])
        #    vg.line_to(bx+br, by)
        #    vg.fill_color(col)
        #    vg.fill
        #end
        
        circ_text(vg, bx, by, lab)
    ;end
  def draw2(vg);
        fixedpad = 5
        pad  = 0
        pad2 = (1-2*pad)
        box = Rect.new(w*pad  + fixedpad,   h*pad  + fixedpad,
                       w*pad2 - 2*fixedpad, h*pad2 - 2*fixedpad)
        vg.path do
            vg.rect(box.x, box.y, box.w, box.h)
            vg.fill_color Theme::VisualBackground
            vg.fill
        end
        Draw::Grid::log_x(vg, 1, 10000, box)
        vg.stroke_width 1.5
        Draw::Grid::linear_y(vg, 1, 50, box)

        radius = 10
        green = color("00ff00", 100)
        lab   = "1"
        bx1 = box.x + 0.1*box.w
        by1 = box.y + 0.8*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 30
        green = color("00ffff", 100)
        lab   = "2"
        bx1 = box.x + 0.2*box.w
        by1 = box.y + 0.75*box.h
        
        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 20
        green = color("ffff00", 100)
        lab   = "3"
        bx1 = box.x + 0.6*box.w
        by1 = box.y + 0.2*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)

        radius = 90
        green = color("ff0000", 100)
        lab   = "4"
        bx1 = box.x + 0.8*box.w
        by1 = box.y + 0.4*box.h

        circ_filter(vg, bx1, by1, radius, lab, green, box)
    ;end
  attr_reader :properties
  attr_accessor :db, :ui_path
  def initialize(database, ui_path="/ui/")
    #t1 = Time.new
    super(database, ui_path)
    @db         ||= database
    @ui_path      = ui_path
    @properties ||= Hash.new
    #t1 = Time.new
    @anonymous1176 = self

    
context = Hash.new

    #puts "ZynWipTwo, #{1000000*(Time.new-t1)}, 123456"
end
end
